/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstdlib>
#include <cerrno>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <classlib/Xml.hpp>
#include <classlib/String.hpp>

#ifdef QWS
#include <qpe/config.h>
#else
#include <myqpe/Config.hpp>
#endif

#include <qmessagebox.h>

#include "BrowserConfigurationManager.hpp"
#include "MainFormImpl.hpp"
#include "BrowserResource.hpp"
#include "BrowserSnmpProfile.hpp"
#include "BrowserNode.hpp"

BrowserConfigurationManager::BrowserConfigurationManager(MainFormImpl *parent)
               : parent(parent), newConfiguration(0), xml(0) {
  }

BrowserConfigurationManager::~BrowserConfigurationManager() {
  delete newConfiguration;
  delete xml;
  }

static void loadPreferences(BrowserConfiguration *configuration) {
  Config preferences(BrowserResource::applicationName(),Config::User);
  preferences.setGroup("General");

  configuration -> preferSystemIcons =
                       preferences.readBoolEntry("preferSystemIcons",false);

  QString applicationFontFamily =
                 preferences.readEntry("applicationFontFamily");
  if (applicationFontFamily.isNull())
    configuration -> applicationFontFamily = 0;
  else {
    configuration -> applicationFontFamily =
                            new String(applicationFontFamily);
    }

  QString applicationFontSize = preferences.readEntry("applicationFontSize");
  if (applicationFontSize.isNull())
    configuration -> applicationFontSize = 0;
  else
    configuration -> applicationFontSize = applicationFontSize.toInt();

  QString textFontFamily = preferences.readEntry("textFontFamily");
  if (textFontFamily.isNull())
    configuration -> textFontFamily = 0;
  else
    configuration -> textFontFamily = new String(textFontFamily);

  QString textFontSize = preferences.readEntry("textFontSize");
  if (textFontSize.isNull())
    configuration -> textFontSize = 0;
  else
    configuration -> textFontSize = textFontSize.toInt();
  }

static void savePreferences(BrowserConfiguration *configuration) {
  Config preferences(BrowserResource::applicationName(),Config::User);
  preferences.setGroup("General");

  preferences.writeEntry("preferSystemIcons",
                         configuration -> preferSystemIcons);

  if (configuration -> applicationFontFamily) {
    preferences.writeEntry("applicationFontFamily",
                           configuration -> applicationFontFamily ->
                                                     getChars());
    }
  else preferences.removeEntry("applicationFontFamily");

  if (configuration -> applicationFontSize != 0) {
    preferences.writeEntry("applicationFontSize",
                           QString().setNum(configuration ->
                                                  applicationFontSize));
    }
  else preferences.removeEntry("applicationFontSize");

  if (configuration -> textFontFamily) {
    preferences.writeEntry("textFontFamily",
                           configuration -> textFontFamily -> getChars());
    }
  else preferences.removeEntry("textFontFamily");

  if (configuration -> textFontSize != 0) {
    preferences.writeEntry("textFontSize",
                           QString().setNum(configuration -> textFontSize));
    }
  else preferences.removeEntry("textFontSize");
  }

bool BrowserConfigurationManager::parseProfileXml(XmlElement profileElement) {

  const char *profileName = xml -> getElementAttributeValue(profileElement,0);

  BrowserSnmpProfile *profile = new BrowserSnmpProfile();
  profile -> setName(profileName);

  int childCount = xml -> getElementChildCount(profileElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement element = xml -> getElementChild(profileElement,i);
    const char *elementName = xml -> getElementName(element);
    if (strcmp(elementName,"transport") == 0) {
      int transportParameterCount = xml -> getElementChildCount(element);
      for (int p = 0 ; p < transportParameterCount ; p++ ) {
        XmlElement parameterElement = xml -> getElementChild(element,p);
        const char *parameterName = xml -> getElementName(parameterElement);
        if (strcmp(parameterName,"timeout") == 0) {
          profile ->
             setTimeout(atoi(xml -> getElementContents(parameterElement)));
          }
        else if (strcmp(parameterName,"retries") == 0) {
          profile ->
             setRetries(atoi(xml -> getElementContents(parameterElement)));
          }
        else if (strcmp(parameterName,"port") == 0) {
          profile ->
             setPort(atoi(xml -> getElementContents(parameterElement)));
          }
        }
      }
    else if (strcmp(elementName,"snmp") == 0) {
      const char *version = xml -> getElementAttributeValue(element,0);
      if (strcmp(version,"1") == 0)
        profile -> setVersion(BrowserSnmpProfile::v1);
      else if (strcmp(version,"2c") == 0)
        profile -> setVersion(BrowserSnmpProfile::v2c);
      else if (strcmp(version,"3") == 0)
        profile -> setVersion(BrowserSnmpProfile::v3);

      int keyCount = xml -> getElementChildCount(element);
      for (int k = 0 ; k < keyCount ; k++ ) {
        XmlElement keyElement = xml -> getElementChild(element,k);
        const char *keyName = xml -> getElementName(keyElement);
        if (strcmp(keyName,"readCommunity") == 0)
          profile -> setReadCommunity(xml -> getElementContents(keyElement));
        else if (strcmp(keyName,"writeCommunity") == 0)
          profile -> setWriteCommunity(xml -> getElementContents(keyElement));
        else if (strcmp(keyName,"security") == 0) {
          int securityChildCount = xml -> getElementChildCount(keyElement);
          for (int s = 0 ; s < securityChildCount ; s++ ) {
            XmlElement secElement = xml -> getElementChild(keyElement,s);
            const char *secElementName = xml -> getElementName(secElement);

            if (strcmp(secElementName,"name") == 0) {
              profile -> setSecurityName(xml -> getElementContents(secElement));
              }
            else if (strcmp(secElementName,"authentication") == 0) {
              const char *typeName =
                            xml -> getElementAttributeValue(secElement,0);
              if (strcmp(typeName,"none") == 0)
                profile -> setAuthProtocol(BrowserSnmpProfile::authNone);
              else {
                if (strcmp(typeName,"hmacMd5") == 0)
                  profile -> setAuthProtocol(BrowserSnmpProfile::authHmacMd5);
                else if (strcmp(typeName,"hmacSha") == 0)
                  profile -> setAuthProtocol(BrowserSnmpProfile::authHmacSha);

                XmlElement passwordElement =
                                 xml -> getElementChild(secElement,0);
                const char *password =
                               xml -> getElementContents(passwordElement);
                if (password) profile -> setAuthPassword(password);
                }
              }
            else if (strcmp(secElementName,"privacy") == 0) {
              const char *typeName =
                              xml -> getElementAttributeValue(secElement,0);
              if (strcmp(typeName,"none") == 0)
                profile -> setPrivProtocol(BrowserSnmpProfile::privNone);
              else {
                if (strcmp(typeName,"des") == 0)
                  profile -> setPrivProtocol(BrowserSnmpProfile::privDes);
                else if (strcmp(typeName,"aes128") == 0)
                  profile -> setPrivProtocol(BrowserSnmpProfile::privAes128);
                else if (strcmp(typeName,"aes192") == 0)
                  profile -> setPrivProtocol(BrowserSnmpProfile::privAes192);
                else if (strcmp(typeName,"aes256") == 0)
                  profile -> setPrivProtocol(BrowserSnmpProfile::privAes256);

                XmlElement passwordElement =
                                    xml -> getElementChild(secElement,0);
                const char *password =
                               xml -> getElementContents(passwordElement);
                if (password) profile -> setPrivPassword(password);
                }
              }
            }
          }
        }
      }
    }

  newConfiguration -> profiles -> add(profile);

  return true;
  }

bool BrowserConfigurationManager::parseProfilesXml(XmlElement profilesElement) {

  int childCount = xml -> getElementChildCount(profilesElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement profileElement = xml -> getElementChild(profilesElement,i);
    if (! parseProfileXml(profileElement)) {
      return false;
      }
    }

  return true;
  }

bool BrowserConfigurationManager::parseNodeXml(XmlElement nodeElement) {

  BrowserNode *node = new BrowserNode();

  int childCount = xml -> getElementChildCount(nodeElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement element = xml -> getElementChild(nodeElement,i);
    const char *elementName = xml -> getElementName(element);
    if (strcmp(elementName,"name") == 0) {
      node -> setName(xml -> getElementContents(element));
      }
    else if (strcmp(elementName,"ip") == 0) {
      node -> setIpAddress(xml -> getElementContents(element));
      }
    else if (strcmp(elementName,"profile") == 0) {
      node -> setProfile(xml -> getElementContents(element));
      }
    }

  newConfiguration -> nodes -> add(node);

  return true;
  }

bool BrowserConfigurationManager::parseNodesXml(XmlElement nodesElement) {

  int childCount = xml -> getElementChildCount(nodesElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement nodeElement = xml -> getElementChild(nodesElement,i);
    if (! parseNodeXml(nodeElement)) {
      return false;
      }
    }

  return true;
  }

void BrowserConfigurationManager::parseXml() {
  XmlElement root = xml -> getRootElement();
  const char *elementName = xml -> getElementName(root);
  if (strcmp(elementName,BrowserResource::applicationName()) != 0) {
    QMessageBox::critical(parent,BrowserResource::applicationTitle(),
                          "Unrecognised structure in the\n"
                          "configuration settings file!\n"
                          "Aborting.",
                          QMessageBox::Ok,0);
    return;
    }

  const char *version = xml -> getElementAttributeValue(root,0);
  if (strcmp(version,"1") != 0) {
    QMessageBox::critical(parent,BrowserResource::applicationTitle(),
                          "Unrecognised configuration version\n"
                          "in the configuration settings file!\n"
                          "Aborting.",
                          QMessageBox::Ok,0);
    return;
    }

  int childCount = xml -> getElementChildCount(root);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement element = xml -> getElementChild(root,i);
    elementName = xml -> getElementName(element);
    if (strcmp(elementName,"profiles") == 0) {
      if (! parseProfilesXml(element)) {
        return;
        }
      }
    else if (strcmp(elementName,"nodes") == 0) {
      if (! parseNodesXml(element)) {
        return;
        }
      }
    }
  }

static bool setupPath(MainFormImpl *parent) {
  //
  // Check whether the configuration directory exists
  //
  struct stat status;
  if (::stat(BrowserResource::getApplicationDirUser(),&status) == -1) {
    //
    // The directory does not exist, or there is an access problem
    //
    if (errno != ENOENT) {
      QMessageBox::critical(parent,BrowserResource::applicationTitle(),
                            "Cannot access the directory\n"
                            "which holds configuration\n"
                            "settings! Aborting.",
                            QMessageBox::Ok,0);
      return false;
      }

    //
    // The directory does not exist, so create it
    //
    if (mkdir(BrowserResource::getApplicationDirUser(),0755) == -1) {
      QMessageBox::critical(parent,BrowserResource::applicationTitle(),
                            "Cannot create a directory\n"
                            "to hold configuration\n"
                            "settings! Aborting.",
                            QMessageBox::Ok,0);
      return false;
      }
    }

  //
  // The configuration directory now exists
  //
  return true;
  }

void BrowserConfigurationManager::load() {
  //
  // Start with an empty configuration
  //
  delete newConfiguration;
  newConfiguration = new BrowserConfiguration();

  //
  // First, load the simple (non-XML) preferences
  //
  loadPreferences(newConfiguration);

  //
  // Create the path to the directory holding configuration files
  // for this application
  //
  if (! setupPath(parent)) {
    std::exit(1);
    }

  //
  // See if the main configuration file can be read
  //
  struct stat status;
  QString path = BrowserResource::getApplicationDirUser() +
                 BrowserResource::configurationFileName();
  if (::stat(path,&status) == -1) {
    //
    // The main configuration file does not exist, or there is an
    // access problem
    //
    if (errno != ENOENT) {
      QMessageBox::critical(parent,BrowserResource::applicationTitle(),
                            "Cannot access the application\n"
                            "configuration settings file!\n"
                            "Aborting.",
                            QMessageBox::Ok,0);
      std::exit(1);
      }

    //
    // The configuration file does not exist. Create a dummy file just
    // to check whether the configuration should be saveable later..
    //
    int fd = ::creat(path,0600);
    if (fd == -1) {
      QMessageBox::critical(parent,BrowserResource::applicationTitle(),
                            "Cannot create the application\n"
                            "configuration settings file!\n"
                            "Aborting.",
                            QMessageBox::Ok,0);
      std::exit(1);
      }

    //
    // Remove the test file because an empty file is not valid XML
    //
    ::close(fd);
    ::unlink(path);

    //
    // Run with an empty configuration
    //
    delete parent -> configuration;
    parent -> configuration = newConfiguration;
    newConfiguration = 0;
    return;
    }

  //
  // The configuration file exists, so populate the configuration
  // from it
  //
  xml = new Xml();
  xml -> loadXmlFromFile(path);
  if (xml -> exception) {
    int code = QMessageBox::critical(parent,
                                     BrowserResource::applicationTitle(),
                                     "The application configuration file\n"
                                     "is invalid.\n\n"
                                     "Continue with empty configuration,\n"
                                     "or abort?",
                                     parent -> tr("Continue"),
                                     parent -> tr("Abort"),QString::null,1,1);
    if (code == 1) { // Abort
      std::exit(1);
      }

    rename(path,path + ".bad");

    delete parent -> configuration;
    parent -> configuration = newConfiguration;
    newConfiguration = 0;
    return;
    }

  parseXml();

  //
  // Give the parent its new configuration, replacing whatever it had
  //
  delete parent -> configuration;
  parent -> configuration = newConfiguration;
  newConfiguration = 0;
  }

void BrowserConfigurationManager::addNodesXml(XmlElement root) {

  XmlElement nodesElement = xml -> addElement(root,"nodes");

  ListIterator it = parent -> configuration -> nodes -> iterator();
  while (it.hasNext()) {
    BrowserNode *node = (BrowserNode *)it.next();
    XmlElement nodeElement = xml -> addElement(nodesElement,"node");

    XmlElement nameElement = xml -> addElement(nodeElement,"name");
    xml -> setElementContents(nameElement,node -> getName());

    XmlElement ipAddressElement = xml -> addElement(nodeElement,"ip");
    xml -> setElementContents(ipAddressElement,node -> getIpAddress());

    XmlElement profileElement = xml -> addElement(nodeElement,"profile");
    xml -> setElementContents(profileElement,node -> getProfile());
    }
  }

void BrowserConfigurationManager::addProfilesXml(XmlElement root) {
  XmlElement profilesElement = xml -> addElement(root,"profiles");
  ListIterator it = parent -> configuration -> profiles -> iterator();
  while (it.hasNext()) {
    BrowserSnmpProfile *profile = (BrowserSnmpProfile *)it.next();

    XmlElement profileElement = xml -> addElement(profilesElement,"profile");
    xml -> addElementAttribute(profileElement,"name",profile -> getName());

    XmlElement transportElement = xml -> addElement(profileElement,"transport");

    XmlElement retriesElement = xml -> addElement(transportElement,"retries");
    char value[16];
    std::sprintf(value,"%d",profile -> getRetries());
    xml -> setElementContents(retriesElement,value);

    XmlElement timeoutElement = xml -> addElement(transportElement,"timeout");
    std::sprintf(value,"%d",profile -> getTimeout());
    xml -> setElementContents(timeoutElement,value);

    XmlElement portElement = xml -> addElement(transportElement,"port");
    std::sprintf(value,"%d",profile -> getPort());
    xml -> setElementContents(portElement,value);

    XmlElement snmpElement = xml -> addElement(profileElement,"snmp");
    switch (profile -> getVersion()) {
      case BrowserSnmpProfile::v1:
        xml -> addElementAttribute(snmpElement,"version","1");
        break;
      case BrowserSnmpProfile::v2c:
        xml -> addElementAttribute(snmpElement,"version","2c");
        break;
      case BrowserSnmpProfile::v3:
        xml -> addElementAttribute(snmpElement,"version","3");
        break;
      }

    if (profile -> getVersion() != BrowserSnmpProfile::v3) {
      XmlElement readCommunityElement = xml -> addElement(snmpElement,
                                                          "readCommunity");
      xml -> setElementContents(readCommunityElement,
                                profile -> getReadCommunity());

      XmlElement writeCommunityElement = xml -> addElement(snmpElement,
                                                           "writeCommunity");
      xml -> setElementContents(writeCommunityElement,
                                profile -> getWriteCommunity());
      }
    else {
      XmlElement secElement = xml -> addElement(snmpElement,"security");

      XmlElement secNameElement = xml -> addElement(secElement,"name");
      xml -> setElementContents(secNameElement,profile -> getSecurityName());

      XmlElement authElement = xml -> addElement(secElement,"authentication");

      switch (profile -> getAuthProtocol()) {
        case BrowserSnmpProfile::authNone:
          xml -> addElementAttribute(authElement,"type","none");
          break;
        case BrowserSnmpProfile::authHmacMd5: {
          xml -> addElementAttribute(authElement,"type","hmacMd5");
          XmlElement authPasswordElement = xml -> addElement(authElement,
                                                             "password");
          xml -> setElementContents(authPasswordElement,
                                    profile -> getAuthPassword());
          break;
          }
        case BrowserSnmpProfile::authHmacSha: {
          xml -> addElementAttribute(authElement,"type","hmacSha");
          XmlElement authPasswordElement = xml -> addElement(authElement,
                                                             "password");
          xml -> setElementContents(authPasswordElement,
                                    profile -> getAuthPassword());
          break;
          }
        }

      XmlElement privElement = xml -> addElement(secElement,"privacy");

      switch (profile -> getPrivProtocol()) {
        case BrowserSnmpProfile::privNone:
          xml -> addElementAttribute(privElement,"type","none");
          break;
        case BrowserSnmpProfile::privDes: {
          xml -> addElementAttribute(privElement,"type","des");
          XmlElement privPasswordElement = xml -> addElement(privElement,
                                                             "password");
          xml -> setElementContents(privPasswordElement,
                                    profile -> getPrivPassword());
          break;
          }
        case BrowserSnmpProfile::privAes128: {
          xml -> addElementAttribute(privElement,"type","aes128");
          XmlElement privPasswordElement = xml -> addElement(privElement,
                                                             "password");
          xml -> setElementContents(privPasswordElement,
                                    profile -> getPrivPassword());
          break;
          }
        case BrowserSnmpProfile::privAes192: {
          xml -> addElementAttribute(privElement,"type","aes192");
          XmlElement privPasswordElement = xml -> addElement(privElement,
                                                             "password");
          xml -> setElementContents(privPasswordElement,
                                    profile -> getPrivPassword());
          break;
          }
        case BrowserSnmpProfile::privAes256: {
          xml -> addElementAttribute(privElement,"type","aes256");
          XmlElement privPasswordElement = xml -> addElement(privElement,
                                                             "password");
          xml -> setElementContents(privPasswordElement,
                                    profile -> getPrivPassword());
          break;
          }
        }
      }
    }
  }

void BrowserConfigurationManager::createConfigurationXml() {
  delete xml;
  xml = new Xml;

  XmlElement root = xml -> setRootElement(BrowserResource::applicationName());
  xml -> addElementAttribute(root,"version","1");

  addNodesXml(root);
  addProfilesXml(root);
  }

static bool writeData(int fd,const char *data,int length) {
  while (length != 0) {
    //
    // Write as much out as possible
    //
    int state = write(fd,data,length);

    //
    // Check to see if something went wrong
    //
    if (state == -1) {
      int savedErrno = errno;
      (void)::close(fd);
      errno = savedErrno;
      return false;
      }

    //
    // The return value tells us how much of the remaining data
    // was written
    //
    data += state;
    length -= state;
    }

  return true;
  }

void BrowserConfigurationManager::save() {
  //
  // First save the simple user preferences (non-XML)
  //
  savePreferences(parent -> configuration);

  //
  // We know the directory path has been setup, because we did that
  // on startup. So, we go straight to an attempt to open the
  // file for writing with new content
  //
  QString path = BrowserResource::getApplicationDirUser() +
                 BrowserResource::configurationFileName();

  //
  // Convert the configuration into an XML structure
  //
  createConfigurationXml();
  char *xmlString = xml -> xmlString();

  while (true) {
    //
    // Try to open the file for writing
    //
    int fd;
    while (true) {
      fd = open(path,O_WRONLY | O_TRUNC | O_CREAT,0600);
      if (fd == -1) {
        int code = QMessageBox::critical(parent,
                                         BrowserResource::applicationTitle(),
                                         "The application configuration file\n"
                                         "could not be opened for writing.\n\n"
                                         "Retry or or continue without saving?",
                                         parent -> tr("Retry"),
                                         parent -> tr("Continue"),
                                         QString::null,0,1);
        if (code == 0) continue;
        else {
          //
          // Give up trying to save
          //
          Xml::free(xmlString);
          return;
          }
        }

      break;
      }

    //
    // The file is open for writing, so write the XML to it.
    //
    bool writeOk = writeData(fd,xmlString,strlen(xmlString));
    if ((! writeOk) || (::close(fd) == -1)) {
      if (! writeOk) (void)close(fd);
      int code = QMessageBox::critical(parent,
                                       BrowserResource::applicationTitle(),
                                       "The application configuration file\n"
                                       "could not be completely written.\n\n"
                                       "Retry or or continue without saving?",
                                       parent -> tr("Retry"),
                                       parent -> tr("Continue"),
                                       QString::null,0,1);
      if (code == 0) continue;
      else {
        //
        // Give up trying to save. Since the file may be only partially
        // written, we delete it. The user has given permission..
        //
        ::unlink(path);
        Xml::free(xmlString);
        return;
        }
      }

    //
    // Successfully completed writing to the file (still possible for
    // there to be a problem, but the application's job is successful)
    //
    Xml::free(xmlString);
    return;
    }
  }
