/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>

#include "BrowserGroupViewItemWidget.hpp"

#include <qsimplerichtext.h>
#include <qpainter.h>

void BrowserGroupViewItemWidget::adjustTextSizes(int widthGoal) {
  //
  // Start by forcing both name and value to their maximum widths
  //
  nameText -> setWidth(nameWidthBound);
  valueText -> setWidth(valueWidthBound);

  //
  // Reduce the width of the widest until we're inside our width goal
  // or we have reduced the width as much as possible
  //
  int nameWidth = nameText -> widthUsed();
  int valueWidth = valueText -> widthUsed();
  int nameHeight = nameText -> height();
  int valueHeight = valueText -> height();
  int newWidth;
  int change = widthGoal / 10;
  bool nameReducible = true;
  bool valueReducible = true;
  while (((nameWidth + valueWidth) > widthGoal) &&
         (nameReducible || valueReducible)) {

    if ((nameWidth > valueWidth) || (! valueReducible)) {
      newWidth = nameWidth - change;
      nameText -> setWidth(newWidth);
      nameWidth = nameText -> widthUsed();
      nameHeight = nameText -> height();
      if (nameWidth > newWidth) nameReducible = false;

      //
      // If the name height changed as a result of reducing the width
      // (which it normally will), then reduce the value width until
      // it reaches the same height or it stops being reducible
      //
      if (valueReducible) {
        if (nameHeight != valueHeight) {
          while (nameHeight != valueHeight) {
            newWidth = valueWidth - change;
            valueText -> setWidth(newWidth);
            valueWidth = valueText -> widthUsed();
            if (valueWidth > newWidth) {
              valueReducible = false;
              break;
              }

            valueHeight = valueText -> height();
            }

          nameText -> setWidth(nameText -> widthUsed());
          }
        }
      }

    if ((valueWidth > nameWidth) || ((! nameReducible) && valueReducible)) {
      newWidth = valueWidth - change;
      valueText -> setWidth(newWidth);
      valueWidth = valueText -> widthUsed();
      valueHeight = valueText -> height();
      if (valueWidth > newWidth) valueReducible = false;

      //
      // If the value height changed as a result of reducing the width
      // (which it often will), then reduce the name width until
      // it changes to match
      //
      if (nameReducible) {
        if (nameHeight != valueHeight) {
          while (nameHeight != valueHeight) {
            newWidth = nameWidth - change;
            nameText -> setWidth(newWidth);
            nameWidth = nameText -> widthUsed();
            if (nameWidth > newWidth) {
              nameReducible = false;
              break;
              }

            nameHeight = nameText -> height();
            }


          nameText -> setWidth(nameText -> widthUsed());
          }
        }
      }
    }

  //
  // Set the size of the entire widget according to the component sizes,
  // margins and padding
  //
  adjustSize();
  }

BrowserGroupViewItemWidget::BrowserGroupViewItemWidget(QWidget *parent,
                                                       const char *name,
                                                       const char *value,
                                                       int widthGoal)
              : QWidget(parent),
                namePaddedWidth(-1), valuePaddedWidth(-1),
                separatorMargin(0) {

  nameText = new QSimpleRichText(name,font());
  separatorText = new QSimpleRichText(":",font());
  valueText = new QSimpleRichText(value,font());

  //
  // Adjust the size of the separator widget to fit its permanent
  // content
  //
  separatorText -> adjustSize();

  //
  // Remember the maximum widths possible for the texts in the current font
  //
  nameWidthBound = fontMetrics().boundingRect(name).width();
  valueWidthBound = fontMetrics().boundingRect(value).width();

  //
  // Adjust the sizes of the name and value widgets to be within
  // the initial width goal
  //
  adjustTextSizes(widthGoal);
  }

BrowserGroupViewItemWidget::~BrowserGroupViewItemWidget() {
  delete nameText;
  delete separatorText;
  delete valueText;
  }

int BrowserGroupViewItemWidget::nameWidth() const {
  return nameText -> widthUsed();
  }

int BrowserGroupViewItemWidget::separatorWidth() const {
  return separatorText -> widthUsed();
  }

int BrowserGroupViewItemWidget::valueWidth() const {
  return valueText -> widthUsed();
  }

void BrowserGroupViewItemWidget::setSeparatorMargin(int separatorMargin) {
  this -> separatorMargin = separatorMargin;
  }

void BrowserGroupViewItemWidget::adjustSize() {
  if (namePaddedWidth == -1) namePaddedWidth = nameText -> widthUsed();
  if (valuePaddedWidth == -1) valuePaddedWidth = valueText -> widthUsed();
  int h = QMAX(nameText -> height(),valueText -> height());
  setFixedSize(namePaddedWidth +
               2 * separatorMargin + separatorText -> widthUsed() +
               valuePaddedWidth,h);
  }

void BrowserGroupViewItemWidget::setNamePaddedWidth(int namePaddedWidth) {
  this -> namePaddedWidth = namePaddedWidth;
  }

void BrowserGroupViewItemWidget::setValuePaddedWidth(int valuePaddedWidth) {
  this -> valuePaddedWidth = valuePaddedWidth;
  }

void BrowserGroupViewItemWidget::paintEvent(QPaintEvent *e) {

  QPainter painter(this);

  int widgetHeight = height();
#ifdef QWS
  QRegion clip(0,0,width(),widgetHeight);
#else
  QRect clip(0,0,width(),widgetHeight);
#endif

  int textHeight = nameText -> height();
  int y = (widgetHeight - textHeight) / 2;
  nameText -> draw(&painter,0,y,clip,colorGroup(),0);

  textHeight = separatorText -> height();
  y = (widgetHeight - textHeight) / 2;
  separatorText -> draw(&painter,namePaddedWidth + separatorMargin,y,
                        clip,colorGroup(),0);

  textHeight = valueText -> height();
  y = (widgetHeight - textHeight) / 2;
  valueText -> draw(&painter,namePaddedWidth + 2 * separatorMargin +
                    separatorText -> widthUsed(),y,clip,colorGroup(),0);
  }
