/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>

#include "BrowserGroupViewLayout.hpp"
#include "BrowserGroupViewItemWidget.hpp"

const int MaxColumns    = 4;
const int MinSepMargin  = 4;
const int MinColSpacing = 10;
const int MinRowSpacing = 0;

const int RedrawDelay   = 100; // ms

#include <qlabel.h>
#include <qscrollview.h>
#include <qtimer.h>

class BrowserGroupViewLayoutData {
  public:
    BrowserGroupViewLayoutData() :
            separatorWidth(0),
            lastColumns(0), lastRows(0),
            currentPreferredSize(QSize()),
            visibleSize(QSize()),
            calculatedColumns(0), calculatedRows(0) { }

    QList<QLayoutItem> items;

    int separatorWidth;

    //
    // Last geometry drawn
    //
    int lastColumns;
    int lastRows;

    //
    // Cached preferred size
    //
    QSize currentPreferredSize;

    //
    // Most recent visible size hint from our parent
    //
    QSize visibleSize;

    //
    // Side-effects of calling heightForWidth()
    //
    int calculatedColumns;
    int calculatedRows;
    int calculatedHeight;
    int maxNameWidth[MaxColumns];
    int maxValueWidth[MaxColumns];
  };

BrowserGroupViewLayoutItem::BrowserGroupViewLayoutItem(QWidget *parent,
                                                       const char *name,
                                                       const char *value,
                                                       int widthGoal) {

  widget = new BrowserGroupViewItemWidget(parent,name,value,widthGoal);
  }

void BrowserGroupViewLayoutItem::move(int x,int y) {
  widget -> move(x,y);
  }

void BrowserGroupViewLayoutItem::setBackgroundColor(
                                      const QColor& backgroundColor) {

  widget -> setBackgroundColor(backgroundColor);
  }

int BrowserGroupViewLayoutItem::nameWidth() const {
  return widget -> nameWidth();
  }

int BrowserGroupViewLayoutItem::separatorWidth() const {
  return widget -> separatorWidth();
  }

int BrowserGroupViewLayoutItem::valueWidth() const {
  return widget -> valueWidth();
  }

int BrowserGroupViewLayoutItem::height() const {
  return widget -> height();
  }

int BrowserGroupViewLayoutItem::width() const {
  return widget -> width();
  }

void BrowserGroupViewLayoutItem::setWidths(int nameColumnWidth,
                                           int valueColumnWidth) {

  widget -> setNamePaddedWidth(nameColumnWidth);
  widget -> setValuePaddedWidth(valueColumnWidth);
  widget -> adjustSize();
  }

void BrowserGroupViewLayoutItem::setSize(int width,int height) {
  widget -> setFixedSize(width,height);
  }

QSize BrowserGroupViewLayoutItem::sizeHint() const {
  return widget -> sizeHint();
  }

QSize BrowserGroupViewLayoutItem::minimumSize() const {
  return widget -> sizeHint();
  }

QSize BrowserGroupViewLayoutItem::maximumSize() const {
  return widget -> sizeHint();
  }

bool BrowserGroupViewLayoutItem::isEmpty() const {
  return false;
  }

QSizePolicy::ExpandData BrowserGroupViewLayoutItem::expanding() const {
  return QSizePolicy::BothDirections;
  }

void BrowserGroupViewLayoutItem::setGeometry(const QRect& rect) {
  widget -> move(rect.x(),rect.y());
  widget -> resize(rect.width(),rect.height());
  }

QRect BrowserGroupViewLayoutItem::geometry() const {
  return QRect(widget -> x(),widget -> y(),
               widget -> width(),widget -> height());
  }

class BrowserGroupViewLayoutIterator : public QGLayoutIterator {
  public:
    BrowserGroupViewLayoutIterator(QList<QLayoutItem> *l)
                : idx(0), list(l)  {}

    QLayoutItem *current() {
      return idx < int(list->count()) ? list->at(idx) : 0;
      }

    QLayoutItem *next() {
      idx++; return current();
      }

    QLayoutItem *takeCurrent() {
      return idx < int(list -> count()) ? list->take( idx ) : 0;
      }

  private:
    int idx;
    QList<QLayoutItem> *list;
  };


BrowserGroupViewLayout::BrowserGroupViewLayout(QWidget *parent)
          : QLayout(parent) {

  layoutData = new BrowserGroupViewLayoutData();

  }

BrowserGroupViewLayout::~BrowserGroupViewLayout() {
  deleteAllItems();
  delete layoutData;
  }

QLayoutIterator BrowserGroupViewLayout::iterator() {
  return QLayoutIterator(new BrowserGroupViewLayoutIterator(
                                             &layoutData -> items));
  }

void BrowserGroupViewLayout::addItem(QLayoutItem *item) {

  layoutData -> items.append(item);

  //
  // Invalidate the previous preferred size, forcing a re-calculation
  //
  layoutData -> currentPreferredSize = QSize();

  ((BrowserGroupViewLayoutItem *)item) -> widget ->
                                            setSeparatorMargin(MinSepMargin);

  //
  // Cache some needed metrics (that must be common to all added items)
  //
  if (! layoutData -> separatorWidth) {
    layoutData -> separatorWidth =
          ((BrowserGroupViewLayoutItem *)item) -> separatorWidth();
    }

  layoutData -> lastColumns = -1;
  QTimer::singleShot(RedrawDelay,(BrowserGroupViewLayout *)this,SLOT(redraw()));
  }

void BrowserGroupViewLayout::visibleSizeChange(const QSize& size) {
  //
  // If the width changed, we will need to recalculate our preferred
  // size and re-do the layout (but only because the current algorithm
  // doesn't really take the allowed height into account. In theory, it
  // could see that the width is OK, but the height is not and then
  // elect to increase the width to some extent (with the goal of
  // occupying the smallest area)
  //
  if (size.width() != layoutData -> visibleSize.width()) {
    //
    // Force a recalculation of our preferred size (which will be able
    // to make the most of the available visible size)
    //
    layoutData -> currentPreferredSize = QSize();
    }

  //
  // Remember the size we can use
  //
  layoutData -> visibleSize = QSize(size);
  }

QSize BrowserGroupViewLayout::sizeHint() const {

  if (layoutData -> currentPreferredSize.isValid()) {
    return layoutData -> currentPreferredSize;
    }

  if (layoutData -> items.count() == 0) {
    layoutData -> currentPreferredSize = QSize(0,0);
    return layoutData -> currentPreferredSize;
    }

  //
  // Right here is an opportunity to see if the items would like to
  // resize themselves based on the current visible width (i.e. the
  // width goal has changed) TBD
  //

  //
  // Calculate the height needed for this width
  //
  int height = heightForWidth(layoutData -> visibleSize.width());

  //
  // Calculate the actual width required for this allowed width
  //
  int w = 0;
  for (int i = 0 ; i < layoutData -> calculatedColumns ; i++ ) {
    w += layoutData -> maxNameWidth[i] +
         2 * MinSepMargin + layoutData -> separatorWidth +
         layoutData -> maxValueWidth[i] + MinColSpacing;
    }
  if (w) w -= MinColSpacing;

  layoutData -> currentPreferredSize = QSize(w,height);

  //
  // Since we don't move and resize items immediately when they are
  // added to the layout, we use the new size hint calculation to
  // trigger a redraw if necessary (this way, we have a chance of
  // reducing the number of redraws to less than one per addItem)
  //
  QTimer::singleShot(RedrawDelay,(BrowserGroupViewLayout *)this,SLOT(redraw()));

  return layoutData -> currentPreferredSize;
  }

QSize BrowserGroupViewLayout::minimumSize() const {
  //
  // Our preferred and minimum size are currently the same
  //
  if (! layoutData -> currentPreferredSize.isValid()) {
    (void)sizeHint();
    }

  return layoutData -> currentPreferredSize;
  }

void BrowserGroupViewLayout::calculateColumns(int columns) const {
  //
  // Clear the current idea of the maximum width of each column
  //
  memset(layoutData -> maxNameWidth,0,sizeof(int) * MaxColumns);
  memset(layoutData -> maxValueWidth,0,sizeof(int) * MaxColumns);

  //
  // Iterate through the items, widening the column that each is mapped to
  //
  int column = 0;
  int width;
  layoutData -> calculatedRows = 0;
  layoutData -> calculatedHeight = 0;
  int rowHeight = 0;
  BrowserGroupViewLayoutItem *item;
  QListIterator<QLayoutItem> colCountIterator(layoutData -> items);
  while ((item = (BrowserGroupViewLayoutItem *)colCountIterator.current())) {
    width = item -> nameWidth();
    if (width > layoutData -> maxNameWidth[column])
      layoutData -> maxNameWidth[column] = width;

    width = item -> valueWidth();
    if (width > layoutData -> maxValueWidth[column])
      layoutData -> maxValueWidth[column] = width;

    int itemHeight = item -> height();
    if (itemHeight > rowHeight) rowHeight = itemHeight;

    column++ ;
    if (column == columns) {
      layoutData -> calculatedRows++ ;
      layoutData -> calculatedHeight += rowHeight;
      rowHeight = 0;
      column = 0;
      }

    ++colCountIterator;
    }

  if (column != 0) {
    layoutData -> calculatedRows++ ;
    layoutData -> calculatedHeight += rowHeight;
    }

  layoutData ->
      calculatedHeight += MinRowSpacing * (layoutData -> calculatedRows - 1);
  }

void BrowserGroupViewLayout::calculateOptimumColumns(int width) const {

  layoutData -> calculatedColumns = 0;
  for (int i = MaxColumns ; i > 0 ; i-- ) {
    //
    // Calculate column item widths for this number of columns, with
    // a side-effect of calculating the number of rows
    //
    calculateColumns(i);

    //
    // Calculate the width of this set of columns
    //
    int w = 0;
    for (int j = 0 ; j < i ; j++ ) {
      w += layoutData -> maxNameWidth[j] +
           2 * MinSepMargin + layoutData -> separatorWidth +
           layoutData -> maxValueWidth[j] + MinColSpacing;
      }
    w -= MinColSpacing;

    //
    // See if this width is acceptable
    //
    if (w <= width) {
      layoutData -> calculatedColumns = i;
      break;
      }
    }
  }


void BrowserGroupViewLayout::setGeometry(const QRect& rect) {

  if (layoutData -> items.count() == 0) {
    return;
    }

  //
  // If it hasn't been done, calculate the arrangement we will use
  //
  if (! layoutData -> currentPreferredSize.isValid()) {
    (void)sizeHint();
    }

  QTimer::singleShot(RedrawDelay,this,SLOT(redraw()));
  }

void BrowserGroupViewLayout::redraw() {

  QRect rect = mainWidget() -> rect();

  //
  // If this is the same arrangement as last time, don't do anything
  // since we don't clip our output to the given rectangle. The only
  // thing that gets missed is the extension of the line colors to
  // the new edge if the width is increased (but we could do something
  // special for that case)
  //
  if ((layoutData -> lastColumns == layoutData -> calculatedColumns) &&
      (layoutData -> lastRows == layoutData -> calculatedRows)) {

    return;
    }

  layoutData -> lastColumns = layoutData -> calculatedColumns;
  layoutData -> lastRows = layoutData -> calculatedRows;

  mainWidget() -> hide();

  int x = rect.x();
  int leftX = x;
  int y = rect.y();
  int width = rect.width();
  QColor backgroundColor1(255,255,190);
  QColor backgroundColor2(255,255,255);
  BrowserGroupViewLayoutItem *item;
  QListIterator<QLayoutItem> it(layoutData -> items);
  for (int row = 0 ; row < layoutData -> calculatedRows ; row++ ) {
    //
    // Build a convenient array of items on this new line, calculating
    // the highest on the way
    //
    BrowserGroupViewLayoutItem *rowItems[MaxColumns];
    int highest = 0;
    int columnsOnRow = 0;
    for (int c = 0 ; c < layoutData -> calculatedColumns ; c++ ) {
      item = (BrowserGroupViewLayoutItem *)it.current();
      if (! item) break;

      rowItems[c] = item;

      int height = item -> height();
      if (height > highest) highest = height;

      columnsOnRow++ ;
      ++it;
      }

    //
    // Now resize and move each item to the right place on the line
    //
    for (int c = 0 ; c < columnsOnRow ; c++ ) {
      item = rowItems[c];

      if ((row & 1) == 0) {
        item -> setBackgroundColor(backgroundColor1);
        }
      else {
        item -> setBackgroundColor(backgroundColor2);
        }

      item -> move(x,y);

      item -> setWidths(layoutData -> maxNameWidth[c],
                        layoutData -> maxValueWidth[c]);
      if (c == (columnsOnRow - 1)) {
        item -> setSize(width - (x - rect.x()),highest);
        }
      else {
        item -> setSize(item -> width() + MinColSpacing,highest);
        }

      item -> widget -> show();

      x += item -> width();
      }

    x = leftX;
    y += highest + MinRowSpacing;
    }

  mainWidget() -> show();
  }

bool BrowserGroupViewLayout::hasHeightForWidth() const {
  return true;
  }

int BrowserGroupViewLayout::heightForWidth(int width) const {
  //
  // If there are no items, tell the caller we have no
  // suggestion for the height
  //
  if (! layoutData -> items.count()) return 0;

  //
  // Calculate the optimum number of columns for this width
  //
  calculateOptimumColumns(width);

  if (layoutData -> calculatedColumns < 1) {
    //
    // We can't even fit one column in the given width. This means
    // our parent will go with a horizontal scrollbar
    //
    layoutData -> calculatedColumns = 1;
    }

  return layoutData -> calculatedHeight;
  }

QLayout::ResizeMode BrowserGroupViewLayout::resizeMode() const {
  return FreeResize;
  }

QSizePolicy::ExpandData BrowserGroupViewLayout::expanding() const {
  return QSizePolicy::NoDirection;
  }
