/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>

#include "BrowserGroupViewWidget.hpp"
#include "BrowserGroupViewLayout.hpp"

#include <qabstractlayout.h>
#include <qlayout.h>
#include <qframe.h>
#include <qlabel.h>
#include <qlist.h>
#include <qscrollbar.h>
#include <qevent.h>

BrowserGroupViewWidget::BrowserGroupViewWidget(QWidget *parent) 
              : QScrollView(parent), updating(true) {

  //
  // Create the content frame on which added items will be drawn
  //
  contentFrame = new QFrame(viewport());

  //
  // sizeHint() is the perfect size, but we can shrink if necessary. But
  // there's no point in growing us beyond sizeHint()
  //
  setSizePolicy(QSizePolicy(QSizePolicy::Preferred,
                            QSizePolicy::Preferred,false));

  //
  // Resize the content frame to expand it to the available space
  // (unfortunately, QScrollView doesn't have a resize policy for
  // shrinking the content)
  //
  setResizePolicy(AutoOneFit);

  //
  // Arrange for the default background colour to be the same as for text
  // widgets
  //
  contentFrame -> setBackgroundMode(PaletteBase);

  //
  // Give the content frame our custom layout manager
  //
  contentLayout = new BrowserGroupViewLayout(contentFrame);

  //
  // Register the content frame as something to be managed
  //
  addChild(contentFrame);
  }

BrowserGroupViewWidget::~BrowserGroupViewWidget() {
  }

void BrowserGroupViewWidget::setUpdating(bool enabled) {
  updating = enabled;
  setUpdatesEnabled(enabled);
  }

void BrowserGroupViewWidget::add(const char *name, const char *value) {
  //
  // Create a new layout item and add it to the custom layout
  //
  contentLayout ->
       addItem(new BrowserGroupViewLayoutItem(contentFrame,name,value,
                                              parentWidget() -> width()));

  updateGeometry();
  }

void BrowserGroupViewWidget::viewportResizeEvent(QResizeEvent *e) {

  if ((e -> size().width() == e -> oldSize().width()) &&
      (e -> size().height() == e -> oldSize().height())) {
#ifdef TRACE_GEOMETRY
    printf("BrowserGroupViewWidget::viewportResizeEvent - No Change!\n");
    return;
#endif
    }

#ifdef TRACE_GEOMETRY
  printf("BrowserGroupViewWidget::viewportResizeEvent(%d x %d)\n",
         e -> size().width(),e -> size().height());
#endif

  //
  // The view port has been resized (typically by a user dragging an
  // edge or corner), so we take this opportunity to adjust the
  // size of the entire widget so that it just fits around the
  // content frame
  //

  //
  // Make sure the event is seen by the base class
  //
  QScrollView::viewportResizeEvent(e);

  //
  // If updates are not enabled, don't do any work here
  //
  if (! updating) return;

  //
  // Notify the content layout manager that the visible size has changed
  //
  contentLayout -> visibleSizeChange(e -> size());

  //
  // Get the size hint of the content frame layout, which should tell us
  // the current size of the content in the content frame
  //
  QSize layoutSizeHint = contentLayout -> sizeHint();

#ifdef TRACE_GEOMETRY
  printf("BrowserGroupViewWidget::viewportResizeEvent: content size changing to  %d x %d\n",layoutSizeHint.width(),layoutSizeHint.height());
#endif

  //
  // Resize the content widget to agree with its contents
  //
  if ((layoutSizeHint.width() != contentsWidth()) ||
      (layoutSizeHint.height() != contentsHeight())) {

#ifdef TRACE_GEOMETRY
  printf("BrowserGroupViewWidget::viewportResizeEvent: resizing contents\n");
#endif

    resizeContents(layoutSizeHint.width(),layoutSizeHint.height());

    //
    // Tell parent layout managers that we have a new size hint (this
    // should cause adjustments in the allocation of space to our
    // siblings)
    //
    updateGeometry();

#ifdef TRACE_GEOMETRY
    printf("BrowserGroupViewWidget::viewportResizeEvent: updated geometry\n");
#endif

    }
  else {
#ifdef TRACE_GEOMETRY
    printf("BrowserGroupViewWidget::viewportResizeEvent: no change in contents size\n");
#endif
    }
  }
