/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstring>

#include <classlib/Xml.hpp>
#include <classlib/ObjectList.hpp>

#include "BrowserException.hpp"
#include "BrowserReportFunctionParametersLoader.hpp"
#include "BrowserReportFunctionParameter.hpp"

BrowserReportFunctionParametersLoader::BrowserReportFunctionParametersLoader(
                                                      const Xml& xml)
                 : XmlLoader(xml), parameters(0) {
  }

BrowserReportFunctionParametersLoader::~BrowserReportFunctionParametersLoader(
                                                                    ) {
  delete parameters;
  }

void BrowserReportFunctionParametersLoader::load(XmlElement parametersElement) {

  parameters = new ObjectList();

  int childCount = xml.getElementChildCount(parametersElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement childElement = xml.getElementChild(parametersElement,i);
    const char *childElementName = xml.getElementName(childElement);
    if (std::strcmp(childElementName,"parameter") == 0) {
      const char *name = 0;
      const char *value = 0;
      const char *defaultValue = 0;
      const char *prompt = 0;
      BrowserReportFunctionParameter::Type
                  type = BrowserReportFunctionParameter::StringParameterType;
      int attributeCount = xml.getElementAttributeCount(childElement);
      for (int a = 0 ; a < attributeCount ; a++ ) {
        const char *attributeName = xml.getElementAttributeName(childElement,a);
        if (std::strcmp(attributeName,"name") == 0) {
          name = xml.getElementAttributeValue(childElement,a);
          }
        else if (std::strcmp(attributeName,"value") == 0) {
          value = xml.getElementAttributeValue(childElement,a);
          }
        else if (std::strcmp(attributeName,"default") == 0) {
          defaultValue = xml.getElementAttributeValue(childElement,a);
          }
        else if (std::strcmp(attributeName,"prompt") == 0) {
          prompt = xml.getElementAttributeValue(childElement,a);
          }
        else if (std::strcmp(attributeName,"type") == 0) {
          const char *typeName = xml.getElementAttributeValue(childElement,a);
          if (std::strcmp(typeName,"String") == 0) {
            type = BrowserReportFunctionParameter::StringParameterType;
            }
          else if (std::strcmp(typeName,"ObjectIdentifier") == 0) {
            type = BrowserReportFunctionParameter::
                                       ObjectIdentifierParameterType;
            }
          else if (std::strcmp(typeName,"Integer") == 0) {
            type = BrowserReportFunctionParameter::IntegerParameterType;
            }
          else if (std::strcmp(typeName,"TableName") == 0) {
            type = BrowserReportFunctionParameter::TableNameParameterType;
            }
          else {
            char diagnostic[128];
            std::sprintf(diagnostic,"Unrecognised parameter type '%s'",
                         typeName);
            THROW(new BrowserException("Error in report definition",
                                       diagnostic),);
            }
          }
        else {
          char diagnostic[128];
          std::sprintf(diagnostic,"Unrecognised parameter attribute '%s'",
                       attributeName);
          THROW(new BrowserException("Error in report definition",diagnostic),);
          }
        }

      if (! name) {
        char diagnostic[128];
        std::sprintf(diagnostic,"Missing parameter name attribute");
        THROW(new BrowserException("Error in report definition",diagnostic),);
        }

      if ((! prompt) && (! value)) {
        char diagnostic[128];
        std::sprintf(diagnostic,"Missing parameter prompt or value attribute");
        THROW(new BrowserException("Error in report definition",diagnostic),);
        }

      BrowserReportFunctionParameter *parameter =
                                  new BrowserReportFunctionParameter(name);
      if (value) parameter -> setValue(value);
      if (defaultValue) parameter -> setDefaultValue(defaultValue);
      parameter -> setPrompt(prompt);
      parameter -> setType(type);

      parameters -> add(parameter);
      }
    else {
      char diagnostic[128];
      std::sprintf(diagnostic,"Unrecognised element '%s' in parameters",
                   childElementName);
      THROW(new BrowserException("Error in report definition",diagnostic),);
      }
    }
  }

ObjectList *BrowserReportFunctionParametersLoader::takeParameterList() {

  ObjectList *returnedParameters = parameters;
  parameters = 0;
  return returnedParameters;
  }
