/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstring>

#include <classlib/ObjectList.hpp>
#include <classlib/Xml.hpp>

#include "BrowserReportFunctionTableWalkLoader.hpp"
#include "BrowserReportFunctionParametersLoader.hpp"

BrowserReportFunctionTableWalkLoader::BrowserReportFunctionTableWalkLoader(
                                                        const Xml& xml)
                 : XmlLoader(xml), function(0) {
  }

BrowserReportFunctionTableWalkLoader::~BrowserReportFunctionTableWalkLoader() {
  delete function;
  }

BrowserReportFunctionTableWalkColumn *
         BrowserReportFunctionTableWalkLoader::parseTableWalkColumn(
                                                    XmlElement columnElement) {

  int attributeCount = xml.getElementAttributeCount(columnElement);
  if (attributeCount < 2) {
    THROW(new BrowserException("Incorrect number of attributes",
                               "Table column must have either "
                               "2 or 3 attributes"),0);
    }

  if ((strcmp(xml.getElementAttributeName(columnElement,0),"name") != 0) ||
      (strcmp(xml.getElementAttributeName(columnElement,1),"object") != 0)) {
    THROW(new BrowserException("Incorrect attributes",
                               "Table column must have 'name' and "
                               "'object' attributes"),0);
    }

  if (attributeCount == 2) {
    return new BrowserReportFunctionTableWalkColumn(
                    xml.getElementAttributeValue(columnElement,0),
                    xml.getElementAttributeValue(columnElement,1));
    }
  else {
    if (strcmp(xml.getElementAttributeName(columnElement,2),"type") != 0) {
      THROW(new BrowserException("Incorrect attributes",
                                 "Table column third attribute "
                                 "must be 'type'"),0);
      }
    else {
      const char *typeName = xml.getElementAttributeValue(columnElement,2);

      BrowserReportFunctionTableWalkColumn::Type type;

      if (strcmp(typeName,"implicit") == 0)
        type = BrowserReportFunctionTableWalkColumn::Implicit;
      else if (strcmp(typeName,"control") == 0)
        type = BrowserReportFunctionTableWalkColumn::Control;
      else if (strcmp(typeName,"normal") == 0)
        type = BrowserReportFunctionTableWalkColumn::Normal;
      else {
        THROW(new BrowserException("Incorrect attribute value",
                                   "Table column third attribute "
                                   "must be a known type"),0);
        }

      return new BrowserReportFunctionTableWalkColumn(
                    xml.getElementAttributeValue(columnElement,0),
                    xml.getElementAttributeValue(columnElement,1),
                    type);
      }
    }
  }

ObjectList *BrowserReportFunctionTableWalkLoader::parseTableWalkColumnsXml(
                    XmlElement columnsElement) {

  ObjectList *columns = new ObjectList();

  int childCount = xml.getElementChildCount(columnsElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement element = xml.getElementChild(columnsElement,i);
    const char *elementName = xml.getElementName(element);
    if (std::strcmp(elementName,"column") == 0) {
      BrowserReportFunctionTableWalkColumn *column =
                     parseTableWalkColumn(element);
      if (exception) {
        delete columns;
        return 0;
        }

      columns -> add(column);
      }
    }

  return columns;
  }

void BrowserReportFunctionTableWalkLoader::load(XmlElement functionElement) {

  function = new BrowserReportFunctionTableWalk();

  int childCount = xml.getElementChildCount(functionElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement element = xml.getElementChild(functionElement,i);
    const char *elementName = xml.getElementName(element);
    if (std::strcmp(elementName,"columns") == 0) {
      ObjectList *columns = parseTableWalkColumnsXml(element);
      if (exception) {
        delete function;
        function = 0;
        }

      function -> setColumnList(columns);
      }
    else if (std::strcmp(elementName,"parameters") == 0) {
      BrowserReportFunctionParametersLoader loader(xml);
      loader.load(element);
      if (loader.exception) {
        RETHROW(loader.exception,);
        }

      function -> setParameterList(loader.takeParameterList());
      }
    else {
      char diagnostic[128];
      std::sprintf(diagnostic,
                   "Unrecognised element '%s' in TableWalk function",
                   elementName);
      THROW(new BrowserException("Error in report definition",diagnostic),);
      }
    }
  }

BrowserReportFunction *BrowserReportFunctionTableWalkLoader::takeFunction() {

  BrowserReportFunction *returnedFunction = (BrowserReportFunction *)function;
  function = 0;
  return returnedFunction;
  }
