/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstdlib>
#include <cerrno>
#include <cstring>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <qmessagebox.h>
#include <qstring.h>
#include <qdir.h>

#include <classlib/Xml.hpp>
#include <classlib/ObjectList.hpp>

#include "BrowserReportManager.hpp"

#include "BrowserResource.hpp"
#include "BrowserUtility.hpp"

#include "BrowserReportFunctionGroup.hpp"
#include "BrowserReportFunctionTableWalkLoader.hpp"
#include "BrowserReportFunctionWalkLoader.hpp"
#include "BrowserReportFunctionGraphLoader.hpp"

#include "MainFormImpl.hpp"

BrowserReportManager::BrowserReportManager(MainFormImpl *parent)
                : parent(parent) {
  }

BrowserReportManager::~BrowserReportManager() {
  }

BrowserReportFunctionGroupEntry *
           BrowserReportManager::parseGroupEntry(const Xml& xml,
                                                 XmlElement entryElement) {

  if (xml.getElementAttributeCount(entryElement) != 2) {
    THROW(new BrowserException("Incorrect number of attributes",
                               "Group entry must have exactly "
                               "2 attributes"),0);
    }

  if ((strcmp(xml.getElementAttributeName(entryElement,0),"name") != 0) ||
      (strcmp(xml.getElementAttributeName(entryElement,1),"object") != 0)) {
    THROW(new BrowserException("Incorrect attributes",
                               "Group entry must have 'name' and "
                               "'object' attributes"),0);
    }

  return new BrowserReportFunctionGroupEntry(
                  xml.getElementAttributeValue(entryElement,0),
                  xml.getElementAttributeValue(entryElement,1));
  }

ObjectList *
         BrowserReportManager::parseGroupEntryList(const Xml& xml,
                                                   XmlElement entriesElement) {

  ObjectList *entries = new ObjectList();

  int childCount = xml.getElementChildCount(entriesElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement element = xml.getElementChild(entriesElement,i);
    const char *elementName = xml.getElementName(element);
    if (strcmp(elementName,"entry") == 0) {
      BrowserReportFunctionGroupEntry *entry = parseGroupEntry(xml,element);
      if (exception) {
        delete entries;
        return 0;
        }

      entries -> add(entry);
      }
    }

  return entries;
  }

BrowserReportFunctionGroup *
      BrowserReportManager::parseGroupFunctionXml(const Xml& xml,
                                                  XmlElement functionElement) {

  BrowserReportFunctionGroup *function = new BrowserReportFunctionGroup();

  int childCount = xml.getElementChildCount(functionElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement element = xml.getElementChild(functionElement,i);
    const char *elementName = xml.getElementName(element);
    if (strcmp(elementName,"entries") == 0) {
      ObjectList *entries = parseGroupEntryList(xml,element);
      if (exception) {
        delete function;
        return 0;
        }

      function -> setEntryList(entries);
      }
    }

  return function;
  }

BrowserReportFunction *
        BrowserReportManager::parseFunctionXml(const Xml& xml,
                                               XmlElement functionElement) {

  if ((xml.getElementAttributeCount(functionElement) < 1) ||
      (strcmp(xml.getElementAttributeName(functionElement,0),"name") != 0)) {
    THROW(new BrowserException("Function does not have name attribute "
                               "as first attribute"),0);
    }

  const char *functionName = xml.getElementAttributeValue(functionElement,0);
  if (strcmp(functionName,"Group") == 0) {
    return parseGroupFunctionXml(xml,functionElement);
    }
  else if (strcmp(functionName,"TableWalk") == 0) {
    BrowserReportFunctionTableWalkLoader loader(xml);
    loader.load(functionElement);
    if (loader.exception) {
      RETHROW(loader.exception,0);
      }

    return loader.takeFunction();
    }
  else if (strcmp(functionName,"Walk") == 0) {
    BrowserReportFunctionWalkLoader loader(xml);
    loader.load(functionElement);
    if (loader.exception) {
      RETHROW(loader.exception,0);
      }

    return loader.takeFunction();
    }
  else if (strcmp(functionName,"Graph") == 0) {
    BrowserReportFunctionGraphLoader loader(xml);
    loader.load(functionElement);
    if (loader.exception) {
      RETHROW(loader.exception,0);
      }

    return loader.takeFunction();
    }
  else {
    THROW(new BrowserException("Unrecognised function name",functionName),0);
    }
  }

BrowserReportSection *
       BrowserReportManager::parseSectionXml(const Xml& xml,
                                             XmlElement sectionElement) {

  BrowserReportSection *section = new BrowserReportSection();

  int childCount = xml.getElementChildCount(sectionElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement element = xml.getElementChild(sectionElement,i);
    const char *elementName = xml.getElementName(element);
    if (strcmp(elementName,"function") == 0) {
      BrowserReportFunction *function = parseFunctionXml(xml,element);
      if (exception) {
        delete section;
        return 0;
        }

      section -> setFunction(function);
      }
    }

  return section;
  }

ObjectList *
      BrowserReportManager::parseSectionsXml(const Xml& xml,
                                             XmlElement sectionsElement) {

  ObjectList *sections = new ObjectList();

  int childCount = xml.getElementChildCount(sectionsElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement sectionElement = xml.getElementChild(sectionsElement,i);
    BrowserReportSection *section = parseSectionXml(xml,sectionElement);
    if (exception) {
      delete sections;
      return 0;
      }

    sections -> add(section);
    }

  return sections;
  }

BrowserReport *
      BrowserReportManager::parseReportXml(const Xml& xml,
                                           XmlElement reportElement) {

  if ((xml.getElementAttributeCount(reportElement) != 1) ||
      (strcmp(xml.getElementAttributeName(reportElement,0),"path") != 0)) {
    THROW(new BrowserException("Incorrect attributes",
                               "Sole report element attribute must "
                               "be 'path'"),0);
    }

  const char *path = xml.getElementAttributeValue(reportElement,0);

  BrowserReport *report = new BrowserReport(path);

  int childCount = xml.getElementChildCount(reportElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement element = xml.getElementChild(reportElement,i);
    const char *elementName = xml.getElementName(element);
    if (strcmp(elementName,"sections") == 0) {
      ObjectList *sections = parseSectionsXml(xml,element);
      if (exception) {
        delete report;
        return 0;
        }

      report -> setSectionList(sections);
      }
    }

  return report;
  }

ObjectList *
      BrowserReportManager::parseReportsXml(const Xml& xml,
                                            XmlElement reportsElement) {

  ObjectList *reports = new ObjectList();

  int childCount = xml.getElementChildCount(reportsElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement reportElement = xml.getElementChild(reportsElement,i);
    BrowserReport *report = parseReportXml(xml,reportElement);
    if (exception) {
      delete reports;
      return 0;
      }

    reports -> add(report);
    }

  return reports;
  }

void BrowserReportManager::parseXml(const Xml& xml) {
  XmlElement root = xml.getRootElement();
  const char *elementName = xml.getElementName(root);
  if (strcmp(elementName,BrowserResource::applicationName()) != 0) {
    THROW(new BrowserException("Incorrect top-level XML element name"),);
    }

  const char *version = xml.getElementAttributeValue(root,0);
  if (strcmp(version,"1") != 0) {
    THROW(new BrowserException("Unrecognised XML file version number"),);
    }

  int childCount = xml.getElementChildCount(root);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement element = xml.getElementChild(root,i);
    elementName = xml.getElementName(element);
    if (strcmp(elementName,"reports") == 0) {
      ObjectList *reports = parseReportsXml(xml,element);
      if (exception) return;

      if (parent -> reports) {
        parent -> reports -> add(reports);
        }
      else parent -> reports = reports;
      }
    }
  }

void BrowserReportManager::load(const char *path) {
  Xml xml;
  xml.loadXmlFromFile(path);
  if (xml.exception) {
    const char *name = std::strrchr(path,'/');
    if (name) name++ ;
    else name = path;
    QMessageBox::warning(parent,BrowserResource::applicationTitle(),
                         BrowserUtility::formatErrorText(
                               "A report configuration file is invalid.",
                               xml.exception -> getMessage(),0,name));
    return;
    }

  parseXml(xml);
  if (exception) {
    const char *name = std::strrchr(path,'/');
    if (name) name++ ;
    else name = path;
    QMessageBox::warning(parent,BrowserResource::applicationTitle(),
                         BrowserUtility::formatErrorText(
                             "A report configuration file has "
                             "invalid content.",exception -> getMessage(),
                             exception -> getDiagnostic(),name));
    return;
    }
  }

void BrowserReportManager::load() {
  //
  // We load all report-defining XML files in the shared reports directory
  //
  QString path = BrowserResource::getApplicationSharedDir() +
                 BrowserResource::reportsDirName() + "/";

  QDir directory(path,"*.xml *.XML");
  int fileCount = directory.count();
  for (int i = 0 ; i < fileCount ; i++ ) {
    load(path + directory[i]);
    }

  //
  // Plus one per-user reports file
  //
  struct stat status;
  path = BrowserResource::getApplicationDirUser() +
                 BrowserResource::reportsFileName();
  if (::stat(path,&status) != -1) load(path);
  }
