/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstring>

#include <qlistview.h>

#include <classlib/List.hpp>

#include "BrowserReportMenuTree.hpp"

#include "MainFormImpl.hpp"
#include "BrowserReport.hpp"
#include "BrowserReportListViewItem.hpp"

BrowserReportMenuTreeNode::BrowserReportMenuTreeNode() : item(0), children(0) {
  }

BrowserReportMenuTreeNode::~BrowserReportMenuTreeNode() {
  //
  // We don't delete the item since it is a widget owned by the Qt
  // widget hierarchy, which deletes itself
  //

  if (children) {
    ListIterator it = children -> iterator();
    while (it.hasNext()) delete (BrowserReportMenuTreeNode *)it.next();

    delete children;
    }
  }

BrowserReportMenuTree::BrowserReportMenuTree(MainFormImpl *parent)
            : parent(parent), root(0) {

  root = new BrowserReportMenuTreeNode();
  }

BrowserReportMenuTree::~BrowserReportMenuTree() {
  delete root;
  }

static BrowserReportMenuTreeNode *addNode(QListView *top,
                                          BrowserReportMenuTreeNode *parent,
                                          const char *title,
                                          const BrowserReport *report=0) {

  //
  // Iterate through the parent's children to see if a node
  // already exists with this title
  //
  if (parent -> children) {
    ListIterator it = parent -> children -> iterator();
    while (it.hasNext()) {
      BrowserReportMenuTreeNode *node = (BrowserReportMenuTreeNode *)it.next();
      if (strcmp(title,node -> item -> text(0)) == 0) {
        //
        // Found a match. If we are looking for an intermediate node,
        // we can stop right here
        //
        if (! report) return node;

        //
        // We are adding a leaf. If the existing node has no report
        // (i.e. it was previously created as an intermediate node),
        // we can give it the report
        //
        if (! node -> item -> getReport()) {
          node -> item -> setReport(report);
          return node;
          }

        //
        // The existing node has a report, and the same name as
        // the report we want to add. The easiest way to resolve
        // this is to replace the existing report with the new
        // one
        // 
        node -> item -> setReport(report);
        return node;
        }
      }
    }

  //
  // Create a new child (possibly with a modified title) as either
  // a top-level child of the list, or as a child of another node
  //
  BrowserReportMenuTreeNode *child = new BrowserReportMenuTreeNode();
  if (parent -> item) {
    child -> item = new BrowserReportListViewItem(parent -> item,title,report);
    }
  else {
    child -> item = new BrowserReportListViewItem(top,title,report);
    }

  //
  // Give the child to its parent
  //
  if (! parent -> children) parent -> children = new List();
  parent -> children -> add(child);

  return child;
  }

void BrowserReportMenuTree::add(const BrowserReport *report) {
  //
  // The report path defines the report's position in the menu structure
  //
  const char *path = report -> getPath();

  //
  // Remove leading separators from the path
  //
  while (*path == '/') path++ ;

  BrowserReportMenuTreeNode *currentNode = root;
  while (*path) {
    const char *delimPtr = std::strchr(path,'/');
    if (! delimPtr) {
      //
      // The remainder of the path the title of the report, and
      // is to be associated with a child of the current node.
      //
      currentNode = addNode(parent -> ReportList,currentNode,path,report);
      break;
      }

    //
    // There is a delimiter, so we need to find the child of the
    // current node with the indicated name or create it
    //
    int length = delimPtr - path;
    char intermediateName[length + 1];
    std::memcpy(intermediateName,path,length);
    intermediateName[length] = '\0';

    //
    // The intermediate name is the name of a node above the one to
    // contain the report. It needs to be created if it does not already
    // exist
    //
    currentNode = addNode(parent -> ReportList,currentNode,intermediateName);

    path = delimPtr + 1;
    }
  }

void BrowserReportMenuTree::add(ObjectList *reports) {
  ListIterator it = reports -> iterator();
  while (it.hasNext()) {
    BrowserReport *report = (BrowserReport *)it.next();
    add(report);
    }
  }

const BrowserReport *BrowserReportMenuTree::selectedReport() const {
  BrowserReportListViewItem *selectedItem =
          (BrowserReportListViewItem *)parent -> ReportList -> selectedItem();
  if (! selectedItem) return 0;

  return selectedItem -> getReport();
  }
