/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>

#include "BrowserSnmpSession.hpp"
#include "BrowserResource.hpp"
#include "BrowserSnmpProfile.hpp"
#include "BrowserSnmpTarget.hpp"

BrowserSnmpSession::BrowserSnmpSession() : session(0), v3MessageProcessor(0),
                                           status(0) {

  session = new Snmp(status);
  if (status != SNMP_CLASS_SUCCESS) {
    THROW(new BrowserSnmpSessionException(getMessage()),);
    }

  //
  // Attach an SNMPv3 message processor to the session. We use the
  // application and user names to form a unique engine ID (unique
  // within this administrative domain, which isn't very big for a
  // handheld). We should probably add the external network IP
  // address to distinguish between identical handhelds within a
  // network
  //
  char engineID[256];
  std::sprintf(engineID,"%s-%s",BrowserResource::applicationName(),
               BrowserResource::getUserName());
  int mpStatus;
  v3MessageProcessor = new v3MP(session,engineID,
                                BrowserResource::bootCount,mpStatus);
  }

BrowserSnmpSession::~BrowserSnmpSession() {
  delete session;
  delete v3MessageProcessor;
  }

Snmp *BrowserSnmpSession::getSession() const {
  return session;
  }

const char *BrowserSnmpSession::getMessage() const {
  return session -> error_msg(status);
  }

void BrowserSnmpSession::addUser(const BrowserSnmpProfile *profile) {

  if (profile -> getVersion() == BrowserSnmpProfile::v3) {
    long authProtocol = 0;
    switch (profile -> getAuthProtocol()) {
      case BrowserSnmpProfile::authNone:
        authProtocol = SNMPv3_usmNoAuthProtocol;
        break;
      case BrowserSnmpProfile::authHmacMd5:
        authProtocol = SNMPv3_usmHMACMD5AuthProtocol;
        break;
      case BrowserSnmpProfile::authHmacSha:
        authProtocol = SNMPv3_usmHMACSHAAuthProtocol;
        break;
      }

    long privProtocol = 0;
    switch (profile -> getPrivProtocol()) {
      case BrowserSnmpProfile::privNone:
        privProtocol = SNMPv3_usmNoPrivProtocol;
        break;
      case BrowserSnmpProfile::privDes:
        privProtocol = SNMPv3_usmDESPrivProtocol;
        break;
      case BrowserSnmpProfile::privAes128:
        privProtocol = SNMPv3_usmAES128PrivProtocol;
        break;
      case BrowserSnmpProfile::privAes192:
        privProtocol = SNMPv3_usmAES192PrivProtocol;
        break;
      case BrowserSnmpProfile::privAes256:
        privProtocol = SNMPv3_usmAES256PrivProtocol;
        break;
      }

    USM *usm = v3MessageProcessor -> get_usm();
    if (usm -> add_usm_user(profile -> getSecurityName(),
                            authProtocol,privProtocol,
                            profile -> getAuthPassword(),
                            profile -> getPrivPassword()) != SNMPv3_USM_OK) {
      THROW(new BrowserException("Failed to add USM user"),);
      }
    }
  }

Pdu *BrowserSnmpSession::createPdu(const BrowserSnmpTarget *target,
                                   int securityLevel) const {
  Pdu *pdu = new Pdu();

  if (target -> getTarget() -> get_type() == SnmpTarget::type_utarget) {
    if (securityLevel == -1) {
      securityLevel = SecurityLevel_noAuthNoPriv;
      if (target -> getProfile() -> getAuthProtocol() !=
                                         BrowserSnmpProfile::authNone) {
        if (target -> getProfile() -> getPrivProtocol() ==
                                         BrowserSnmpProfile::privNone)
          securityLevel = SecurityLevel_authNoPriv;
        else
          securityLevel = SecurityLevel_authPriv;
        }
      }

    pdu -> set_security_level(securityLevel);
    pdu -> set_context_name("");
    pdu -> set_context_engine_id("");
    }

  return pdu;
  }

