/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstring>

#include "BrowserSnmpTarget.hpp"
#include "BrowserSnmpSession.hpp"
#include "BrowserConfiguration.hpp"
#include "BrowserNode.hpp"
#include "BrowserSnmpProfile.hpp"

BrowserSnmpTarget::BrowserSnmpTarget(const char *addressString)
            : target(0), profile(0) {

  UdpAddress address(addressString);
  if (! address.valid()) {
    char message[64];
    std::sprintf(message,"Unknown address '%s'",addressString);
    THROW(new BrowserSnmpTargetException(message),);
    }

  address.set_port(161);

  target = new CTarget(address,"public","public");
  target -> set_version(version1);
  target -> set_retry(1);
  target -> set_timeout(100);
  }

BrowserSnmpTarget::BrowserSnmpTarget(
                       BrowserSnmpSession *session,
                       const BrowserNode *node,
                       const BrowserConfiguration *configuration)
             : target(0), profile(0) {
  //
  // If the node has an IP address, use that field to build the
  // address. Otherwise, use the name field
  //
  const char *ipAddress = node -> getIpAddress();
  if ((! ipAddress) || (! *ipAddress)) {
    ipAddress = node -> getName();
    }

  UdpAddress address(ipAddress);
  if (! address.valid()) {
    char message[64];
    std::sprintf(message,"Unknown address '%s'",ipAddress);
    THROW(new BrowserSnmpTargetException(message),);
    }

  //
  // Get the profile for this node
  //
  const char *profileName = node -> getProfile();
  ListIterator it = configuration -> profiles -> iterator();
  while (it.hasNext()) {
    profile = (BrowserSnmpProfile *)it.next();
    if (std::strcmp(profile -> getName(),profileName) == 0) break;
    }

  const char *readCommunity = profile -> getReadCommunity();
  if ((! readCommunity) || (! *readCommunity)) readCommunity = "public";
  const char *writeCommunity = profile -> getWriteCommunity();
  if ((! writeCommunity) || (! *writeCommunity)) writeCommunity = readCommunity;

  address.set_port(profile -> getPort());

  switch (profile -> getVersion()) {
    case BrowserSnmpProfile::v1:
      target = new CTarget(address,readCommunity,writeCommunity);
      target -> set_version(version1);
      break;
    case BrowserSnmpProfile::v2c:
      target = new CTarget(address,readCommunity,writeCommunity);
      target -> set_version(version2c);
      break;
    case BrowserSnmpProfile::v3:
      UTarget *usmTarget = new UTarget(address);
      usmTarget -> set_version(version3);
      usmTarget -> set_security_model(SecurityModel_USM);
      usmTarget -> set_security_name(profile -> getSecurityName());

      target = usmTarget;
      break;
    }

  target -> set_retry(profile -> getRetries());
  target -> set_timeout(profile -> getTimeout());

  session -> addUser(profile);
  if (session -> exception) RETHROW(session -> exception,);
  }

BrowserSnmpTarget::~BrowserSnmpTarget() {
  delete target;
  }

SnmpTarget *BrowserSnmpTarget::getTarget() const {
  return target;
  }

const BrowserSnmpProfile *BrowserSnmpTarget::getProfile() const {
  return profile;
  }
