/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>

#include <qlabel.h>
#include <qcombobox.h>
#include <qtoolbutton.h>
#include <qmessagebox.h>
#include <qlistview.h>
#include <qcheckbox.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <qstringlist.h>

#ifdef QWS
#include <qpe/fontdatabase.h>
const int FONT_SIZE_DIVISOR = 10; // Fix for Qt/E (is there better detection?)
#else
#include <qfontdatabase.h>
const int FONT_SIZE_DIVISOR = 1;
#endif

#include "BrowserResource.hpp"
#include "ConfigFormImpl.hpp"
#include "ProfileConfigFormImpl.hpp"
#include "BrowserProfileListViewItem.hpp"
#include "BrowserNode.hpp"

static void setSampleFont(QComboBox *fontFamilyComboBox,
                          QComboBox *fontSizeComboBox,
                          QLabel *sampleLabel) {
  QFont font(fontFamilyComboBox -> currentText(),
             fontSizeComboBox -> currentText().toInt());
  sampleLabel -> setFont(font);
  }

static void loadFontSizeConfiguration(QComboBox *familyComboBox,
                                      QComboBox *sizeComboBox,
                                      QLabel *sampleLabel,
                                      int preferedFontSize) {
#ifdef QWS
  FontDatabase fontDatabase;
#else
  QFontDatabase fontDatabase;
#endif

  sizeComboBox -> clear();
  QValueList<int> pointSizes =
                     fontDatabase.pointSizes(familyComboBox -> currentText());
  QValueList<int>::Iterator it;
  for (it = pointSizes.begin() ; it != pointSizes.end() ; ++it) {
    sizeComboBox -> insertItem(QString().setNum(*it / FONT_SIZE_DIVISOR));
    }

  int defaultSize = familyComboBox -> fontInfo().pointSize();
  int sizeCount = sizeComboBox -> count();
  for (int i = 0 ; i < sizeCount ; i++ ) {
    if (preferedFontSize != 0) {
      if (preferedFontSize == sizeComboBox -> text(i).toInt()) {
        sizeComboBox -> setCurrentItem(i);
        break;
        }
      }
    else if (defaultSize == sizeComboBox -> text(i).toInt()) {
      sizeComboBox -> setCurrentItem(i);
      break;
      }
    }

  setSampleFont(familyComboBox,sizeComboBox,sampleLabel);
  }

static void loadFontConfiguration(QCheckBox *checkBox,QFrame *fontFrame,
                                  QComboBox *familyComboBox,
                                  QComboBox *sizeComboBox,QLabel *sampleLabel,
                                  String *configuredFontFamily,
                                  int configuredFontSize) {

  //
  // First we load up the family list with the available font families
  //

#ifdef QWS
  FontDatabase fontDatabase;
#else
  QFontDatabase fontDatabase;
#endif

  QStringList families = fontDatabase.families();

  familyComboBox -> insertStringList(families);
  const char *configuredFamily = 0;
  if (configuredFontFamily) {
    configuredFamily = configuredFontFamily -> getChars();
    checkBox -> setChecked(true);
    fontFrame -> setEnabled(true);
    }
  else {
    checkBox -> setChecked(false);
    fontFrame -> setEnabled(false);
    }
  const char *defaultFamily = familyComboBox -> fontInfo().family();
  int familyCount = familyComboBox -> count();
  for (int i = 0 ; i < familyCount ; i++ ) {
    if (configuredFamily) {
      if (strcmp(configuredFamily,familyComboBox -> text(i)) == 0) {
        familyComboBox -> setCurrentItem(i);
        break;
        }
      }
    else if (defaultFamily &&
             (strcmp(defaultFamily,familyComboBox -> text(i)) == 0)) {
      familyComboBox -> setCurrentItem(i);
      break;
      }
    }

  //
  // Now we load the font sizes list with the available font sizes
  // for the current family
  //
  loadFontSizeConfiguration(familyComboBox,sizeComboBox,
                            sampleLabel,configuredFontSize);
  }

ConfigFormImpl::ConfigFormImpl(QWidget* parent,
                               const BrowserConfiguration *configuration)
    : ConfigForm(parent,0,true,0), newConfiguration(0) {

#ifndef QWS
  QFrame *buttonFrame = new QFrame(this);

  QHBoxLayout *buttonFrameLayout = new QHBoxLayout(buttonFrame);

  QPushButton *okButton = new QPushButton(buttonFrame);
  okButton -> setText("Ok");
  okButton -> adjustSize();
  okButton -> setDefault(true);

  QPushButton *cancelButton = new QPushButton(buttonFrame);
  cancelButton -> setText("Cancel");
  cancelButton -> adjustSize();

  buttonFrameLayout -> addStretch(1);
  buttonFrameLayout -> addWidget(okButton);
  buttonFrameLayout -> addWidget(cancelButton);

  layout() -> add(buttonFrame);

  connect(okButton,SIGNAL(clicked()),this,SLOT(accept()));
  connect(cancelButton,SIGNAL(clicked()),this,SLOT(reject()));
#endif

  //
  // Make a copy of the configuration. This copies everything,
  // even things which cannot be changed in the configuration
  // dialog (like the node list).
  //
  newConfiguration = new BrowserConfiguration(*configuration);

  //
  // Iterate through the profile list and add/create list view
  // entries for each profile
  //
  ListIterator it = newConfiguration -> profiles -> iterator();
  while (it.hasNext()) {
    BrowserSnmpProfile *profile = (BrowserSnmpProfile *)it.next();

    (void) new BrowserProfileListViewItem(ProfileList,profile);
    }

  IconLocationCheckBox -> setChecked(configuration -> preferSystemIcons);

  loadFontConfiguration(CustomTextFontCheckBox,CustomTextFontFrame,
                        TextFontFamilyComboBox,TextFontSizeComboBox,
                        TextFontSampleLabel,configuration -> textFontFamily,
                        configuration -> textFontSize);

  loadFontConfiguration(CustomApplicationFontCheckBox,
                        CustomApplicationFontFrame,
                        ApplicationFontFamilyComboBox,
                        ApplicationFontSizeComboBox,
                        ApplicationFontSampleLabel,
                        configuration -> applicationFontFamily,
                        configuration -> applicationFontSize);
  }

ConfigFormImpl::~ConfigFormImpl() {
  delete newConfiguration;

  // no need to delete child widgets, Qt does it all for us
  }

void ConfigFormImpl::applicationFontFamilyChangedSlot() {
  delete newConfiguration -> applicationFontFamily;
  newConfiguration -> applicationFontFamily =
               new String(ApplicationFontFamilyComboBox -> currentText());

  loadFontSizeConfiguration(ApplicationFontFamilyComboBox,
                            ApplicationFontSizeComboBox,
                            ApplicationFontSampleLabel,0);

  newConfiguration -> applicationFontSize =
                         ApplicationFontSizeComboBox -> currentText().toInt();
  }

void ConfigFormImpl::textFontFamilyChangedSlot() {
  delete newConfiguration -> textFontFamily;
  newConfiguration -> textFontFamily =
               new String(TextFontFamilyComboBox -> currentText());

  loadFontSizeConfiguration(TextFontFamilyComboBox,
                            TextFontSizeComboBox,
                            TextFontSampleLabel,0);

  newConfiguration -> textFontSize =
                          TextFontSizeComboBox -> currentText().toInt();
  }

void ConfigFormImpl::applicationFontSizeChangedSlot() {
  newConfiguration -> applicationFontSize =
                         ApplicationFontSizeComboBox -> currentText().toInt();

  setSampleFont(ApplicationFontFamilyComboBox,ApplicationFontSizeComboBox,
                ApplicationFontSampleLabel);
  }

void ConfigFormImpl::textFontSizeChangedSlot() {
  newConfiguration -> textFontSize =
                           TextFontSizeComboBox -> currentText().toInt();

  setSampleFont(TextFontFamilyComboBox,TextFontSizeComboBox,
                TextFontSampleLabel);
  }

void ConfigFormImpl::applicationFontCheckBoxChangedSlot() {
  if (CustomApplicationFontCheckBox -> isChecked()) {
    CustomApplicationFontFrame -> setEnabled(true);
    delete newConfiguration -> applicationFontFamily;
    newConfiguration -> applicationFontFamily =
               new String(ApplicationFontFamilyComboBox -> currentText());
    newConfiguration -> applicationFontSize =
                          ApplicationFontSizeComboBox -> currentText().toInt();

    setSampleFont(ApplicationFontFamilyComboBox,ApplicationFontSizeComboBox,
                  ApplicationFontSampleLabel);
    }
  else {
    CustomApplicationFontFrame -> setEnabled(false);
    delete newConfiguration -> applicationFontFamily;
    newConfiguration -> applicationFontFamily = 0;
    newConfiguration -> applicationFontSize = 0;
    }
  }

void ConfigFormImpl::customTextFontCheckBoxChangedSlot() {
  if (CustomTextFontCheckBox -> isChecked()) {
    CustomTextFontFrame -> setEnabled(true);
    delete newConfiguration -> textFontFamily;
    newConfiguration -> textFontFamily =
               new String(TextFontFamilyComboBox -> currentText());
    newConfiguration -> textFontSize =
                           TextFontSizeComboBox -> currentText().toInt();

    setSampleFont(TextFontFamilyComboBox,TextFontSizeComboBox,
                  TextFontSampleLabel);
    }
  else {
    CustomTextFontFrame -> setEnabled(false);
    delete newConfiguration -> textFontFamily;
    newConfiguration -> textFontFamily = 0;
    newConfiguration -> textFontSize = 0;
    }
  }

void ConfigFormImpl::enableProfileDeleteButton() {
  DeleteProfileButton -> setEnabled(true);
  }

void ConfigFormImpl::disableProfileDeleteButton() {
  DeleteProfileButton -> setEnabled(false);
  }

void ConfigFormImpl::enableProfileEditButton() {
  EditProfileButton -> setEnabled(true);
  }

void ConfigFormImpl::disableProfileEditButton() {
  EditProfileButton -> setEnabled(false);
  }

void ConfigFormImpl::iconLocationChangedSlot() {
  newConfiguration -> preferSystemIcons = IconLocationCheckBox -> isChecked();
  }

void ConfigFormImpl::editSelectedProfileSlot() {
  //
  // Get the profile for this selected item
  //
  BrowserProfileListViewItem *profileListViewItem =
             (BrowserProfileListViewItem *)ProfileList -> selectedItem();
  BrowserSnmpProfile *profile =
             (BrowserSnmpProfile *)profileListViewItem -> getProfile();

  //
  // Create the empty profile dialog
  //
  ProfileConfigFormImpl profileDialog(this);

  //
  // Let the empty profile dialog populate itself
  //
  profileDialog.setProfile(profile);

  //
  // Let the games commence and complete
  //
  if (BrowserResource::maximiseDialogs) profileDialog.showMaximized();
  else profileDialog.show();
  int code = profileDialog.exec(); // Cancel=0, OK=1

  //
  // If the changes are accepted, make the changes to the profile
  // in the profile list
  //
  if (code == 1) {
    //
    // We need to remember the old name of the profile in
    // case it changed with the edit
    //
    String oldProfileNameString(profile -> getName());
    const char *oldProfileName = oldProfileNameString.getChars();

    //
    // Get the new profile from the dialog
    //
    profileDialog.getProfile((BrowserSnmpProfile *)profile);

    //
    // Rename uses of the old name to make them use the new name
    //
    const char *newProfileName = profile -> getName();
    if (strcmp(newProfileName,oldProfileName) != 0) {
      ListIterator it = newConfiguration -> nodes -> iterator();
      while (it.hasNext()) {
        BrowserNode *node = (BrowserNode *)it.next();
        if (strcmp(node -> getProfile(),oldProfileName) == 0) {
          node -> setProfile(newProfileName);
          }
        }
      }
 
    //
    // Refresh the GUI
    //
    profileListViewItem -> refreshProfile();
    }
  }

void ConfigFormImpl::newProfileSlot() {
  ProfileConfigFormImpl profileDialog(this);
  if (BrowserResource::maximiseDialogs) profileDialog.showMaximized();
  else profileDialog.show();
  int code = profileDialog.exec(); // Cancel=0, OK=1
  if (code == 1) {
    //
    // We have a new profile, and it has been validated (including
    // duplicate name checking), so we add it to the displayed list
    //
    BrowserSnmpProfile *profile = profileDialog.takeProfile();
    newConfiguration -> profiles -> add(profile);
    (void) new BrowserProfileListViewItem(ProfileList,profile);
    }
  }

void ConfigFormImpl::profileSelectionChangedSlot() {
  //
  // If a profiles is selected, the toolbar delete button
  // can be enabled and the toolbar edit button can be enabled
  //
  if (ProfileList -> selectedItem()) {
    enableProfileDeleteButton();
    enableProfileEditButton();
    }
  else {
    disableProfileDeleteButton();
    disableProfileEditButton();
    }
  }

void ConfigFormImpl::deleteSelectedProfileSlot() {
  //
  // Confirm the deletion
  //
  int code = QMessageBox::warning(this,
                                  BrowserResource::applicationTitle(),
                                  "Please confirm deletion\n"
                                  "of the selected profile.",
                                  QMessageBox::Ok,
                                  QMessageBox::Cancel |
                                        QMessageBox::Default |
                                        QMessageBox::Escape);
  if (code == QMessageBox::Cancel) {
    return;
    }

  BrowserProfileListViewItem *profileListViewItem =
             (BrowserProfileListViewItem *)ProfileList -> selectedItem();

  BrowserSnmpProfile *profile =
              (BrowserSnmpProfile *)profileListViewItem -> getProfile();

  //
  // Find out if the profile is in use by a node definition and
  // disallow the deletion
  //
  const char *profileName = profile -> getName();
  ListIterator nodeIterator = newConfiguration -> nodes -> iterator();
  bool found = false;
  while (nodeIterator.hasNext()) {
    BrowserNode *node = (BrowserNode *)nodeIterator.next();
    if (strcmp(node -> getProfile(),profileName) == 0) {
      found = true;
      break;
      }
    }
  if (found) {
    QString string = "Profile '";
    string = string + profileName + "'\n"
                     "is in use by one or\n"
                     "more nodes.\n"
                     "Not deleting it.";

    QMessageBox::warning(this,BrowserResource::applicationTitle(),
                         string,QMessageBox::Ok,0);
    return;
    }

  //
  // Find the profile in the list and remove it
  //
  ListIterator it = newConfiguration -> profiles -> iterator();
  while (it.hasNext()) {
    if ((BrowserSnmpProfile *)it.next() == profile) {
      it.remove();
      delete profile;
      break;
      }
    }

  //
  // Delete the visible selected item
  //
  delete profileListViewItem;
  }

BrowserConfiguration *ConfigFormImpl::takeNewConfiguration() {
  BrowserConfiguration *returnedConfiguration = newConfiguration;
  newConfiguration = 0;
  return returnedConfiguration;
  }
