/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdlib>
#include <cstdio>
#include <cstring>

#include <sys/time.h>

#include <qstring.h>
#include <qtextview.h>
#include <qtabwidget.h>
#include <qlabel.h>
#include <qtoolbutton.h>
#include <qmessagebox.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qobjectlist.h>
#include <qapplication.h>
#include <qregexp.h>
#include <qtimer.h>

#ifdef USE_SPLITTER
#include <qsplitter.h>
#endif

#include <smidb/Smi.hpp>

#include "SnmpOperationFactory.hpp"

#include "BrowserConfigurationManager.hpp"
#include "BrowserReportManager.hpp"
#include "BrowserReportMenuTree.hpp"
#include "MainFormImpl.hpp"
#include "ConfigFormImpl.hpp"
#include "NodeConfigFormImpl.hpp"
#include "ObjectDetailFormImpl.hpp"
#include "SearchMIBFormImpl.hpp"
#include "HelpFormImpl.hpp"
#include "BrowserResource.hpp"
#include "BrowserUtility.hpp"
#include "BrowserNodeListViewItem.hpp"
#include "BrowserReportListViewItem.hpp"
#include "BrowserMIBTreeItem.hpp"
#include "BrowserReportFunctionParameter.hpp"
#include "FunctionParameterFormImpl.hpp"

static void setApplicationFont(QApplication *application,
                               BrowserConfiguration *configuration) {

  if (configuration -> applicationFontFamily) {
    QFont font = application -> font();
    font.setFamily(configuration -> applicationFontFamily -> getChars());
    if (configuration -> applicationFontSize != 0) {
      font.setPointSize(configuration -> applicationFontSize);
      }

    application -> setFont(font,true);
    }
  }

//
// Update the visible widget hierarchy with any changes to the
// configured font families and sizes. Note that we currently
// only do this for the application-wide font and not for the
// set of widgets that have a specific text font. To do that,
// we'd have to maintain a list of those widgets or traverse
// the widget hierarchy ourselves (starting at the Results tab)
//
void MainFormImpl::updateFonts() {
  //
  // Update all widgets in the application (at least, the ones
  // that don't have a specific font)
  //
  setApplicationFont(application,configuration);

  //
  // It's possible the user has requested to return to the default font
  //
  if (! configuration -> applicationFontFamily) {
    application -> setFont(systemDefaultFont,true);
    }
  }

MainFormImpl::MainFormImpl(QApplication *application)
    : MainForm(),
      configuration(0), reports(0), reportMenuTree(0),
      application(application), smi(0), session(0), target(0),
      operationList(0), operationTimer(0), searchMIBForm(0) {

  //
  // Do some setup that wasn't possible in the designer
  //
  connect(TabCollection,SIGNAL(currentChanged(QWidget *)),
          this,SLOT(tabSwitchedSlot()));

  //
  // If the application close button is pushed, tell running operations to
  // stop (which will then allow the application to complete its
  // shutdown)
  //
  application -> connect(application,SIGNAL(lastWindowClosed()),
                         this,SLOT(cancelCurrentOperationSlot()));

  //
  // Remember the default font so we can go back to it
  //
  systemDefaultFont = application -> font();

  //
  // Read the browser configuration from persistent storage
  //
  readConfiguration();

  //
  // Set the default font for the application
  //
  setApplicationFont(application,configuration);

  //
  // Create an empty list of running operations
  //
  operationList = new ObjectList();

  //
  // Create a timer that will cause operations to be run
  //
  operationTimer = new QTimer(this);
  connect(operationTimer,SIGNAL(timeout()),this,SLOT(checkOperationSlot()));

  //
  // Set the initial values for the multi-tab icons
  //

  //
  // A new node can only be created if there is at least one profile
  //
  ListIterator it = configuration -> profiles -> iterator();
  iconStates[TabIdNodes][IconIdNew] = it.hasNext();
  iconStates[TabIdNodes][IconIdInfo] = false;
  iconStates[TabIdNodes][IconIdEdit] = false;
  iconStates[TabIdNodes][IconIdDelete] = false;
  iconStates[TabIdNodes][IconIdSearch] = false;

  iconStates[TabIdMIB][IconIdNew] = false;
  iconStates[TabIdMIB][IconIdInfo] = false;
  iconStates[TabIdMIB][IconIdEdit] = false;
  iconStates[TabIdMIB][IconIdDelete] = false;
  iconStates[TabIdMIB][IconIdSearch] = true;

  iconStates[TabIdReports][IconIdNew] = false;
  iconStates[TabIdReports][IconIdInfo] = false;
  iconStates[TabIdReports][IconIdEdit] = false;
  iconStates[TabIdReports][IconIdDelete] = false;
  iconStates[TabIdReports][IconIdSearch] = false;

  iconStates[TabIdResults][IconIdNew] = false;
  iconStates[TabIdResults][IconIdInfo] = false;
  iconStates[TabIdResults][IconIdEdit] = false;
  iconStates[TabIdResults][IconIdDelete] = false;
  iconStates[TabIdResults][IconIdSearch] = false;

  setIconStates();

  rebuildNodeList();

  reportMenuTree = new BrowserReportMenuTree(this);

  if (reports) {
    reportMenuTree -> add(reports);
    }

  setStatusText("Opening database");
  QString databasePath = BrowserResource::getApplicationGroupDirUser() +
                         BrowserResource::databaseFileName();
  smi = new Smi(databasePath);
  if (smi -> exception) {
    QMessageBox::critical(0,BrowserResource::applicationTitle(),
                          BrowserUtility::formatErrorText(
                                  "Failed to open the SMI database.",
                                  smi -> exception -> getMessage(),
                                  smi -> exception -> getDiagnostic(),
                                  "Aborting."),QMessageBox::Ok,0);
    std::exit(1);
    }
  else {
    setStatusText("Building MIB tree");

    //
    // Sort using the OID column
    //
    Tree -> setSorting(2);

    //
    // Reduce the offset between parent and child nodes
    //
    Tree -> setTreeStepSize(10);

    rebuildMIBTree();
    }
  }

MainFormImpl::~MainFormImpl() {
  delete smi;
  delete session;
  delete target;
  delete operationList;
  delete operationTimer;
  delete configuration;
  delete reports;
  delete reportMenuTree;
  delete searchMIBForm;
  }

void MainFormImpl::rebuildMIBTree(BrowserMIBTreeItem *parentItem,
                                  const SmiOidTreeNode *parentNode) {
  ObjectList *children = parentNode -> getChildren();
  if (! children) return;

  ListIterator it = children -> iterator();
  while (it.hasNext()) {
    const SmiOidTreeNode *node = (SmiOidTreeNode *)it.next();

    const char *nodeName = smi -> getOidTree() -> getName(node);
    const SmiOid oid(node);
    const String *oidString = oid.toString();

    if (nodeName) {
      char moduleName[64];
      const char *dotPtr = std::strchr(nodeName,'.');
      memcpy(moduleName,nodeName,dotPtr - nodeName);
      moduleName[dotPtr - nodeName] = '\0';
      nodeName = dotPtr + 1;
      BrowserMIBTreeItem *item =
                new BrowserMIBTreeItem(parentItem,moduleName,nodeName,
                                       oidString -> getChars(),node);
      rebuildMIBTree(item,node);
      }
    else {
      rebuildMIBTree(parentItem,node);
      }

    delete oidString;
    }
  }

void MainFormImpl::rebuildMIBTree(const SmiOidTreeNode *parentNode) {
  ObjectList *children = parentNode -> getChildren();
  if (! children) return;

  ListIterator it = children -> iterator();
  while (it.hasNext()) {
    const SmiOidTreeNode *node = (SmiOidTreeNode *)it.next();

    const char *nodeName = smi -> getOidTree() -> getName(node);
    const SmiOid oid(node);
    const String *oidString = oid.toString();

    if (nodeName) {
      char moduleName[64];
      const char *dotPtr = std::strchr(nodeName,'.');
      std::memcpy(moduleName,nodeName,dotPtr - nodeName);
      moduleName[dotPtr - nodeName] = '\0';
      nodeName = dotPtr + 1;
      BrowserMIBTreeItem *item =
                  new BrowserMIBTreeItem(Tree,moduleName,nodeName,
                                         oidString -> getChars(),node);
      rebuildMIBTree(item,node);
      }
    else {
      rebuildMIBTree(node);
      }

    delete oidString;
    }
  }

void MainFormImpl::rebuildMIBTree() {
  Tree -> clear();
  const SmiOidTree *oidTree = smi -> getOidTree();
  const SmiOidTreeNode *root = oidTree -> getRootNode();
  rebuildMIBTree(root);
  }

void MainFormImpl::rebuildNodeList() {
  NodeList -> clear();
  ListIterator it = configuration -> nodes -> iterator();
  while (it.hasNext()) {
    (void) new BrowserNodeListViewItem(NodeList,(BrowserNode *)it.next());
    }

  //
  // Rebuilding the node list cleared the current selection, without
  // resulting in a signal, so we invoke the handler manually (Qt 2.3.x)
  //
  nodeSelectionChangedSlot();
  }

void MainFormImpl::setIconStates() {
  //
  // If a report is running, only the Cancel button is enabled
  //
  if (operationList -> count()) {
    CancelButton -> setEnabled(true);
    ConfigureButton -> setEnabled(false);

    NewButton -> setEnabled(false);
    InfoButton -> setEnabled(false);
    EditButton -> setEnabled(false);
    DeleteButton -> setEnabled(false);
    SearchButton -> setEnabled(false);

    GoButton -> setEnabled(false);
    }
  else {
    CancelButton -> setEnabled(false);
    ConfigureButton -> setEnabled(true);

    int currentTab = TabCollection -> currentPageIndex();
    NewButton -> setEnabled(iconStates[currentTab][IconIdNew]);
    InfoButton -> setEnabled(iconStates[currentTab][IconIdInfo]);
    EditButton -> setEnabled(iconStates[currentTab][IconIdEdit]);
    DeleteButton -> setEnabled(iconStates[currentTab][IconIdDelete]);
    SearchButton -> setEnabled(iconStates[currentTab][IconIdSearch]);

    //
    // The Go button is conditional, based on what is selected
    //
    if (iconStates[TabIdNodes][IconIdEdit] &&
        reportMenuTree -> selectedReport()) {
      GoButton -> setEnabled(true);
      }
    else {
      GoButton -> setEnabled(false);
      }
    }
  }

void MainFormImpl::clearStatusText() {
  setStatusText("Ready");
  }

void MainFormImpl::setStatusText(const char *text) {
  QString string(text);
  Status -> setText(string);
  application -> processEvents();
  }

void MainFormImpl::switchToResultsPage() {
  TabCollection -> showPage(Results);
  }

void MainFormImpl::executeSelectedOperationSlot() {
  //
  // Delete old children of the Results widget
  //
  QObjectList *children = Results -> queryList(0,0,true,false);
  QObjectListIt childIterator(*children);
  QObject *child;
  while ((child = childIterator.current()) != 0) {
    ++childIterator;
    delete child;
    }

  //
  // Get the selected node
  //
  const BrowserNode *node =
        ((BrowserNodeListViewItem *)NodeList -> selectedItem()) -> getNode();

  if (! session) {
    setStatusText("Creating SNMP session");
    session = new BrowserSnmpSession();
    if (session -> exception) {
      QMessageBox::warning(this,BrowserResource::applicationTitle(),
                           session -> exception -> getMessage());

      setStatusText("Aborted");

      //
      // Delete the session, since it is not usable
      //
      delete session;
      session = 0;

      return;
      }
    }

  if (! target) {
    setStatusText("Creating SNMP target");
    target = new BrowserSnmpTarget(session,node,configuration);
    if (target -> exception) {
      QMessageBox::warning(this,BrowserResource::applicationTitle(),
                           target -> exception -> getMessage());
      setStatusText("Aborted");

      delete session;
      session = 0;

      delete target;
      target = 0;

      return;
      }
    }

  //
  // Get the selected report
  //
  const BrowserReport *report = reportMenuTree -> selectedReport();

  //
  // No sections means no report!
  //
  ObjectList *sections = report -> getSectionList();
  if (! sections) {
    clearStatusText();
    return;
    }

  setStatusText("Building report parameters..");

  const char *defaultObjectIdentifier = 0;
  const char *defaultTableName = 0;
  char fullName[128];
  if (Tree -> selectedItem()) {
    defaultObjectIdentifier = Tree -> selectedItem() -> text(0);

    const char *moduleName = Tree -> selectedItem() -> text(1);
    std::sprintf(fullName,"%s.%s",moduleName,defaultObjectIdentifier);
    const SmiOidTreeNode *node = smi -> getOidTree() -> getNode(fullName);
    if (node) {
      ObjectList *children = node -> getChildren();
      if (children) {
        SmiModule *module = smi -> getModuleList() -> getModule(moduleName);
        if (module) {
          const SmiAssignment *assignment =
                              module -> getAssignment(defaultObjectIdentifier);
          if (assignment &&
              (assignment -> getType() == SmiAssignmentInfo::OBJECT_TYPE)) {
            //
            // The selected node is an object with children, which is a
            // pretty good sign. At this point the node is either the Table
            // or the Entry. Either way, we want the Table to be used as
            // the default for this operation
            //
            SmiOidTreeNode *child =
                      (SmiOidTreeNode *)children -> iterator().next();
            if (child -> getChildren()) {
              //
              // The node is a grandparent object, so assume it is a table
              //
              defaultTableName = fullName;
              }
            else {
              //
              // The node is a grandchild-less parent object, so assume it
              // is an Entry and its parent is therefore a table
              //
              defaultTableName =
                        smi -> getOidTree() -> getName(node -> getParent());
              }
            }
          }
        }
      }
    }

  //
  // Prompt for any parameters that have no preset value
  //
  FunctionParameterFormImpl *parameterForm = 0;
  ListIterator sectionIt = sections -> iterator();
  while (sectionIt.hasNext()) {
    const BrowserReportSection *section =
                    (BrowserReportSection *)sectionIt.next();
    const BrowserReportFunction *function = section -> getFunction();

    ObjectList *parameters = function -> getParameterList();
    if (parameters) {
      ListIterator parameterIterator = parameters -> iterator();
      while (parameterIterator.hasNext()) {
        BrowserReportFunctionParameter *parameter =
                    (BrowserReportFunctionParameter *)parameterIterator.next();

        const char *value = parameter -> getValue();
        if (! value) {
          if (! parameterForm) {
            parameterForm = new FunctionParameterFormImpl(this);
            if (defaultObjectIdentifier) {
              parameterForm ->
                      setDefaultObjectIdentifier(defaultObjectIdentifier);
              }
            if (defaultTableName) {
              parameterForm -> setDefaultTableName(defaultTableName);
              }
            }

          parameterForm -> addParameter(parameter);
          }
        }
      }
    }

  //
  // If there are editable parameters, run the form
  //
  if (parameterForm) {
    if (BrowserResource::maximiseDialogs)
      parameterForm -> showMaximized();
    else
      parameterForm -> show();
    if (parameterForm -> exec() == 0) {
      //
      // Cancelled the form, so we cancel the report
      //
      delete parameterForm;
      clearStatusText();
      return;
      }
    }

  //
  // Now build a new parameter list with all the values present
  //
  ObjectList *parameterValues = new ObjectList();
  int row = 0;
  sectionIt = sections -> iterator();
  while (sectionIt.hasNext()) {
    const BrowserReportSection *section =
                    (BrowserReportSection *)sectionIt.next();
    const BrowserReportFunction *function = section -> getFunction();

    ObjectList *parameters = function -> getParameterList();
    if (parameters) {
      ListIterator parameterIterator = parameters -> iterator();
      while (parameterIterator.hasNext()) {
        BrowserReportFunctionParameter *parameter =
                    (BrowserReportFunctionParameter *)parameterIterator.next();

        BrowserReportFunctionParameter *valueParameter =
                     new BrowserReportFunctionParameter(parameter -> getName());

        const char *value = parameter -> getValue();
        if (value) {
          valueParameter -> setValue(value);
          }
        else {
          valueParameter -> setValue(parameterForm -> getParameterValue(row++));
          }

        parameterValues -> add(valueParameter);
        }
      }
    }

  //
  // Done with the parameter form
  //
  delete parameterForm;

  //
  // Create the (vertical) layout manager for the results of these
  // operations
  //
  QVBoxLayout *resultsLayout = new QVBoxLayout(Results);
  resultsLayout -> setSpacing(0);
  resultsLayout -> setMargin(1);
#ifdef USE_SPLITTER
  QSplitter *resultsSplitter = new QSplitter(Vertical,Results);
  resultsLayout -> addWidget(resultsSplitter);
  resultsSplitter -> show();
#endif

  //
  // Execute the sections of the report
  //
  sections = report -> getSectionList();
  bool first = true;
  sectionIt = sections -> iterator();
  while (sectionIt.hasNext()) {
    //
    // Each section specifies a function to be invoked (a function
    // generates data to be displayed)
    //
    const BrowserReportSection *section =
                    (BrowserReportSection *)sectionIt.next();
    const BrowserReportFunction *function = section -> getFunction();

    SnmpOperation *operation = SnmpOperationFactory::manufactureOperation(
                                            function -> getName());
    if (! operation) {
      QString message;
      message = message + "Operation '" + function -> getName() + "'\n"
                          "does not exist!";
      QMessageBox::warning(this,BrowserResource::applicationTitle(),
                           message);
      setStatusText("Aborted");

      setIconStates();
      return;
      }

    //
    // Add this new operation to the list of running operations
    //
    operationList -> add(operation);

    //
    // An operation is now in progress (we set operation)
    //
    if (first) {
      setIconStates();
      first = false;

      //
      // Show the (empty) results page.
      //
      switchToResultsPage();
      setStatusText("Running..");
      }

    //
    // Give the operation its working environment
    //
    operation -> setEnvironment(configuration,smi,
                                session,target,function,
                                parameterValues,0);

    //
    // Create the appropriate widget for displaying the results of this
    // operation
    //
#ifdef USE_SPLITTER
    QWidget *widget = operation -> createWidget(resultsSplitter);
    resultsSplitter -> setResizeMode(widget,QSplitter::FollowSizeHint);
#else
    QWidget *widget = operation -> createWidget(Results);
    resultsLayout -> addWidget(widget);
#endif
    widget -> show();

    //
    // Initialise the operation
    //
    operation -> initialise();
    }

  //
  // Arrange for some work to be done soon
  //
  operationTimer -> start(0,true);
  }

//
// Perform some work on each running operation
//
void MainFormImpl::checkOperationSlot() {

  //
  // Ensure there is a reasonable maximum time between checks of operations
  //
  int nextDelay = 3000;

  ListIterator operationIterator = operationList -> iterator();
  while (operationIterator.hasNext()) {
    SnmpOperation *operation = (SnmpOperation *)operationIterator.next();

    if (operation -> getState() == SnmpOperation::Running)
      operation -> doWork();

    SnmpOperation::State state = operation -> getState();
    switch (state) {
      case SnmpOperation::Running: {
        //
        // See if the operation has an opinion about the next time
        // to call its doWork() method
        //
        int delay = operation -> getSleepTime();
        if (delay < nextDelay) nextDelay = delay;
        break;
        }
      case SnmpOperation::Error:
        QMessageBox::warning(this,BrowserResource::applicationTitle(),
                             BrowserUtility::formatErrorText(
                                  "SNMP operation error.",
                                  operation -> exception -> getMessage(),
                                  operation -> exception -> getDiagnostic()));
        setStatusText("Aborted");
        break;
      case SnmpOperation::Cancelled:
        setStatusText("Stopped");
        break;
      case SnmpOperation::Finished:
        break;
      default:
        setStatusText("Unknown completion state");
        break;
      }

    //
    // If this operation is no longer running, remove it from the
    // list.
    //
    if (state != SnmpOperation::Running) {

      operationIterator.remove();
      delete operation;

      //
      // If this operation did not finish normally, cancel all other
      // operations. We also delete them here because we want the
      // status of this operation to be the final status
      //
      if (state != SnmpOperation::Finished) {
        ListIterator operationCancelIterator = operationList -> iterator();
        while (operationCancelIterator.hasNext()) {
          SnmpOperation *cancelledOperation =
                              (SnmpOperation *)operationCancelIterator.next();
          operationCancelIterator.remove();
          cancelledOperation -> cancel();
          delete cancelledOperation;
          }

        //
        // Don't visit the next operation, because there isn't one anymore
        // (we have invalidated the iterator)
        //
        break;
        }
      }
    }

  if (operationList -> count() == 0) {
    //
    // No longer running an operation, so update the icon states
    //
    clearStatusText();
    setIconStates();
    }
  else {
    //
    // An operation is still running, so schedule the next check
    //
    operationTimer -> start(nextDelay,true);
    }
  }

//
// Cancel the currently-running operation
//
void MainFormImpl::cancelCurrentOperationSlot() {
  //
  // Notify each running operation that it is being cancelled. It will notice
  // this the next time doWork() is called.
  //
  ListIterator operationIterator = operationList -> iterator();
  while (operationIterator.hasNext()) {
    ((SnmpOperation *)operationIterator.next()) -> cancel();
    }

  //
  // Advance the operation timer, so that the operations are checked early
  //
  operationTimer -> start(0,true);
  }

void MainFormImpl::openConfigFormSlot() {
  ConfigFormImpl configForm(this,configuration);

  if (BrowserResource::maximiseDialogs) configForm.showMaximized();
  else configForm.show();
  int code = configForm.exec();
  if (code == 1) {
    //
    // Get the updated configuration from the dialog
    //
    delete configuration;
    configuration = configForm.takeNewConfiguration();

    //
    // Update the configuration in persistent storage
    //
    writeConfiguration();

    //
    // Activate the new fonts
    //
    updateFonts();

    //
    // Rebuild the node list (the re-configuration made a new copy of
    // the node list)
    //
    rebuildNodeList();

    //
    // Clear the target cache
    //
    delete target;
    target = 0;

    //
    // If there is a profile, the new node button can be enabled
    //
    ListIterator it = configuration -> profiles -> iterator();
    iconStates[TabIdNodes][IconIdNew] = it.hasNext();

    setIconStates();
    }
  }

//
// The selection changed in the MIB tree tab, so we need to update 
// the state of the action buttons
//
void MainFormImpl::mibTreeSelectionChangedSlot() {
  iconStates[TabIdMIB][IconIdInfo] = (Tree -> selectedItem() != 0);

  setIconStates();
  }

//
// The selection changed in the reports tab, so we need to update 
// the state of the action buttons
//
void MainFormImpl::reportSelectionChangedSlot() {
  setIconStates();
  }

void MainFormImpl::nodeSelectionChangedSlot() {
  //
  // If a node is selected, the toolbar delete button
  // and the toolbar edit button can be enabled. This assumes
  // single selection mode for the node list.
  //
  if (NodeList -> selectedItem()) {
    iconStates[TabIdNodes][IconIdDelete] = true;
    iconStates[TabIdNodes][IconIdEdit] = true;
    }
  else {
    iconStates[TabIdNodes][IconIdDelete] = false;
    iconStates[TabIdNodes][IconIdEdit] = false;
    }

  //
  // Clear the target cache
  //
  delete target;
  target = 0;

  //
  // It may now be possible to run a query
  //
  setIconStates();
  }

void MainFormImpl::openNewNodeForm() {
  NodeConfigFormImpl nodeConfigForm(this,configuration);

  if (BrowserResource::maximiseDialogs) nodeConfigForm.showMaximized();
  else nodeConfigForm.show();
  int code = nodeConfigForm.exec();
  if (code == 1) {
    //
    // Get the updated node from the dialog
    //
    BrowserNode *node = nodeConfigForm.takeNode();
    configuration -> nodes -> add(node);

    (void) new BrowserNodeListViewItem(NodeList,node);

    //
    // Update the configuration in persistent storage
    //
    writeConfiguration();
    }
  }

bool MainFormImpl::matchMIBObject(const SmiOidTreeNode *node) const {

  const char *searchText = searchMIBForm -> getSearchText();

  QRegExp regExp(searchText,true,true);

  const char *fullName = smi -> getOidTree() -> getName(node);

  return regExp.match(std::strchr(fullName,'.') + 1) != -1;
  }

BrowserMIBTreeItem *MainFormImpl::findNextMIBMatchChild(
                                       BrowserMIBTreeItem *currentItem) const {

  BrowserMIBTreeItem *child = (BrowserMIBTreeItem *)currentItem -> firstChild();
  while (child) {
    //
    // Check for a match with this child
    //
    if (matchMIBObject(child -> getNode())) return child;

    //
    // No match, so continue with the children of this child before
    // moving to the next sibling of the child
    //
    BrowserMIBTreeItem *grandChild = findNextMIBMatchChild(child);
    if (grandChild) return grandChild;

    child = (BrowserMIBTreeItem *)child -> nextSibling();
    }

  return 0;
  }

BrowserMIBTreeItem *MainFormImpl::findNextMIBMatch(
                                      BrowserMIBTreeItem *currentItem) const {
  while (true) {
    //
    // The search first goes through all the children of the current node
    //
    BrowserMIBTreeItem *child = findNextMIBMatchChild(currentItem);
    if (child) return child;

    //
    // No match found in the children, so move on to the next sibling of
    // the current node
    //
    BrowserMIBTreeItem *nextCurrentItem =
                           (BrowserMIBTreeItem *)currentItem -> nextSibling();
    if (! nextCurrentItem) {
      //
      // Next up is our parent's next sibling
      //
      if (! currentItem -> parent()) return 0;
      currentItem =
           (BrowserMIBTreeItem *)currentItem -> parent() -> nextSibling();
      if (! currentItem) return 0;
      }
    else currentItem = nextCurrentItem;

    //
    // Check for a match on this next sibling
    //
    if (matchMIBObject(currentItem -> getNode())) return currentItem;
    }
  }

void MainFormImpl::openSearchMIBForm() {
  if (! searchMIBForm) searchMIBForm = new SearchMIBFormImpl(this);

  searchMIBForm -> show();

  if (searchMIBForm -> exec() == 1) {
    //
    // The search begins just after the currently selected item (or from
    // the beginning if no item is selected), so that the form can be
    // re-used iteratively very quickly. The search order is in 
    // OID tree order
    //
    BrowserMIBTreeItem *currentItem =
                           (BrowserMIBTreeItem *)Tree -> selectedItem();
    if (! currentItem) {
      currentItem = (BrowserMIBTreeItem *)Tree -> firstChild();
      if (! currentItem) {
        QMessageBox::warning(this,BrowserResource::applicationTitle(),
                             "No match (No items to search).");
        }
      }

    currentItem = findNextMIBMatch(currentItem);
    if (currentItem) {
      Tree -> setSelected(currentItem,true);
      Tree -> ensureItemVisible(currentItem);
      return;
      }

    //
    // We've exhausted the tree looking for a match, so tell the user
    //
    QMessageBox::warning(this,BrowserResource::applicationTitle(),
                         "No match.");
    }
  }

void MainFormImpl::openInfoOnSelectedObjectForm() {
  //
  // Get the selected MIB tree node (assumes single selection mode for the list)
  //
  const SmiOidTreeNode *node =
          ((BrowserMIBTreeItem *)Tree -> selectedItem()) -> getNode();

  ObjectDetailFormImpl detailForm(this,*smi,configuration,node);

  if (BrowserResource::maximiseDialogs) detailForm.showMaximized();
  else detailForm.show();

  (void)detailForm.exec();
  }

//
// Edit the selected node
//
void MainFormImpl::openEditSelectedNodeForm() {
  //
  // Get the selected node (assumes single selection mode for the list)
  //
  BrowserNodeListViewItem *nodeListViewItem =
               (BrowserNodeListViewItem *)NodeList -> selectedItem();
  BrowserNode *node = (BrowserNode *)nodeListViewItem -> getNode();

  //
  // Create the empty node dialog
  //
  NodeConfigFormImpl nodeDialog(this,configuration);

  //
  // Let the empty node dialog populate itself
  //
  nodeDialog.setNode(node);

  //
  // Let the games commence and complete
  //
  if (BrowserResource::maximiseDialogs) nodeDialog.showMaximized();
  else nodeDialog.show();
  int code = nodeDialog.exec(); // Cancel=0, OK=1

  //
  // If the changes are accepted, update the list
  //
  if (code == 1) {
    nodeDialog.getNode(node);
    nodeListViewItem -> refreshNode();

    //
    // Clear the target cache
    //
    delete target;
    target = 0;
    }

  //
  // Update the configuration in persistent storage
  //
  writeConfiguration();
  }

void MainFormImpl::deleteSelectedNode() {
  //
  // Confirm the deletion
  //
  int code = QMessageBox::warning(this,
                                  BrowserResource::applicationTitle(),
                                  "Please confirm deletion\n"
                                  "of the selected node.",
                                  QMessageBox::Ok,
                                  QMessageBox::Cancel |
                                        QMessageBox::Default |
                                        QMessageBox::Escape);
  if (code == QMessageBox::Cancel) {
    return;
    }

  //
  // Get the selected node (assumes single selection mode for the list)
  //
  BrowserNodeListViewItem *nodeListViewItem =
               (BrowserNodeListViewItem *)NodeList -> selectedItem();
  BrowserNode *node = (BrowserNode *)nodeListViewItem -> getNode();

  //
  // We need to find this node on the node list so we can remove it from
  // the list
  //
  ListIterator it = configuration -> nodes -> iterator();
  while (it.hasNext()) {
    if ((BrowserNode *)it.next() == node) {
      it.remove();
      delete node;
      break;
      }
    }

  //
  // Delete the visible selected node (which will remove it from the
  // visible list)
  //
  delete nodeListViewItem;

  //
  // Update the configuration in persistent storage
  //
  writeConfiguration();

  //
  // Clear the target cache
  //
  delete target;
  target = 0;
  }

void MainFormImpl::tabSwitchedSlot() {
  setIconStates();

  if (operationList -> count()) setStatusText("Running..");
  else {
    clearStatusText();
    }
  }

//
// Create a new object appropriate for the current tab. Currently, only
// the node list supports creation
//
void MainFormImpl::newSlot() {
  switch (TabCollection -> currentPageIndex()) {
    case TabIdNodes:
      openNewNodeForm();
      break;
    }
  }

void MainFormImpl::searchSlot() {
  switch (TabCollection -> currentPageIndex()) {
    case TabIdMIB:
      openSearchMIBForm();
      break;
    }
  }

//
// Get information appropriate for the selected item in the current tab.
// Currently, only the MIB tree supports this
//
void MainFormImpl::infoSlot() {
  switch (TabCollection -> currentPageIndex()) {
    case TabIdMIB:
      openInfoOnSelectedObjectForm();
      break;
    }
  }

//
// Edit the selected object in the current tab. Currently, only
// the node list supports editing
//
void MainFormImpl::editSlot() {
  switch (TabCollection -> currentPageIndex()) {
    case TabIdNodes:
      openEditSelectedNodeForm();
      break;
    }
  }

//
// Delete the selected object in the current tab. Currently, only
// the node list supports deletion
//
void MainFormImpl::deleteSlot() {
  switch (TabCollection -> currentPageIndex()) {
    case TabIdNodes:
      deleteSelectedNode();
      break;
    }
  }

//
// Show the application help dialog
//
void MainFormImpl::helpSlot() {
  HelpFormImpl helpForm(this);

  helpForm.VersionLabel -> setText(VERSION_STRING);

  if (BrowserResource::maximiseDialogs) helpForm.showMaximized();
  else helpForm.show();

  (void)helpForm.exec();
  }

void MainFormImpl::readConfiguration() {
  BrowserConfigurationManager configurationManager(this);
  configurationManager.load();

  BrowserReportManager reportManager(this);
  reportManager.load();
  }

void MainFormImpl::writeConfiguration() {
  BrowserConfigurationManager configurationManager(this);
  setStatusText("Saving..");
  configurationManager.save();
  clearStatusText();
  }
