/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cctype>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <qcombobox.h>
#include <qmessagebox.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qlineedit.h>
#include <qvalidator.h>

#include "NodeConfigFormImpl.hpp"

#include "BrowserSnmpProfile.hpp"
#include "BrowserResource.hpp"

class NameValidator : public QValidator {
  public:
    NameValidator(QWidget *parent) : QValidator(parent) {
      }

    State validate(QString& value,int& pos) const {
      const char *chars = value;
      while (*chars) {
        if ((! isalnum(*chars)) && (strchr("_-.",*chars) == 0)) {
          return Invalid;
          }

        chars++ ;
        }

      return Acceptable;
      }
  };

class IpAddressValidator : public QValidator {
  public:
    IpAddressValidator(QWidget *parent) : QValidator(parent) {
      }

    State validate(QString& value,int& pos) const {
      const char *chars = value;
      bool numeric = isdigit(*chars);
      while (*chars) {
        if (numeric) {
          if (! isdigit(*chars) && (*chars != '.')) return Invalid;
          }
        else {
          if ((! isalnum(*chars)) && (strchr("_-.",*chars) == 0)) {
            return Invalid;
            }
          }

        chars++ ;
        }

      return Acceptable;
      }
  };

NodeConfigFormImpl::NodeConfigFormImpl(QWidget *parent,
                                       BrowserConfiguration *configuration)
              : NodeConfigForm(parent,0,true,0),
                configuration(configuration),originalNode(0) {
#ifndef QWS
  QFrame *buttonFrame = new QFrame(this);

  QHBoxLayout *buttonFrameLayout = new QHBoxLayout(buttonFrame);

  QPushButton *okButton = new QPushButton(buttonFrame);
  okButton -> setText("Ok");
  okButton -> adjustSize();
  okButton -> setDefault(true);

  QPushButton *cancelButton = new QPushButton(buttonFrame);
  cancelButton -> setText("Cancel");
  cancelButton -> adjustSize();

  buttonFrameLayout -> addStretch(1);
  buttonFrameLayout -> addWidget(okButton);
  buttonFrameLayout -> addWidget(cancelButton);

  layout() -> add(buttonFrame);

  connect(okButton,SIGNAL(clicked()),this,SLOT(accept()));
  connect(cancelButton,SIGNAL(clicked()),this,SLOT(reject()));
#endif

  //
  // Populate the profile list from the configuration
  //
  ListIterator it = configuration -> profiles -> iterator();
  while (it.hasNext()) {
    const BrowserSnmpProfile *profile = (BrowserSnmpProfile *)it.next();
    NodeProfileCombo -> insertItem(profile -> getName());
    }

  //
  // Add validators for the name and IP Address fields.
  //
  NodeNameText -> setValidator(new NameValidator(this));
  NodeIPAddressText -> setValidator(new IpAddressValidator(this));
  }

NodeConfigFormImpl::~NodeConfigFormImpl() {
  // no need to delete child widgets, Qt does it all for us
  }

void NodeConfigFormImpl::accept() {
  //
  // Validate the form contents (at form level. Field-level
  // validation has already been done)
  //

  //
  // Disallow the form if both name and IP address fields are empty
  //
  const char *name = NodeNameText -> text();
  const char *ipAddress = NodeIPAddressText -> text();
  if ((! *name) && (! *ipAddress)) {
    QMessageBox::warning(this,BrowserResource::applicationTitle(),
                         "One or both of Name and\n"
                         "IP Address must be\n"
                         "provided.");
    return;
    }

  //
  // Normalise the IP address (checking it for complete validity)
  //
  if (isdigit(*ipAddress)) {
    struct in_addr in;
    if (inet_aton(ipAddress,&in) == 0) {
      QMessageBox::warning(this,BrowserResource::applicationTitle(),
                           "The IP address is not\n"
                           "valid!");
      return;
      }
    else ipAddress = inet_ntoa(in);
    }

  //
  // If a new node is being created, or an existing node is
  // being edited, then we need to look for duplicates. For
  // nodes, the tuple {name,ip,profile} must be unique
  //
  const char *profileName = NodeProfileCombo -> currentText();
  ListIterator it = configuration -> nodes -> iterator();
  while (it.hasNext()) {
    const BrowserNode *node = (BrowserNode *)it.next();

    if (originalNode != node) {
      if ((strcmp(node -> getName(),name) == 0) &&
          (strcmp(node -> getIpAddress(),ipAddress) == 0) &&
          (strcmp(node -> getProfile(),profileName) == 0)) {

        QMessageBox::warning(this,BrowserResource::applicationTitle(),
                             "Duplicate node! The\n"
                             "combination of Name,\n"
                             "IP Address and Profile\n"
                             "must be unique.");
        return;
        }
      }
    }

  QDialog::accept();
  }

void NodeConfigFormImpl::setNode(const BrowserNode *node) {
  originalNode = node;

  NodeNameText -> setText(node -> getName());
  NodeIPAddressText -> setText(node -> getIpAddress());

  const char *profileName = node -> getProfile();
  int profileCount = NodeProfileCombo -> count();
  for (int i = 0 ; i < profileCount ; i++ ) {
    if (strcmp(NodeProfileCombo -> text(i),profileName) == 0) {
      NodeProfileCombo -> setCurrentItem(i);
      break;
      }
    }
  }

BrowserNode *NodeConfigFormImpl::takeNode() const {
  BrowserNode *node = new BrowserNode();

  getNode(node);

  return node;
  }

void NodeConfigFormImpl::getNode(BrowserNode *node) const {
  node -> setName(NodeNameText -> text());
  node -> setIpAddress(NodeIPAddressText -> text());
  node -> setProfile(NodeProfileCombo -> currentText());
  }
