/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstring>

#include <classlib/StringBuffer.hpp>

#include <smidb/SmiOid.hpp>
#include <smidb/SmiModule.hpp>
#include <smidb/SmiIndexItem.hpp>
#include <smidb/SmiObjectType.hpp>
#include <smidb/SmiObjectIdentity.hpp>
#include <smidb/SmiNotificationType.hpp>
//#include <smidb/SmiModuleIdentity.hpp>
//#include <smidb/SmiModuleCompliance.hpp>
//#include <smidb/SmiAgentCapabilities.hpp>

#include <qlabel.h>
#include <qtextbrowser.h>
#include <qtabwidget.h>
#include <qlistview.h>
#include <qlistbox.h>
#include <qcheckbox.h>
#include <qgroupbox.h>

#include "ObjectDetailFormImpl.hpp"

static void loadAssignment(ObjectDetailFormImpl& form,
                           const Smi& smi,const SmiModule *module,
                           const SmiAssignment *assignment);

static void disableNotificationTab(ObjectDetailFormImpl& form) {
  delete form.Notification;
  form.Notification = 0;
  }

static void disableIndexTab(ObjectDetailFormImpl& form) {
  delete form.Index;
  form.Index = 0;
  }

static void disableDescriptionTab(ObjectDetailFormImpl& form) {
  delete form.Description;
  form.Description = 0;
  }

static void disableSyntaxTab(ObjectDetailFormImpl& form) {
  delete form.Syntax;
  form.Syntax = 0;
  }

static void disableAllTabs(ObjectDetailFormImpl& form) {
  disableSyntaxTab(form);
  disableNotificationTab(form);
  disableIndexTab(form);
  disableDescriptionTab(form);
  }

ObjectDetailFormImpl::ObjectDetailFormImpl(QWidget* parent,
                                           const Smi& smi,
                                           BrowserConfiguration *configuration,
                                           const SmiOidTreeNode *node)
         : ObjectDetailForm(parent,0,true,0) {

  if (configuration -> textFontFamily) {
    QFont font = DescriptionText -> font();
    font.setFamily(configuration -> textFontFamily -> getChars());
    if (configuration -> textFontSize != 0) {
      font.setPointSize(configuration -> textFontSize);
      }

    DescriptionText -> setFont(font);
    }

  const char *objectName = smi.getOidTree() -> getName(node);
  char moduleName[64];
  const char *dotPtr = std::strchr(objectName,'.');
  memcpy(moduleName,objectName,dotPtr - objectName);
  moduleName[dotPtr - objectName] = '\0';
  objectName = dotPtr + 1;

  ObjectNameLabel -> setText(objectName);
  setCaption(caption() + " - " + objectName);
  ModuleNameLabel -> setText(moduleName);

  SmiOid oid(node);
  String *oidString = oid.toString();
  OIDLabel -> setText(oidString -> getChars());
  delete oidString;

  if (*moduleName) {
    const SmiModule *module = smi.getModuleList() -> getModule(moduleName);
    if (! module) {
      ObjectTypeLabel -> setText("Unknown");
      disableAllTabs(*this);
      return;
      }

    const SmiAssignment *assignment = module -> getAssignment(objectName);
    if (! assignment) {
      ObjectTypeLabel -> setText("Unknown");
      disableAllTabs(*this);
      return;
      }

    loadAssignment(*this,smi,module,assignment);
    }
  else {
    //
    // This should be the iso object identifier (no module)
    //
    ObjectTypeLabel -> setText("Object Identifier");
    disableAllTabs(*this);
    }
  }

ObjectDetailFormImpl::~ObjectDetailFormImpl() {
  // no need to delete child widgets, Qt does it all for us
  }

//
// Convert raw text from the original SMI clauses into formatted
// rich text for display in a text browser
//
static void setFormattedText(QTextBrowser *textBrowser,
                             const char *rawText) {
  StringBuffer buffer;
  buffer.append(rawText);

  char *text = buffer.detach();
  textBrowser -> setText(text);

  StringBuffer::free(text);
  }

static void loadSyntax(ObjectDetailFormImpl& form,
                       const Smi& smi,const SmiModule *module,
                       const SmiSyntax *syntax) {
  QString type = syntax -> getTyperef();
  const char *subtype = syntax -> getSubtype();
  if (subtype) type.append(" ").append(subtype);
  form.TypeRefLabel -> setText(type);

  SmiSizeList *sizes = syntax -> getSizeList();
  if (sizes) {
    ListIterator it = sizes -> iterator();
    bool first = true;
    QString value;
    while (it.hasNext()) {
      SmiSize *size = (SmiSize *)it.next();
      if (! first) value.append(", ");
      else first = false;

      char scratch[128];
      std::sprintf(scratch,"%lld",size -> getMin());
      value.append(scratch);
      if (size -> getMin() != size -> getMax()) {
        value.append("..");
        std::sprintf(scratch,"%lld",size -> getMax());
        value.append(scratch);
        }
      }

    form.SizeLabel -> setText(value);
    }
  else delete form.SizeFrame;

  SmiRangeList *ranges = syntax -> getRangeList();
  if (ranges) {
    ListIterator it = ranges -> iterator();
    bool first = true;
    QString value;
    while (it.hasNext()) {
      SmiRange *range = (SmiRange *)it.next();
      if (! first) value.append(", ");
      else first = false;

      char scratch[128];
      std::sprintf(scratch,"%lld",range -> getMin());
      value.append(scratch);
      if (range -> getMin() != range -> getMax()) {
        value.append("..");
        std::sprintf(scratch,"%lld",range -> getMax());
        value.append(scratch);
        }
      }

    form.RangeLabel -> setText(value);
    }
  else delete form.RangeFrame;

  SmiEnumerationList *enumerations = syntax -> getEnumerationList();
  if (enumerations) {
    form.EnumerationList -> setSorting(1);
    ListIterator it = enumerations -> iterator();
    while (it.hasNext()) {
      SmiEnumeration *enumeration = (SmiEnumeration *)it.next();
      QListViewItem *item = new QListViewItem(form.EnumerationList);
      item -> setText(0,enumeration -> getName());
      item -> setText(1,QString::number(enumeration -> getNumber()));
      }
    }
  else delete form.EnumerationsFrame;
  }

static void loadAssignment(ObjectDetailFormImpl& form,
                           const Smi& smi,const SmiModule *module,
                           const SmiObjectIdentity *assignment) {
  //
  // Enable the description tab only if there is a description (object
  // identity includes OID assignments, which don't have a description)
  //
  const char *description = assignment -> getDescription();
  if (description && *description) {
    form.ObjectTypeLabel -> setText("Object Identity");
    setFormattedText(form.DescriptionText,assignment -> getDescription());

    const char *reference = assignment -> getReference();
    if (reference && *reference) {
      QString referenceText = "\nReference:\n";
      referenceText += assignment -> getReference();
      form.DescriptionText -> append(referenceText);
      }
    }
  else {
    form.ObjectTypeLabel -> setText("Object Identifier");
    disableDescriptionTab(form);
    }

  disableSyntaxTab(form);
  disableIndexTab(form);
  disableNotificationTab(form);
  }

static void loadAssignment(ObjectDetailFormImpl& form,
                           const Smi& smi,const SmiModule *module,
                           const SmiNotificationType *assignment) {

  form.ObjectTypeLabel -> setText("Notification");

  setFormattedText(form.DescriptionText,assignment -> getDescription());

  const char *reference = assignment -> getReference();
  if (reference && *reference) {
    QString referenceText = "\nReference:\n";
    referenceText += assignment -> getReference();
    form.DescriptionText -> append(referenceText);
    }

  ObjectList *objects = assignment -> getObjectList();
  if (objects) {
    //
    // Disable sorting for this list, since the objects
    // are ordered by their position in the notification
    //
    form.NotificationObjectList -> setSorting(-1);

    QListViewItem *previous = 0;
    ListIterator it = objects -> iterator();
    while (it.hasNext()) {
      SmiValueReference *object = (SmiValueReference *)it.next();
      QListViewItem *listItem =
               new QListViewItem(form.NotificationObjectList,
                                 previous,object -> getName());
      if (object -> getModule()) listItem -> setText(1,object -> getModule());

      previous = listItem;
      }
    }
  else disableNotificationTab(form); // !!

  disableIndexTab(form);
  disableSyntaxTab(form);
  }

static void loadAssignment(ObjectDetailFormImpl& form,
                           const Smi& smi,const SmiModule *module,
                           const SmiObjectType *assignment) {

  form.ObjectTypeLabel -> setText("Object Type");

  setFormattedText(form.DescriptionText,assignment -> getDescription());

  const char *reference = assignment -> getReference();
  if (reference && *reference) {
    QString referenceText = "\nReference:\n";
    referenceText += assignment -> getReference();
    form.DescriptionText -> append(referenceText);
    }

  loadSyntax(form,smi,module,assignment -> getSyntax());

  const char *units = assignment -> getUnits();
  if (units && *units) form.UnitsLabel -> setText(units);
  else delete form.UnitsFrame;

  const SmiValueReference *augments = assignment -> getAugments();
  if (augments) {
    form.AugmentsNameLabel -> setText(augments -> getName());
    if (augments -> getModule())
      form.AugmentsModuleLabel -> setText(augments -> getModule());
    else form.AugmentsModuleFrame -> setEnabled(false);
    }
  else delete form.AugmentsFrame;

  SmiIndexList *indices = assignment -> getIndexList();
  if (indices) {
    //
    // Disable sorting for this list, since the indices
    // are ordered by their position in OIDs
    //
    form.IndexList -> setSorting(-1);

    QListViewItem *previous = 0;
    ListIterator it = indices -> iterator();
    while (it.hasNext()) {
      SmiIndexItem *index = (SmiIndexItem *)it.next();
      QListViewItem *listItem =
                new QListViewItem(form.IndexList,previous,index -> getName());
      if (index -> getModule()) listItem -> setText(1,index -> getModule());

      previous = listItem;
      }

    form.ImpliedIndexCheckBox -> setChecked(indices -> getImplied());
    }
  else delete form.IndexListFrame;

  if ((! augments) && (! indices)) disableIndexTab(form);
  disableNotificationTab(form);
  }

static void loadAssignment(ObjectDetailFormImpl& form,
                           const Smi& smi,const SmiModule *module,
                           const SmiAssignment *assignment) {
  switch (assignment -> getType()) {
    case SmiAssignmentInfo::OBJECT_TYPE:
      loadAssignment(form,smi,module,(SmiObjectType *)assignment);
      break;
    case SmiAssignmentInfo::OBJECT_IDENTITY:
      loadAssignment(form,smi,module,(SmiObjectIdentity *)assignment);
      break;
    case SmiAssignmentInfo::NOTIFICATION_TYPE:
      loadAssignment(form,smi,module,(SmiNotificationType *)assignment);
      break;
    // Other cases TBD
    default:
      form.ObjectTypeLabel -> setText("Not handled yet");
      disableAllTabs(form);
      break;
    }
  }
