/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cctype>

#include <qstring.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qradiobutton.h>
#include <qspinbox.h>
#include <qvalidator.h>
#include <qtabwidget.h>
#include <qcombobox.h>

#include "BrowserResource.hpp"
#include "BrowserUtility.hpp"
#include "ConfigFormImpl.hpp"
#include "ProfileConfigFormImpl.hpp"
#include "BrowserProfileListViewItem.hpp"

class NameValidator : public QValidator {
  public:
    NameValidator(QWidget *parent) : QValidator(parent) {
      }

    State validate(QString& value,int& pos) const {
      const char *chars = value;
      while (*chars) {
        if ((! isalnum(*chars)) && (strchr("_-.",*chars) == 0)) {
          return Invalid;
          }

        chars++ ;
        }

      return Acceptable;
      }
  };

ProfileConfigFormImpl::ProfileConfigFormImpl(ConfigFormImpl *parent)
    : ProfileConfigForm(parent,0,true,0),
      configForm(parent), originalProfile(0) {

#ifndef QWS
  QFrame *buttonFrame = new QFrame(this);

  QHBoxLayout *buttonFrameLayout = new QHBoxLayout(buttonFrame);

  QPushButton *okButton = new QPushButton(buttonFrame);
  okButton -> setText("Ok");
  okButton -> adjustSize();
  okButton -> setDefault(true);

  QPushButton *cancelButton = new QPushButton(buttonFrame);
  cancelButton -> setText("Cancel");
  cancelButton -> adjustSize();

  buttonFrameLayout -> addStretch(1);
  buttonFrameLayout -> addWidget(okButton);
  buttonFrameLayout -> addWidget(cancelButton);

  layout() -> add(buttonFrame);

  connect(okButton,SIGNAL(clicked()),this,SLOT(accept()));
  connect(cancelButton,SIGNAL(clicked()),this,SLOT(reject()));
#endif

  ProfileNameText -> setValidator(new NameValidator(this));

  //
  // Use v1/v2c by default (must match default radio button selection)
  //
  selectedCommunityBasedVersionSlot();
  }

ProfileConfigFormImpl::~ProfileConfigFormImpl() {
  // no need to delete child widgets, Qt does it all for us
  }

void ProfileConfigFormImpl::accept() {
  //
  // Validate the form contents (at form level. Field-level
  // validation has already been done)
  //

  //
  // Disallow empty names (if we wanted a default profile, we'd make the
  // name 'default' special, or use a checkbox)
  //
  const char *text = ProfileNameText -> text();
  if (! *text) {
    QMessageBox::warning(this,BrowserResource::applicationTitle(),
                         BrowserUtility::formatErrorText(
                  "Missing profile name"));
    return;
    }

  if (SNMPv3Button -> isChecked()) {
    if (SecurityNameText -> text().isEmpty()) {
      QMessageBox::warning(this,BrowserResource::applicationTitle(),
                           BrowserUtility::formatErrorText(
                  "Missing security name"));
      return;
      }

    if ((AuthProtocolComboBox -> currentItem() != 0) &&
        AuthPasswordText -> text().isEmpty()) {
      QMessageBox::warning(this,BrowserResource::applicationTitle(),
                           BrowserUtility::formatErrorText(
                  "Missing authentication password"));
      return;
      }

    if ((PrivProtocolComboBox -> currentItem() != 0) &&
        PrivPasswordText -> text().isEmpty()) {
      QMessageBox::warning(this,BrowserResource::applicationTitle(),
                           BrowserUtility::formatErrorText(
                  "Missing privacy password"));
      return;
      }

    if ((AuthProtocolComboBox -> currentItem() == 0) &&
        (PrivProtocolComboBox -> currentItem() != 0)) {
      QMessageBox::warning(this,BrowserResource::applicationTitle(),
                           BrowserUtility::formatErrorText(
                  "Cannot use privacy without authentication"));
      return;
      }
    }

  //
  // If a new profile is being created, or an existing profile is
  // being edited and the name changed, then we need to look for
  // duplicates
  //
  if ((! originalProfile) ||
      (strcmp(originalProfile -> getName(),text) != 0)) {
    BrowserProfileListViewItem *profileListViewItem =
       (BrowserProfileListViewItem *)configForm -> ProfileList -> firstChild();
    while (profileListViewItem) {
      const BrowserSnmpProfile *profile = profileListViewItem -> getProfile();
      if (strcmp(profile -> getName(),text) == 0) {
        QMessageBox::warning(this,BrowserResource::applicationTitle(),
                             BrowserUtility::formatErrorText(
                   "Duplicate profile name"));
        return;
        }

      profileListViewItem =
             (BrowserProfileListViewItem *)profileListViewItem -> nextSibling();
      }
    }

  QDialog::accept();
  }

void ProfileConfigFormImpl::setProfile(const BrowserSnmpProfile *profile) {
  //
  // Remember that an existing profile is being edited. When the dialog
  // is finished, we use the original profile to determine whether
  // the profile name was changed (we're allowed to save the pointer
  // because this is a modal dialog)
  //
  originalProfile = profile;

  QString string(profile -> getName());
  ProfileNameText -> setText(string);

  TimeoutSpinBox -> setValue(profile -> getTimeout());
  RetriesSpinBox -> setValue(profile -> getRetries());
  PortSpinBox -> setValue(profile -> getPort());

  switch (profile -> getVersion()) {
    case BrowserSnmpProfile::v1:
      SNMPv1Button -> setChecked(true);
      selectedCommunityBasedVersionSlot();
      break;
    case BrowserSnmpProfile::v2c:
      SNMPv2cButton -> setChecked(true);
      selectedCommunityBasedVersionSlot();
      break;
    case BrowserSnmpProfile::v3:
      SNMPv3Button -> setChecked(true);
      selectedUSMBasedVersionSlot();
      break;
    }

  ReadCommunityText -> setText(profile -> getReadCommunity());
  WriteCommunityText -> setText(profile -> getWriteCommunity());

  SecurityNameText -> setText(profile -> getSecurityName());

  switch (profile -> getAuthProtocol()) {
    case BrowserSnmpProfile::authNone:
      AuthProtocolComboBox -> setCurrentItem(0);
      break;
    case BrowserSnmpProfile::authHmacMd5:
      AuthProtocolComboBox -> setCurrentItem(1);
      break;
    case BrowserSnmpProfile::authHmacSha:
      AuthProtocolComboBox -> setCurrentItem(2);
      break;
    }

  AuthPasswordText -> setText(profile -> getAuthPassword());

  switch (profile -> getPrivProtocol()) {
    case BrowserSnmpProfile::privNone:
      PrivProtocolComboBox -> setCurrentItem(0);
      break;
    case BrowserSnmpProfile::privDes:
      PrivProtocolComboBox -> setCurrentItem(1);
      break;
    case BrowserSnmpProfile::privAes128:
      PrivProtocolComboBox -> setCurrentItem(2);
      break;
    case BrowserSnmpProfile::privAes192:
      PrivProtocolComboBox -> setCurrentItem(3);
      break;
    case BrowserSnmpProfile::privAes256:
      PrivProtocolComboBox -> setCurrentItem(4);
      break;
    }

  PrivPasswordText -> setText(profile -> getPrivPassword());
  }

void ProfileConfigFormImpl::getProfile(BrowserSnmpProfile *profile) const {
  profile -> setName(ProfileNameText -> text());

  profile -> setTimeout(TimeoutSpinBox -> value());
  profile -> setRetries(RetriesSpinBox -> value());
  profile -> setPort(PortSpinBox -> value());

  if (SNMPv1Button -> isChecked())
    profile -> setVersion(BrowserSnmpProfile::v1);
  else if (SNMPv2cButton -> isChecked())
    profile -> setVersion(BrowserSnmpProfile::v2c);
  else if (SNMPv3Button -> isChecked())
    profile -> setVersion(BrowserSnmpProfile::v3);

  profile -> setReadCommunity(ReadCommunityText -> text());
  profile -> setWriteCommunity(WriteCommunityText -> text());

  profile -> setSecurityName(SecurityNameText -> text());

  switch (AuthProtocolComboBox -> currentItem()) {
    case 0:
      profile -> setAuthProtocol(BrowserSnmpProfile::authNone);
      break;
    case 1:
      profile -> setAuthProtocol(BrowserSnmpProfile::authHmacMd5);
      break;
    case 2:
      profile -> setAuthProtocol(BrowserSnmpProfile::authHmacSha);
      break;
    }

  profile -> setAuthPassword(AuthPasswordText -> text());

  switch (PrivProtocolComboBox -> currentItem()) {
    case 0:
      profile -> setPrivProtocol(BrowserSnmpProfile::privNone);
      break;
    case 1:
      profile -> setPrivProtocol(BrowserSnmpProfile::privDes);
      break;
    case 2:
      profile -> setPrivProtocol(BrowserSnmpProfile::privAes128);
      break;
    case 3:
      profile -> setPrivProtocol(BrowserSnmpProfile::privAes192);
      break;
    case 4:
      profile -> setPrivProtocol(BrowserSnmpProfile::privAes256);
      break;
    }

  profile -> setPrivPassword(PrivPasswordText -> text());
  }

BrowserSnmpProfile *ProfileConfigFormImpl::takeProfile() const {
  BrowserSnmpProfile *profile = new BrowserSnmpProfile();

  getProfile(profile);

  return profile;
  }

void ProfileConfigFormImpl::selectedCommunityBasedVersionSlot() {
  SNMPTabCollection -> showPage(CommunitiesTab);
  SNMPTabCollection -> setTabEnabled(CommunitiesTab,true);
  SNMPTabCollection -> setTabEnabled(SecurityTab,false);
  SNMPTabCollection -> setTabEnabled(AuthTab,false);
  SNMPTabCollection -> setTabEnabled(PrivTab,false);
  }

void ProfileConfigFormImpl::selectedUSMBasedVersionSlot() {
  SNMPTabCollection -> showPage(SecurityTab);
  SNMPTabCollection -> setTabEnabled(CommunitiesTab,false);
  SNMPTabCollection -> setTabEnabled(SecurityTab,true);
  SNMPTabCollection -> setTabEnabled(AuthTab,true);
  SNMPTabCollection -> setTabEnabled(PrivTab,true);
  }
