/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstring>
#include <cctype>
#include <cstdlib>
#include <unistd.h>
#include <stdlib.h>
#include <climits>
#include <sys/time.h>

#include <classlib/String.hpp>
#include <classlib/Object.hpp>

#include <smidb/SmiModule.hpp>
#include <smidb/SmiSyntax.hpp>
#include <smidb/SmiObjectType.hpp>
#include <smisnmp/SmiSnmpPPAdapter.hpp>

#include "SnmpGraphOperation.hpp"
#include "BrowserReportFunctionParameter.hpp"
#include "BrowserReportFunctionGraph.hpp"

#include "GraphFormImpl.hpp"

class EntryInfo : public Object {
  public:
    EntryInfo(const char *oidString,int traceIdArg,int samplePeriodArg)
         : traceId(traceIdArg), samplePeriod(samplePeriodArg),
           lastValueValid(false) {
      oid = new Oid(oidString);
      nextTime.tv_sec = 0;
      nextTime.tv_usec = 0;
      }

    ~EntryInfo() {
      delete oid;
      }

    Oid *oid;
    struct timeval nextTime;
    int traceId;
    int samplePeriod;
    bool sent;
    bool lastValueValid;
    unsigned long lastValue;
  };

class SnmpGraphOperationPrivate {
  public:
    SnmpGraphOperationPrivate() : pdu(0), entryInfoList(0) {
      entryInfoList = new ObjectList();
      }

    ~SnmpGraphOperationPrivate() {
      delete pdu;
      delete entryInfoList;
      }

    Pdu *pdu;
    ObjectList *entryInfoList;
  };

SnmpGraphOperation::SnmpGraphOperation() : data(0) {
  data = new SnmpGraphOperationPrivate();
  }

SnmpGraphOperation::~SnmpGraphOperation() {
  delete data;
  }

//
// Build a numeric OID string for the given name string
//
static String *findOidString(const Smi *smi,const char *name) {
  //
  // First check to see if we already have a digit string
  //
  if (isdigit(*name)) {
    return new String(name);
    }

  //
  // Not numeric, so we have a module, an object name and instance
  // values
  //
  const char *dotPtr = std::strchr(name,'.');
  if (! dotPtr) return 0;
  dotPtr = std::strchr(dotPtr + 1,'.');
  if (! dotPtr) return 0;
  int length = dotPtr - name;
  char nameBuffer[length + 1];
  std::memcpy(nameBuffer,name,length);
  nameBuffer[length] = '\0';

  //
  // We've collected what should be a fully-qualified object name
  //
  const SmiOidTreeNode *node = smi -> getOidTree() -> getNode(nameBuffer);
  if (! node) {
    return 0;
    }

  //
  // Look up the object name and get its numeric representation
  //
  SmiOid smiOid(node);
  String *oidString = smiOid.toString();
  StringBuffer result;
  result.append(oidString -> getChars());

  //
  // Now start appending instance values (if any)
  //
  const char *namePtr = dotPtr + 1;
  while (*namePtr) {
    if (isdigit(*namePtr)) {
      dotPtr = std::strchr(namePtr,'.');
      if (! dotPtr) {
        result.append(namePtr - 1);
        break;
        }

      length = dotPtr - namePtr;
      result.append(namePtr - 1,length + 1);
      namePtr = dotPtr + 1;
      }
    else return 0; // Can't cope with non-numeric indexes yet
    }

  return new String(result);
  }

QWidget *SnmpGraphOperation::createWidget(QWidget *parent) {
  widget = new GraphFormImpl(parent);

  return widget;
  }

void SnmpGraphOperation::initialise() {

  //
  // Collect the graphing parameters
  //
  int shiftPeriod = -1;
  if (parameterValues) {
    ListIterator it = parameterValues -> iterator();
    while (it.hasNext()) {
      BrowserReportFunctionParameter *parameter =
                           (BrowserReportFunctionParameter *)it.next();

      if ((strcmp(parameter -> getName(),"shiftTime") == 0) &&
          parameter -> getValue()) {
        shiftPeriod = std::atoi(parameter -> getValue());
        }
      }
    }

  //
  // We have a skeleton graph form, so we create the required
  // strip charts in it, plus the trace list
  //

  ObjectList *entries =
                 ((BrowserReportFunctionGraph *)function) -> getEntryList();
  if (! entries) {
    state = Finished;
    }
  else {
    int smallestSamplePeriod = INT_MAX;
    ListIterator entryIterator = entries -> iterator();
    while (entryIterator.hasNext()) {
      const BrowserReportFunctionGraphEntry *entry =
                   (BrowserReportFunctionGraphEntry *)entryIterator.next();

      int traceId = ((GraphFormImpl *)widget) -> addTrace(entry -> getName());

      String *oidString = findOidString(smi,entry -> getObjectName());
      if (! oidString) {
        state = Error;

        char scratch[256];
        std::sprintf(scratch,"OID not found for object %s",
                     entry -> getObjectName());
        THROW(new BrowserException(scratch),);
        }

      EntryInfo *entryInfo = new EntryInfo(oidString -> getChars(),traceId,
                                           entry -> getSamplePeriod());
      data -> entryInfoList -> add(entryInfo);
      delete oidString;

      if ((shiftPeriod == -1) &&
          (entry -> getSamplePeriod() < smallestSamplePeriod)) {
        smallestSamplePeriod = entry -> getSamplePeriod();
        }
      }

    if (shiftPeriod == -1) shiftPeriod = smallestSamplePeriod;
    ((GraphFormImpl *)widget) -> setShiftPeriod(shiftPeriod);

    state = Running;
    }
  }

void SnmpGraphOperation::doWork() {
  //
  // See if it is time to do something.
  //
  struct timeval now;
  struct timezone tz;
  gettimeofday(&now,&tz);

  SnmpTarget& snmpTarget = *target -> getTarget();
  Snmp& snmpSession = *session -> getSession();

  delete data -> pdu;
  data -> pdu = 0;

  int variableCount = 0;
  ListIterator entryInfoIterator = data -> entryInfoList -> iterator();
  while (entryInfoIterator.hasNext()) {
    EntryInfo *entryInfo = (EntryInfo *)entryInfoIterator.next();

    if ((now.tv_sec < entryInfo -> nextTime.tv_sec) ||
        ((now.tv_sec == entryInfo -> nextTime.tv_sec) &&
         (now.tv_usec < entryInfo -> nextTime.tv_usec))) {
      entryInfo -> sent = false;
      continue;
      }

    Vb vb;
    vb.set_oid(*entryInfo -> oid);

    if (! data -> pdu) {
      data -> pdu = session -> createPdu(target);
      }

    *data -> pdu += vb;

    entryInfo -> sent = true;

    variableCount++ ;
    }

  //
  // If nothing was ready for re-sampling, return immediately. The
  // caller is supposed to query us for a suitable sleep time.
  //
  if (! variableCount) {
    return;
    }

  int status = snmpSession.get(*data -> pdu,snmpTarget);
  if (status != SNMP_CLASS_SUCCESS) {
    state = Error;
    THROW(new BrowserException(snmpSession.error_msg(status)),);
    }

  //
  // See if this is an SNMPv3 report message complaining about the
  // PDU
  //
  if (data -> pdu -> get_type() == REPORT_MSG) {
    state = Error;
    processReportMessage(data -> pdu);
    return;
    }

  if (data -> pdu -> get_vb_count() != variableCount) {
    state = Error;
    THROW(new BrowserException("Target returned wrong number of variables!"),);
    }

  int variable = 0;
  entryInfoIterator = data -> entryInfoList -> iterator();
  while (entryInfoIterator.hasNext()) {
    EntryInfo *entryInfo = (EntryInfo *)entryInfoIterator.next();

    if (entryInfo -> sent) {
      const Vb& vb = data -> pdu -> get_vb(variable++);
      SmiUINT32 snmpSyntax = vb.get_syntax();
      switch (snmpSyntax) {
        case sNMP_SYNTAX_INT32: {
          long value;
          vb.get_value(value);

          ((GraphFormImpl *)widget) -> setValue(entryInfo -> traceId,value);
          break;
          }
        case sNMP_SYNTAX_UINT32: {
          unsigned long value;
          vb.get_value(value);

          ((GraphFormImpl *)widget) -> setValue(entryInfo -> traceId,value);
          break;
          }
        case sNMP_SYNTAX_CNTR32: {
          unsigned long value;
          vb.get_value(value);

          //
          // This is a counter, so it only makes sense to graph the
          // difference between consecutive samples (which requires
          // us to delay the trace by one sample time, although if
          // we were really smart, we'd figure this out at initialisation
          // and delay non-counter sampling by one period)
          //
          if (entryInfo -> lastValueValid) {
            ((GraphFormImpl *)widget) ->
                                setValue(entryInfo -> traceId,
                                         value - entryInfo -> lastValue);
            }
          else entryInfo -> lastValueValid = true;
          entryInfo -> lastValue = value;
          }
        case sNMP_SYNTAX_NOSUCHOBJECT:
        case sNMP_SYNTAX_NOSUCHINSTANCE:
          break;
        }

      int next_msec = now.tv_usec / 1000 + entryInfo -> samplePeriod;
      entryInfo -> nextTime.tv_sec = now.tv_sec + next_msec / 1000;
      entryInfo -> nextTime.tv_usec = (next_msec % 1000) * 1000;
      }
    }
  }

void SnmpGraphOperation::cancel() {
  ((GraphFormImpl *)widget) -> stop();
  SnmpOperation::cancel();
  }

int SnmpGraphOperation::getSleepTime() const {
  //
  // Find out the earliest next time for the traces
  //
  struct timeval nextTime = {0, 0};
  ListIterator entryInfoIterator = data -> entryInfoList -> iterator();
  while (entryInfoIterator.hasNext()) {
    EntryInfo *entryInfo = (EntryInfo *)entryInfoIterator.next();

    if (entryInfo -> nextTime.tv_sec != 0) {
      if (nextTime.tv_sec == 0) nextTime = entryInfo -> nextTime;
      else if ((entryInfo -> nextTime.tv_sec < nextTime.tv_sec) ||
               ((entryInfo -> nextTime.tv_sec == nextTime.tv_sec) &&
                (entryInfo -> nextTime.tv_usec < nextTime.tv_usec))) {
        nextTime = entryInfo -> nextTime;
        }
      }
    }

  //
  // The sleep time is the time from now to the earliest next time
  //
  struct timeval now;
  struct timezone tz;
  gettimeofday(&now,&tz);
  int ms = (nextTime.tv_sec - now.tv_sec) * 1000 +
           (nextTime.tv_usec - now.tv_usec) / 1000;
  if (ms < 0) ms = 0;

  return ms;
  }
