/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstring>
#include <cctype>

#include <classlib/String.hpp>

#include <smidb/SmiModule.hpp>
#include <smidb/SmiSyntax.hpp>
#include <smidb/SmiObjectType.hpp>
#include <smisnmp/SmiSnmpPPAdapter.hpp>

#include "SnmpGroupOperation.hpp"
#include "BrowserReportFunctionGroup.hpp"

#include "BrowserGroupViewWidget.hpp"

const int OBJECTS_PER_PDU = 16;

SnmpGroupOperation::SnmpGroupOperation() : entryIterator(0), pdu(0) {
  }

SnmpGroupOperation::~SnmpGroupOperation() {
  delete pdu;
  }

//
// Build a numeric OID string for the given name string
//
static String *findOidString(const Smi *smi,const char *name) {
  //
  // First check to see if we already have a digit string
  //
  if (isdigit(*name)) {
    return new String(name);
    }

  //
  // Not numeric, so we have a module, an object name and instance
  // values
  //
  const char *dotPtr = std::strchr(name,'.');
  if (! dotPtr) return 0;
  dotPtr = std::strchr(dotPtr + 1,'.');
  if (! dotPtr) return 0;
  int length = dotPtr - name;
  char nameBuffer[length + 1];
  std::memcpy(nameBuffer,name,length);
  nameBuffer[length] = '\0';

  //
  // We've collected what should be a fully-qualified object name
  //
  const SmiOidTreeNode *node = smi -> getOidTree() -> getNode(nameBuffer);
  if (! node) {
    return 0;
    }

  //
  // Look up the object name and get its numeric representation
  //
  SmiOid smiOid(node);
  String *oidString = smiOid.toString();
  StringBuffer result;
  result.append(oidString -> getChars());

  //
  // Now start appending instance values (if any)
  //
  const char *namePtr = dotPtr + 1;
  while (*namePtr) {
    if (isdigit(*namePtr)) {
      dotPtr = std::strchr(namePtr,'.');
      if (! dotPtr) {
        result.append(namePtr - 1);
        break;
        }

      length = dotPtr - namePtr;
      result.append(namePtr - 1,length + 1);
      namePtr = dotPtr + 1;
      }
    else return 0; // Can't cope with non-numeric indexes yet
    }

  return new String(result);
  }

QWidget *SnmpGroupOperation::createWidget(QWidget *parent) {
  widget = new BrowserGroupViewWidget(parent);

  //
  // If a text font has been specified, set it now (before its
  // too late). We have to set it before the widget is shown.
  // (Note that the widget either gets its font metrics when
  // text is added, or it listens to fontChange signals)
  //
  if (configuration -> textFontFamily) {
    QFont font = widget -> font();
    font.setFamily(configuration -> textFontFamily -> getChars());
    if (configuration -> textFontSize != 0) {
      font.setPointSize(configuration -> textFontSize);
      }

    widget -> setFont(font);
    }

  return widget;
  }

void SnmpGroupOperation::initialise() {
  entryIterator =
      ((BrowserReportFunctionGroup *)function) -> getEntryList() -> iterator();

  state = Running;
  }

void SnmpGroupOperation::doWork() {
  SnmpTarget& snmpTarget = *target -> getTarget();
  Snmp& snmpSession = *session -> getSession();

  delete pdu;
  pdu = session -> createPdu(target);
  
  int variableCount = 0;
  ListIterator startingEntryIterator = entryIterator;
  while (entryIterator.hasNext()) {
    const BrowserReportFunctionGroupEntry *entry =
                 (BrowserReportFunctionGroupEntry *)entryIterator.next();

    String *oidString = findOidString(smi,entry -> getObjectName());
    if (! oidString) {
      state = Error;
      THROW(new BrowserException("OID not found for object",
                                 entry -> getObjectName()),);
      }

    Oid oid(oidString -> getChars());
    delete oidString;

    Vb vb;
    vb.set_oid(oid);
    *pdu += vb;

    variableCount++ ;
    if (variableCount == OBJECTS_PER_PDU) {
      break;
      }
    }

  if (! variableCount) {
    state = Finished;
    return;
    }

  int status = snmpSession.get(*pdu,snmpTarget);
  if (status != SNMP_CLASS_SUCCESS) {
    state = Error;
    THROW(new BrowserException(snmpSession.error_msg(status)),);
    }

  //
  // See if this is an SNMPv3 report message complaining about the
  // PDU
  //
  if (pdu -> get_type() == REPORT_MSG) {
    state = Error;
    processReportMessage(pdu);
    return;
    }

  if (pdu -> get_vb_count() != variableCount) {
    state = Error;
    THROW(new BrowserException("Target returned wrong number of variables!"),);
    }

  int variable = 0;
  entryIterator = startingEntryIterator;
  while (variable < variableCount) {
    const BrowserReportFunctionGroupEntry *entry =
                 (BrowserReportFunctionGroupEntry *)entryIterator.next();

    const Vb& vb = pdu -> get_vb(variable++);
    if (vb.get_syntax() == sNMP_SYNTAX_NOSUCHOBJECT) {
      ((BrowserGroupViewWidget *)widget) -> add(entry -> getName(),
                                                "[No object]");
      }
    else if (vb.get_syntax() == sNMP_SYNTAX_NOSUCHINSTANCE) {
      ((BrowserGroupViewWidget *)widget) -> add(entry -> getName(),
                                                "[No instance]");
      }
    else {
      String *value = SmiSnmpPPAdapter::snmpVbToString(*smi,vb);

      ((BrowserGroupViewWidget *)widget) -> add(entry -> getName(),
                                                value -> getChars());
      delete value;
      }
    }
  }

int SnmpGroupOperation::getSleepTime() const {
  //
  // We want to be called all the time
  //
  return 0;
  }
