/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstring>
#include <cctype>

#include <classlib/String.hpp>
#include <classlib/ObjectPair.hpp>

#include <smidb/Smi.hpp>
#include <smidb/SmiModule.hpp>
#include <smidb/SmiSyntax.hpp>
#include <smidb/SmiObjectType.hpp>

#include <smisnmp/SmiSnmpPPAdapter.hpp>

#include <qfont.h>

#include "SnmpTableWalkOperation.hpp"
#include "BrowserReportFunctionTableWalk.hpp"
#include "BrowserReportFunctionParameter.hpp"
#include "BrowserTableViewWidget.hpp"

const int PdusPerWorkCycle = 10;

class ColumnInfo : public Object {
  public:
    ColumnInfo() : objectName(0), oid(0) { }
    ~ColumnInfo() {
      delete objectName;
      delete oid;
      }

    void setOid(SmiOid *oidArg) { oid = oidArg; }
    SmiOid *getOid() const { return oid; }

    void setObjectName(const char *objectNameArg) {
      objectName = new String(objectNameArg);
      }
    const char *getObjectName() {
      if (objectName) return objectName -> getChars();
      else return 0;
      }

    void setType(BrowserReportFunctionTableWalkColumn::Type typeArg) {
      type = typeArg;
      }
    BrowserReportFunctionTableWalkColumn::Type getType() const {
      return type;
      }

  private:
    String *objectName;
    BrowserReportFunctionTableWalkColumn::Type type;
    SmiOid *oid;
  };

class SnmpTableWalkOperationPrivate {
  public:
    SnmpTableWalkOperationPrivate() :
              columns(0), pdu(0), returnedControlOid(0), controlColumnInfo(0),
              controlVbIndex(-1) {

      columns = new ObjectList();
      }

    ~SnmpTableWalkOperationPrivate() {
      delete columns;
      delete pdu;
      delete returnedControlOid;
      }

    ObjectList *columns;
    Pdu *pdu;
    SmiOid *returnedControlOid;
    ColumnInfo *controlColumnInfo;
    int controlVbIndex;
  };

SnmpTableWalkOperation::SnmpTableWalkOperation() {
  data = new SnmpTableWalkOperationPrivate();
  }

SnmpTableWalkOperation::~SnmpTableWalkOperation() {
  delete data;
  }

QWidget *SnmpTableWalkOperation::createWidget(QWidget *parent) {
  widget = new BrowserTableViewWidget(parent);

  //
  // If a text font has been specified, set it now (before its
  // too late). We have to set it before the widget is shown.
  // (Note that the widget either gets its font metrics when
  // text is added, or it listens to fontChange signals)
  //
  if (configuration -> textFontFamily) {
    QFont font = widget -> font();
    font.setFamily(configuration -> textFontFamily -> getChars());
    if (configuration -> textFontSize != 0) {
      font.setPointSize(configuration -> textFontSize);
      }

    widget -> setFont(font);
    }

  return widget;
  }

void SnmpTableWalkOperation::initialise() {
  //
  // Build a list of columns for this operation and create the table heading
  //

  //
  // See if we were given a table name as a parameter
  //
  if (parameterValues) {
    ListIterator it = parameterValues -> iterator();
    while (it.hasNext()) {
      BrowserReportFunctionParameter *parameter =
                           (BrowserReportFunctionParameter *)it.next();

      if (parameter -> getValue()) {
        //
        // The leaf descendents of the table will be our columns, and
        // the first readable column will be our control column (TBD -
        // add a parameter for the user to choose the control column?)
        //
        const SmiOidTreeNode *node =
                smi -> getOidTree() -> getNode(parameter -> getValue());
        if (node == 0) {
          state = Error;

          char scratch[256];
          std::sprintf(scratch,"Table '%s' was not found in the MIB",
                       parameter -> getValue());
          THROW(new BrowserException(scratch),);
          }

        //
        // The first (and only) child of the table is the Entry node
        //
        ObjectList *tableChildren = node -> getChildren();
        if ((! tableChildren) || (tableChildren -> count() != 1)) {
          state = Error;

          char scratch[256];
          std::sprintf(scratch,"'%s' is not a table",
                       parameter -> getValue());
          THROW(new BrowserException(scratch),);
          }
        ListIterator tableIterator = node -> getChildren() -> iterator();
        SmiOidTreeNode *entryNode = (SmiOidTreeNode *)tableIterator.next();
        ObjectList *entryChildren = entryNode -> getChildren();
        if (! entryChildren) {
          state = Error;
          char scratch[256];

          std::sprintf(scratch,"'%s' is not a table",
                       parameter -> getValue());
          THROW(new BrowserException(scratch),);
          }

        bool foundControlColumn = false;
        ListIterator entryIterator = entryChildren -> iterator();
        while (entryIterator.hasNext()) {
          SmiOidTreeNode *childNode = (SmiOidTreeNode *)entryIterator.next();

          const char *childName = smi -> getOidTree() -> getName(childNode);
          if (childName) {
            const char *dotPtr = std::strchr(childName,'.');
            char moduleName[256];
            int length = dotPtr - childName;
            std::memcpy(moduleName,childName,length);
            moduleName[length] = '\0';
            const char *objectName = dotPtr + 1;

            //
            // Now we need to determine the column type, based on the
            // object accessibility
            //
            SmiModule *module = smi -> getModuleList() -> getModule(moduleName);
            if (module) {
              const SmiAssignment *assignment =
                                      module -> getAssignment(objectName);
              if (assignment &&
                  (assignment -> getType() == SmiAssignmentInfo::OBJECT_TYPE)) {

                ((BrowserTableViewWidget *)widget) -> addColumn(objectName);

                ColumnInfo *columnInfo = new ColumnInfo();
                columnInfo -> setObjectName(objectName);

                const char *access =
                         ((SmiObjectType *)assignment) -> getAccess();
                if ((strcmp(access,"read-only") == 0) ||
                    (strcmp(access,"read-write") == 0) ||
                    (strcmp(access,"read-create") == 0)) {
                  if (! foundControlColumn) {
                    columnInfo ->
                         setType(BrowserReportFunctionTableWalkColumn::Control);
                    foundControlColumn = true;
                    }
                  else {
                    columnInfo ->
                         setType(BrowserReportFunctionTableWalkColumn::Normal);
                    }
                  }
                else {
                  columnInfo ->
                       setType(BrowserReportFunctionTableWalkColumn::Implicit);
                  }

                data -> columns -> add(columnInfo);
                }
              }
            }
          }
        }
      }
    }

  ObjectList *columns =
           ((BrowserReportFunctionTableWalk *)function) -> getColumnList();
  if (columns) {
    ListIterator it = columns -> iterator();
    while (it.hasNext()) {
      BrowserReportFunctionTableWalkColumn *column =
                            (BrowserReportFunctionTableWalkColumn *)it.next();
      ((BrowserTableViewWidget *)widget) -> addColumn(column -> getName());
  
      ColumnInfo *columnInfo = new ColumnInfo();
      columnInfo -> setObjectName(column -> getObjectName());
      columnInfo -> setType(column -> getType());
      data -> columns -> add(columnInfo);
      }
    }

  //
  // Build a list of OIDs corresponding to the column object names. These
  // will be used to generate the request variables
  //
  int vbIndex = 0;
  ListIterator it = data -> columns -> iterator();
  while (it.hasNext()) {
    ColumnInfo *columnInfo = (ColumnInfo *)it.next();

    if (columnInfo -> getType() !=
                        BrowserReportFunctionTableWalkColumn::Implicit) {
      if (isdigit(*columnInfo -> getObjectName())) {
        columnInfo -> setOid(new SmiOid(columnInfo -> getObjectName()));
        }
      else {
        const SmiOidTreeNode *node =
                smi -> getOidTree() -> getNode(columnInfo -> getObjectName());
        if (node == 0) {
          state = Error;

          char scratch[256];
          std::sprintf(scratch,"Object '%s' was not found in the MIB",
                       columnInfo -> getObjectName());
          THROW(new BrowserException(scratch),);
          }

        columnInfo -> setOid(new SmiOid(node));
        }

      if (columnInfo -> getType() ==
                          BrowserReportFunctionTableWalkColumn::Control) {
        data -> controlColumnInfo = columnInfo;
        data -> controlVbIndex = vbIndex;
        }

      vbIndex++ ;
      }
    }

  if (! data -> controlColumnInfo) {
    state = Error;

    char scratch[256];
    std::sprintf(scratch,"No control column was given for this function");
    THROW(new BrowserException(scratch),);
    }

  //
  // We now have a list of OIDs to request next values for
  //
  generateNextRequest();

  state = Running;
  }

void SnmpTableWalkOperation::generateNextRequest() {

  int controlSidCount;
  int returnedControlSidCount;
  const SmiTypes::SubIdentifier *returnedControlSids;
  Oid indexOid;

  //
  // If we're regenerating a PDU based on a previous response,
  // construct an index OID from the previous control OID instance
  //
  if (data -> returnedControlOid) {
    // We'll need the number of sub-identifiers in the original control
    // OID (this will tell us where the index sub-identifiers are
    // located in the returned control instance)
    //
    controlSidCount = data -> controlColumnInfo -> getOid() -> getLength();

    //
    // We also need the number of sub-identifiers in the returned control
    // OID. The difference between this and the length of the base OID
    // gives us the number of instance sub-identifiers returned
    //
    returnedControlSidCount = data -> returnedControlOid -> getLength();

    //
    // The index OID (to be appended to each column OID) is the trailing
    // part of the returned control OID instance
    //
    returnedControlSids = data -> returnedControlOid -> getSids();
    indexOid = Oid(&returnedControlSids[controlSidCount],
                   returnedControlSidCount - controlSidCount);
    }

  delete data -> pdu;
  data -> pdu = session -> createPdu(target);

  ListIterator it = data -> columns -> iterator();
  while (it.hasNext()) {
    ColumnInfo *columnInfo = (ColumnInfo *)it.next();

    if (columnInfo -> getType() !=
                    BrowserReportFunctionTableWalkColumn::Implicit) {

      Oid snmpOid(columnInfo -> getOid() -> getSids(),
                  columnInfo -> getOid() -> getLength());

      if (data -> returnedControlOid) {
        snmpOid += indexOid;
        }

      Vb vb(snmpOid);
      *data -> pdu += vb;
      }
    }
  }

void SnmpTableWalkOperation::processResults() {
  //
  // We'll need the control column index length, so we can discard other column
  // values if they're on the wrong row
  //
  int indexLength = data -> returnedControlOid -> getLength() -
                    data -> controlColumnInfo -> getOid() -> getLength();

  //
  // Get a list of index name/value pairs in case we have any implicit
  // columns
  //
  ObjectList *indexValues =
            smi -> getIndexValues(*data -> returnedControlOid,0);

  QListViewItem *item = new QListViewItem((BrowserTableViewWidget *)widget);

  int vbIndex = 0;
  int columnIndex = 0;
  ListIterator it = data -> columns -> iterator();
  while (it.hasNext()) {
    ColumnInfo *columnInfo = (ColumnInfo *)it.next();

    if (columnInfo -> getType() ==
                    BrowserReportFunctionTableWalkColumn::Implicit) {
      //
      // The value for this column is part of the index
      //
      if (indexValues) {
        const char *columnName = columnInfo -> getObjectName();
        bool unqualified = false;
        const char *dotPtr = std::strchr(columnName,'.');
        if (! dotPtr) unqualified = true;

        ListIterator indexValueIterator = indexValues -> iterator();
        while (indexValueIterator.hasNext()) {
          ObjectPair *pair = (ObjectPair *)indexValueIterator.next();
          const char *indexName = ((String *)pair -> first) -> getChars();

          if (! unqualified) {
            if (strcmp(columnName,indexName) == 0) {
              item -> setText(columnIndex,
                              ((String *)pair -> second) -> getChars());
              break;
              }
            }
          else {
            dotPtr = std::strchr(indexName,'.');
            if (dotPtr && (strcmp(dotPtr + 1,columnName) == 0)) {
              item -> setText(columnIndex,
                              ((String *)pair -> second) -> getChars());
              break;
              }
            }
          }
        }
      }
    else {
      //
      // Get the variable binding for this column and its OID
      //
      const Vb& vb = data -> pdu -> get_vb(vbIndex++);
      SmiOid *valueOid = SmiSnmpPPAdapter::snmpOid(vb.get_oid());

      //
      // See if the returned OID is still in the right column and
      // the index matches the control index
      //
      if (valueOid ->
             sameBeginning(columnInfo -> getOid(),
                           columnInfo -> getOid() -> getLength())) {
        if (valueOid -> sameEnding(data -> returnedControlOid,indexLength)) {
          //
          // Yep, we can use the value
          //
          String *value =
                   SmiSnmpPPAdapter::snmpVbToString(*smi,vb,Smi::ShortValues);
          item -> setText(columnIndex,value -> getChars());
          delete value;
          }
        }
      }

    columnIndex++ ;
    }

  delete indexValues;
  }

void SnmpTableWalkOperation::doWork() {
  //
  // On entry, we have a PDU ready to send (the first one is the initial PDU
  // created in initialise(), and subsequent ones we created the last time
  // we did work)
  //
  int status;
  SnmpTarget& snmpTarget = *target -> getTarget();
  Snmp& snmpSession = *session -> getSession();
  for (int request = 0 ; request < PdusPerWorkCycle ; request++) {
    //
    // Send the next request and see what we get
    //
    status = snmpSession.get_next(*data -> pdu,snmpTarget);

    if (status == SNMP_CLASS_SUCCESS) {
      //
      // See if this is an SNMPv3 report message complaining about the
      // PDU
      //
      if (data -> pdu -> get_type() == REPORT_MSG) {
        state = Error;
        processReportMessage(data -> pdu);
        return;
        }

      //
      // We have some values (or exceptions) in the response. The
      // first thing to do is to see if the control column response
      // is still in the table (We assume for now that the response
      // is valid, i.e. matches the request)
      //
      const Vb& returnedControlVb =
                      data -> pdu -> get_vb(data -> controlVbIndex);
      if (returnedControlVb.get_syntax() == sNMP_SYNTAX_ENDOFMIBVIEW) {
        state = Finished;
        return;
        }
      else {
        delete data -> returnedControlOid;
        data -> returnedControlOid =
                     SmiSnmpPPAdapter::snmpOid(returnedControlVb.get_oid());

        if (! data -> returnedControlOid ->
                      isUnder(data -> controlColumnInfo -> getOid())) {
          state = Finished;
          return;
          }

        //
        // Extract values for the next results column (this method will
        // exclude any values that aren't for the current row)
        //
        processResults();

        //
        // Now build the next request, based on the returned control
        // instance
        //
        generateNextRequest();

        //
        // The PDU is now ready for sending out
        //
        }
      }
    else if (status == SNMP_ERROR_NO_SUCH_NAME) {
      //
      // No more objects in the MIB view
      //
      state = Finished;
      return;
      }
    else {
      //
      // Unexpected error
      //
      state = Error;
      THROW(new BrowserException(snmpSession.error_msg(status)),);
      }
    }
  }

int SnmpTableWalkOperation::getSleepTime() const {
  //
  // We want to be called all the time
  //
  return 0;
  }
