/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstring>
#include <cctype>

#include <classlib/String.hpp>

#include <smidb/Smi.hpp>
#include <smidb/SmiModule.hpp>
#include <smidb/SmiSyntax.hpp>
#include <smidb/SmiObjectType.hpp>

#include <smisnmp/SmiSnmpPPAdapter.hpp>

#include <qfont.h>

#include "SnmpWalkOperation.hpp"
#include "BrowserReportFunctionWalk.hpp"
#include "BrowserTextViewWidget.hpp"
#include "BrowserReportFunctionParameter.hpp"

//
// The number of requests to send per call to doWork(). This nasty thing
// will go away once we convert over to an asynchronous model and tie it
// in to the Qt event system.
//
const int PdusPerWorkCycle = 10;

SnmpWalkOperation::SnmpWalkOperation() : startingOid(0), pdu(0) {
  }

SnmpWalkOperation::~SnmpWalkOperation() {
  delete startingOid;
  delete pdu;
  }

QWidget *SnmpWalkOperation::createWidget(QWidget *parent) {
  widget = new BrowserTextViewWidget(parent);

  //
  // If a text font has been specified, set it now (before its
  // too late). We have to set it before the widget is shown.
  // (Note that the widget either gets its font metrics when
  // text is added, or it listens to fontChange signals)
  //
  if (configuration -> textFontFamily) {
    QFont font = widget -> font();
    font.setFamily(configuration -> textFontFamily -> getChars());
    if (configuration -> textFontSize != 0) {
      font.setPointSize(configuration -> textFontSize);
      }

    widget -> setFont(font);
    }

  return widget;
  }

void SnmpWalkOperation::initialise() {
  //
  // Get the name or number for the starting point
  //
  const char *start = "1";
  if (parameterValues) {
    ListIterator it = parameterValues -> iterator();
    if (it.hasNext()) {
      BrowserReportFunctionParameter *parameter =
                       (BrowserReportFunctionParameter *)it.next();
      if (strcmp(parameter -> getName(),"walkStart") == 0) {
        start = parameter -> getValue();
        }
      }
    }

  //
  // If the starting node is not numeric, convert it to numeric form
  //
  Oid snmpOid;
  if (isdigit(*start)) {
    snmpOid = start;
    startingOid = new SmiOid(start);
    }
  else {
    startingOid = smi -> getOid(start);
    if (startingOid == 0) {
      state = Error;

      char scratch[256];
      std::sprintf(scratch,"Starting OID '%s' not found",start);
      THROW(new BrowserException(scratch),);
      }

    String *oidString = startingOid -> toString();
    snmpOid = oidString -> getChars();
    delete oidString;
    }

  Vb vb;
  vb.set_oid(snmpOid);

  pdu = session -> createPdu(target);
  *pdu += vb;

  state = Running;
  }

static String *printableOid(const Smi& smi,const SmiOid& oid,
                            const SmiModule **module,
                            const SmiSyntax **syntax,int options) {
  *syntax = 0;
  const SmiOidTreeNode *node = smi.getOidTree() -> findDeepest(&oid);
  if (! node) {
    return smi.stringValue(oid,options);
    }

  const char *nodeName = smi.getOidTree() -> getName(node);
  if (! nodeName) {
    return smi.stringValue(oid,options);
    }

  char *delimPtr = std::strchr(nodeName,'.');
  int length = delimPtr - nodeName;
  char moduleName[128];
  std::strncpy(moduleName,nodeName,length);
  moduleName[length] = '\0';
  char *assignmentName = delimPtr + 1;

  *module = smi.getModuleList() -> getModule(moduleName);
  if (! *module) {
    return smi.stringValue(oid,options);
    }

  const SmiAssignment *assignment = (*module) -> getAssignment(assignmentName);
  if (assignment) {
    if (assignment -> getType() == SmiAssignmentInfo::OBJECT_TYPE) {
      const SmiObjectType *objectType = (SmiObjectType *)assignment;
      *syntax = objectType -> getSyntax();
      }
    }

  return smi.stringValue(oid,options);
  }

static void printVb(const Smi& smi,BrowserTextViewWidget& resultTextView,
                    const SmiOid& oid,const Vb& vb,int options) {
  const SmiSyntax *syntax;
  const SmiModule *module;
  String *oidString = printableOid(smi,oid,&module,&syntax,options);
  String *value =
       SmiSnmpPPAdapter::snmpVbToString(smi,module,syntax,vb,options);

  QString line(oidString -> getChars());
  line += " = ";
  line += value -> getChars();
  line += '\n';
  resultTextView.appendString(line);

  delete value;
  delete oidString;
  }

//
// This method performs a few SNMP requests each time it is called. At the
// moment, it is completely synchronous and could block the UI (and other
// operations) for some time. TBD - Convert to an asynchronous model, use
// QSocketNotifier, etc.
//
void SnmpWalkOperation::doWork() {
  int status;
  SnmpTarget& snmpTarget = *target -> getTarget();
  Snmp& snmpSession = *session -> getSession();
  for (int request = 0 ; request < PdusPerWorkCycle ; request++) {
    status = snmpSession.get_next(*pdu,snmpTarget);
    if (status == SNMP_CLASS_SUCCESS) {
      //
      // See if this is an SNMPv3 report message complaining about the
      // PDU
      //
      if (pdu -> get_type() == REPORT_MSG) {
        state = Error;
        processReportMessage(pdu);
        return;
        }

      const Vb& returnedVb = pdu -> get_vb(0);
      if (returnedVb.get_syntax() == sNMP_SYNTAX_ENDOFMIBVIEW) {
        ((BrowserTextViewWidget *)widget) -> appendString("[End of MIB view]");
        state = Finished;
        return;
        }
      else {
        SmiOid *returnedOid = SmiSnmpPPAdapter::snmpOid(returnedVb.get_oid());
        if (! returnedOid -> isUnder(startingOid)) {
          ((BrowserTextViewWidget *)widget) ->
                       appendString("[End of results under starting point]");
          state = Finished;
          return;
          }

        printVb(*smi,*(BrowserTextViewWidget *)widget,
                *returnedOid,returnedVb,options);
        delete returnedOid;
        }
      }
    else if (status == SNMP_ERROR_NO_SUCH_NAME) {
      //
      // No more objects in the MIB view
      //
      ((BrowserTextViewWidget *)widget) -> appendString("[End of MIB reached]");
      state = Finished;
      return;
      }
    else {
      //
      // Unexpected error
      //
      state = Error;
      THROW(new BrowserException(snmpSession.error_msg(status)),);
      }
    }
  }

int SnmpWalkOperation::getSleepTime() const {
  //
  // We want to be called all the time
  //
  return 0;
  }
