/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstring>
#include <climits>

#include <qpainter.h>

#include "StripChartWidget.hpp"

const int MaxTraces = 8; // Must match colour list!
const int MaxSamples = 640;

QColor traceColours[MaxTraces] = { QColor(255,0,0),        // red
                                   QColor(0,255,0),        // green
                                   QColor(0,0,255),        // blue
                                   QColor(255,255,0),      // yellow
                                   QColor(255,0,255),      // purple
                                   QColor(0,255,255),      // cyan
                                   QColor(255,128,0),      // orange
                                   QColor(128,255,0),      // lime
                                 };

class Trace {
  public:
    Trace(int maxSamples);
    ~Trace();

    void closeSample();
    void getRange(int& min,int& max);
    int getLatestSample();

    int maxSamples;
    int validSamples;
    int first;
    int base;
    int current;
    int *samples;
  };

class StripChartWidgetPrivate {
  public:
    StripChartWidgetPrivate() : autoScale(true),
                                shiftPeriod(1000), ticks(false),
                                zeroLine(false), timerId(-1), traceCount(0) {
      traces = new Trace *[MaxTraces];
      std::memset(traces,0,sizeof(Trace *) * MaxTraces);
      }

    ~StripChartWidgetPrivate() {
      delete[] traces;
      }

    int rangeMin;
    int rangeMax;
    int maxSamples;
    bool autoScale;
    int shiftPeriod;
    bool ticks;
    bool zeroLine;
    int timerId;
    int traceCount;
    Trace **traces;
  };

Trace::Trace(int maxSamplesArg) : maxSamples(maxSamplesArg),
                                  validSamples(0), first(0), base(0),
                                  current(0) {
  samples = new int[maxSamples];
  }

Trace::~Trace() {
  delete[] samples;
  }

void Trace::closeSample() {
  //
  // Adjust 'first' to point at a new bucket
  //
  first++ ;
  if (first == maxSamples) first = 0;

  //
  // Empty the current sample into the new bucket
  //
  samples[first] = base + current;
  current = 0;

  //
  // Indicate that we have another valid sample
  //
  validSamples++ ;
  if (validSamples > maxSamples) validSamples = maxSamples;
  }

int Trace::getLatestSample() {
  if (validSamples) return samples[first];

  return 0;
  }

void Trace::getRange(int& min, int& max) {
  min = INT_MAX;
  max = INT_MIN;

  int si = first;
  for (int s = 0 ; s < validSamples ; s++ ) {
    if (samples[si] > max) max = samples[si];
    if (samples[si] < min) min = samples[si];

    si-- ;
    if (si < 0) si = maxSamples - 1;
    }
  }

StripChartWidget::StripChartWidget(QWidget *parent,int maxSamplesArg,
                                   int rangeMinArg,int rangeMaxArg)
              : QWidget(parent), data(0) {

  data = new StripChartWidgetPrivate();

  if (maxSamplesArg == -1) data -> maxSamples = size().width();
  else data -> maxSamples = maxSamplesArg;
  if (data -> maxSamples > MaxSamples) data -> maxSamples = MaxSamples;

  data -> rangeMin = rangeMinArg;
  data -> rangeMax = rangeMaxArg;

  data -> timerId = startTimer(data -> shiftPeriod);
  }

StripChartWidget::~StripChartWidget() {
  delete data;
  }

void StripChartWidget::setAutoScale(bool autoScaleArg) {
  data -> autoScale = autoScaleArg;
  }

void StripChartWidget::setShiftPeriod(int shiftPeriodArg) {
  data -> shiftPeriod = shiftPeriodArg;

  killTimer(data -> timerId);
  data -> timerId = -1;
  if (data -> shiftPeriod) data -> timerId = startTimer(data -> shiftPeriod);
  }

void StripChartWidget::setTicks(bool ticksArg) {
  data -> ticks = ticksArg;
  }

void StripChartWidget::setZeroLine(bool zeroLineArg) {
  data -> zeroLine = zeroLineArg;
  }

int StripChartWidget::addTrace() {
  for (int i = 0 ; i < MaxTraces ; i++ ) {
    if (data -> traces[i] == 0) {
      data -> traces[i] = new Trace(data -> maxSamples);
      data -> traceCount++ ;
      updateGeometry();
      return i;
      }
    }

  return -1;
  }

void StripChartWidget::deleteTrace(int traceId) {
  if ((traceId >= 0) && (traceId < MaxTraces)) {
    delete data -> traces[traceId];
    data -> traces[traceId] = 0;
    data -> traceCount-- ;
    }
  }

void StripChartWidget::addSample(int traceId,int sample) {
  data -> traces[traceId] -> current += sample;
  }

void StripChartWidget::setValue(int traceId,int value) {
  data -> traces[traceId] -> base = value;
  }

QColor StripChartWidget::getColor(int traceId) const {
  return traceColours[traceId];
  }

QSize StripChartWidget::minimumSizeHint() const {
  return QSize(40,15 + 8 * data -> traceCount);
  }

QSize StripChartWidget::sizeHint() const {
  return minimumSizeHint();
  }

void StripChartWidget::timerEvent(QTimerEvent *event) {
  //
  // Convert the current value into a new sample for all traces
  //
  for (int i = 0 ; i < MaxTraces ; i++ ) {
    if (data -> traces[i]) {
      data -> traces[i] -> closeSample();
      emit newSample(i,data -> traces[i] -> getLatestSample());
      }
    }

  //
  // A repaint is needed to display the new sample
  //
  update();
  }

void StripChartWidget::paintEvent(QPaintEvent *event) {
  //
  // Find the greatest range in the traces
  //
  int displayMin = INT_MAX;
  int displayMax = INT_MIN;
  if (data -> autoScale) {
    for (int i = 0 ; i < MaxTraces ; i++ ) {
      if (data -> traces[i]) {
        int traceMin, traceMax;
        data -> traces[i] -> getRange(traceMin,traceMax);
        if (traceMin < displayMin) displayMin = traceMin;
        if (traceMax > displayMax) displayMax = traceMax;
        }
      }
    }
  else {
    displayMin = data -> rangeMin;
    displayMax = data -> rangeMax;
    }

  //
  // If the autoscale did not produce a usable minimum, there's nothing
  // to show
  //
  if (displayMin == INT_MAX) return;

  //
  // Calculate the scale factor for making the samples fit in
  // the entire widget height
  //
  float heightMultiplier = 1.0;
  int widgetHeight = size().height();
  if (displayMax != displayMin) {
    heightMultiplier = (float)widgetHeight / (displayMax - displayMin + 1);
    }

  //
  // Calculate the x-axis scale factor to make the samples fit all the
  // way across the widget
  //
  int widgetWidth = size().width();
  float xStep = (float)widgetWidth / data -> maxSamples;

  //printf("widget %d x %d, hm=%f, xs=%f\n",
  //       widgetWidth,widgetHeight,heightMultiplier,xStep);

  QPainter painter;
  painter.begin(this);

  painter.setPen(QColor(255,255,255));
  painter.drawText(0,widgetHeight,QString().setNum(displayMin));
  QString text;
  text.setNum(displayMax);
  painter.drawText(0,fontMetrics().boundingRect(text).height(),text);

  for (int i = 0 ; i < MaxTraces ; i++ ) {
    if (data -> traces[i]) {

      int x = widgetWidth - 1;

      painter.setPen(traceColours[i]);

      int si = data -> traces[i] -> first;
      for (int s = 0 ; s < data -> traces[i] -> validSamples ; s++ ) {
        //
        // Force the sample to fit the display range
        //
        int adjustedSample = data -> traces[i] -> samples[si];
        if (adjustedSample < displayMin) adjustedSample = displayMin;
        if (adjustedSample > displayMax) adjustedSample = displayMax;

        //
        // Shift and scale the sample
        //
        adjustedSample = (int)((adjustedSample - displayMin) *
                               heightMultiplier);

        //
        // Convert to the widget coordinate system
        //
        int y = (widgetHeight - 1) - adjustedSample;

        //printf("t=%d s=%d, x=%d, y=%d, as=%d, min=%d, max=%d\n",
        //       i,s,x,y,adjustedSample,displayMin,displayMax);

        if (s == 0) {
          painter.moveTo(x,y);
          }
        else {
          painter.lineTo(x,y);
          }

        x = (int)(x - xStep);

        si-- ;
        if (si < 0) si = data -> maxSamples - 1;
        }
      }
    }
  }
