/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <malloc.h>
#include <errno.h>
#include <string.h>

#include "DbFlatDir.hpp"
#include "DetailedException.hpp"

//
// The flat file directory database stores each record in a separate
// file in the database directory. At the moment, the database is
// simplified in that it does not support deletion of the entire
// database, has no index and does not support enumeration of records.
// Hence it can share a directory with other files (there's also no naming
// convention for files/records). It's used for the case where
// an application knows exactly what the records are.
//
DbFlatDir::DbFlatDir(const char *dbPath,Db::Mode mode) : Db(dbPath,mode) {

  if (mode == Db::Write) {
    if ((mkdir(dbPath,0755) == -1) && (errno != EEXIST)) {
      THROW(new DetailedException("Failed to create database",
                                  strerror(errno)),);
      }
    }
  }

DbFlatDir::~DbFlatDir() {
  }

void *DbFlatDir::fetch(const char *recordName,int *length) const {
  char path[256];
  std::sprintf(path,"%s/%s",dbPath,recordName);
  int fd = ::open(path,O_RDONLY);
  if (fd == -1) {
    if (errno == ENOENT) return 0;

    char scratch[256];
    std::sprintf(scratch,
                 "Failed to open file database for reading record %s (in %s)",
                 recordName,dbPath);
    THROW(new DetailedException(scratch,strerror(errno)),0);
    }

  int left = 4096;
  char *record = (char *)malloc(left);
  char *recordFree = record;
  *length = 0;
  ssize_t consumed;
  while (true) {
    consumed = ::read(fd,recordFree,left);
    if (consumed == -1) {
      char scratch[256];
      std::sprintf(scratch,"Failed to read file record %s",recordName);
      THROW(new DetailedException(scratch,strerror(errno)),0);
      }

    if (consumed == 0) break;

    left -= consumed;
    *length += consumed;

    if (left < 128) {
      left = 4096;
      record = (char *)realloc(record,*length + left);
      }

    recordFree = record + *length;
    }

  ::close(fd);

  return record;
  }

void DbFlatDir::store(const char *recordName,const void *record,int length) {
  char path[256];
  std::sprintf(path,"%s/%s",dbPath,recordName);
  int fd = ::open(path,O_WRONLY | O_TRUNC | O_CREAT,0644);
  if (fd == -1) {
    char scratch[256];
    std::sprintf(scratch,
                 "Failed to open file database for writing record %s (in %s)",
                 recordName,dbPath);
    THROW(new DetailedException(scratch,strerror(errno)),);
    }
  while (length) {
    ssize_t wrote = ::write(fd,record,length);
    if (wrote == -1) {
      ::close(fd);

      char scratch[256];
      std::sprintf(scratch,"Failed to write file record %s",recordName);
      THROW(new DetailedException(scratch,strerror(errno)),);
      }

    (char *)record += wrote;
    length -= wrote;
    }

  if (::close(fd) == -1) {
    char scratch[256];
    std::sprintf(scratch,"Failed to write file record %s",recordName);
    THROW(new DetailedException(scratch,strerror(errno)),);
    }
  }

void DbFlatDir::deleteRecord(const char *recordName) {
  char path[256];
  std::sprintf(path,"%s/%s",dbPath,recordName);
  if (::unlink(path) == -1) {
    char scratch[256];
    std::sprintf(scratch,"Failed to delete file record %s",recordName);
    THROW(new DetailedException(scratch,strerror(errno)),);
    }
  }

void DbFlatDir::deleteDatabase() {
  THROW(new DetailedException("Flat file database deletion not implemented"),);
  }
