/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstring>
#include <cstdlib>

#include "List.hpp"

ListEntry::ListEntry(const void *item) : item(item), next(0) {
  }

ListEntry::~ListEntry() {
  }

List::List()
    : first(0), last(0), entryCount(0) {
  }

List::~List() {
  //
  // Delete each entry in the list
  //
  if (first != 0) {
    ListEntry *current = first;
    ListEntry *next;
    while (current) {
      next = current -> next;
      delete current;
      current = next;
      }
    }
  }

void List::add(const void *item) {
  ListEntry *newEntry = new ListEntry(item);

  if (last) {
    last -> next = newEntry;
    last = newEntry;
    if (! first) first = newEntry;
    }
  else {
    first = newEntry;
    last = newEntry;
    }

  entryCount++ ;
  }

int List::count() const {
  return entryCount;
  }

ListIterator List::iterator() {
  ListIterator it(this);

  return it;
  }

ListIterator::~ListIterator() {
  }

ListIterator::ListIterator() : list(0), previous(0), entry(0) {
  }

ListIterator::ListIterator(const ListIterator& other) {
  list = other.list;
  entry = other.entry;
  previous = other.previous;
  }

ListIterator::ListIterator(List *list) : list(list), previous(0), entry(0) {
  }

ListIterator& ListIterator::operator=(const ListIterator& other) {
  list = other.list;
  entry = other.entry;
  previous = other.previous;

  return *this;
  }

const void *ListIterator::next() {
  if ((! previous) && (! entry)) entry = list -> first;
  else {
    if (! entry) entry = previous -> next;
    else {
      previous = entry;
      entry = entry -> next;
      }
    }

  if (! entry) return 0;

  return entry -> item;
  }

bool ListIterator::hasNext() const {
  //
  // No list means no next item (empty iterator)
  //
  if (! list) return false;

  //
  // If both the current and previous are 0, we are just before
  // the start of the list, so the next item is the first one
  // on the list
  //
  if ((! entry) && (! previous)) return (list -> first != 0);

  //
  // If the current entry is 0, it means we deleted it, so whatever
  // follows the previous entry is next (and we already know there
  // is a previous entry)
  //
  if (! entry) return (previous -> next != 0);

  //
  // The entry after the current entry is next
  //
  return (entry -> next != 0);
  }

void ListIterator::remove() {

  //
  // If the entry we're removing is the last on the list, then
  // whatever is before this entry is the new last entry
  //
  if (entry == list -> last) list -> last = previous;

  ListEntry *thisEntry = entry;
  if (previous) {
    previous -> next = entry -> next;
    }
  else {
    list -> first = entry -> next;
    }

  //
  // Mark the current entry as 0. The hasNext() and next() methods use
  // this to key off previous->next rather than the now non-existent
  // current entry
  //
  entry = 0;

  delete thisEntry;

  list -> entryCount-- ;
  }

void ListIterator::insert(const void *item) {
  ListEntry *newEntry = new ListEntry(item);

  if (previous == 0) {
    //
    // No previous entry for the iterator means we're inserting
    // at the front of the list
    //
    newEntry -> next = list -> first;
    list -> first = newEntry;

    //
    // If the list was previously empty, the new item is also now
    // the last on the list
    //
    if (! list -> last) list -> last = newEntry;
    }
  else {
    //
    // Inserting just after the previous entry
    //
    newEntry -> next = previous -> next;
    previous -> next = newEntry;

    //
    // If the previous entry was the last on the list, then the
    // entry we just linked after the previous becomes the new
    // last entry
    //
    if (previous == list -> last) list -> last = newEntry;
    }

  list -> entryCount++ ;
  }
