/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstring>
#include <cstdlib>

#include "ObjectHashMap.hpp"
#include "String.hpp"

const int HASH_SIZE = 4095;

static unsigned long hash(const char *name) {
  const char *np = name;
  unsigned long h;

  h = 0;
  while (*np) {
    h = (h << 1) + *np++;
    }

  return h % HASH_SIZE;
  }

class ObjectChain {
  public:
    ObjectChain() : next(0) {
      mapping.name = 0;
      mapping.value = 0;
      }

    ObjectChain(ObjectMapping mapping) : next(0) {
      this -> mapping = mapping;
      }

    ~ObjectChain() {
      delete mapping.name;
      delete mapping.value;
      if (next) delete next;
      }

    ObjectMapping mapping;
    ObjectChain *next;

  private:
    ObjectChain(const ObjectChain& other);
    const ObjectChain& operator=(const ObjectChain& other);
  };

bool ObjectHashMapIterator::hasNext() {
  if ((! current) || (! current -> next)) {
    current = 0;
    while (! current) {
      chainIndex++ ;
      if (chainIndex == HASH_SIZE) return false;
      current = hashMap -> table[chainIndex];
      }

    //
    // Found a base chain that's not empty, so we make it appear to
    // next() that we are at the end of the previous base index
    //
    current = 0;
    previous = 0;
    chainIndex-- ;
    }

  return true;
  }

ObjectMapping ObjectHashMapIterator::next() {
  ObjectMapping mapping;

  //
  // Use hasNext() to step through to the next entry, and tell us if
  // we're off the end
  //
  if (! hasNext()) {
    mapping.name = 0;
    mapping.value = 0;
    return mapping;
    }

  //
  // hasNext() returned true, which means we are positioned just
  // before the next entry. Either we are at the null entry at
  // the end of the previous chain (current == 0) or we are at the
  // previous current entry
  //
  if (! current) {
    chainIndex++ ;
    previous = 0; // Already done, really
    current = hashMap -> table[chainIndex];
    }
  else {
    //
    // hasNext() guarantees that current -> next is not null..
    //
    previous = current;
    current = current -> next;
    }

  return current -> mapping;
  }

void ObjectHashMapIterator::detach() {
  if (! previous) {
    //
    // We are detaching the first entry in this chain
    //
    hashMap -> table[chainIndex] = current -> next;

    //
    // Delete the current mapping (the contents are already
    // owned by the caller, so we must not delete them too)
    //
    current -> mapping.name = 0;
    current -> mapping.value = 0;
    current -> next = 0;
    delete current;

    //
    // We want next() to return the (new) first entry in the
    // chain, so we have to point off the end of the previous
    // chain
    //
    current = 0;
    chainIndex-- ;
    }
  else {
    previous -> next = current -> next;
    current -> mapping.name = 0;
    current -> mapping.value = 0;
    current -> next = 0;
    delete current;
    current = previous;
    }

  hashMap -> mappingCount-- ;
  }

ObjectHashMap::ObjectHashMap() : mappingCount(0) {
  int tablesize = sizeof(ObjectChain *) * HASH_SIZE;
  table = new (ObjectChain *)[HASH_SIZE];
  std::memset(table,0,tablesize);
  }

ObjectHashMap::~ObjectHashMap() {
  for (int i = 0 ; i < HASH_SIZE ; i++ ) {
    delete table[i];
    }

  delete[] table;
  }

ObjectHashMapIterator ObjectHashMap::iterator() {
  ObjectHashMapIterator it;

  it.hashMap = this;
  it.current = 0;
  it.chainIndex = -1;

  return it;
  }

int ObjectHashMap::size() {
  return mappingCount;
  }

void ObjectHashMap::replaceValue(ObjectHashMapIterator it,
                                 const Object *newValue) {
  delete it.current -> mapping.value;
  it.current -> mapping.value = newValue;
  }

ObjectMapping ObjectHashMap::find(const char *name) const {
  unsigned long hashValue = hash(name);
  ObjectChain *chain = table[hashValue];
  while (chain) {
    if (chain -> mapping.name) {
       String *thisName = chain -> mapping.name -> toString();
       if (thisName -> equals(name)) {
        delete thisName;
        return chain -> mapping;
        }

      delete thisName;
      }

    chain = chain -> next;
    }

  ObjectMapping mapping;
  mapping.name = 0;
  mapping.value = 0;
  return mapping;
  }

void ObjectHashMap::remove(const char *name) {
  unsigned long hashValue = hash(name);
  ObjectChain *previous = 0;
  ObjectChain *chain = table[hashValue];
  while (chain) {
    if (chain -> mapping.name) {
       String *thisName = chain -> mapping.name -> toString();
       if (thisName -> equals(name)) {
        delete thisName;

        if (previous) {
          previous -> next = chain -> next;
          }
        else {
          table[hashValue] = chain -> next;
          }

        chain -> next = 0;
        delete chain;

        return;
        }

      delete thisName;
      }

    previous = chain;
    chain = chain -> next;
    }
  }

ObjectMapping ObjectHashMap::get(const Object *name) const {
  String *thisName = name -> toString();
  unsigned long hashValue = hash(thisName -> getChars());
  delete thisName;

  ObjectChain *chain = table[hashValue];
  while (chain) {
    if (chain -> mapping.name == name) return chain -> mapping;

    chain = chain -> next;
    }

  ObjectMapping mapping;
  mapping.name = 0;
  mapping.value = 0;
  return mapping;
  }

char *ObjectHashMap::getStats() const {
  int totalChainLength = 0;
  int emptyChains = 0;
  int longestChainLength = 0;
  for (int i = 0 ; i < HASH_SIZE ; i++ ) {
    ObjectChain *chain = table[i];
    if (! chain) emptyChains++ ;
    else {
      int chainLength = 1;
      while (chain -> next) {
        chainLength++ ;
        chain = chain -> next;
        }
      if (chainLength > longestChainLength) longestChainLength = chainLength;
      totalChainLength += chainLength;
      }
    }

  char *statString = new char[128];
  if (emptyChains == HASH_SIZE) {
    std::sprintf(statString,"No mappings in the table (100%% empty)");
    }
  else {
    std::sprintf(statString,
            "Average chain length is %.2f, longest %d with %d empty chains",
           (float)totalChainLength / (HASH_SIZE - emptyChains),
           longestChainLength,emptyChains);
    }

  return statString;
  }

void ObjectHashMap::add(Object *name,Object *value) {
  String *thisName = name -> toString();
  unsigned long hashValue = hash(thisName -> getChars());
  delete thisName;

  ObjectMapping mapping;
  mapping.name = name;
  mapping.value = value;
  ObjectChain *chain = new ObjectChain(mapping);

  if (table[hashValue] == 0) {
    table[hashValue] = chain;
    }
  else {
    chain -> next = table[hashValue];
    table[hashValue] = chain;
    }

  mappingCount++ ;
  }

void ObjectHashMap::add(const char *name,const char *value) {
  add(new String(name),new String(value));
  }

void ObjectHashMap::invalidate() {
  ObjectChain *chain, *nextChain;
  for (int i = 0 ; i < HASH_SIZE ; i++ ) {
    chain = table[i];
    table[i] = 0;
    while (chain) {
      //
      // Invalidate the contents of the mapping so that delete will
      // not do anything
      //
      chain -> mapping.name = 0;
      chain -> mapping.value = 0;;

      //
      // Delete only this chain entry, by truncating the chain here
      //
      nextChain = chain -> next;
      chain -> next = 0;
      delete chain;

      chain = nextChain;
      }
    }

  mappingCount = 0;
  }
