/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cerrno>

#include "Xml.hpp"
#include "StringBuffer.hpp"
#include "DetailedException.hpp"

#include <scew/scew.h>

Xml::Xml(void) {
  tree = scew_tree_create();
  }

Xml::Xml(const char *xmlPath) : tree(0) {
  loadXmlFromFile(xmlPath);
  }

Xml::~Xml() {
  if (tree) scew_tree_free((scew_tree *)tree);
  }

static const char *xml_errmsg(const char *text,int sys_errno) {
  static char message[512];

  if (sys_errno == 0) {
    std::sprintf(message,"%s (not well-formed?)",text);
    }
  else {
    std::sprintf(message,"%s : %s",text,strerror(sys_errno));
    }

  return message;
  }

void Xml::loadXmlFromFile(const char *xmlPath) {
  if (tree) {
    scew_tree_free((scew_tree *)tree);
    tree = 0; // In case we throw an exception before givng new value..
    }
  scew_parser *parser = scew_parser_create();
  errno = 0;
  if (! scew_parser_load_file(parser,xmlPath)) {
    int errnoCopy = errno;
    scew_parser_free(parser);
    const char *message = xml_errmsg("Cannot parse XML file",errnoCopy);
    THROW(new DetailedException(message),);
    }

  tree = (void *)scew_parser_tree(parser);

  scew_parser_free(parser);
  }

void Xml::loadXmlFromBuffer(const char *buffer,int length) {
  if (tree) {
    scew_tree_free((scew_tree *)tree);
    tree = 0; // In case we throw an exception before setting a new value..
    }
  scew_parser *parser = scew_parser_create();

  if (! scew_parser_load_buffer(parser,(const char *)buffer,length)) {
    scew_parser_free(parser);
    THROW(new DetailedException("Cannot parse XML buffer"),);
    }

  tree = (void *)scew_parser_tree(parser);

  scew_parser_free(parser);
  }

void Xml::loadXmlFromString(const char *xmlString) {
  loadXmlFromBuffer(xmlString,strlen(xmlString));
  }

XmlElement Xml::setRootElement(const char *name) {
  return scew_tree_add_root((scew_tree *)tree,name);
  }

XmlElement Xml::getRootElement(void) const {
  return (XmlElement)scew_tree_root((scew_tree *)tree);
  }

void Xml::deleteElement(XmlElement element) {
  scew_element_free((scew_element *)element); // Unlink and free
  }

XmlElement Xml::addElement(XmlElement parent,const char *name) {
  return (XmlElement)scew_element_add((scew_element *)parent,name);
  }

void Xml::addElementAttribute(XmlElement element,const char *name,
                              const char *value) {
  scew_element_add_attr_pair((scew_element *)element,name,value);
  }

int Xml::getElementChildCount(XmlElement parent) const {
  return scew_element_count((scew_element *)parent);
  }

XmlElement Xml::getElementChild(XmlElement parent,int i) const {
  return (XmlElement)scew_element_by_index((scew_element *)parent,i);
  }

const char *Xml::getElementContents(XmlElement element) const {
  return (const char *)scew_element_contents((scew_element *)element);
  }

void Xml::setElementContents(XmlElement element,const char *contents) {
  (void)scew_element_set_contents((scew_element *)element,(XML_Char *)contents);
  }

const char *Xml::getElementName(XmlElement element) const {
  return (const char *)scew_element_name((scew_element *)element);
  }

int Xml::getElementAttributeCount(XmlElement element) const {
  return scew_attribute_count((scew_element *)element);
  }

const char *Xml::getElementAttributeName(XmlElement element,int i) const {
  if (i < (int)scew_attribute_count((scew_element *)element)) {
    return (const char *)scew_attribute_name(
                   scew_attribute_by_index((scew_element *)element,i));
    }
  else return 0;
  }

const char *Xml::getElementAttributeValue(XmlElement element,int i) const {
  if (i < (int)scew_attribute_count((scew_element *)element)) {
    return (const char *)scew_attribute_value(
                   scew_attribute_by_index((scew_element *)element,i));
    }
  else return 0;
  }

static void addElementToBuffer(const Xml& xml,XmlElement element,
                               StringBuffer &buffer) {
  const char *elementName = xml.getElementName(element);
  buffer.append("<");
  buffer.append(elementName);
  int attributeCount = xml.getElementAttributeCount(element);
  for (int i = 0 ; i < attributeCount ; i++ ) {
    buffer.append(" ");
    buffer.append(xml.getElementAttributeName(element,i));
    buffer.append("=\"");
    buffer.append(xml.getElementAttributeValue(element,i));
    buffer.append("\"");
    }

  int childElementCount = xml.getElementChildCount(element);
  if (childElementCount == 0) {
    const char *elementContents = xml.getElementContents(element);
    if (elementContents == NULL) {
      buffer.append("/>");
      }
    else {
      buffer.append(">");

      const char *ePtr = elementContents;
      int length;
      while (*ePtr) {
        length = strcspn(ePtr,"<>&");
        if (length != 0) {
          buffer.append(ePtr,length);
          ePtr += length;
          }
        switch (*ePtr) {
          case '<': {
            buffer.append("&lt;");
            ePtr++ ;
            break;
            }
          case '>': {
            buffer.append("&gt;");
            ePtr++ ;
            break;
            }
          case '&': {
            buffer.append("&amp;");
            ePtr++ ;
            break;
            }
          default: break;
          }
        }

      buffer.append("</");
      buffer.append(elementName);
      buffer.append(">");
      }
    }
  else {
    buffer.append(">");

    for (int i = 0 ; i < childElementCount ; i++ ) {
      addElementToBuffer(xml,xml.getElementChild(element,i),buffer);
      }

    buffer.append("</");
    buffer.append(elementName);
    buffer.append(">");
    }
  }

char *Xml::xmlString(XmlElement element) const {
  StringBuffer buffer;
  addElementToBuffer(*this,element,buffer);

  return buffer.detach();
  }

char *Xml::xmlString() const {
  return xmlString(getRootElement());
  }

void Xml::free(char *string) {
  StringBuffer::free(string);
  }
