/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>

#include "ManagerResource.hpp"
#include "ConfigFormImpl.hpp"

#include <qlabel.h>
#include <qcombobox.h>
#include <qtoolbutton.h>
#include <qmessagebox.h>
#include <qlistview.h>
#include <qcheckbox.h>
#include <qstringlist.h>
#include <qpushbutton.h>
#include <qlayout.h>

#ifdef QWS
#include <qpe/fontdatabase.h>
const int FONT_SIZE_DIVISOR = 10; // Fix for Qt/E (is there better detection?)
#else
#include <qfontdatabase.h>
const int FONT_SIZE_DIVISOR = 1;
#endif

static void setSampleFont(QComboBox *fontFamilyComboBox,
                          QComboBox *fontSizeComboBox,
                          QLabel *sampleLabel) {
  QFont font(fontFamilyComboBox -> currentText(),
             fontSizeComboBox -> currentText().toInt());
  sampleLabel -> setFont(font);
  }

static void loadFontSizeConfiguration(QComboBox *familyComboBox,
                                      QComboBox *sizeComboBox,
                                      QLabel *sampleLabel,
                                      int preferedFontSize) {
#ifdef QWS
  FontDatabase fontDatabase;
#else
  QFontDatabase fontDatabase;
#endif

  sizeComboBox -> clear();
  QValueList<int> pointSizes =
                     fontDatabase.pointSizes(familyComboBox -> currentText());
  QValueList<int>::Iterator it;
  for (it = pointSizes.begin() ; it != pointSizes.end() ; ++it) {
    sizeComboBox -> insertItem(QString().setNum(*it / FONT_SIZE_DIVISOR));
    }

  int defaultSize = familyComboBox -> fontInfo().pointSize();
  int sizeCount = sizeComboBox -> count();
  for (int i = 0 ; i < sizeCount ; i++ ) {
    if (preferedFontSize != 0) {
      if (preferedFontSize == sizeComboBox -> text(i).toInt()) {
        sizeComboBox -> setCurrentItem(i);
        break;
        }
      }
    else if (defaultSize == sizeComboBox -> text(i).toInt()) {
      sizeComboBox -> setCurrentItem(i);
      break;
      }
    }

  setSampleFont(familyComboBox,sizeComboBox,sampleLabel);
  }

static void loadFontConfiguration(QCheckBox *checkBox,QFrame *fontFrame,
                                  QComboBox *familyComboBox,
                                  QComboBox *sizeComboBox,QLabel *sampleLabel,
                                  String *configuredFontFamily,
                                  int configuredFontSize) {

  //
  // First we load up the family list with the available font families
  //

#ifdef QWS
  FontDatabase fontDatabase;
#else
  QFontDatabase fontDatabase;
#endif

  QStringList families = fontDatabase.families();

  familyComboBox -> insertStringList(families);
  const char *configuredFamily = 0;
  if (configuredFontFamily) {
    configuredFamily = configuredFontFamily -> getChars();
    checkBox -> setChecked(true);
    fontFrame -> setEnabled(true);
    }
  else {
    checkBox -> setChecked(false);
    fontFrame -> setEnabled(false);
    }
  const char *defaultFamily = familyComboBox -> fontInfo().family();
  int familyCount = familyComboBox -> count();
  for (int i = 0 ; i < familyCount ; i++ ) {
    if (configuredFamily) {
      if (strcmp(configuredFamily,familyComboBox -> text(i)) == 0) {
        familyComboBox -> setCurrentItem(i);
        break;
        }
      }
    else if (defaultFamily &&
             (strcmp(defaultFamily,familyComboBox -> text(i)) == 0)) {
      familyComboBox -> setCurrentItem(i);
      break;
      }
    }

  //
  // Now we load the font sizes list with the available font sizes
  // for the current family
  //
  loadFontSizeConfiguration(familyComboBox,sizeComboBox,
                            sampleLabel,configuredFontSize);
  }

ConfigFormImpl::ConfigFormImpl(QWidget* parent,
                               const ManagerConfiguration *configuration)
    : ConfigForm(parent,0,true,0), newConfiguration(0) {

#ifndef QWS
  QFrame *buttonFrame = new QFrame(this);

  QHBoxLayout *buttonFrameLayout = new QHBoxLayout(buttonFrame);

  QPushButton *okButton = new QPushButton(buttonFrame);
  okButton -> setText("Ok");
  okButton -> adjustSize();
  okButton -> setDefault(true);

  QPushButton *cancelButton = new QPushButton(buttonFrame);
  cancelButton -> setText("Cancel");
  cancelButton -> adjustSize();

  buttonFrameLayout -> addStretch(1);
  buttonFrameLayout -> addWidget(okButton);
  buttonFrameLayout -> addWidget(cancelButton);

  layout() -> add(buttonFrame);

  connect(okButton,SIGNAL(clicked()),this,SLOT(accept()));
  connect(cancelButton,SIGNAL(clicked()),this,SLOT(reject()));
#endif

  //
  // Make a copy of the configuration. This copies everything,
  // even things which cannot be changed in the configuration
  // dialog
  //
  newConfiguration = new ManagerConfiguration(*configuration);

  IconLocationCheckBox -> setChecked(configuration -> preferSystemIcons);

  loadFontConfiguration(CustomApplicationFontCheckBox,
                        CustomApplicationFontFrame,
                        ApplicationFontFamilyComboBox,
                        ApplicationFontSizeComboBox,
                        ApplicationFontSampleLabel,
                        configuration -> applicationFontFamily,
                        configuration -> applicationFontSize);
  }

ConfigFormImpl::~ConfigFormImpl() {
  delete newConfiguration;

  // no need to delete child widgets, Qt does it all for us
  }

void ConfigFormImpl::applicationFontFamilyChangedSlot() {
  delete newConfiguration -> applicationFontFamily;
  newConfiguration -> applicationFontFamily =
               new String(ApplicationFontFamilyComboBox -> currentText());

  loadFontSizeConfiguration(ApplicationFontFamilyComboBox,
                            ApplicationFontSizeComboBox,
                            ApplicationFontSampleLabel,0);
  }

void ConfigFormImpl::applicationFontSizeChangedSlot() {
  newConfiguration -> applicationFontSize =
                         ApplicationFontSizeComboBox -> currentText().toInt();

  setSampleFont(ApplicationFontFamilyComboBox,ApplicationFontSizeComboBox,
                ApplicationFontSampleLabel);
  }

void ConfigFormImpl::applicationFontCheckBoxChangedSlot() {
  if (CustomApplicationFontCheckBox -> isChecked()) {
    CustomApplicationFontFrame -> setEnabled(true);
    delete newConfiguration -> applicationFontFamily;
    newConfiguration -> applicationFontFamily =
               new String(ApplicationFontFamilyComboBox -> currentText());
    newConfiguration -> applicationFontSize =
                          ApplicationFontSizeComboBox -> currentText().toInt();

    setSampleFont(ApplicationFontFamilyComboBox,ApplicationFontSizeComboBox,
                  ApplicationFontSampleLabel);
    }
  else {
    CustomApplicationFontFrame -> setEnabled(false);
    delete newConfiguration -> applicationFontFamily;
    newConfiguration -> applicationFontFamily = 0;
    newConfiguration -> applicationFontSize = 0;
    }
  }

void ConfigFormImpl::iconLocationChangedSlot() {
  newConfiguration -> preferSystemIcons = IconLocationCheckBox -> isChecked();
  }

ManagerConfiguration *ConfigFormImpl::takeNewConfiguration() {
  ManagerConfiguration *returnedConfiguration = newConfiguration;
  newConfiguration = 0;
  return returnedConfiguration;
  }
