/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstdlib>
#include <cerrno>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <qapplication.h>
#include <qlistview.h>
#include <qmessagebox.h>
#include <qtoolbutton.h>
#include <qlabel.h>

#ifndef QWS
#include <qfiledialog.h>
#endif

#include <classlib/List.hpp>

#include "ManagerFormImpl.hpp"
#include "ModuleLoadSelectionDialogImpl.hpp"
#include "ManagerResource.hpp"
#include "ModuleManager.hpp"
#include "XmlDbOperation.hpp"
#include "ManagerUtility.hpp"
#include "ProgressDialogImpl.hpp"
#include "ManagerConfigurationManager.hpp"
#include "ConfigFormImpl.hpp"
#include "HelpFormImpl.hpp"

static void setApplicationFont(QApplication *application,
                               ManagerConfiguration *configuration) {

  if (configuration -> applicationFontFamily) {
    QFont font = application -> font();
    font.setFamily(configuration -> applicationFontFamily -> getChars());
    if (configuration -> applicationFontSize != 0) {
      font.setPointSize(configuration -> applicationFontSize);
      }

    application -> setFont(font,true);
    }
  }

void ManagerFormImpl::updateFonts() {
  //
  // Set the new application font, if one is configured
  //
  setApplicationFont(application,configuration);

  //
  // It's possible the user has requested to return to the default font
  //
  if (! configuration -> applicationFontFamily) {
    application -> setFont(systemDefaultFont,true);
    }
  }

static bool setupDatabasePath(QWidget *parent) {
  //
  // Check whether the application group directory exists
  //
  struct stat status;
  if (::stat(ManagerResource::getApplicationGroupDirUser(),&status) == -1) {
    //
    // The directory does not exist, or there is an access problem
    //
    if (errno != ENOENT) {
      QMessageBox::critical(parent,ManagerResource::applicationTitle(),
                            "Cannot access the directory\n"
                            "which will hold the SMI\n"
                            "database! Aborting.",
                            QMessageBox::Ok,0);
      return false;
      }

    //
    // The directory does not exist, so create it
    //
    if (mkdir(ManagerResource::getApplicationGroupDirUser(),0755) == -1) {
      QMessageBox::critical(parent,ManagerResource::applicationTitle(),
                            "Cannot create a directory\n"
                            "to hold the SMI\n"
                            "database! Aborting.",
                            QMessageBox::Ok,0);
      return false;
      }
    }

  //
  // The database directory now exists
  //
  return true;
  }

void ManagerFormImpl::readConfiguration() {
  ManagerConfigurationManager configurationManager(this);
  configurationManager.load();
  }

void ManagerFormImpl::writeConfiguration() {
  ManagerConfigurationManager configurationManager(this);
  //setStatusText("Saving..");
  configurationManager.save();
  //clearStatusText();
  }

ManagerFormImpl::ManagerFormImpl(QApplication *application)
    : ManagerForm(), application(application), configuration(0),
      moduleManager(0) {

  //
  // Remember the application default font in case we need to return to it
  //
  systemDefaultFont = application -> font();

  //
  // Read the user preferences
  //
  readConfiguration();

  //
  // Set the default font for the application
  //
  setApplicationFont(application,configuration);

  if (! setupDatabasePath(this)) {
    std::exit(1);
    }

  moduleManager = new ModuleManager();
  QString databasePath = ManagerResource::getApplicationGroupDirUser() +
                         ManagerResource::databaseFileName();

  moduleManager -> openDb(databasePath);
  if (moduleManager -> exception) {
    QMessageBox::critical(0,ManagerResource::applicationTitle(),
                          ManagerUtility::formatErrorText(
                                 "Cannot access the SMI database.",
                                 moduleManager -> exception -> getMessage(),
                                 moduleManager -> exception -> getDiagnostic(),
                                 "Aborting."),
                            QMessageBox::Ok,0);
    std::exit(1);
    }

  rebuildModuleListView();

  setIconStates();
  }

ManagerFormImpl::~ManagerFormImpl() {
  // no need to delete child widgets, Qt does it all for us

  delete moduleManager;
  }

void ManagerFormImpl::setIconStates() {
  DeleteButton -> setEnabled(ModuleList -> selectedItem() != 0);
  ConfigureButton -> setEnabled(true);
  HelpButton -> setEnabled(true);
  InfoButton -> setEnabled(false);
  }

void ManagerFormImpl::rebuildModuleListView() {
  ModuleList -> clear();
  ListIterator it = moduleManager -> getModuleList() -> iterator();
  while (it.hasNext()) {
    const char *moduleName = (char *)it.next();
    QListViewItem *item = new QListViewItem(ModuleList);
    item -> setText(0,moduleName);
    }

  //
  // Rebuilding the module list cleared the current selection, without
  // resulting in a signal, so we invoke the handler manually (Qt 2.3.x)
  //
  mibSelectionChangedSlot();
  }

void ManagerFormImpl::newSlot() {
#ifdef QWS
  ModuleLoadSelectionDialogImpl selectionDialog(this);
#else
  QFileDialog selectionDialog(QString::null,QString::null,this,0,true);
  selectionDialog.setMode(QFileDialog::ExistingFiles);
#endif

  if (ManagerResource::maximiseDialogs) selectionDialog.showMaximized();
  else selectionDialog.show();

  int code = selectionDialog.exec();
  if (code == 1) {
    QStringList paths = selectionDialog.selectedFiles();
    if (paths.count() != 0) {
      //
      // We have at least one selected file, so convert the Qt StringList
      // into our own List (Qt independence) and load the files on it
      //
      List pathList;
      for (QStringList::ConstIterator it = paths.begin() ;
           it != paths.end() ; ++it) {
        pathList.add((*it).latin1());
        }

      XmlDbOperation *loader = moduleManager -> dbLoader(&pathList);

      ProgressDialogImpl *progressDialog = 0;
      if (loader -> getState() == XmlDbOperation::Running) {
        progressDialog = new ProgressDialogImpl(this);
        progressDialog -> show();
        }
      while (loader -> getState() == XmlDbOperation::Running) {
        progressDialog -> setModule(loader -> getCurrentModuleName());
        progressDialog -> setProgress(loader -> getProgress());
        application -> processEvents();

        loader -> doWork();
        }

      //
      // If all files were loaded successfully, commit to the new
      // database, else revert to the previous state
      //
      moduleManager -> finishedEditing(loader -> getState() ==
                                            XmlDbOperation::Completed);

      if (loader -> exception) {
        QString string = ManagerUtility::formatErrorText(
                            "Module loading error.",
                            loader -> exception -> getMessage(),
                            loader -> exception -> getDiagnostic());

        QMessageBox::warning(this,ManagerResource::applicationTitle(),
                             string,QMessageBox::Ok,0);
        }

      if (moduleManager -> exception) {
        QString string = ManagerUtility::formatErrorText(
                            "Module loading error.",
                            moduleManager -> exception -> getMessage(),
                            moduleManager -> exception -> getDiagnostic());

        QMessageBox::warning(this,ManagerResource::applicationTitle(),
                             string,QMessageBox::Ok,0);
        }

      delete progressDialog;
      delete loader;

      rebuildModuleListView();
      }
    }
  }

void ManagerFormImpl::helpSlot() {
  HelpFormImpl helpForm(this);

  helpForm.VersionLabel -> setText(VERSION_STRING);

  if (ManagerResource::maximiseDialogs) helpForm.showMaximized();
  else helpForm.show();

  (void)helpForm.exec();
  }

void ManagerFormImpl::configureSlot() {
  ConfigFormImpl configForm(this,configuration);

  if (ManagerResource::maximiseDialogs) configForm.showMaximized();
  else configForm.show();
  int code = configForm.exec();
  if (code == 1) {
    //
    // Get the updated configuration from the dialog
    //
    delete configuration;
    configuration = configForm.takeNewConfiguration();

    //
    // Update the configuration in persistent storage
    //
    writeConfiguration();

    //
    // Update the visible widgets with any changes to the font configuration
    //
    updateFonts();
    }
  }

void ManagerFormImpl::mibSelectionChangedSlot() {
  setIconStates();
  }

void ManagerFormImpl::deleteSelectedSlot() {
  //
  // Confirm the deletion
  //
  int code = QMessageBox::warning(this,
                                  ManagerResource::applicationTitle(),
                                  "Please confirm deletion\n"
                                  "of the selected module.",
                                  QMessageBox::Ok,
                                  QMessageBox::Cancel |
                                        QMessageBox::Default |
                                        QMessageBox::Escape);
  if (code == QMessageBox::Cancel) {
    return;
    }

  QListViewItem *moduleListViewItem = ModuleList -> selectedItem();
  const char *moduleName = moduleListViewItem -> text(0);

  XmlDbOperation *unloader = moduleManager -> dbUnloader(moduleName);
  while (unloader -> getState() == XmlDbOperation::Running) {
    unloader -> doWork();
    application -> processEvents();
    }
  if (unloader -> exception) {
    QString string = ManagerUtility::formatErrorText(
                            "Module unloading error.",
                            unloader -> exception -> getMessage(),
                            unloader -> exception -> getDiagnostic());

    QMessageBox::warning(this,ManagerResource::applicationTitle(),
                         string,QMessageBox::Ok,0);
    }
  else delete moduleListViewItem;

  delete unloader;
  }

void ManagerFormImpl::trashDatabaseSlot() {
  //
  // Confirm the deletion
  //
  int code = QMessageBox::warning(this,
                                  ManagerResource::applicationTitle(),
                                  "Please confirm deletion\n"
                                  "of ALL modules.",
                                  QMessageBox::Ok,
                                  QMessageBox::Cancel |
                                        QMessageBox::Default |
                                        QMessageBox::Escape);
  if (code == QMessageBox::Cancel) {
    return;
    }

  moduleManager -> deleteDatabase();
  if (moduleManager -> exception) {
    QString string = ManagerUtility::formatErrorText(
                            "Failed to delete the database.",
                            moduleManager -> exception -> getMessage(),
                            moduleManager -> exception -> getDiagnostic());

    QMessageBox::warning(this,ManagerResource::applicationTitle(),
                         string,QMessageBox::Ok,0);
    }

  QString databasePath = ManagerResource::getApplicationGroupDirUser() +
                         ManagerResource::databaseFileName();
  moduleManager -> openDb(databasePath);

  rebuildModuleListView();
  }

void ManagerFormImpl::showInfoOnSelectedSlot() {
  }
