/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <cctype>

#include "ModuleManager.hpp"

#include "XmlLoadOperation.hpp"
#include "XmlUnloadOperation.hpp"
#include "SmiLoadOperation.hpp"
#include "FileLoadOperation.hpp"

#include <classlib/List.hpp>
#include <classlib/Xml.hpp>
#include <classlib/StringBuffer.hpp>
#include <classlib/ObjectHashMap.hpp>
#include <classlib/String.hpp>

#include <smidb/SmiDb.hpp>

class ModuleInfo : public Object {
  public:
    ModuleInfo() {
      oidMap = new ObjectHashMap();

#ifdef CHECK_MAPPING_DUPLICATES
      reverseOidMap = new ObjectHashMap();
#endif
      }

    ~ModuleInfo() {
      }

    ObjectHashMap *oidMap;

#ifdef CHECK_MAPPING_DUPLICATES
    ObjectHashMap *reverseOidMap; 
#endif
  };

class ModuleManagerPrivate {
  public:
    ModuleManagerPrivate() : db(0), dbPath(0), moduleList(0), moduleMap(0) {
      }

    ~ModuleManagerPrivate() {
      delete db;
      delete dbPath;
      delete moduleList;
      delete moduleMap;
      }

    SmiDb *db;
    String *dbPath;
    List *moduleList;
    ObjectHashMap *moduleMap;
  };

ModuleManager::ModuleManager() : data(0) {
  data = new ModuleManagerPrivate();
  }

ModuleManager::~ModuleManager() {
  delete data;
  }

static char *currentTime() {
  return "now";
  }

void ModuleManager::openDb(const char *path) {

  delete data -> moduleMap;
  data -> moduleMap = new ObjectHashMap();

  delete data -> moduleList;
  data -> moduleList = new List();

  delete data -> db;
  delete data -> dbPath;
  data -> dbPath = new String(path);

  data -> db = new SmiDb(path,Db::Write);
  if (data -> db -> exception) RETHROW(data -> db -> exception,);

  char *xmlData;
  int xmlLength;
  Xml moduleListXml;
  if ((xmlData = (char *)data -> db -> fetch("/MODULES",&xmlLength)) != 0) {
    moduleListXml.loadXmlFromBuffer(xmlData,xmlLength);
    SmiDb::free((unsigned char *)xmlData);

    XmlElement root = moduleListXml.getRootElement();
    XmlElement moduleElement;
    int childCount = moduleListXml.getElementChildCount(root);
    for (int i = 0 ; i < childCount ; i++ ) {
      moduleElement = moduleListXml.getElementChild(root,i);
      addModule(moduleListXml.getElementAttributeValue(moduleElement,0));
      }
    }

  if ((xmlData = (char *)data -> db -> fetch("/OID",&xmlLength)) != 0) {
   Xml xml;
    xml.loadXmlFromBuffer(xmlData,xmlLength);
    if (xml.exception) {
      RETHROW(xml.exception,);
      }

    SmiDb::free((unsigned char *)xmlData);

    XmlElement root = xml.getRootElement();
    int childCount = xml.getElementChildCount(root);
    for (int i = 0 ; i < childCount ; i++ ) {
      XmlElement oidElement = xml.getElementChild(root,i);

      addOidMapping(xml.getElementAttributeValue(oidElement,0),
                    xml.getElementAttributeValue(oidElement,1));
      }
    }
  }

void ModuleManager::addOidMapping(const char *fullOidName,const char *value) {

  char moduleName[64];
  const char *dotPtr = std::strchr(fullOidName,'.');
  int length = dotPtr - fullOidName;
  std::memcpy(moduleName,fullOidName,length);
  moduleName[length] = '\0';
  const char *oidName = dotPtr + 1;

  ObjectMapping moduleMapping = data -> moduleMap -> find(moduleName);
  if (! moduleMapping.value) {
    char errBuffer[512];
    std::sprintf(errBuffer,"Reference to non-loaded module %s",
                 moduleName);
    THROW(new XmlDbOperationException(errBuffer),);
    }

  ModuleInfo *moduleInfo = (ModuleInfo *)moduleMapping.value;

#ifdef CHECK_MAPPING_DUPLICATES
  ObjectMapping mapping = moduleInfo -> reverseOidMap -> find(value);
  if (mapping.value) {
    printf("OID mapping value %s already exists (%s vs %s)\n",value,
           ((String *)mapping.value) -> getChars(),name);
    }

  mapping = moduleInfo -> oidMap -> find(name);
  if (mapping.value) {
    printf("OID mapping name %s already exists (%s vs %s)\n",name,
           ((String *)mapping.value) -> getChars(),value);
    }

  moduleInfo -> reverseOidMap -> add(value,fullOidName);

#endif

  moduleInfo -> oidMap -> add(oidName,value);
  }

void ModuleManager::deleteDatabase() {

  data -> db -> deleteDatabase();
  if (data -> db -> exception) {
    RETHROW(data -> db -> exception,);
    }

  delete data -> db;
  data -> db = 0;
  }

XmlDbOperation *ModuleManager::dbLoader(const char *path) {
  XmlDbLoadOperation *op = new FileLoadOperation(this);
  op -> initialise(path);
  return op;
  }

XmlDbOperation *ModuleManager::dbLoader(List *pathList) {
  XmlDbLoadOperation *op = new FileLoadOperation(this);
  op -> initialise(pathList);
  return op;
  }

XmlDbOperation *ModuleManager::dbUnloader(const char *moduleName) {
  XmlUnloadOperation *op = new XmlUnloadOperation(this);
  op -> initialise(moduleName);
  return op;
  }

Db* ModuleManager::getDb() const {
  return data -> db;
  }

List *ModuleManager::getModuleList() const {
  return data -> moduleList;
  }

void ModuleManager::addModule(const char *moduleName) {
  ModuleInfo *moduleInfo = new ModuleInfo();
  String *savedModuleName = new String(moduleName);
  data -> moduleMap -> add(savedModuleName,moduleInfo);
  data -> moduleList -> add(savedModuleName -> getChars());
  }

void ModuleManager::deleteModule(const char *moduleName) {

  //
  // Delete from the module list first, since the name pointers are into the
  // module map
  //
  ListIterator it = data -> moduleList -> iterator();
  while (it.hasNext()) {
    if (std::strcmp(moduleName,(char *)it.next()) == 0) {
      it.remove();
      break;
      }
    }

  //
  // Remove the module from the module map, which also deleted all OIDs
  // defined by the module
  //
  data -> moduleMap -> remove(moduleName);
  }

void ModuleManager::writeXml(Xml& xml,XmlElement element,const char *key) {
  char *xmlData = xml.xmlString(element);
  data -> db -> store(key,xmlData,std::strlen(xmlData));
  Xml::free(xmlData);

  if (data -> db -> exception) {
    RETHROW(data -> db -> exception,);
    }
  }

void ModuleManager::writeXml(Xml& xml,const char *key) {
  char *xmlData = xml.xmlString();
  data -> db -> store(key,xmlData,std::strlen(xmlData));
  Xml::free(xmlData);

  if (data -> db -> exception) {
    RETHROW(data -> db -> exception,);
    }
  }

void ModuleManager::saveModuleList() {
  Xml moduleListXml;

  XmlElement root = moduleListXml.setRootElement("modules");

  ObjectHashMapIterator it = data -> moduleMap -> iterator();
  while (it.hasNext()) {
    ObjectMapping mapping = it.next();
    //ModuleInfo *moduleInfo = (ModuleInfo *)mapping.value;
    const char *moduleName = ((String *)mapping.name) -> getChars();

    XmlElement moduleElement = moduleListXml.addElement(root,"module");
    moduleListXml.addElementAttribute(moduleElement,"name",moduleName);

    moduleListXml.addElementAttribute(moduleElement,"insertionDate",
                                      currentTime());
    }

  writeXml(moduleListXml,"/MODULES");
  }

void ModuleManager::saveOidMap() {
  StringBuffer buffer;
  buffer.append("<oids>");

  ObjectHashMapIterator moduleIterator = data -> moduleMap -> iterator();
  while (moduleIterator.hasNext()) {
    ObjectMapping moduleMapping = moduleIterator.next();
    ModuleInfo *moduleInfo = (ModuleInfo *)moduleMapping.value;
    const char *moduleName = ((String *)moduleMapping.name) -> getChars();

    ObjectHashMapIterator it = moduleInfo -> oidMap -> iterator();
    while (it.hasNext()) {
      ObjectMapping mapping = it.next();

      buffer.append("<oid name=\"");
      buffer.append(moduleName);
      buffer.append(".");
      buffer.append(((String *)mapping.name) -> getChars());
      buffer.append("\" number=\"");
      buffer.append(((String *)mapping.value) -> getChars());
      buffer.append("\"/>");
      }
    }
  buffer.append("</oids>");

  char *xmlData = buffer.detach();
  data -> db -> store("/OID",xmlData,std::strlen(xmlData));
  StringBuffer::free(xmlData);

  if (data -> db -> exception) {
    RETHROW(data -> db -> exception,);
    }
  }

void ModuleManager::replaceOidValue(ModuleInfo *moduleInfo,
                                    ObjectHashMapIterator it,
                                    ObjectMapping mapping,
                                    const char *newValue) {
#ifdef CHECK_MAPPING_DUPLICATES
  moduleInfo -> reverseOidMap ->
                    remove(((String *)mapping.value) -> getChars());

  moduleInfo -> reverseOidMap ->
                    add(newValue,((String *)mapping.name) -> getChars());
#endif

  moduleInfo -> oidMap -> replaceValue(it,new String(newValue));
  }

//
// Return the fully numeric form of the given OID value
//
const char *ModuleManager::resolveOid(const char *value) {
  char *resolveBuffer = new char[256];
  char *bp = resolveBuffer;
  char sidBuffer[128];
  char *sidPtr;
  const char *vp;

  *bp = '\0';
  vp = value;
  while (*vp) {
    //
    // Copy the next sub-identifier to the sub-identifier buffer
    //
    sidPtr = sidBuffer;
    while (*vp && (*vp != ' ')) {
      *sidPtr++ = *vp++ ;
      }
    *sidPtr = '\0';

    //
    // Resolve the sub-identifier to the resolve buffer
    //
    if (bp != resolveBuffer) *bp++ = '.';
    if (isdigit(sidBuffer[0])) std::strcpy(bp,sidBuffer);
    else if (sidBuffer[0] == '.') {
      if (std::strcmp(sidBuffer,".iso") == 0) {
        std::strcpy(bp,"1");
        }
      else {
        char errBuffer[512];
        std::sprintf(errBuffer,"Cannot resolve %s to a numeric OID\n",
                     sidBuffer);
        THROW(new XmlDbOperationException(errBuffer),0);
        }
      }
    else {
      //
      // Get the module name of the symbol being referred to
      //
      char moduleName[64];
      const char *dotPtr = std::strchr(sidBuffer,'.');
      int length = dotPtr - sidBuffer;
      std::memcpy(moduleName,sidBuffer,length);
      moduleName[length] = '\0';
      const char *oidName = dotPtr + 1;

      //
      // Get the module information for this module
      //
      ObjectMapping moduleMapping = data -> moduleMap -> find(moduleName);
      if (! moduleMapping.value) {
        char errBuffer[512];
        std::sprintf(errBuffer,"Cannot resolve %s to a numeric OID\n",
                     sidBuffer);
        THROW(new XmlDbOperationException(errBuffer),0);
        }
      ModuleInfo *moduleInfo = (ModuleInfo *)moduleMapping.value;

      //
      // Find the symbol in the OID map for the module
      //
      ObjectMapping sidMapping = moduleInfo -> oidMap -> find(oidName);
      if (! sidMapping.value) {
        char errBuffer[512];
        std::sprintf(errBuffer,"Cannot resolve %s to a numeric OID\n",
                     sidBuffer);
        THROW(new XmlDbOperationException(errBuffer),0);
        }
      else {
        //
        // The sub-identifier has resolved to a value, but this value
        // may not be fully numeric so we resolve it
        //
        const char *mappingValue = ((String *)sidMapping.value) -> getChars();
        if (std::strspn(mappingValue,"0123456789.") !=
            std::strlen(mappingValue)) {
          const char *rv = resolveOid(mappingValue);
          if (exception) return 0;

          std::strcpy(bp,rv);
          delete[] rv;
          }
        else std::strcpy(bp,mappingValue);
        }
      }
    bp += std::strlen(bp);

    //
    // Prepare to work on the next word
    //
    if (*vp) vp++ ;
    }

  return resolveBuffer;
  }

//
// This method is called after loading one or more modules. Its job
// is to scan the OID maps and convert values into the simplest form
// possible (numeric)
//
void ModuleManager::resolveOids() {
  //
  // Iterator through the module map
  //
  ObjectHashMapIterator moduleIterator = data -> moduleMap -> iterator();
  while (moduleIterator.hasNext()) {
    ObjectMapping moduleMapping = moduleIterator.next();
    ModuleInfo *moduleInfo = (ModuleInfo *)moduleMapping.value;
    //const char *moduleName = ((String *)moduleMapping.name) -> getChars();

    //
    // Now iterator through the OID map for this module
    //
    ObjectHashMapIterator it = moduleInfo -> oidMap -> iterator();
    while (it.hasNext()) {
      ObjectMapping mapping = it.next();

      const char *mappingValue = ((String *)mapping.value) -> getChars();
      if (std::strspn(mappingValue,"0123456789.") !=
          std::strlen(mappingValue)) {

        const char *newValue = resolveOid(mappingValue);
        if (exception) return;

        replaceOidValue(moduleInfo,it,mapping,newValue);

        delete[] newValue;
        }
      }
    }
  }

void ModuleManager::finishedEditing(bool ok) {
  if (ok) {
    resolveOids();
    if (! exception) {
      saveModuleList();
      saveOidMap();
      }

    // TBD - remove undo information
    }
  else {
    // TBD - use undo information
    }
  }
