/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <cctype>

#include "SmiFileReader.hpp"

#include <classlib/String.hpp>
#include <classlib/DetailedException.hpp>

class SmiFileReaderPrivate {
  public:
    SmiFileReaderPrivate() : tokenStartLineNumber(0) {
      }

    off_t fileSize;
    int tokenStartLineNumber;
    SmiToken putbackList;
  };

SmiFileReader::SmiFileReader(const char *fileNameArg)
         : LineNumberReader(fileNameArg) {
  data = new SmiFileReaderPrivate();

  struct stat fileInfo;
  if (::stat(fileNameArg,&fileInfo) == 0) {
    data -> fileSize = fileInfo.st_size;
    }
  }

SmiFileReader::~SmiFileReader() {
  delete data;
  }

bool SmiFileReader::isInitialIdentifierCharacter(char ch) const {
  return ((ch >= 'a') && (ch <= 'z')) ||
          ((ch >= 'A') && (ch <= 'Z'));
  }

bool SmiFileReader::isIdentifierCharacter(char ch) const {
  //
  // Hyphen is allowed for backwards-compatibility. In the
  // future, hyphen will be used for subtraction
  //
  return ((ch >= 'a') && (ch <= 'z')) ||
         ((ch >= 'A') && (ch <= 'Z')) ||
         ((ch >= '0') && (ch <= '9')) ||
         (ch == '-');
  }

int SmiFileReader::getTokenStartLineNumber() const {
  return data -> tokenStartLineNumber;
  }

void SmiFileReader::putbackToken(SmiToken token) {
  if (data -> putbackList.getTokenType() != SmiToken::TT_NULL) {
    THROW(new DetailedException("Token putback depth exceeded!"),);
    }

  data -> putbackList = token;
  }

SmiToken SmiFileReader::nextToken() {
  //
  // If there is a putback token, return it
  //
  if (data -> putbackList.getTokenType() != SmiToken::TT_NULL) {
    SmiToken token = data -> putbackList;
    data -> putbackList = SmiToken();
    return token;
    }

  //
  // First, skip leading whitespace until we get to the next
  // interesting character, if there is one
  //
  int ch = read();
  while ((ch != -1) && (isspace((char)ch))) {
    ch = read();
    }

  //
  // Remember the line number this token starts on
  //
  data -> tokenStartLineNumber = getLineNumber();

  //
  // Let's get end-of-stream out of the way
  //
  if (ch == -1) return SmiToken(SmiToken::TT_EOF,data -> tokenStartLineNumber);

  if (ch == '-') {
    //
    // Could be the start of a comment, or the start of a negative
    // number (unary minus) so we mark the stream before reading
    // the next character
    //
    mark(0);
    ch = read();
    if (ch == -1) {
      THROW(new DetailedException("Premature end of file"),SmiToken());
      }

    if (ch == '-') {
      //
      // Start of comment, so we need to read all characters up
      // to the end of comment or the end of the line, and return
      // them as the token
      //
      StringBuffer buffer;
      ch = read();
      if (ch != '\n') {
        int ch2;
        ch2 = read();
        while (((ch != '-') || (ch2 != '-')) && (ch != '\n')) {
          if ((ch == -1) || (ch2 == -1)) {
            //
            // End of file in the middle of a comment
            //
            THROW(new DetailedException("Premature end of file"),SmiToken());
            }

          //
          // Append the first character to the buffer and read the
          // next lucky contestant
          //
          buffer.append((char)ch);
          ch = ch2;
          if (ch != '\n') ch2 = read();
          }
        }

      //
      // Reached the end of the comment so return a token for it
      //
      return SmiToken(SmiToken::TT_COMMENT,data -> tokenStartLineNumber,
                      buffer.getChars());
      }

    //
    // Unary minus, so un-read the determining character and
    // fall through (where we'll treat it as unary minus, or
    // as the start of a negative number)
    //
    reset();
    ch = '-';
    }

  //
  // See if we are looking at an identifier or keyword
  //
  if (isInitialIdentifierCharacter((char)ch)) {
    mark(0);
    StringBuffer buffer;
    while (isIdentifierCharacter((char)ch)) {
      buffer.append((char)ch);
      mark(0);
      ch = read();
      }

    reset();
    return SmiToken(SmiToken::TT_IDENTIFIER,
                    data -> tokenStartLineNumber,buffer.getChars());
    }

  //
  // Maybe it's a number?
  //
  if (isdigit((char)ch) || ch == '-') {
    StringBuffer buffer;
    buffer.append((char)ch);
    mark(0);
    ch = read();
    while (isdigit((char)ch)) {
      buffer.append((char)ch);
      mark(0);
      ch = read();
      }

    reset();
    return SmiToken(SmiToken::TT_NUMBER,
                    data -> tokenStartLineNumber,buffer.getChars());
    }

  //
  // See if it is a quoted string
  //
  if (ch == '"') {
    StringBuffer buffer;
    int indent = -1;
    ch = read();
    while (ch != '"') {
      if (ch == -1)
        THROW(new DetailedException("Premature end of file"),SmiToken());

      //
      // See if we are measuring or stripping leading spaces
      //
      if (ch == '\n') {
        //
        // Append the newline, since we need it in the token anyway
        //
        buffer.append((char)ch);
        if (indent == -1) {
          //
          // We have not yet measured the indent, so measure it now
          //
          ch = read();
          indent = 0;
          while (ch == ' ') {
            indent++ ;
            ch = read();
            }

          //
          // If the first non-space character on the line is the end
          // of line, we throw away the spaces and the indent
          // measurement
          //
          if (ch == '\n') {
            indent = -1;
            continue;
            }
          }
        else {
          //
          // Skip the indent
          //
          ch = read();
          int i = 0;
          while ((i < indent) && (ch == ' ')) {
            i++ ;
            ch = read();
            }
          if (ch == -1)
            THROW(new DetailedException("Premature end of file"),SmiToken());

          //
          // If we found a non-space before skipping all the expected
          // spaces, and this is not a blank line, reset the indent
          //
          if ((i != indent) && (ch != '\n')) {
            indent = i;
            }

          //
          // If this was a blank line, we need to continue skipping
          // leading spaces
          //
          if (ch == '\n') continue;
          }

        //
        // If we skipped to a (closing) quote, we need to break out of the
        // loop
        //
        if (ch == '"') break;
        }

      buffer.append((char)ch);
      ch = read();
      }

    return SmiToken(SmiToken::TT_STRING,
                    data -> tokenStartLineNumber,buffer.getChars());
    }

  //
  // See if it is a digit string (binary or hexadecimal)
  //
  if (ch == '\'') {
    StringBuffer buffer;
    ch = read();
    while (ch != '\'') {
      if (ch == -1)
        THROW(new DetailedException("Premature end of file"),SmiToken());
      buffer.append((char)ch);
      ch = read();
      }

    ch = read();
    if ((ch == 'b') || (ch == 'B')) {
      return SmiToken(SmiToken::TT_BSTRING,
                      data -> tokenStartLineNumber,buffer.getChars());
      }
    else if ((ch == 'h') || (ch == 'H')) {
      return SmiToken(SmiToken::TT_HSTRING,data -> tokenStartLineNumber,
                      buffer.getChars());
      }
    else {
      THROW(new DetailedException("Unrecognised digit string"),SmiToken());
      }
    }

  //
  // Miscellaneous single-character symbols
  //
  if ((ch == '{') || (ch == '}') || (ch == '(') || (ch == ')') ||
      (ch == '|') || (ch == '[') || (ch == ']') || (ch == ';') ||
      (ch == ',')) {
    char c[] = {ch, 0};
    return SmiToken(SmiToken::TT_SYMBOL,data -> tokenStartLineNumber,c);
    }

  //
  // Might be "." or ".."
  //
  if (ch == '.') {
    mark(0);
    ch = read();
    if (ch == -1)
      THROW(new DetailedException("Premature end of file"),SmiToken());

    if (ch != '.') {
      reset();
      return SmiToken(SmiToken::TT_SYMBOL,data -> tokenStartLineNumber,".");
      }

    return SmiToken(SmiToken::TT_SYMBOL,data -> tokenStartLineNumber,"..");
    }

  //
  // Now we're into multiple-character symbols
  //
  if (ch == ':') {
    ch = read();
    if (ch == -1)
      THROW(new DetailedException("Premature end of file"),SmiToken());

    if (ch != ':') {
      THROW(new DetailedException("Unrecognised token"),SmiToken());
      }
    ch = read();
    if (ch == -1)
      THROW(new DetailedException("Premature end of file"),SmiToken());
    if (ch != '=') {
      THROW(new DetailedException("Unrecognised token"),SmiToken());
      }

    return SmiToken(SmiToken::TT_SYMBOL,data -> tokenStartLineNumber,"::=");
    }

  // ???
  THROW(new DetailedException("Unrecognised token"),SmiToken());
  }

int SmiFileReader::getProgress() const {
  off_t pos = ::lseek(fd,0,SEEK_CUR);
  return (int)(((float)pos / data -> fileSize) * 100);
  }
