/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstring>
#include <cctype>
#include <cstdlib>

#include <classlib/Xml.hpp>
#include <classlib/String.hpp>
#include <classlib/ObjectList.hpp>
#include <classlib/ObjectHashMap.hpp>

#include <smidb/SmiObjectIdentity.hpp>
#include <smidb/SmiIndexList.hpp>
#include <smidb/SmiIndexItem.hpp>
#include <smidb/SmiValueReference.hpp>
#include <smidb/SmiSyntax.hpp>
#include <smidb/SmiSizeList.hpp>
#include <smidb/SmiRangeList.hpp>
#include <smidb/SmiOidValue.hpp>
#include <smidb/SmiSubIdentifier.hpp>
#include <smidb/SmiDefaultValue.hpp>
#include <smidb/SmiIdentifierList.hpp>

#include "SmiLoadOperation.hpp"
#include "XmlUnloadOperation.hpp"
#include "ModuleManager.hpp"
#include "SmiFileReader.hpp"

#include "SmiMibModule.hpp"
#include "SmiImportsFrom.hpp"

class SmiLoadOperationPrivate {
  public:
    enum SubState { Global, Definitions };

    SmiLoadOperationPrivate() : subState(Global), currentModule(0), reader(0),
                                currentModuleImportsMap(0),
                                currentModuleAssignmentsXml(0),
                                currentModuleTypedefsXml(0) {
      }
    ~SmiLoadOperationPrivate() {
      delete currentModule;
      delete reader;
      delete currentModuleImportsMap;
      delete currentModuleAssignmentsXml;
      delete currentModuleTypedefsXml;
      }

    SubState subState;

    SmiMibModule *currentModule;
    SmiFileReader *reader;
    ObjectHashMap *currentModuleImportsMap;
    Xml *currentModuleAssignmentsXml;
    XmlElement currentModuleAssignmentsRoot;
    Xml *currentModuleTypedefsXml;
    XmlElement currentModuleTypedefsRoot;
  };

class SmiLoadOperationException : public DetailedException {
  public:
    SmiLoadOperationException(SmiLoadOperation *operation,
                              const SmiToken *token,const char *message)
               : DetailedException(message) {
      if (token) {
        char diagnostic[512];
        std::sprintf(diagnostic,"near '%s' on line %d",
                     token -> getTokenValue(),
                     token -> getSourceLineNumber());
        setDiagnostic(diagnostic);
        }
      else {
        char diagnostic[512];
        std::sprintf(diagnostic,"near line %d",
                     operation -> data -> reader -> getTokenStartLineNumber());
        setDiagnostic(diagnostic);
        }
      }
    ~SmiLoadOperationException() {}
  };

SmiLoadOperation::SmiLoadOperation(ModuleManager *manager)
          : XmlDbLoadOperation(manager), data(0) {
  data = new SmiLoadOperationPrivate();
  }

SmiLoadOperation::~SmiLoadOperation() {
  delete data;
  }

void SmiLoadOperation::initialise(const char *path) {
  XmlDbOperation::initialise();

  data -> reader = new SmiFileReader(path);
  if (data -> reader -> exception) {
    state = Error;
    RETHROW(data -> reader -> exception,);
    }
  }

const char *SmiLoadOperation::buildOidString(SmiOidValue *oidValue) {
  static char oidString[512];
  char *oidStringPtr = oidString;

  *oidStringPtr = '\0';
  bool first = true;
  ListIterator it = oidValue -> iterator();
  while (it.hasNext()) {
    SmiSubIdentifier *subIdentifier = (SmiSubIdentifier *)it.next();

    if (! first) *oidStringPtr++ = ' ';
    first = false;

    if (subIdentifier -> getSidPresent()) {
      std::sprintf(oidStringPtr,"%lu",subIdentifier -> getSid());
      }
    else if (subIdentifier -> getModule()) {
      std::sprintf(oidStringPtr,"%s.%s",subIdentifier -> getModule(),
                                        subIdentifier -> getName());
      }
    else {
      const char *name = subIdentifier -> getName();
      ObjectMapping moduleMapping =
                         data -> currentModuleImportsMap -> find(name);
      if (moduleMapping.name) {
        std::sprintf(oidStringPtr,"%s.%s",
                     ((String *)moduleMapping.value) -> getChars(),name);
        }
      else if (std::strcmp(name,"iso") == 0) {
        std::sprintf(oidStringPtr,".%s",name);
        }
      else {
        //
        // The name is local (or a mistake) so use the current module
        //
        std::sprintf(oidStringPtr,"%s.%s",
                     data -> currentModule -> getName(),name);
        }
      }

    oidStringPtr += std::strlen(oidStringPtr);
    }

  return oidString;
  }

//
// Read the next token from the reader, skipping comments (later,
// we'll start collecting comments, and make them available)
//
SmiToken SmiLoadOperation::nextToken() {
  SmiToken token = data -> reader -> nextToken();
  while (token.getTokenType() == SmiToken::TT_COMMENT) {
    token = data -> reader -> nextToken();
    }

  return token;
  }

//
// Read the next token, throwing an exception if it is not a word
// (identifier, symbol, number)
//
SmiToken SmiLoadOperation::readWordToken() {
  SmiToken::Type tokenType;

  SmiToken token = nextToken();
  tokenType = token.getTokenType();
  if (tokenType == SmiToken::TT_EOF) {
    state = Error;
    exception = new SmiLoadOperationException(this,0,"Premature end of file");
    THROW(exception,SmiToken());
    }
  if (tokenType == SmiToken::TT_STRING) {
    state = Error;
    exception = new SmiLoadOperationException(this,&token,"Unexpected input");
    THROW(exception,SmiToken());
    }

  return token;
  }

SmiToken SmiLoadOperation::readStringToken() {
  SmiToken::Type tokenType;

  SmiToken token = nextToken();
  tokenType = token.getTokenType();
  if (tokenType == SmiToken::TT_EOF) {
    state = Error;
    exception = new SmiLoadOperationException(this,0,"Premature end of file");
    THROW(exception,SmiToken());
    }
  if (tokenType != SmiToken::TT_STRING) {
    state = Error;
    exception = new SmiLoadOperationException(this,&token,"Unexpected input");
    THROW(exception,SmiToken());
    }

  return token;
  }

void SmiLoadOperation::parseImports() {

  SmiToken token = readWordToken();
  if (exception) return;
  while (std::strcmp(token.getTokenValue(),";") != 0) {
    SmiIdentifierList *importList = new SmiIdentifierList;
    while (std::strcmp(token.getTokenValue(),"FROM") != 0) {
      importList -> add(token.getTokenValue());

      token = readWordToken();
      if (exception) {
        delete importList;
        return;
        }
      if (std::strcmp(token.getTokenValue(),",") == 0) {
        token = readWordToken();
        if (exception) {
          delete importList;
          return;
          }
        }
      }

    //
    // Read the name of the module these symbols are imported from
    //
    token = readWordToken();
    if (exception) {
      delete importList;
      return;
      }

    ListIterator it = importList -> iterator();
    while (it.hasNext()) {
      String *import = (String *)it.next();

      data -> currentModuleImportsMap -> add(import -> getChars(),
                                             token.getTokenValue());
      }
    delete importList;

    token = readWordToken();
    if (exception) return;
    }
  }

void SmiLoadOperation::parseExports() {
  SmiToken token = readWordToken();
  if (exception) return;
  //ObjectList *exportList = new ObjectList();
  while (std::strcmp(token.getTokenValue(),";") != 0) {
    //exportList -> add(new String(token.getTokenValue()));

    token = readWordToken();
    if (exception) {
      //delete exportList;
      return;
      }
    if (std::strcmp(token.getTokenValue(),",") == 0) {
      token = readWordToken();
      if (exception) {
        //delete exportList;
        return;
        }
      }
    }

  //data -> currentModule -> addExportList(exportList);
  }

SmiOidValue *SmiLoadOperation::parseOidValue() {
  SmiToken token = readWordToken();
  if (exception) return 0;
  if (std::strcmp(token.getTokenValue(),"{") != 0) {
    exception = new SmiLoadOperationException(this,&token,"Expected {");
    THROW(exception,0);
    }

  token = readWordToken();
  if (exception) return 0;
  SmiOidValue *oidValue = new SmiOidValue();
  while (std::strcmp(token.getTokenValue(),"}") != 0) {
    SmiSubIdentifier *subIdentifier = 0;
    SmiToken nextToken = readWordToken();
    if (exception) {
      delete oidValue;
      return 0;
      }
    if (std::strcmp(nextToken.getTokenValue(),"(") == 0) {
      nextToken = readWordToken();
      if (exception) {
        delete oidValue;
        return 0;
        }
      subIdentifier =
            new SmiSubIdentifier(0,token.getTokenValue(),
                                 std::atoi(nextToken.getTokenValue()));
      nextToken = readWordToken();
      if (exception) {
        delete oidValue;
        return 0;
        }
      if (std::strcmp(nextToken.getTokenValue(),")") != 0) {
        exception = new SmiLoadOperationException(this,&nextToken,"Expected )");
        delete oidValue;
        return 0;
        }

      token = readWordToken();
      }
    else if (std::strcmp(token.getTokenValue(),".") == 0) {
      nextToken = readWordToken();
      if (exception) {
        delete oidValue;
        return 0;
        }
      subIdentifier = new SmiSubIdentifier(token.getTokenValue(),
                                           nextToken.getTokenValue());
      token = readWordToken();
      }
    else {
      subIdentifier = new SmiSubIdentifier(token.getTokenValue());
      token = nextToken;
      }

    if (exception) {
      delete oidValue;
      return 0;
      }

    oidValue -> appendSubIdentifier(subIdentifier);
    }

  return oidValue;
  }

SmiValueReference *SmiLoadOperation::parseValueReference() {
  SmiToken token = readWordToken();
  if (exception) return 0;

  if (std::strcmp(token.getTokenValue(),"{") == 0) {
    data -> reader -> putbackToken(token);
    return new SmiValueReference(parseOidValue());
    }
  else {
    SmiToken second = readWordToken();
    if (exception) return 0;
    if (std::strcmp(second.getTokenValue(),".") == 0) {
      second = readWordToken();
      if (exception) return 0;
      return new SmiValueReference(token.getTokenValue(),
                                   second.getTokenValue());
      }
    else {
      data -> reader -> putbackToken(second);
      return new SmiValueReference(token.getTokenValue());
      }
    }
  }

ObjectList *SmiLoadOperation::parseValueReferenceList() {
  SmiToken token = readWordToken();
  if (exception) return 0;

  if (std::strcmp(token.getTokenValue(),"{") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected {"),0);

  token = readWordToken();
  if (exception) return 0;
  ObjectList *valueReferenceList = new ObjectList;
  while (std::strcmp(token.getTokenValue(),"}") != 0) {
    data -> reader -> putbackToken(token);
    SmiValueReference *valueReference = parseValueReference();
    if (exception) {
      delete valueReferenceList;
      return 0;
      }

    valueReferenceList -> add(valueReference);

    token = readWordToken();
    if (exception) return 0;
    if (std::strcmp(token.getTokenValue(),",") == 0) {
      token = readWordToken();
      if (exception) {
        delete valueReferenceList;
        return 0;
        }
      }
    }

  return valueReferenceList;
  }

SmiIndexList *SmiLoadOperation::parseIndexList() {
  SmiToken token = readWordToken();
  if (exception) return 0;

  if (std::strcmp(token.getTokenValue(),"{") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected {"),0);

  token = readWordToken();
  if (exception) return 0;
  SmiIndexList *indices = new SmiIndexList();
  while (std::strcmp(token.getTokenValue(),"}") != 0) {
    if (std::strcmp(token.getTokenValue(),"IMPLIED") == 0) {
      indices -> setImplied(true);
      token = readWordToken();
      if (exception) {
        delete indices;
        return 0;
        }
      }
    else if (std::strcmp(token.getTokenValue(),"{") == 0) {
      data -> reader -> putbackToken(token);
      indices -> add(parseOidValue());
      if (exception) {
        delete indices;
        return 0;
        }
      }
    else {
      SmiToken second = readWordToken();
      if (exception) {
        delete indices;
        return 0;
        }
      if (std::strcmp(second.getTokenValue(),".") == 0) {
        second = readWordToken();
        if (exception) {
          delete indices;
          return 0;
          }
        indices -> add(new SmiIndexItem(second.getTokenValue(),
                                        token.getTokenValue()));
        token = readWordToken();
        if (exception) {
          delete indices;
          return 0;
          }
        }
      else {
        indices -> add(new SmiIndexItem(0,token.getTokenValue()));
        token = second;
        }

      if (std::strcmp(token.getTokenValue(),",") == 0) {
        token = readWordToken();
        if (exception) {
          delete indices;
          return 0;
          }
        }
      }
    }

  return indices;
  }

SmiIdentifierList *SmiLoadOperation::parseIdentifierList() {
  SmiToken token = readWordToken();
  if (exception) return 0;

  if (std::strcmp(token.getTokenValue(),"{") != 0) {
    THROW(new SmiLoadOperationException(this,&token,"Expected {"),0);
    }

  token = readWordToken();
  if (exception) return 0;
  SmiIdentifierList *idlist = new SmiIdentifierList();
  while (std::strcmp(token.getTokenValue(),"}") != 0) {
    idlist -> add(token.getTokenValue());

    token = readWordToken();
    if (exception) {
      delete idlist;
      return 0;
      }
    if (std::strcmp(token.getTokenValue(),",") == 0) {
      token = readWordToken();
      if (exception) {
        delete idlist;
        return 0;
        }
      }
    }

  return idlist;
  }

SmiRangeList *SmiLoadOperation::parseRangeList() {
  SmiToken token = readWordToken();
  if (exception) return 0;

  SmiRangeList *rangeList = new SmiRangeList();
  while (std::strcmp(token.getTokenValue(),")") != 0) {
    long long value1 = token.getBigIntegerValue();
    if (exception) {
      delete rangeList;
      return 0;
      }

    token = readWordToken();
    if (exception) {
      delete rangeList;
      return 0;
      }
    if (std::strcmp(token.getTokenValue(),"..") == 0) {
      //
      // We have upper and lower bounds
      //
      token = readWordToken();
      if (exception) {
        delete rangeList;
        return 0;
        }
      long long value2 = token.getBigIntegerValue();
      if (exception) {
        delete rangeList;
        return 0;
        }
      SmiRange *range = new SmiRange(value1,value2);
      rangeList -> add(range);

      token = readWordToken();
      if (exception) {
        delete rangeList;
        return 0;
        }
      }
    else {
      SmiRange *range = new SmiRange(value1);
      rangeList -> add(range);
      }

    if (std::strcmp(token.getTokenValue(),"|") == 0) {
      token = readWordToken();
      if (exception) {
        delete rangeList;
        return 0;
        }
      }
    }

  return rangeList;
  }

SmiSizeList *SmiLoadOperation::parseSizeList() {
  SmiToken token = readWordToken();
  if (exception) return 0;
  if (std::strcmp(token.getTokenValue(),"(") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected SIZE("),0);

  token = readWordToken();
  if (exception) return 0;

  SmiSizeList *sizeList = new SmiSizeList();
  while (std::strcmp(token.getTokenValue(),")") != 0) {
    long long value1 = token.getBigIntegerValue();
    if (exception) {
      delete sizeList;
      return 0;
      }

    token = readWordToken();
    if (exception) {
      delete sizeList;
      return 0;
      }
    if (std::strcmp(token.getTokenValue(),"..") == 0) {
      //
      // We have upper and lower bounds
      //
      token = readWordToken();
      if (exception) {
        delete sizeList;
        return 0;
        }
      long long value2 = token.getBigIntegerValue();
      if (exception) {
        delete sizeList;
        return 0;
        }
      SmiSize *size = new SmiSize(value1,value2);
      sizeList -> add(size);
      token = readWordToken();
      if (exception) {
        delete sizeList;
        return 0;
        }
      }
    else {
      SmiSize *size = new SmiSize(value1);
      sizeList -> add(size);
      }

    if (std::strcmp(token.getTokenValue(),"|") == 0) {
      token = readWordToken();
      if (exception) {
        delete sizeList;
        return 0;
        }
      }
    }

  return sizeList;
  }

SmiComponentList *SmiLoadOperation::parseSyntaxList() {
  SmiToken token = readWordToken();
  if (exception) return 0;

  if (std::strcmp(token.getTokenValue(),"{") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected {"),0);

  token = readWordToken();
  if (exception) return 0;
  SmiComponentList *syntaxList = new SmiComponentList();
  while (std::strcmp(token.getTokenValue(),"}") != 0) {
    SmiSyntax *syntax = parseSyntax();
    if (exception) {
      delete syntaxList;
      return 0;
      }

    syntax -> setName(token.getTokenValue());
    syntaxList -> add(syntax);

    token = readWordToken();
    if (exception) {
      delete syntaxList;
      return 0;
      }
    if (std::strcmp(token.getTokenValue(),",") == 0) {
      token = readWordToken();
      if (exception) {
        delete syntaxList;
        return 0;
        }
      }
    }

  return syntaxList;
  }

SmiEnumerationList *SmiLoadOperation::parseEnumerationList() {
  SmiToken token = readWordToken();
  if (exception) return 0;

  if (std::strcmp(token.getTokenValue(),"{") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected {"),0);

  token = readWordToken();
  if (exception) return 0;
  SmiEnumerationList *enums = new SmiEnumerationList();
  while (std::strcmp(token.getTokenValue(),"}") != 0) {
    SmiToken second = readWordToken();
    if (exception) {
      delete enums;
      return 0;
      }
    if (std::strcmp(second.getTokenValue(),"(") != 0) {
      delete enums;
      THROW(new SmiLoadOperationException(this,&second,"Expected ("),0);
      }
    second = readWordToken();
    if (exception) {
      delete enums;
      return 0;
      }

    SmiEnumeration *enumeration =
          new SmiEnumeration(token.getTokenValue(),second.getTokenValue());
    enums -> add(enumeration);

    second = readWordToken();
    if (exception) {
      delete enums;
      return 0;
      }
    if (std::strcmp(second.getTokenValue(),")") != 0) {
      delete enums;
      THROW(new SmiLoadOperationException(this,&second,"Expected )"),0);
      }

    token = readWordToken();
    if (exception) {
      delete enums;
      return 0;
      }
    if (std::strcmp(token.getTokenValue(),",") == 0) {
      token = readWordToken();
      if (exception) {
        delete enums;
        return 0;
        }
      }
    }

  return enums;
  }

SmiSyntax *SmiLoadOperation::parseSyntax() {
  SmiToken token = readWordToken();
  if (exception) return 0;

  SmiSyntax *syntax = new SmiSyntax();
  if (std::strcmp(token.getTokenValue(),"[") == 0) {
    token = readWordToken();
    if (exception) {
      delete syntax;
      return 0;
      }

    SmiAsn1Tag *tag = new SmiAsn1Tag(token.getTokenValue());

    token = readWordToken();
    if (exception) {
      delete tag;
      delete syntax;
      return 0;
      }

    tag -> setClass(token.getTokenValue());
    syntax -> setTag(tag);

    token = readWordToken();
    if (exception) {
      delete syntax;
      return 0;
      }
    if (std::strcmp(token.getTokenValue(),"]") != 0) {
      delete syntax;
      THROW(new SmiLoadOperationException(this,&token,"Expected ]"),0);
      }

    token = readWordToken();
    if (exception) {
      delete syntax;
      return 0;
      }
    }

  if (std::strcmp(token.getTokenValue(),"IMPLICIT") == 0) {
    SmiAsn1Tag *tag = syntax -> getTag();
    if (tag == 0) {
      delete syntax;
      THROW(new SmiLoadOperationException(this,&token,
                                "Unexpected IMPLICIT type modifier"),0);
      }

    tag -> setImplicit(true);

    token = readWordToken();
    if (exception) {
      delete syntax;
      return 0;
      }
    }

  if (std::strcmp(token.getTokenValue(),"INTEGER") == 0) {
    syntax -> setTyperef("INTEGER");
    }
  else if (std::strcmp(token.getTokenValue(),"OCTET") == 0) {
    token = readWordToken();
    if (exception) {
      delete syntax;
      return 0;
      }
    if (std::strcmp(token.getTokenValue(),"STRING") != 0) {
      delete syntax;
      THROW(new SmiLoadOperationException(this,&token,
                                          "Expected OCTET STRING"),0);
      }

    syntax -> setTyperef("OCTET STRING");
    }
  else if (strcmp(token.getTokenValue(),"OBJECT") == 0) {
    token = readWordToken();
    if (exception) {
      delete syntax;
      return 0;
      }

    if (std::strcmp(token.getTokenValue(),"IDENTIFIER") != 0) {
      delete syntax;
      THROW(new SmiLoadOperationException(this,&token,
                                          "Expected OBJECT IDENTIFIER"),0);
      }

    syntax -> setTyperef("OBJECT IDENTIFIER");
    }
  else if (std::strcmp(token.getTokenValue(),"SEQUENCE") == 0) {
    token = readWordToken();
    if (exception) {
      delete syntax;
      return 0;
      }

    if (std::strcmp(token.getTokenValue(),"OF") == 0) {
      syntax -> setTyperef("SEQUENCE OF");
      token = readWordToken();
      if (exception) {
        delete syntax;
        return 0;
        }

      syntax -> setSubtype(token.getTokenValue());
      }
    else {
      syntax -> setTyperef("SEQUENCE");
      data -> reader -> putbackToken(token);
      syntax -> setComponentList(parseSyntaxList());
      if (exception) {
        delete syntax;
        return 0;
        }
      }
    }
  else if (std::strcmp(token.getTokenValue(),"CHOICE") == 0) {
    syntax -> setTyperef("CHOICE");
    syntax -> setComponentList(parseSyntaxList());
    if (exception) {
      delete syntax;
      return 0;
      }
    }
  else {
    syntax -> setTyperef(token.getTokenValue());
    }

  token = nextToken();
  if (exception) {
    delete syntax;
    return 0;
    }

  if ((token.getTokenType() == SmiToken::TT_SYMBOL) &&
         ((std::strcmp(token.getTokenValue(),"(") == 0) ||
          (std::strcmp(token.getTokenValue(),"{") == 0))) {
    if (std::strcmp(token.getTokenValue(),"{") == 0) {
      //
      // It's an enumeration list
      //
      data -> reader -> putbackToken(token);
      syntax -> setEnumerationList(parseEnumerationList());
      if (exception) {
        delete syntax;
        return 0;
        }
      }
    else {
      //
      // Could be either a (SIZE (xxx)) or a range (xxx)
      //
      token = readWordToken();
      if (exception) {
        delete syntax;
        return 0;
        }
      if (std::strcmp(token.getTokenValue(),"SIZE") == 0) {
        syntax -> setSizeList(parseSizeList());
        if (exception) {
          delete syntax;
          return 0;
          }

        //
        // Consume the final closing )
        //
        token = readWordToken();
        if (exception) {
          delete syntax;
          return 0;
          }
        if (std::strcmp(token.getTokenValue(),")") != 0) {
          delete syntax;
          THROW(new SmiLoadOperationException(this,&token,
                                              "Expected )"),0);
          }
        }
      else {
        data -> reader -> putbackToken(token);
        syntax -> setRangeList(parseRangeList());
        if (exception) {
          delete syntax;
          return 0;
          }
        }
      }
    }
  else data -> reader -> putbackToken(token);

  return syntax;
  }

SmiDefaultValue *SmiLoadOperation::parseDefaultValue() {
  SmiToken token = readWordToken();
  if (exception) return 0;

  if (std::strcmp(token.getTokenValue(),"{") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected {"),0);

  token = nextToken();
  if (exception) return 0;

  SmiDefaultValue *defval = 0;
  switch (token.getTokenType()) {
    case SmiToken::TT_NUMBER:
      defval = new SmiIntegerDefaultValue(token.getTokenValue());
      break;
    case SmiToken::TT_STRING:
      defval = new SmiStringDefaultValue(token.getTokenValue());
      break;
    case SmiToken::TT_HSTRING:
      defval = new SmiHexStringDefaultValue(token.getTokenValue());
      break;
    case SmiToken::TT_BSTRING:
      defval = new SmiBinaryStringDefaultValue(token.getTokenValue());
      break;
    default:
      if (std::strcmp(token.getTokenValue(),"{") == 0) {
        data -> reader -> putbackToken(token);
        defval = new SmiIdentifierListDefaultValue(parseIdentifierList());
        if (exception) {
          delete defval;
          return 0;
          }
        }
      else defval = new SmiIdentifierDefaultValue(token.getTokenValue());
      break;
    }

  token = readWordToken();
  if (std::strcmp(token.getTokenValue(),"}") != 0) {
    delete defval;
    THROW(new SmiLoadOperationException(this,&token,"Expected }"),0);
    }

  return defval;
  }

void SmiLoadOperation::generateOidValueXml(Xml& xml,XmlElement element,
                                           SmiOidValue *oidValue) {

  XmlElement oidElement = xml.addElement(element,"oid");

  ListIterator it = oidValue -> iterator();
  while (it.hasNext()) {
    SmiSubIdentifier *subIdentifier = (SmiSubIdentifier *)it.next();

    XmlElement sidElement = xml.addElement(oidElement,"subid");
    const char *module = subIdentifier -> getModule();
    if (module) xml.addElementAttribute(sidElement,"module",module);
    const char *name = subIdentifier -> getName();
    if (name) xml.addElementAttribute(sidElement,"name",name);
    if (subIdentifier -> getSidPresent()) {
      char scratch[32];
      std::sprintf(scratch,"%lu",subIdentifier -> getSid());
      xml.addElementAttribute(sidElement,"value",scratch);
      }
    }
  }

void SmiLoadOperation::generateValueReferenceXml(
                                          Xml& xml,XmlElement parentElement,
                                          const char *tag,
                                          SmiValueReference *valueReference) {

  XmlElement element = xml.addElement(parentElement,tag);

  const char *name = valueReference -> getName();
  if (name) {
    const char *module = valueReference -> getModule();
    if (module) xml.addElementAttribute(element,"module",module);
    xml.addElementAttribute(element,"name",name);
    }
  else {
    generateOidValueXml(xml,element,valueReference -> getOidValue());
    }
  }

void SmiLoadOperation::generateSyntaxXml(Xml& xml,XmlElement parentElement,
                                         const char *tag, SmiSyntax *syntax) {

  XmlElement syntaxElement = xml.addElement(parentElement,tag);

  const char *name = syntax -> getName();
  if (name) xml.addElementAttribute(syntaxElement,"name",name);

  xml.addElementAttribute(syntaxElement,"typeref",
                          syntax -> getTyperef());

  const char *subtype = syntax -> getSubtype();
  if (subtype) xml.addElementAttribute(syntaxElement,"subtype",subtype);

  SmiAsn1Tag *asn1tag = syntax -> getTag();
  if (asn1tag) {
    XmlElement tagElement = xml.addElement(syntaxElement,"tag");

    const char *tagClass = asn1tag -> getClass();
    if (tagClass) xml.addElementAttribute(tagElement,"class",tagClass);

    //const char *module = asn1tag -> getModule();
    //if (module) xml.addElementAttribute(tagElement,"module",module);

    const char *value = asn1tag -> getValue();
    if (value) xml.addElementAttribute(tagElement,"value",value);

    if (asn1tag -> getImplicit())
      xml.addElementAttribute(tagElement,"implicit","true");
    }

  SmiComponentList *components = syntax -> getComponentList();
  if (components) {
    XmlElement componentsElement = xml.addElement(syntaxElement,"components");
    ListIterator it = components -> iterator();
    while (it.hasNext()) {
      SmiSyntax *syntax = (SmiSyntax *)it.next();
      generateSyntaxXml(xml,componentsElement,"syntax",syntax);
      }
    }

  SmiRangeList *ranges = syntax -> getRangeList();
  if (ranges) {
    XmlElement rangesElement = xml.addElement(syntaxElement,"ranges");
    ListIterator it = ranges -> iterator();
    while (it.hasNext()) {
      SmiRange *range = (SmiRange *)it.next();

      XmlElement rangeElement = xml.addElement(rangesElement,"range");
      char scratch[32];
      std::sprintf(scratch,"%lld",range -> getMin());
      xml.addElementAttribute(rangeElement,"min",scratch);
      std::sprintf(scratch,"%lld",range -> getMax());
      xml.addElementAttribute(rangeElement,"max",scratch);
      }
    }

  SmiSizeList *sizes = syntax -> getSizeList();
  if (sizes) {
    XmlElement sizesElement = xml.addElement(syntaxElement,"sizes");
    ListIterator it = sizes -> iterator();
    while (it.hasNext()) {
      SmiSize *size = (SmiSize *)it.next();

      XmlElement sizeElement = xml.addElement(sizesElement,"size");
      char scratch[32];
      std::sprintf(scratch,"%lld",size -> getMin());
      xml.addElementAttribute(sizeElement,"min",scratch);
      std::sprintf(scratch,"%lld",size -> getMax());
      xml.addElementAttribute(sizeElement,"max",scratch);
      }
    }

  SmiEnumerationList *enumerations = syntax -> getEnumerationList();
  if (enumerations) {
    XmlElement enumerationsElement = xml.addElement(syntaxElement,
                                                    "enumerations");
    ListIterator it = enumerations -> iterator();
    while (it.hasNext()) {
      SmiEnumeration *enumeration = (SmiEnumeration *)it.next();

      XmlElement enumerationElement = xml.addElement(enumerationsElement,
                                                     "enumeration");
      char scratch[32];
      std::sprintf(scratch,"%d",enumeration -> getNumber());
      xml.addElementAttribute(enumerationElement,"name",
                              enumeration -> getName());
      xml.addElementAttribute(enumerationElement,"value",scratch);
      }
    }
  }

void SmiLoadOperation::generateDefaultValueXml(Xml& xml,
                                               XmlElement parentElement,
                                               SmiDefaultValue *defval) {
  XmlElement defvalElement = xml.addElement(parentElement,"defval");
  XmlElement valueElement;
  switch (defval -> getType()) {
    case SmiDefaultValue::INTEGER_VALUE:
      char scratch[32];
      std::sprintf(scratch,"%lld",
                   ((SmiIntegerDefaultValue *)defval) -> getValue());
      valueElement = xml.addElement(defvalElement,"integer");
      xml.setElementContents(valueElement,scratch);
      break;
    case SmiDefaultValue::STRING_VALUE:
      valueElement = xml.addElement(defvalElement,"cstring");
      xml.setElementContents(valueElement,
                             ((SmiStringDefaultValue *)defval) -> getValue());
      break;
    case SmiDefaultValue::HEX_STRING_VALUE:
      valueElement = xml.addElement(defvalElement,"hstring");
      xml.setElementContents(valueElement,
                        ((SmiHexStringDefaultValue *)defval) -> getValue());
      break;
    case SmiDefaultValue::BINARY_STRING_VALUE:
      valueElement = xml.addElement(defvalElement,"bstring");
      xml.setElementContents(valueElement,
                        ((SmiBinaryStringDefaultValue *)defval) -> getValue());
      break;
    case SmiDefaultValue::IDENTIFIER_VALUE:
      valueElement = xml.addElement(defvalElement,"identifier");
      xml.setElementContents(valueElement,
                        ((SmiIdentifierDefaultValue *)defval) -> getValue());
      break;
    case SmiDefaultValue::OID_VALUE:
      //generateOidValueXml(xml,defvalElement,
      //                    ((SmiOidDefaultValue *)defval) -> getValue());
      break;
    case SmiDefaultValue::IDLIST_VALUE:
      valueElement = xml.addElement(defvalElement,"idlist");
      SmiIdentifierList *idlist =
                ((SmiIdentifierListDefaultValue *)defval) -> getValue();
      ListIterator it = idlist -> iterator();
      while (it.hasNext()) {
        String *identifier = (String *)it.next();
        XmlElement idElement = xml.addElement(valueElement,"identifier");
        xml.setElementContents(idElement,identifier -> getChars());
        }

      break;
    }
  }

void SmiLoadOperation::parseObjectType(const char *identifier) {
  //
  // Create the empty assignment
  //
  Xml xml;
  XmlElement root = xml.setRootElement("objecttype");
  xml.addElementAttribute(root,"name",identifier);

  SmiToken token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"SYNTAX") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected SYNTAX clause"),);
  SmiSyntax *syntax = parseSyntax();
  if (exception) return;
  generateSyntaxXml(xml,xml.getRootElement(),"syntax",syntax);

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"UNITS") == 0) {
    token = readStringToken();
    if (exception) return;

    XmlElement unitsElement = xml.addElement(root,"units");
    xml.setElementContents(unitsElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"MAX-ACCESS") == 0) {
    token = readWordToken();
    if (exception) return;

    XmlElement accessElement = xml.addElement(root,"access");
    xml.setElementContents(accessElement,token.getTokenValue());
    }
  else if (std::strcmp(token.getTokenValue(),"ACCESS") == 0) {
    token = readWordToken();
    if (exception) return;

    XmlElement accessElement = xml.addElement(root,"access");
    xml.setElementContents(accessElement,token.getTokenValue());
    }
  else THROW(new SmiLoadOperationException(this,&token,
                         "Expected ACCESS or MAX-ACCESS clause"),);

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"STATUS") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected STATUS clause"),);
  token = readWordToken();
  if (exception) return;
  xml.addElementAttribute(root,"status",token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected DESCRIPTION clause"),);
  token = readStringToken();
  if (exception) return;

  //
  // Write the description to the database in its own record
  //
  Xml descriptionXml;
  XmlElement descriptionRoot = descriptionXml.setRootElement("description");
  descriptionXml.setElementContents(descriptionRoot,token.getTokenValue());
  char key[128];
  std::sprintf(key,"/MODULE/%s/assignments/%s/description",
               data -> currentModule -> getName(),identifier);
  manager -> writeXml(descriptionXml,key);

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"REFERENCE") == 0) {
    token = readStringToken();
    if (exception) return;

    XmlElement referenceElement = xml.addElement(root,"reference");
    xml.setElementContents(referenceElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }
  if (std::strcmp(token.getTokenValue(),"INDEX") == 0) {
    SmiIndexList *indices = parseIndexList();
    if (exception) return;

    XmlElement indicesElement = xml.addElement(root,"indices");
    if (indices -> getImplied())
      xml.addElementAttribute(indicesElement,"implied","true");
    ListIterator it = indices -> iterator();
    while (it.hasNext()) {
      SmiIndexItem *index = (SmiIndexItem *)it.next();

      XmlElement indexElement = xml.addElement(indicesElement,"index");

      const char *name = index -> getName();
      if (name) {
        const char *module = index -> getModule();
        if (module) xml.addElementAttribute(indexElement,"module",module);
        xml.addElementAttribute(indexElement,"name",name);
        }
      }

    delete indices;

    token = readWordToken();
    if (exception) return;
    }
  if (std::strcmp(token.getTokenValue(),"AUGMENTS") == 0) {
    token = readWordToken();
    if (exception) return;

    if (std::strcmp(token.getTokenValue(),"{") != 0)
      THROW(new SmiLoadOperationException(this,&token,"Expected {"),);

    SmiValueReference *augments = parseValueReference();
    if (exception) return;

    generateValueReferenceXml(xml,root,"augments",augments);
    delete augments;

    token = readWordToken();
    if (exception) return;
    if (std::strcmp(token.getTokenValue(),"}") != 0)
      THROW(new SmiLoadOperationException(this,&token,"Expected }"),);

    token = readWordToken();
    if (exception) return;
    }
  if (std::strcmp(token.getTokenValue(),"DEFVAL") == 0) {
    SmiDefaultValue *defaultValue = parseDefaultValue();
    if (exception) return;

    generateDefaultValueXml(xml,root,defaultValue);
    delete defaultValue;

    token = readWordToken();
    if (exception) return;
    }
  if (std::strcmp(token.getTokenValue(),"::=") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected ::="),);

  processAssignmentTail(xml,identifier,"objecttype");
  }

void SmiLoadOperation::parseObjectIdentifier(const char *identifier) {
  SmiToken token = readWordToken();
  if (exception) return;

  if (std::strcmp(token.getTokenValue(),"::=") != 0) {
    THROW(new SmiLoadOperationException(this,&token,"Unrecognized keyword"),);
    }

  //
  // Create the empty assignment
  //
  Xml xml;
  XmlElement root = xml.setRootElement("objectid");
  xml.addElementAttribute(root,"name",identifier);

  //
  // Add this assignment to the module's assignments list
  //
  processAssignmentTail(xml,identifier,"objectid");
  }

void SmiLoadOperation::parseObjectIdentity(const char *identifier) {
  //
  // Parse the mandatory status clause
  //
  SmiToken token = readWordToken();
  if (std::strcmp(token.getTokenValue(),"STATUS") == 0) {
    token = readWordToken();
    if (exception) return;
    }
  else {
    THROW(new SmiLoadOperationException(this,0,"Missing STATUS clause"),);
    }

  //
  // Create the empty assignment
  //
  Xml xml;
  XmlElement root = xml.setRootElement("objectid");
  xml.addElementAttribute(root,"name",identifier);
  xml.addElementAttribute(root,"status",token.getTokenValue());

  //
  // Parse the mandatory description clause
  //
  token = readWordToken();
  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") == 0) {
    token = readStringToken();
    if (exception) return;
    }
  else {
    THROW(new SmiLoadOperationException(this,0,"Missing DESCRIPTION clause"),);
    }

  //
  // Write the description to the database in its own record
  //
  Xml descriptionXml;
  XmlElement descriptionRoot = descriptionXml.setRootElement("description");
  descriptionXml.setElementContents(descriptionRoot,token.getTokenValue());
  char key[128];
  std::sprintf(key,"/MODULE/%s/assignments/%s/description",
               data -> currentModule -> getName(),identifier);
  manager -> writeXml(descriptionXml,key);

  //
  // See if we have a reference clause
  //
  token = readWordToken();
  if (std::strcmp(token.getTokenValue(),"REFERENCE") == 0) {
    token = readStringToken();
    if (exception) return;

    //
    // Add the reference clause to the assignment
    //
    XmlElement referenceElement = xml.addElement(root,"reference");
    xml.setElementContents(referenceElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }
  if (std::strcmp(token.getTokenValue(),"::=") != 0) {
    THROW(new SmiLoadOperationException(this,&token,"Unrecognized keyword"),);
    }

  processAssignmentTail(xml,identifier,"objectid");
  }

void SmiLoadOperation::parseNotificationType(const char *identifier) {
  //
  // Create the empty assignment
  //
  Xml xml;
  XmlElement root = xml.setRootElement("notification");
  xml.addElementAttribute(root,"name",identifier);

  SmiToken token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"OBJECTS") == 0) {
    ObjectList *objects = parseValueReferenceList();
    if (exception) return;

    XmlElement objectsElement = xml.addElement(root,"objects");
    ListIterator it = objects -> iterator();
    while (it.hasNext()) {
      SmiValueReference *valueReference = (SmiValueReference *)it.next();
      generateValueReferenceXml(xml,objectsElement,"object",valueReference);
      }

    delete objects;

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"STATUS") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected STATUS clause"),);
  token = readWordToken();
  if (exception) return;
  xml.addElementAttribute(root,"status",token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected DESCRIPTION clause"),);
  token = readStringToken();
  if (exception) return;

  //
  // Write the description to the database in its own record
  //
  Xml descriptionXml;
  XmlElement descriptionRoot = descriptionXml.setRootElement("description");
  descriptionXml.setElementContents(descriptionRoot,token.getTokenValue());
  char key[128];
  std::sprintf(key,"/MODULE/%s/assignments/%s/description",
               data -> currentModule -> getName(),identifier);
  manager -> writeXml(descriptionXml,key);

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"REFERENCE") == 0) {
    token = readStringToken();
    if (exception) return;

    XmlElement referenceElement = xml.addElement(root,"reference");
    xml.setElementContents(referenceElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"::=") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected ::="),);

  processAssignmentTail(xml,identifier,"notification");
  }

void SmiLoadOperation::parseTrapType(const char *identifier) {
  //
  // Create the empty assignment
  //
  Xml xml;
  XmlElement root = xml.setRootElement("notification");
  xml.addElementAttribute(root,"name",identifier);

  SmiToken token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"ENTERPRISE") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected ENTERPRISE clause"),);
  SmiOidValue *enterprise = parseOidValue();
  if (exception) return;

  token = readWordToken();
  if (exception) {
      delete enterprise;
      return;
      }
  if (std::strcmp(token.getTokenValue(),"VARIABLES") == 0) {
    ObjectList *objects = parseValueReferenceList();
    if (exception) {
      delete enterprise;
      return;
      }

    XmlElement objectsElement = xml.addElement(root,"objects");
    ListIterator it = objects -> iterator();
    while (it.hasNext()) {
      SmiValueReference *valueReference = (SmiValueReference *)it.next();
      generateValueReferenceXml(xml,objectsElement,"object",valueReference);
      }

    delete objects;

    token = readWordToken();
    if (exception) {
      delete enterprise;
      return;
      }
    }

  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") == 0) {
    token = readStringToken();
    if (exception) {
      delete enterprise;
      return;
      }

    //
    // Write the description to the database in its own record
    //
    Xml descriptionXml;
    XmlElement descriptionRoot = descriptionXml.setRootElement("description");
    descriptionXml.setElementContents(descriptionRoot,token.getTokenValue());
    char key[128];
    std::sprintf(key,"/MODULE/%s/assignments/%s/description",
                 data -> currentModule -> getName(),identifier);
    manager -> writeXml(descriptionXml,key);

    token = readWordToken();
    if (exception) {
      delete enterprise;
      return;
      }
    }

  if (std::strcmp(token.getTokenValue(),"REFERENCE") == 0) {
    token = readStringToken();
    if (exception) {
      delete enterprise;
      return;
      }

    XmlElement referenceElement = xml.addElement(root,"reference");
    xml.setElementContents(referenceElement,token.getTokenValue());

    token = readWordToken();
    if (exception) {
      delete enterprise;
      return;
      }
    }

  if (std::strcmp(token.getTokenValue(),"::=") != 0) {
    delete enterprise;
    THROW(new SmiLoadOperationException(this,&token,"Expected ::="),);
    }

  //
  // Add this assignment to the module's assignments list
  //
  XmlElement assignmentElement =
      data -> currentModuleAssignmentsXml ->
               addElement(data -> currentModuleAssignmentsRoot,"notification");
  data -> currentModuleAssignmentsXml ->
                   addElementAttribute(assignmentElement,"name",identifier);

  //
  // Write the assignment to the database
  //
  char key[128];
  std::sprintf(key,"/MODULE/%s/assignments/%s",
               data -> currentModule -> getName(),identifier);
  manager -> writeXml(xml,key);

  //
  // Construct the OID string for this assignment
  //
  token = readWordToken();
  if (exception) return;
  if (token.getTokenType() != SmiToken::TT_NUMBER) {
    delete enterprise;
    THROW(new SmiLoadOperationException(this,&token,"Expected a number"),);
    }
  int trapNumber = token.getIntValue();
  ListIterator it = enterprise -> iterator();
  if (! it.hasNext()) {
    THROW(new SmiLoadOperationException(this,0,"Invalid ENTERPRISE"),);
    delete enterprise;
    }
  SmiSubIdentifier *subIdentifier = (SmiSubIdentifier *)it.next();
  const char *sidName = subIdentifier -> getName();
  if (sidName && (std::strcmp(sidName,"snmp") == 0) && (! it.hasNext())) {
    switch (trapNumber) {
      case 0:
        subIdentifier = new SmiSubIdentifier("SNMPv2-MIB","coldStart");
        break;
      case 1:
        subIdentifier = new SmiSubIdentifier("SNMPv2-MIB","warmStart");
        break;
      case 2:
        subIdentifier = new SmiSubIdentifier("IF-MIB","linkDown");
        break;
      case 3:
        subIdentifier = new SmiSubIdentifier("IF-MIB","linkUp");
        break;
      case 4:
        subIdentifier = new SmiSubIdentifier("SNMPv2-MIB",
                                             "authenticationFailure");
        break;
      case 5:
        subIdentifier = new SmiSubIdentifier("SNMPv2-MIB","egpNeighborLoss");
        break;
      default:
        delete enterprise;
        THROW(new SmiLoadOperationException(this,&token,
                                            "Invalid trap number"),);
      }

    delete enterprise;
    enterprise = new SmiOidValue();
    enterprise -> appendSubIdentifier(subIdentifier);
    }
  else {
    enterprise ->
         appendSubIdentifier(new SmiSubIdentifier((SmiTypes::SubIdentifier)0));
    enterprise -> appendSubIdentifier(new SmiSubIdentifier(trapNumber));
    }

  const char *oidString = buildOidString(enterprise);
  delete enterprise;
  if (exception) return;

  //
  // Add the OID string to the global OID map
  //
  char fullOidName[256];
  std::sprintf(fullOidName,"%s.%s",data -> currentModule -> getName(),
               identifier);
  manager -> addOidMapping(fullOidName,oidString);
  }

void SmiLoadOperation::parseModuleIdentity(const char *identifier) {
  SmiToken token = readWordToken();
  if (exception) return;

  //
  // Create the empty assignment
  //
  Xml xml;
  XmlElement root = xml.setRootElement("moduleid");
  xml.addElementAttribute(root,"name",identifier);

  if (std::strcmp(token.getTokenValue(),"LAST-UPDATED") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected LAST-UPDATED clause"),);
  token = readStringToken();
  if (exception) return;
  xml.addElementAttribute(root,"updated",token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"ORGANIZATION") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected ORGANIZATION clause"),);
  token = readStringToken();
  if (exception) return;
  XmlElement orgElement = xml.addElement(root,"organization");
  xml.setElementContents(orgElement,token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"CONTACT-INFO") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected CONTACT-INFO clause"),);
  token = readStringToken();
  if (exception) return;
  XmlElement contactElement = xml.addElement(root,"contact");
  xml.setElementContents(contactElement,token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected DESCRIPTION clause"),);
  token = readStringToken();
  if (exception) return;
  XmlElement descriptionElement = xml.addElement(root,"description");
  xml.setElementContents(descriptionElement,token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  XmlElement revisionsElement = xml.addElement(root,"revisions");
  while (std::strcmp(token.getTokenValue(),"::=") != 0) {
    if (std::strcmp(token.getTokenValue(),"REVISION") != 0)
      THROW(new SmiLoadOperationException(this,&token,
                                          "Expected REVISION clause"),);
    token = readStringToken();
    XmlElement revisionElement = xml.addElement(revisionsElement,"revision");
    xml.addElementAttribute(revisionElement,"date",token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
      THROW(new SmiLoadOperationException(this,&token,
                                "Expected revision DESCRIPTION clause"),);
    token = readStringToken();
    descriptionElement = xml.addElement(revisionElement,"description");
    xml.setElementContents(descriptionElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }

  processAssignmentTail(xml,identifier,"moduleid");
  }

void SmiLoadOperation::parseComplianceGroup(Xml& xml,
                                            XmlElement groupsElement) {
  SmiToken token = readWordToken();
  if (exception) return;

  XmlElement groupElement = xml.addElement(groupsElement,"group");
  if (std::strcmp(token.getTokenValue(),"{") == 0) {
    data -> reader -> putbackToken(token);
    SmiOidValue *groupOid = parseOidValue();
    generateOidValueXml(xml,groupElement,groupOid);
    delete groupOid;
    }
  else xml.addElementAttribute(groupElement,"name",token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected DESCRIPTION"),);
  token = readStringToken();
  if (exception) return;
  XmlElement descriptionElement = xml.addElement(groupElement,"description");
  xml.setElementContents(descriptionElement,token.getTokenValue());
  }

void SmiLoadOperation::parseComplianceObject(Xml& xml,
                                             XmlElement objectsElement) {
  SmiToken token = readWordToken();
  if (exception) return;

  XmlElement objectElement = xml.addElement(objectsElement,"object");
  if (std::strcmp(token.getTokenValue(),"{") == 0) {
    data -> reader -> putbackToken(token);
    SmiOidValue *objectOid = parseOidValue();
    generateOidValueXml(xml,objectElement,objectOid);
    delete objectOid;
    }
  else xml.addElementAttribute(objectElement,"name",token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"SYNTAX") == 0) {
    SmiSyntax *syntax = parseSyntax();
    if (exception) return;
    generateSyntaxXml(xml,objectElement,"syntax",syntax);
    delete syntax;

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"WRITE-SYNTAX") == 0) {
    SmiSyntax *syntax = parseSyntax();
    if (exception) return;
    generateSyntaxXml(xml,objectElement,"writesyntax",syntax);
    delete syntax;

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"MIN-ACCESS") == 0) {
    token = readWordToken();
    if (exception) return;
    XmlElement accessElement = xml.addElement(objectElement,"minaccess");
    xml.setElementContents(accessElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected DESCRIPTION"),);
  token = readStringToken();
  if (exception) return;
  XmlElement descriptionElement = xml.addElement(objectElement,"description");
  xml.setElementContents(descriptionElement,token.getTokenValue());
  }

void SmiLoadOperation::parseModuleComplianceEntry(Xml& xml,
                                                  XmlElement modulesElement) {

  XmlElement moduleElement = xml.addElement(modulesElement,"module");

  SmiToken token = readWordToken();
  if (exception) return;
  if ((std::strcmp(token.getTokenValue(),"MANDATORY-GROUPS") != 0) &&
      (std::strcmp(token.getTokenValue(),"GROUP") != 0) &&
      (std::strcmp(token.getTokenValue(),"OBJECT") != 0) &&
      (std::strcmp(token.getTokenValue(),"::=") != 0)) {
    xml.addElementAttribute(moduleElement,"name",token.getTokenValue());
    token = readWordToken();
    if (exception) return;
    if (std::strcmp(token.getTokenValue(),"{") == 0) {
      data -> reader -> putbackToken(token);
      SmiOidValue *moduleOidValue = parseOidValue();
      if (exception) return;
      generateOidValueXml(xml,moduleElement,moduleOidValue);
      delete moduleOidValue;

      token = readWordToken();
      if (exception) return;
      }
    }
  else xml.addElementAttribute(moduleElement,"name",
                               data -> currentModule -> getName());

  XmlElement groupsElement = 0;
  XmlElement objectsElement = 0;
  while ((std::strcmp(token.getTokenValue(),"MANDATORY-GROUPS") == 0) ||
         (std::strcmp(token.getTokenValue(),"GROUP") == 0) ||
         (std::strcmp(token.getTokenValue(),"OBJECT") == 0)) {
    if (std::strcmp(token.getTokenValue(),"MANDATORY-GROUPS") == 0) {
      ObjectList *mandatoryGroups = parseValueReferenceList();
      if (exception) return;
      XmlElement requiresElement = xml.addElement(moduleElement,"requires");
      ListIterator it = mandatoryGroups -> iterator();
      while (it.hasNext()) {
        SmiValueReference *valueReference = (SmiValueReference *)it.next();

        generateValueReferenceXml(xml,requiresElement,"group",valueReference);
        }
      delete mandatoryGroups;
      }
    else if (std::strcmp(token.getTokenValue(),"GROUP") == 0) {
      if (groupsElement == 0)
        groupsElement = xml.addElement(moduleElement,"groups");
      parseComplianceGroup(xml,groupsElement);
      }
    else { // OBJECT
      if (objectsElement == 0)
        objectsElement = xml.addElement(moduleElement,"objects");
      parseComplianceObject(xml,objectsElement);
      }

    token = readWordToken();
    if (exception) return;
    }

  data -> reader -> putbackToken(token);
  }

void SmiLoadOperation::parseModuleCompliance(const char *identifier) {
  //
  // Create the empty assignment
  //
  Xml xml;
  XmlElement root = xml.setRootElement("compliance");
  xml.addElementAttribute(root,"name",identifier);

  SmiToken token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"STATUS") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected STATUS clause"),);
  token = readWordToken();
  if (exception) return;
  xml.addElementAttribute(root,"status",token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected DESCRIPTION clause"),);
  token = readStringToken();
  if (exception) return;
  XmlElement descriptionElement = xml.addElement(root,"description");
  xml.setElementContents(descriptionElement,token.getTokenValue());
  
  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"REFERENCE") == 0) {
    token = readStringToken();
    if (exception) return;

    XmlElement referenceElement = xml.addElement(root,"reference");
    xml.setElementContents(referenceElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }

  XmlElement modulesElement = xml.addElement(root,"modules");
  while (std::strcmp(token.getTokenValue(),"::=") != 0) {
    if (std::strcmp(token.getTokenValue(),"MODULE") != 0)
      THROW(new SmiLoadOperationException(this,&token,"Expected MODULE"),);

    parseModuleComplianceEntry(xml,modulesElement);

    token = readWordToken();
    if (exception) return;
    }

  processAssignmentTail(xml,identifier,"moduleid");
  }

void SmiLoadOperation::parseVariation(Xml& xml,XmlElement variationsElement) {

  XmlElement variationElement = xml.addElement(variationsElement,"variation");

  SmiToken token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"{") == 0) {
    data -> reader -> putbackToken(token);
    SmiOidValue *oid = parseOidValue();
    generateOidValueXml(xml,variationElement,oid);
    delete oid;

    token = readWordToken();
    if (exception) return;
    }
  else xml.addElementAttribute(variationElement,"name",token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"SYNTAX") == 0) {
    SmiSyntax *syntax = parseSyntax();
    generateSyntaxXml(xml,variationElement,"syntax",syntax);
    delete syntax;

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"WRITE-SYNTAX") == 0) {
    SmiSyntax *syntax = parseSyntax();
    generateSyntaxXml(xml,variationElement,"writesyntax",syntax);
    delete syntax;

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"ACCESS") == 0) {
    token = readWordToken();
    if (exception) return;
    XmlElement accessElement = xml.addElement(variationElement,"access");
    xml.setElementContents(accessElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"CREATION-REQUIRES") == 0) {
    ObjectList *objects = parseValueReferenceList();
    XmlElement creationElement = xml.addElement(variationElement,"creation");
    ListIterator it = objects -> iterator();
    while (it.hasNext()) {
      SmiValueReference *column = (SmiValueReference *)it.next();

      generateValueReferenceXml(xml,creationElement,"column",column);
      }
    delete objects;

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"DEFVAL") == 0) {
    SmiDefaultValue *defaultValue = parseDefaultValue();
    if (exception) return;

    generateDefaultValueXml(xml,variationElement,defaultValue);
    delete defaultValue;

    token = readWordToken();
    if (exception) return;
    }

  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
    THROW(new SmiLoadOperationException(this,&token,"Expected DESCRIPTION"),);
  token = readStringToken();
  XmlElement descriptionElement =
                      xml.addElement(variationElement,"description");
  xml.setElementContents(descriptionElement,token.getTokenValue());
  }

void SmiLoadOperation::parseAgentCapabilities(const char *identifier) {
  //
  // Create the empty assignment
  //
  Xml xml;
  XmlElement root = xml.setRootElement("capability");
  xml.addElementAttribute(root,"name",identifier);

  SmiToken token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"PRODUCT-RELEASE") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected STATUS clause"),);
  token = readStringToken();
  if (exception) return;
  xml.addElementAttribute(root,"release",token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"STATUS") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected STATUS clause"),);
  token = readWordToken();
  if (exception) return;
  xml.addElementAttribute(root,"status",token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected DESCRIPTION clause"),);
  token = readStringToken();
  if (exception) return;
  XmlElement descriptionElement = xml.addElement(root,"description");
  xml.setElementContents(descriptionElement,token.getTokenValue());
  
  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"REFERENCE") == 0) {
    token = readStringToken();
    if (exception) return;

    XmlElement referenceElement = xml.addElement(root,"reference");
    xml.setElementContents(referenceElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }

  XmlElement modulesElement = xml.addElement(root,"modules");
  while (std::strcmp(token.getTokenValue(),"::=") != 0) {
    if (std::strcmp(token.getTokenValue(),"SUPPORTS") != 0)
      THROW(new SmiLoadOperationException(this,&token,"Expected SUPPORTS"),);

    token = readWordToken();
    if (exception) return;
    XmlElement moduleElement = xml.addElement(modulesElement,"module");
    xml.addElementAttribute(moduleElement,"name",token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    if (std::strcmp(token.getTokenValue(),"{") == 0) {
      data -> reader -> putbackToken(token);
      SmiOidValue *moduleOid = parseOidValue();
      if (exception) return;
      generateOidValueXml(xml,moduleElement,moduleOid);
      delete moduleOid;

      token = readWordToken();
      if (exception) return;
      }

    if (std::strcmp(token.getTokenValue(),"INCLUDES") != 0)
      THROW(new SmiLoadOperationException(this,&token,"Expected INCLUDES"),);
    ObjectList *groups = parseValueReferenceList();
    if (exception) return;
    ListIterator it = groups -> iterator();
    XmlElement includesElement = xml.addElement(moduleElement,"includes");
    while (it.hasNext()) {
      SmiValueReference *group = (SmiValueReference *)it.next();

      generateValueReferenceXml(xml,includesElement,"group",group);
      }
    delete groups;

    token = readWordToken();
    if (exception) return;
    XmlElement variationsElement = xml.addElement(moduleElement,"variations");
    while (std::strcmp(token.getTokenValue(),"VARIATION") == 0) {
      parseVariation(xml,variationsElement);

      token = readWordToken();
      if (exception) return;
      }
    }

  processAssignmentTail(xml,identifier,"capability");
  }

void SmiLoadOperation::parseObjectGroup(const char *identifier) {
  //
  // Create the empty assignment
  //
  Xml xml;
  XmlElement root = xml.setRootElement("objectgroup");
  xml.addElementAttribute(root,"name",identifier);

  SmiToken token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"OBJECTS") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected OBJECTS clause"),);
  XmlElement objectsElement = xml.addElement(root,"objects");
  ObjectList *objects = parseValueReferenceList();
  ListIterator it = objects -> iterator();
  while (it.hasNext()) {
    SmiValueReference *object = (SmiValueReference *)it.next();

    generateValueReferenceXml(xml,objectsElement,"object",object);
    }
  delete objects;

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"STATUS") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected STATUS clause"),);
  token = readWordToken();
  if (exception) return;
  xml.addElementAttribute(root,"status",token.getTokenValue());
 
  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected DESCRIPTION clause"),);
  token = readStringToken();
  if (exception) return;
  XmlElement descriptionElement = xml.addElement(root,"description");
  xml.setElementContents(descriptionElement,token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"REFERENCE") == 0) {
    token = readStringToken();
    if (exception) return;
    XmlElement referenceElement = xml.addElement(root,"reference");
    xml.setElementContents(referenceElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }

  //
  // Add this assignment to the module's assignments list
  //
  processAssignmentTail(xml,identifier,"objectgroup");
  }

void SmiLoadOperation::processAssignmentTail(Xml& xml,const char *identifier,
                                             const char *assignmentTag) {
  XmlElement assignmentElement =
        data -> currentModuleAssignmentsXml ->
             addElement(data -> currentModuleAssignmentsRoot,assignmentTag);
  data -> currentModuleAssignmentsXml ->
                   addElementAttribute(assignmentElement,"name",identifier);

  //
  // Write the assignment to the database
  //
  char key[128];
  std::sprintf(key,"/MODULE/%s/assignments/%s",
               data -> currentModule -> getName(),identifier);
  manager -> writeXml(xml,key);

  //
  // Construct the OID string for this assignment
  //
  SmiOidValue *oidValue = parseOidValue();
  if (exception) return;
  const char *oidString = buildOidString(oidValue);
  delete oidValue;
  if (exception) return;

  //
  // Add the OID string to the global OID map
  //
  char fullOidName[256];
  std::sprintf(fullOidName,"%s.%s",data -> currentModule -> getName(),
               identifier);
  manager -> addOidMapping(fullOidName,oidString);
  }

void SmiLoadOperation::parseNotificationGroup(const char *identifier) {
  //
  // Create the empty assignment
  //
  Xml xml;
  XmlElement root = xml.setRootElement("notifygroup");
  xml.addElementAttribute(root,"name",identifier);

  SmiToken token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"NOTIFICATIONS") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected NOTIFICATIONS clause"),);
  XmlElement objectsElement = xml.addElement(root,"notifications");
  ObjectList *objects = parseValueReferenceList();
  ListIterator it = objects -> iterator();
  while (it.hasNext()) {
    SmiValueReference *object = (SmiValueReference *)it.next();

    generateValueReferenceXml(xml,objectsElement,"notification",object);
    }
  delete objects;

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"STATUS") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected STATUS clause"),);
  token = readWordToken();
  if (exception) return;
  xml.addElementAttribute(root,"status",token.getTokenValue());
 
  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
    THROW(new SmiLoadOperationException(this,&token,
                                        "Expected DESCRIPTION clause"),);
  token = readStringToken();
  if (exception) return;
  XmlElement descriptionElement = xml.addElement(root,"description");
  xml.setElementContents(descriptionElement,token.getTokenValue());

  token = readWordToken();
  if (exception) return;
  if (std::strcmp(token.getTokenValue(),"REFERENCE") == 0) {
    token = readStringToken();
    if (exception) return;
    XmlElement referenceElement = xml.addElement(root,"reference");
    xml.setElementContents(referenceElement,token.getTokenValue());

    token = readWordToken();
    if (exception) return;
    }

  //
  // Add this assignment to the module's assignments list
  //
  processAssignmentTail(xml,identifier,"notifygroup");
  }

void SmiLoadOperation::parseMacro(const char *identifier) {
  SmiToken token = readWordToken();
  if (exception) return;
  while ((token.getTokenType() != SmiToken::TT_IDENTIFIER) ||
         (std::strcmp(token.getTokenValue(),"END") != 0)) {

    token = nextToken();
    if (exception) return;
    }
  }

void SmiLoadOperation::parseTypedef(const char *identifier) {
  SmiToken token = readWordToken();
  if (exception) return;

  //
  // Create the empty typedef
  //
  Xml xml;
  XmlElement root = xml.setRootElement("typedef");
  xml.addElementAttribute(root,"name",identifier);

  if (std::strcmp(token.getTokenValue(),"TEXTUAL-CONVENTION") == 0) {
    token = readWordToken();
    if (exception) return;
    if (std::strcmp(token.getTokenValue(),"DISPLAY-HINT") == 0) {
      token = readStringToken();
      if (exception) return;
      XmlElement hintElement = xml.addElement(root,"format");
      xml.setElementContents(hintElement,token.getTokenValue());

      token = readWordToken();
      if (exception) return;
      }

    if (std::strcmp(token.getTokenValue(),"STATUS") != 0)
      THROW(new SmiLoadOperationException(this,&token,
                                          "Expected STATUS clause"),);
    token = readWordToken();
    if (exception) return;
    xml.addElementAttribute(root,"status",token.getTokenValue());
 
    token = readWordToken();
    if (exception) return;
    if (std::strcmp(token.getTokenValue(),"DESCRIPTION") != 0)
      THROW(new SmiLoadOperationException(this,&token,
                                          "Expected DESCRIPTION clause"),);
    token = readStringToken();
    if (exception) return;

    //
    // Write the description to the database in its own record
    //
    Xml descriptionXml;
    XmlElement descriptionRoot = descriptionXml.setRootElement("description");
    descriptionXml.setElementContents(descriptionRoot,token.getTokenValue());
    char key[128];
    std::sprintf(key,"/MODULE/%s/typedefs/%s/description",
                 data -> currentModule -> getName(),identifier);
    manager -> writeXml(descriptionXml,key);

    token = readWordToken();
    if (exception) return;
    if (std::strcmp(token.getTokenValue(),"REFERENCE") == 0) {
      token = readStringToken();
      if (exception) return;
      XmlElement referenceElement = xml.addElement(root,"reference");
      xml.setElementContents(referenceElement,token.getTokenValue());

      token = readWordToken();
      if (exception) return;
      }

    if (std::strcmp(token.getTokenValue(),"SYNTAX") != 0)
      THROW(new SmiLoadOperationException(this,&token,
                                          "Expected SYNTAX clause"),);
    }
  else {
    data -> reader -> putbackToken(token);
    }

  SmiSyntax *syntax = parseSyntax();
  if (exception) return;
  generateSyntaxXml(xml,root,"syntax",syntax);

  //
  // Add this typedef to the module's typedef list
  //
  XmlElement typedefElement =
        data -> currentModuleTypedefsXml ->
                 addElement(data -> currentModuleTypedefsRoot,"typedef");
  data -> currentModuleTypedefsXml ->
                   addElementAttribute(typedefElement,"name",identifier);

  //
  // Write the typedef to the database
  //
  char key[128];
  std::sprintf(key,"/MODULE/%s/typedefs/%s",
               data -> currentModule -> getName(),identifier);
  manager -> writeXml(xml,key);
  }

bool SmiLoadOperation::parseDefinition() {
  SmiToken token = readWordToken();
  if (exception) return false;

  if (std::strcmp(token.getTokenValue(),"END") == 0) return false;
  else if (std::strcmp(token.getTokenValue(),"IMPORTS") == 0) parseImports();
  else if (std::strcmp(token.getTokenValue(),"EXPORTS") == 0) parseExports();
  else {
    SmiToken identifierToken = token;
    const char *identifier = identifierToken.getTokenValue();

    token = readWordToken();
    if (exception) return false;
    const char *tokenValue = token.getTokenValue();
    if (std::strcmp(tokenValue,"OBJECT-TYPE") == 0)
      parseObjectType(identifier);
    else if (std::strcmp(tokenValue,"OBJECT") == 0) {
      token = readWordToken();
      tokenValue = 0;
      if (exception) return false;
      if (std::strcmp(token.getTokenValue(),"IDENTIFIER") != 0) {
        exception = new SmiLoadOperationException(this,&token,
                                  "Expected OBJECT IDENTIFIER");
        THROW(exception,false);
        }

      parseObjectIdentifier(identifier);
      }
    else if (std::strcmp(tokenValue,"OBJECT-IDENTITY") == 0)
      parseObjectIdentity(identifier);
    else if (std::strcmp(tokenValue,"NOTIFICATION-TYPE") == 0)
      parseNotificationType(identifier);
    else if (std::strcmp(tokenValue,"TRAP-TYPE") == 0)
      parseTrapType(identifier);
    else if (std::strcmp(tokenValue,"MODULE-IDENTITY") == 0)
      parseModuleIdentity(identifier);
    else if (std::strcmp(tokenValue,"MODULE-COMPLIANCE") == 0)
      parseModuleCompliance(identifier);
    else if (std::strcmp(tokenValue,"AGENT-CAPABILITIES") == 0)
      parseAgentCapabilities(identifier);
    else if (std::strcmp(tokenValue,"OBJECT-GROUP") == 0)
      parseObjectGroup(identifier);
    else if (std::strcmp(tokenValue,"NOTIFICATION-GROUP") == 0)
      parseNotificationGroup(identifier);
    else if (std::strcmp(tokenValue,"MACRO") == 0)
      parseMacro(identifier);
    else if (std::strcmp(tokenValue,"::=") == 0)
      parseTypedef(identifier);
    else {
      exception = new SmiLoadOperationException(this,&token,
                                                "Unrecognized keyword");
      THROW(exception,false);
      }
    }

  return exception == 0;
  }

void SmiLoadOperation::writeModuleIndexes() {
  char key[128];

  std::sprintf(key,"/MODULE/%s/assignments",
               data -> currentModule -> getName());
  manager -> writeXml(*data -> currentModuleAssignmentsXml,key);

  std::sprintf(key,"/MODULE/%s/typedefs",
               data -> currentModule -> getName());
  manager -> writeXml(*data -> currentModuleTypedefsXml,key);
  }

void SmiLoadOperation::writeModuleImports() {
  Xml dbXml;
  XmlElement root = dbXml.setRootElement("imports");

  ObjectHashMapIterator it = data -> currentModuleImportsMap -> iterator();
  while (it.hasNext()) {
    ObjectMapping mapping = it.next();
    XmlElement dbSymbolElement = dbXml.addElement(root,"symbol");

    dbXml.addElementAttribute(dbSymbolElement,"name",
                              ((String *)mapping.name) -> getChars());

    dbXml.addElementAttribute(dbSymbolElement,"importsFrom",
                              ((String *)mapping.value) -> getChars());
    }

  char key[128];
  std::sprintf(key,"/MODULE/%s/imports",data -> currentModule -> getName());
  manager -> writeXml(dbXml,root,key);
  }

void SmiLoadOperation::writeModuleHeading() {
  Xml dbXml;
  XmlElement root = dbXml.setRootElement("module");

  char key[128];
  std::sprintf(key,"/MODULE/%s",data -> currentModule -> getName());
  manager -> writeXml(dbXml,root,key);
  }

void SmiLoadOperation::createModule(const char *name) {
  //
  // If the module is already loaded, we unload the old one on the
  // assumption that something might have changed
  //
  ListIterator it = manager -> getModuleList() -> iterator();
  while (it.hasNext()) {
    if (std::strcmp((char *)it.next(),name) == 0) {
      //
      // For now, run the unloader to completion
      //
      XmlDbOperation *unloader = manager -> dbUnloader(name);
      while (unloader -> getState() == Running) {
        unloader -> doWork();
        }
      if (unloader -> getState() == Error) {
        state = Error;
        exception = unloader -> takeException();
        delete unloader;
        THROW(exception,);
        }
      delete unloader;

      break;
      }
    }

  //
  // Create the module in the module manager, so it can receive
  // loaded OID assignments
  //
  manager -> addModule(name);

  //
  // Create a context for the module's definitions
  //
  delete data -> currentModule;
  data -> currentModule = new SmiMibModule(name);

  delete data -> currentModuleImportsMap;
  data -> currentModuleImportsMap = new ObjectHashMap();

  delete data -> currentModuleAssignmentsXml;
  data -> currentModuleAssignmentsXml = new Xml();
  data -> currentModuleAssignmentsRoot =
        data -> currentModuleAssignmentsXml -> setRootElement("assignments");

  delete data -> currentModuleTypedefsXml;
  data -> currentModuleTypedefsXml = new Xml();
  data -> currentModuleTypedefsRoot =
        data -> currentModuleTypedefsXml -> setRootElement("typedefs");
  }

void SmiLoadOperation::doWork() {
  if (data -> subState == SmiLoadOperationPrivate::Global) {
    //
    // We're looking for the next module in this file
    //
    SmiToken lastWordToken;
    while (true) {
      SmiToken token = nextToken();
      if (exception) {
        state = Error;
        }

      if (token.getTokenType() == SmiToken::TT_EOF) {
        if (lastWordToken.getTokenType() != SmiToken::TT_NULL) {
          state = Error;
          exception = new SmiLoadOperationException(this,&lastWordToken,
                                    "Extraneous identifier or keyword");
          THROW(exception,);
          }

        state = Completed;
        return;
        }

      if (token.getTokenType() != SmiToken::TT_IDENTIFIER) {
        state = Error;
        exception = new SmiLoadOperationException(this,&token,
                                         "<identifier> DEFINITIONS expected");
        THROW(exception,);
        }

      if (std::strcmp(token.getTokenValue(),"DEFINITIONS") == 0) {
        //
        // Must be preceded by an identifier
        //
        if (lastWordToken.getTokenType() == SmiToken::TT_NULL) {
          state = Error;
          exception =
              new SmiLoadOperationException(this,0,"Missing identifier");
          THROW(exception,);
          }

        createModule(lastWordToken.getTokenValue());

        //
        // Consume the rest of the prologue
        //
        token = readWordToken();
        if (exception) {
          state = Error;
          return;
          }
        if (std::strcmp(token.getTokenValue(),"::=") != 0) {
          state = Error;
          exception = new SmiLoadOperationException(this,&token,"Expected ::=");
          THROW(exception,);
          }

        token = readWordToken();
        if (exception) {
          state = Error;
          return;
          }
        if (std::strcmp(token.getTokenValue(),"BEGIN") != 0) {
          state = Error;
          exception =
               new SmiLoadOperationException(this,&token,"Expected BEGIN");
          THROW(exception,);
          }

        //
        // Prepare to parse the definitions of this module
        //
        data -> subState = SmiLoadOperationPrivate::Definitions;
        return;
        }

      //
      // We're here because we just read an identifier. Check that the
      // last thing we saw was not also an identifier
      //
      if (lastWordToken.getTokenType() != SmiToken::TT_NULL) {
        state = Error;
        exception = new SmiLoadOperationException(this,&lastWordToken,
                                       "Extraneous identifier or keyword");
        THROW(exception,);
        }

      lastWordToken = token;
      }
    }
  else {
    //
    // We're looking for the next definition within the current module
    //
    if (! parseDefinition()) {
      if (exception) {
        state = Error;
        return;
        }

      //
      // Completed parsing the module, so add module index records to the
      // database
      //
      writeModuleHeading();
      writeModuleIndexes();
      writeModuleImports();

      //
      // We're back at the global level of the file
      //
      data -> subState = SmiLoadOperationPrivate::Global;
      }
    }
  }

const char *SmiLoadOperation::getCurrentModuleName() const {
  if (data -> currentModule) return data -> currentModule -> getName();
  return 0;
  }

int SmiLoadOperation::getProgress() const {
  return data -> reader -> getProgress();
  }
