/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstring>
#include <cctype>

#include <classlib/Xml.hpp>
#include <classlib/List.hpp>
#include <classlib/String.hpp>

#include "XmlLoadOperation.hpp"
#include "XmlUnloadOperation.hpp"

#include "ModuleManager.hpp"

XmlLoadOperation::XmlLoadOperation(ModuleManager *manager)
          : XmlDbLoadOperation(manager), xml(0) {
  }

XmlLoadOperation::~XmlLoadOperation() {
  delete xml;
  }

void XmlLoadOperation::initialise(const char *path) {
  XmlDbOperation::initialise();

  xml = new Xml();
  xml -> loadXmlFromFile(path);
  if (xml -> exception) {
    state = Error;
    RETHROW(xml -> exception,);
    }

  root = xml -> getRootElement();
  moduleCount = xml -> getElementChildCount(root);
  currentModuleIndex = 0; // Just started, not even read its name yet
  currentModuleName = "";
  }

const char *XmlLoadOperation::buildOidString(XmlElement oidElement,
                                             ObjectHashMap& importsMap) {
  static char oidString[512]; // Lazy
  char *oidStringPtr = oidString;

  *oidStringPtr = '\0';
  int childCount = xml -> getElementChildCount(oidElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement sidElement = xml -> getElementChild(oidElement,i);

    const char *module = NULL;
    const char *name = NULL;
    const char *value = NULL;
    int attributeCount = xml -> getElementAttributeCount(sidElement);
    for (int j = 0 ; j < attributeCount ; j++ ) {
      const char *attributeName = xml -> getElementAttributeName(sidElement,j);
      if (strcmp(attributeName,"name") == 0) {
        name = xml -> getElementAttributeValue(sidElement,j);
        }
      else if (strcmp(attributeName,"module") == 0) {
        module = xml -> getElementAttributeValue(sidElement,j);
        }
      else if (strcmp(attributeName,"value") == 0) {
        value = xml -> getElementAttributeValue(sidElement,j);
        }
      }

    if (i != 0) *oidStringPtr++ = ' ';
    if (value != NULL) std::strcpy(oidStringPtr,value);
    else if (module != NULL) std::sprintf(oidStringPtr,"%s.%s",module,name);
    else {
      //
      // No module specified in the reference, so see if we can find
      // an import for the symbol name
      //
      ObjectMapping moduleMapping = importsMap.find(name);
      if (moduleMapping.name) {
        std::sprintf(oidStringPtr,"%s.%s",
                     ((String *)moduleMapping.value) -> getChars(),name);
        }
      else if (strcmp(name,"iso") == 0) {
        //
        // Handle global definitions by using an empty module name (thus
        // code that assumes a dot must be present will not break so
        // badly)
        //
        std::sprintf(oidStringPtr,".%s",name);
        }
      else {
        //
        // The name is local (or a mistake) so use the current module
        //
        std::sprintf(oidStringPtr,"%s.%s",currentModuleName,name);
        }
      }

    oidStringPtr += std::strlen(oidStringPtr);
    }

  return oidString;
  }

void XmlLoadOperation::writeModuleHeading(XmlElement moduleElement) {
  Xml dbXml;
  XmlElement root = dbXml.setRootElement("module");

  const char *language = xml -> getElementAttributeValue(moduleElement,1);
  if (language != NULL) {
    dbXml.addElementAttribute(root,"language",language);
    }

  char key[128];
  std::sprintf(key,"/MODULE/%s",currentModuleName);
  manager -> writeXml(dbXml,root,key);
  }

void XmlLoadOperation::writeModule(XmlElement moduleElement) {

  currentModuleName = xml -> getElementAttributeValue(moduleElement,0);

  //
  // If the module is already loaded, we unload the old one on the
  // assumption that something might have changed
  //
  ListIterator it = manager -> getModuleList() -> iterator();
  while (it.hasNext()) {
    if (strcmp((char *)it.next(),currentModuleName) == 0) {
      //
      // For now, run the unloader to completion
      //
      XmlDbOperation *unloader = manager -> dbUnloader(currentModuleName);
      while (unloader -> getState() == Running) {
        unloader -> doWork();
        }
      if (unloader -> getState() == Error) {
        state = Error;
        exception = unloader -> takeException();
        delete unloader;
        THROW(exception,);
        }
      delete unloader;

      break;
      }
    }

  //
  // Ready to add this module
  //

  //
  // Create the module in the module manager, so it can receive
  // loaded OID assignments
  //
  manager -> addModule(currentModuleName);

  writeModuleHeading(moduleElement);

  ObjectHashMap importsMap;
  int children = xml -> getElementChildCount(moduleElement);
  for (int i = 0 ; i < children ; i++ ) {
    XmlElement element = xml -> getElementChild(moduleElement,i);
    const char *name = xml -> getElementName(element);

    if (strcmp(name,"imports") == 0) {
      processModuleImports(element,importsMap);
      }
    else if (strcmp(name,"assignments") == 0) {
      processModuleAssignments(element,importsMap);
      }
    else if (strcmp(name,"typedefs") == 0) {
      processModuleTypedefs(element);
      }
    }

  //
  // Write the module imports map to the database
  //
  writeModuleImports(importsMap);
  }

void XmlLoadOperation::processModuleAssignment(XmlElement assignmentElement,
                                               const char *assignmentName,
                                               ObjectHashMap& importsMap) {
  //
  // For assignments, we have an index record to the database, plus
  // a record for each assignment, possibly plus a record for its
  // description clause. The index record allow a client
  // to build a basic oid <-> name mapping without spending a lot
  // of time reading from the database, and then to be able to
  // fetch more detail on demand (for example in response to user
  // requests in a GUI).
  //
  // The index record contain XML entries in no special order,
  // and the naming attributes will be fully-qualified
  // in all cases (we add the defining module name).
  //

  //
  // Find the OID element for this assignment and add a mapping for it
  //
  int children = xml -> getElementChildCount(assignmentElement);
  for (int i = 0 ; i < children ; i++ ) {
    XmlElement element = xml -> getElementChild(assignmentElement,i);
    if (strcmp(xml -> getElementName(element),"oid") == 0) {
      const char *oidString = buildOidString(element,importsMap);
      if (exception) return;

      char fullOidName[256];
      std::sprintf(fullOidName,"%s.%s",currentModuleName,assignmentName);
      manager -> addOidMapping(fullOidName,oidString);
      }
    else if (strcmp(xml -> getElementName(element),"description") == 0) {
      //
      // Write the description element to its own database record,
      //
      char key[128];
      std::sprintf(key,"/MODULE/%s/assignments/%s/description",
                   currentModuleName,assignmentName);
      manager -> writeXml(*xml,element,key);

      //
      // Remove the contents of the element (not the element itself)
      //
      xml -> setElementContents(element,"");
      }
    }

  //
  // Write a copy of the assignment XML to the database as a record
  //
  char key[128];
  std::sprintf(key,"/MODULE/%s/assignments/%s",currentModuleName,
                                               assignmentName);
  manager -> writeXml(*xml,assignmentElement,key);
  }

void XmlLoadOperation::processModuleAssignments(XmlElement assignmentsElement,
                                                ObjectHashMap& importsMap) {
  Xml dbIndexXml;
  XmlElement root = dbIndexXml.setRootElement("assignments");

  int children = xml -> getElementChildCount(assignmentsElement);
  for (int i = 0 ; i < children ; i++ ) {
    XmlElement assignmentElement = xml -> getElementChild(assignmentsElement,i);

    XmlElement dbAssignmentElement =
          dbIndexXml.addElement(root,xml -> getElementName(assignmentElement));
    const char *assignmentName =
                    xml -> getElementAttributeValue(assignmentElement,0);
    dbIndexXml.addElementAttribute(dbAssignmentElement,"name",assignmentName);

    processModuleAssignment(assignmentElement,assignmentName,importsMap);
    }

  //
  // Write the assignments index record
  //
  char key[128];
  std::sprintf(key,"/MODULE/%s/assignments",currentModuleName);
  manager -> writeXml(dbIndexXml,root,key);
  }

void XmlLoadOperation::processModuleImportsFrom(XmlElement importModuleElement,
                                                const char *importModuleName,
                                                ObjectHashMap& importsMap) {
  int children = xml -> getElementChildCount(importModuleElement);
  for (int i = 0 ; i < children ; i++ ) {
    XmlElement importSymbolElement =
                         xml -> getElementChild(importModuleElement,i);
    const char *importSymbolName =
                     xml -> getElementAttributeValue(importSymbolElement,0);

    //
    // Add a lookup entry to the imports map, so we can find out
    // which module a given symbol was imported from
    //
    importsMap.add(importSymbolName,importModuleName);
    }
  }

void XmlLoadOperation::processModuleImports(XmlElement importsElement,
                                            ObjectHashMap& importsMap) {
  int children = xml -> getElementChildCount(importsElement);
  for (int i = 0 ; i < children ; i++ ) {
    XmlElement importModuleElement = xml -> getElementChild(importsElement,i);
    const char *importModuleName =
                    xml -> getElementAttributeValue(importModuleElement,0);

    processModuleImportsFrom(importModuleElement,importModuleName,importsMap);
    }
  }

void XmlLoadOperation::processModuleTypedef(XmlElement typedefElement,
                                            const char *typedefName) {
  int children = xml -> getElementChildCount(typedefElement);
  for (int i = 0 ; i < children ; i++ ) {
    XmlElement element = xml -> getElementChild(typedefElement,i);
    if (strcmp(xml -> getElementName(element),"description") == 0) {
      //
      // Write the description element to its own database record,
      //
      char key[128];
      std::sprintf(key,"/MODULE/%s/typedefs/%s/description",
                   currentModuleName,typedefName);
      manager -> writeXml(*xml,element,key);

      //
      // Remove the contents of the element (not the element itself)
      //
      xml -> setElementContents(element,"");
      }
    }

  //
  // Write a copy of the typedef XML to the database as a record
  //
  char key[128];
  std::sprintf(key,"/MODULE/%s/typedefs/%s",currentModuleName,typedefName);
  manager -> writeXml(*xml,typedefElement,key);
  }

void XmlLoadOperation::processModuleTypedefs(XmlElement typedefsElement) {
  Xml dbIndexXml;
  XmlElement root = dbIndexXml.setRootElement("typedefs");

  int children = xml -> getElementChildCount(typedefsElement);
  for (int i = 0 ; i < children ; i++ ) {
    XmlElement typedefElement = xml -> getElementChild(typedefsElement,i);
    XmlElement dbTypedefElement = dbIndexXml.addElement(root,"typedef");

    const char *typedefName =
                    xml -> getElementAttributeValue(typedefElement,0);
    dbIndexXml.addElementAttribute(dbTypedefElement,"name",typedefName);

    const char *typedefStatus =
                    xml -> getElementAttributeValue(typedefElement,1);
    if (typedefStatus != 0) {
      dbIndexXml.addElementAttribute(dbTypedefElement,"status",typedefStatus);
      }

    processModuleTypedef(typedefElement,typedefName);
    }

  //
  // Write the typedefs index record
  //
  char key[128];
  std::sprintf(key,"/MODULE/%s/typedefs",currentModuleName);
  manager -> writeXml(dbIndexXml,root,key);
  }

void XmlLoadOperation::writeModuleImports(ObjectHashMap& importsMap) {
  Xml dbXml;
  XmlElement root = dbXml.setRootElement("imports");

  ObjectHashMapIterator it = importsMap.iterator();
  while (it.hasNext()) {
    ObjectMapping mapping = it.next();
    XmlElement dbSymbolElement = dbXml.addElement(root,"symbol");

    dbXml.addElementAttribute(dbSymbolElement,"name",
                              ((String *)mapping.name) -> getChars());

    dbXml.addElementAttribute(dbSymbolElement,"importsFrom",
                              ((String *)mapping.value) -> getChars());
    }

  char key[128];
  std::sprintf(key,"/MODULE/%s/imports",currentModuleName);
  manager -> writeXml(dbXml,root,key);
  }

void XmlLoadOperation::doWork() {
  if (currentModuleIndex < moduleCount) {
    writeModule(xml -> getElementChild(root,currentModuleIndex++));
    if (exception) {
      state = Error;
      }
    }
  else {
    if (exception) {
      state = Error;
      }
    else {
      state = Completed;
      }
    }
  }

const char *XmlLoadOperation::getCurrentModuleName() const {
  return currentModuleName;
  }

int XmlLoadOperation::getProgress() const {
  if (moduleCount == 0) return 100;
  return 100 * (currentModuleIndex + 1) / moduleCount;
  }
