/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstdio>
#include <cstdlib>

#include "Config.hpp"

#include <qfile.h>
#include <qtextstream.h>
#include <qmap.h>

//
// Private data for the Config class
//
class ConfigPrivate {
  public:
    ConfigPrivate(QString name) : configName(name), currentGroup("General") {}

    typedef QMap<QString,QString> GroupMap;      // Key/value pairs for 1 group
    typedef QMap<QString,GroupMap> ConfigMap;    // Groups

    //
    // Write an entry to the configuration, replacing/creating as necessary
    //
    void putEntry(QString key,QString value) {
      ConfigMap::Iterator it = configMap.find(currentGroup);
      if (it == configMap.end()) {
        it = configMap.insert(currentGroup,GroupMap());
        }
      if (it.data().contains(key)) it.data().replace(key,value);
      else it.data().insert(key,value);
      }

    QString getEntry(QString key) {
      ConfigMap::Iterator it = configMap.find(currentGroup);
      if (it == configMap.end()) return QString();
      GroupMap::Iterator git = it.data().find(key);
      if (git == it.data().end()) return QString();
      return git.data();
      }

    void deleteEntry(QString key) {
      ConfigMap::Iterator it = configMap.find(currentGroup);
      if (it == configMap.end()) return;
      it.data().remove(key);
      }

    QString configName;
    QString currentGroup;
    ConfigMap configMap;
  };

Config::Config(const QString& name,Domain domain) {
  data = new ConfigPrivate(name);
  QString configPath = getenv("HOME");
  configPath = configPath + "/." + data -> configName + ".prefs";
  QFile config(configPath);
  if (config.open(IO_ReadOnly)) {
    QTextStream stream(&config);
    while (! stream.atEnd()) {
      QString line = stream.readLine().stripWhiteSpace();
      if (line.startsWith("#")) continue;
      if (line.startsWith("[")) {
        setGroup(line.mid(1,line.length() - 2));
        }
      else {
        int pos = line.find('=');
        writeEntry(line.left(pos),line.right(line.length() - pos - 1));
        }
      }
    }
  }

Config::~Config() {
  QString configPath = getenv("HOME");
  configPath = configPath + "/." + data -> configName + ".prefs";
  QFile config(configPath);
  if (config.open(IO_WriteOnly)) {
    QTextStream stream(&config);
    ConfigPrivate::ConfigMap::Iterator it = data -> configMap.begin();
    while (it != data -> configMap.end()) {
      stream << "[" << it.key().latin1() << "]\n";
      ConfigPrivate::GroupMap::Iterator git = it.data().begin();
      while (git != it.data().end()) {
        stream << git.key().latin1() << "=" << git.data().latin1() << "\n";

        ++git ;
        }

      ++it ;
      }

    config.close();
    }

  delete data;
  }

void Config::setGroup(const QString& gname) {
  data -> currentGroup = gname;
  }

void Config::writeEntry(const QString& key,int num) {
  data -> putEntry(key,QString().setNum(num));
  }

void Config::writeEntry(const QString& key,const QString& value) {
  data -> putEntry(key,value);
  }

QString Config::readEntry(const QString& key,const QString& deflt) const {
  QString result = data -> getEntry(key);
  if (! result.isNull()) return result;
  return deflt;
  }

bool Config::readBoolEntry(const QString& key,bool deflt) const {
  QString result = data -> getEntry(key);
  if (! result.isNull()) return result.toInt();
  return deflt;
  }

int Config::readNumEntry(const QString& key,int deflt) const {
  QString result = data -> getEntry(key);
  if (! result.isNull()) return result.toInt();
  return deflt;
  }

void Config::removeEntry(const QString& key) {
  data -> deleteEntry(key);
  }
