/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstring>
#include <cctype>
#include <cstdlib>

#include <classlib/String.hpp>
#include <classlib/StringBuffer.hpp>
#include <classlib/ObjectPair.hpp>
#include <classlib/ObjectList.hpp>

#include "Smi.hpp"

#include "SmiTypes.hpp"
#include "SmiReferenceCache.hpp"
#include "SmiUtility.hpp"
#include "SmiObjectType.hpp"
#include "SmiSize.hpp"
#include "SmiIndexItem.hpp"

const int HEXDUMP_OCTETS_PER_LINE = 16;

Smi::Smi() : smiDb(0), moduleList(0), oidTree(0), referenceCache(0) {
  }

Smi::Smi(const char *dbPath) : smiDb(0), moduleList(0), oidTree(0),
                               referenceCache(0) {
  open(dbPath);
  }

Smi::~Smi() {
  delete smiDb;
  delete moduleList;
  delete referenceCache;
  delete oidTree;
  }

void Smi::open(const char *dbPath) {
  smiDb = new SmiDb(dbPath,Db::Read);
  if (smiDb -> exception) {
    RETHROW(smiDb -> exception,);
    }

  moduleList = new SmiModuleList(smiDb);
  referenceCache = new SmiReferenceCache();
  oidTree = new SmiOidTree(smiDb);
  }

SmiOidTree *Smi::getOidTree() const {
  return oidTree;
  }

SmiModuleList *Smi::getModuleList() const {
  return moduleList;
  }

SmiOid *Smi::getOid(const char *fullName) const {

  //
  // If the first character is a digit then the whole argument is numeric
  //
  if (isdigit(fullName[0])) {
    return new SmiOid(fullName);
    }

  //
  // Find out where the first two dots occur. This will help decide what
  // components are present. We already know that an object name is present,
  // possibly preceded by a module name
  //
  const char *dot1Ptr = std::strchr(fullName,'.');
  const char *dot2Ptr = 0;
  if (dot1Ptr) dot2Ptr = std::strchr(dot1Ptr + 1,'.');

  char objectName[256];
  const char *remainder = 0;
  if (dot1Ptr) {
    //
    // We have at least two components
    //
    if (isalpha(dot1Ptr[1])) {
      //
      // The second component begins with a letter, so the first component
      // must be a module name with the second as an object name within
      // that module
      //
      if (dot2Ptr) {
        int length = dot2Ptr - fullName;
        std::memcpy(objectName,fullName,length);
        objectName[length] = '\0';

        remainder = dot2Ptr + 1;
        }
      else {
        std::strcpy(objectName,fullName);
        }
      }
    else {
      //
      // The second component is not the object name, so the first component
      // is the unqualified object name.
      //
      int length = dot1Ptr - fullName;
      std::memcpy(objectName,fullName,length);
      objectName[length] = '\0';

      remainder = dot1Ptr + 1;
      }
    }
  else {
    //
    // Only one component is present, so it is the object name
    //
    std::strcpy(objectName,fullName);
    }

  //
  // We have either the qualified or unqualified object name so
  // we can perform the lookup/search
  //
  const SmiOidTreeNode *node = oidTree -> getNode(objectName);
  if (! node) return 0;

  //
  // Now create an OID from the node we found, and append the remaining
  // sub-identifiers
  //
  SmiOid *oid = new SmiOid(node);
  if (remainder) oid -> append(SmiOid(remainder));

  return oid;
  }

static void splitName(const char *fullName,char *moduleName,int moduleNameSize,
                      char *name,int nameSize) {
  char *splitPtr = std::strchr(fullName,'.');
  if (! splitPtr) {
    int size = std::strlen(fullName) + 1;
    if (size > nameSize) size = nameSize;
    std::memcpy(name,fullName,size - 1);
    name[size - 1] = '\0';
    }
  else {
    int size = splitPtr - fullName + 1;
    if (size > nameSize) size = nameSize;
    std::memcpy(moduleName,fullName,size - 1);
    moduleName[size - 1] = '\0';

    size = std::strlen(splitPtr + 1) + 1;
    if (size > nameSize) size = nameSize;
    std::memcpy(name,splitPtr + 1,size - 1);
    name[size - 1] = '\0';
    }
  }

static String *oidStringValue(const Smi& smi,const SmiOid& oid,
                              const SmiOidTreeNode *node,int options) {
  //
  // The result begins with the node name
  //
  const char *nodeName = smi.getOidTree() -> getName(node);
  if (! nodeName) return oid.toString(); // Shouldn't happen!
  char buffer[512];
  if (options & Smi::LongOidNames) {
    std::strcpy(buffer,nodeName);
    }
  else {
    const char *nodeNamePtr = strchr(nodeName,'.');
    if (nodeNamePtr) std::strcpy(buffer,nodeNamePtr + 1);
    else std::strcpy(buffer,nodeName);
    }
  char *bufferPtr = buffer + std::strlen(buffer);

  //
  // And is followed by the residual sids in numeric form
  //
  SmiOid nodeOid(node);
  int nodeSidCount = nodeOid.getLength();
  int residualSidCount = oid.getLength() - nodeSidCount;
  const SmiTypes::SubIdentifier *sids = oid.getSids();
  for (int i = 0 ; i < residualSidCount ; i++ ) {
    std::sprintf(bufferPtr,".%ld",sids[i + nodeSidCount]);
    bufferPtr += std::strlen(bufferPtr);
    }

  return new String(buffer);
  }

static SmiObjectType *getIndexObject(const Smi& smi,
                                     const SmiModule& referringModule,
                                     const char *indexName,
                                     const SmiModule **indexModule) {

  const SmiAssignment *indexAssignment =
                 smi.findAssignment(&referringModule,indexName,indexModule);
  if ((! indexAssignment) ||
      (indexAssignment -> getType() != SmiAssignmentInfo::OBJECT_TYPE)) {
    return 0;
    }

  return (SmiObjectType *)indexAssignment;
  }

static String *getIndexValue(const Smi& smi,
                             const SmiModule& indexModule,
                             const SmiObjectType& indexObject,
                             const SmiTypes::SubIdentifier* sids,
                             int sidCount,int& sidIndex,int isImpliedLast,
                             int options) {

  //
  // The only syntaxes we provide a mapping for are Integer32/INTEGER
  // (since we may be able to get an enumeration name), OCTET STRING
  // (since we may be able to get some printable ASCII) and OBJECT
  // IDENTIFIER. TBD - Think about TimeTicks and IpAddress (DNS lookup)
  //
  const SmiSyntax *indexSyntax = indexObject.getSyntax();
  const char *hint = 0;
  const SmiTypedef *baseTypedef = 0;
  const SmiSyntax *baseSyntax = 0;
  const SmiEnumerationList *enumerations = 0;
  smi.getTypeInfo(indexModule,*indexSyntax,&baseTypedef,&baseSyntax,&hint,
                  &enumerations);

  //
  // See if the base typedef is one we want to handle specially (the
  // normal formatting for the base type is not appropriate, and
  // there will be no display hint or enumeration)
  //
  if (baseTypedef) {
    const char *name = baseTypedef -> getName();
    if (std::strcmp(name,"IpAddress") == 0) {
      //
      // For now, we just format an IP address as the dotted quad
      //
      if (sidCount - sidIndex < 4) return 0;
      char scratch[32];
      std::sprintf(scratch,"%u.%u.%u.%u",
                   (unsigned)sids[sidIndex],(unsigned)sids[sidIndex + 1],
                   (unsigned)sids[sidIndex + 2],(unsigned)sids[sidIndex + 3]);
      sidIndex += 4;
      return new String(scratch);
      }
    }

  //
  // Format based on the base type, enumerations and display hint
  //
  SmiSyntax::Type type = baseSyntax -> getType();
  if (type == SmiSyntax::ST_Integer) {
    //
    // We know we will consume 1 sub-identifier
    //
    if (sidCount - sidIndex < 1) return 0;
    long value = sids[sidIndex++];

    //
    // See if a DISPLAY-HINT is present
    //
    if (hint) {
      return SmiUtility::hintedSignedInteger(hint,value);
      }

    //
    // No DISPLAY-HINT, so see if there's an enumeration
    //
    if (enumerations) {
      const char *enumerationName = enumerations -> findEnumeration(value);
      if (enumerationName) {
        return new String(enumerationName);
        }
      }

    //
    // Plain integer
    //
    char scratch[32];
    std::sprintf(scratch,"%ld",value);
    return new String(scratch);
    }
  else if (type == SmiSyntax::ST_OctetString) {
    //
    // An octet string is coded as a sub-identifier for each octet,
    // possibly preceded by a sub-identifier containing the length.
    //
    // If isImpliedLast, the length is not present.
    //
    // If the syntax has an ASN.1 tag that specifies an implicit length,
    // then that length is used (IpAddress does this, for example).
    //
    int length;
    if (isImpliedLast) length = sidCount - sidIndex;
    else {
      const SmiAsn1Tag *tag = baseSyntax -> getTag();
      if (tag && tag -> getImplicit()) {
        SmiSizeList *sizes = baseSyntax -> getSizeList();
        ListIterator it = sizes -> iterator();
        SmiSize *size = (SmiSize *)it.next();
        length = size -> getMin();
        }
      else {
        if (sidCount - sidIndex < 1) {
          //
          // Length sub-identifier is missing
          //
          return 0;
          }

        length = sids[sidIndex++ ];
        }
      }

    //
    // Check the length in case the octet string is truncated
    //
    if (length > (sidCount - sidIndex)) {
      //
      // String claims N sub-identifiers, but there aren't that many
      // present!
      //
      length = sidCount - sidIndex;
      }

    //
    // Format the sub-identifiers. First determine if we have a printable
    // string
    //
    int binary = 0;
    for (int i = 0 ; i < length ; i++ ) {
      if (! isprint(sids[sidIndex + i])) {
        binary = 1;
        break;
        }
      }

    if (binary) {
      int resultLength = length * 2 + 3;
      char scratch[resultLength];
      scratch[0] = '\'' ; scratch[resultLength - 2] = '\'';
      scratch[resultLength - 1] = 'h';
      for (int i = 0 ; i < length ; i++ ) {
        int digit = (sids[sidIndex + i] >> 4) & 0x0F;
        if (digit < 10) scratch[2 * i + 1] = '0' + digit;
        else scratch[2 * i + 1] = 'a' + (digit - 10);

        digit = sids[sidIndex + i] & 0x0F;
        if (digit < 10) scratch[2 * i + 2] = '0' + digit;
        else scratch[2 * i + 2] = 'a' + (digit - 10);
        }
      sidIndex += length;
      return new String(scratch,resultLength);
      }
    else {
      int resultLength = length + 2;
      char scratch[resultLength];
      scratch[0] = '"'; scratch[resultLength - 1] = '"';
      for (int i = 0 ; i < length ; i++ ) {
        scratch[i + 1] = sids[sidIndex + i];
        }
      sidIndex += length;
      return new String(scratch,resultLength);
      }
    }
  else if (type == SmiSyntax::ST_ObjectIdentifier) {
    //
    // An object identifier is coded as the number of sub-identifiers
    // followed by the sub-identifiers. If isImpliedLast,
    // the length is not present.
    //
    int length;
    if (isImpliedLast) length = sidCount - sidIndex;
    else {
      if (sidCount - sidIndex < 1) {
        //
        // Length sub-identifier is missing
        //
        return 0;
        }

      length = sids[sidIndex++ ];
      if (length > (sidCount - sidIndex)) {
        //
        // String claims N sub-identifiers, but there aren't that many present!
        //
        length = sidCount - sidIndex;
        }
      }

    SmiOid valueOid(&sids[sidIndex],length);
    sidIndex += length;

    return smi.stringValue(valueOid,options);
    }

  return 0;
  }

static bool addIndexValue(const Smi& smi,const SmiModule& module,
                          const char *indexName,
                          const SmiTypes::SubIdentifier* sids,int sidCount,
                          int& sidIndex,int isImpliedLast,StringBuffer& buffer,
                          int options) {

  const SmiModule *indexModule;
  const SmiObjectType *indexObject = getIndexObject(smi,module,indexName,
                                                    &indexModule);
  if (! indexObject) return false;

  String *value = getIndexValue(smi,*indexModule,*indexObject,
                                sids,sidCount,sidIndex,isImpliedLast,
                                options);
  if (! value) return false;

  buffer.append(".");
  buffer.append(value -> getChars());
  delete value;

  return true;
  }

ObjectList *Smi::getIndexValues(const SmiOid& oid,int options) const {
  //
  // Find the first named node at or above the one with this OID. This
  // should at least find "iso". We won't know for a while whether
  // this node is a leaf node (TBD - add leaf information into node?)
  //
  const SmiOidTreeNode *node = oidTree -> findDeepestNamed(&oid);
  if (! node) return 0; // Shouldn't happen!

  //
  // We need the Entry assignment for this node so we can see how to
  // interpret the residual sids. The Entry assignment gives the
  // table indices, and the index definitions give syntaxes for
  // indexes.
  //
  // The Entry node is the parent of this node
  //
  SmiOidTreeNode *entryNode = node -> getParent();
  if (! entryNode) {
    //
    // The node has no parent, so we will be unable to determine the
    // indexing
    //
    return 0;
    }

  //
  // The assignment is fetched by name from a module, so we need
  // the node's fully-qualified name
  //
  const char *entryNodeName = oidTree -> getName(entryNode);
  if (! entryNodeName) {
    //
    // The parent node is not named, so it doesn't exist as an
    // assignment in our database. That means we can't fetch
    // any information about indexes.
    //
    return 0;
    }

  //
  // Split the name into a module name and an assignment name
  //
  char moduleName[128];
  char assignmentName[128];
  splitName(entryNodeName,moduleName,sizeof(moduleName),
            assignmentName,sizeof(assignmentName));

  //
  // Fetch the named module
  //
  const SmiModule *module = getModuleList() -> getModule(moduleName);
  if (! module) {
    //
    // Reference to a module we don't have, so we can't get index
    // information
    //
    return 0;
    }

  //
  // Fetch the entry assignment from the entry's module and interpret
  // it as an OBJECT TYPE assignment
  //
  const SmiAssignment *assignment = module -> getAssignment(assignmentName);
  if ((! assignment) || (assignment -> getType() !=
                              SmiAssignmentInfo::OBJECT_TYPE)) {
    //
    // Reference to an assignment we don't have (unusual since we have
    // the module. Possibly there is a mis-spelled import or a version
    // discrepancy)
    //
    return 0;
    }
  SmiObjectType *objectType = (SmiObjectType *)assignment;

  //
  // Get the index list for the entry node
  //
  SmiIndexList *indices = objectType -> getIndexList();
  if (! indices) {
    //
    // No indexes, so this object was scalar, or we don't have the Entry
    // assignment for this OID (which happens when we don't have the
    // right MIB loaded)
    //
    return 0;
    }

  //
  // 
  //
  ObjectList *indexValues = new ObjectList();

  //
  // We need to know how many sids of the OID are accounted for by
  // the named node, so we can tell where the residual sids start
  // in the OID
  //
  SmiOid namedOid(node);
  int namedSidCount = namedOid.getLength();
  int oidSidCount = oid.getLength();

  //
  // For each index, find the syntax and format the indicated sub-identifiers
  // into a string, appending it to the result
  //
  int sidIndex = namedSidCount;
  const SmiTypes::SubIdentifier *sids = oid.getSids();
  ListIterator it = indices -> iterator();
  while (it.hasNext()) {
    //
    // We'll need the index name
    //
    const SmiIndexItem *indexItem = (SmiIndexItem *)it.next();
    const char *indexName = indexItem -> getName();

    //
    // We also need to know if this index is coded using the IMPLIED
    // convention. This means that the rightmost index can omit any
    // explicit length from the coding, since the length can be
    // inferred from the OID length
    //
    int isImpliedLast = indices -> getImplied() && (! it.hasNext());

    //
    // Add the index value to the list
    //
    const SmiModule *indexModule;
    SmiObjectType *indexObject = getIndexObject(*this,*module,indexName,
                                                &indexModule);

    if (! indexObject) break;

    String *value = getIndexValue(*this,*indexModule,*indexObject,
                                  sids,oidSidCount,
                                  sidIndex,isImpliedLast,options);
    if (! value) break;

    char scratch[128];
    std::sprintf(scratch,"%s.%s",indexModule -> getName(),indexName);
    ObjectPair *indexValue = new ObjectPair(new String(scratch),value);
    indexValues -> add(indexValue);
    }

  return indexValues;
  }

String *Smi::stringValue(const SmiOid& oid,int options) const {
  //
  // Find the first named node at or above the one with this OID. This
  // should at least find "iso". We won't know for a while whether
  // this node is a leaf node (TBD - add leaf information into node?)
  //
  const SmiOidTreeNode *node = oidTree -> findDeepestNamed(&oid);
  if (! node) return oid.toString(); // Shouldn't happen!

  //
  // We need the Entry assignment for this node so we can see how to
  // interpret the residual sids. The Entry assignment gives the
  // table indices, and the index definitions give syntaxes for
  // indexes.
  //
  // The Entry node is the parent of this node
  //
  SmiOidTreeNode *entryNode = node -> getParent();
  if (! entryNode) {
    //
    // The node has no parent, so we will be unable to determine the
    // indexing
    //
    return oidStringValue(*this,oid,node,options);
    }

  //
  // The assignment is fetched by name from a module, so we need
  // the node's fully-qualified name
  //
  const char *entryNodeName = oidTree -> getName(entryNode);
  if (! entryNodeName) {
    //
    // The parent node is not named, so it doesn't exist as an
    // assignment in our database. That means we can't fetch
    // any information about indexes.
    //
    return oidStringValue(*this,oid,node,options);
    }

  //
  // Split the name into a module name and an assignment name
  //
  char moduleName[128];
  char assignmentName[128];
  splitName(entryNodeName,moduleName,sizeof(moduleName),
            assignmentName,sizeof(assignmentName));

  //
  // Fetch the named module
  //
  const SmiModule *module = getModuleList() -> getModule(moduleName);
  if (! module) {
    //
    // Reference to a module we don't have, so we can't get index
    // information
    //
    return oidStringValue(*this,oid,node,options);
    }

  //
  // Fetch the entry assignment from the entry's module and interpret
  // it as an OBJECT TYPE assignment
  //
  const SmiAssignment *assignment = module -> getAssignment(assignmentName);
  if ((! assignment) || (assignment -> getType() !=
                              SmiAssignmentInfo::OBJECT_TYPE)) {
    //
    // Reference to an assignment we don't have (unusual since we have
    // the module. Possibly there is a mis-spelled import or a version
    // discrepancy)
    //
    return oidStringValue(*this,oid,node,options);
    }
  SmiObjectType *objectType = (SmiObjectType *)assignment;

  //
  // Get the index list for the entry node
  //
  SmiIndexList *indices = objectType -> getIndexList();
  if (! indices) {
    //
    // No indexes, so this object was scalar, or we don't have the Entry
    // assignment for this OID (which happens when we don't have the
    // right MIB loaded)
    //
    return oidStringValue(*this,oid,node,options);
    }

  //
  // The result is the object's node name appended with the string
  // representations of the index values
  //
  StringBuffer buffer;
  const char *nodeName = oidTree -> getName(node);
  if (options & Smi::LongOidNames) {
    buffer.append(nodeName);
    }
  else {
    const char *nodeNamePtr = strchr(nodeName,'.');
    if (nodeNamePtr) buffer.append(nodeNamePtr + 1);
    else buffer.append(nodeName);
    }

  //
  // We need to know how many sids of the OID are accounted for by
  // the named node, so we can tell where the residual sids start
  // in the OID
  //
  SmiOid namedOid(node);
  int namedSidCount = namedOid.getLength();
  int oidSidCount = oid.getLength();

  //
  // For each index, find the syntax and format the indicated sub-identifiers
  // into a string, appending it to the result
  //
  int sidIndex = namedSidCount;
  const SmiTypes::SubIdentifier *sids = oid.getSids();
  ListIterator it = indices -> iterator();
  while (it.hasNext()) {
    //
    // We'll need the index name
    //
    const SmiIndexItem *indexItem = (SmiIndexItem *)it.next();
    const char *indexName = indexItem -> getName();

    //
    // We also need to know if this index is coded using the IMPLIED
    // convention. This means that the rightmost index can omit any
    // explicit length from the coding, since the length can be
    // inferred from the OID length
    //
    int isImpliedLast = indices -> getImplied() && (! it.hasNext());

    //
    // Add the index value (preceded by a dot)
    //
    if (! addIndexValue(*this,*module,indexName,sids,oidSidCount,
                        sidIndex,isImpliedLast,buffer,options)) {
      break;
      }
    }

  //
  // Append any residual sub-identifiers as plain integers
  //
  while (sidIndex < oidSidCount) {
    char scratch[16];
    std::sprintf(scratch,".%ld",sids[sidIndex]);
    buffer.append(scratch);

    sidIndex++ ;
    }

  return new String(buffer);
  }

const SmiAssignment *
           Smi::findAssignment(const SmiModule *referringModule,
                               const char *assignmentName,
                               const SmiModule **definingModule) const {
  //
  // Check to see if the referring module itself defines the assignment
  //
  const SmiAssignment *assignment =
                          referringModule -> getAssignment(assignmentName);
  if (assignment) {
    *definingModule = referringModule;
    return assignment;
    }

  //
  // The definition is not local, so we have to find an import for
  // it
  //
  const char *definingModuleName =
                      referringModule -> getModuleImport(assignmentName);
  if (definingModuleName) {
    *definingModule = moduleList -> getModule(definingModuleName);
    if (*definingModule) {
      return (*definingModule) -> getAssignment(assignmentName);
      }
    }

  return 0;
  }

const SmiTypedef *Smi::findTypedef(const SmiModule *referringModule,
                                   const SmiSyntax *syntax,
                                   const SmiModule **definingModule) const {
  //
  // If the typeref is already a *common* base type, we can avoid going to
  // the database or the module's typedef cache. Another way would be
  // to add these guys to the typedef cache with a null syntax.
  //
  SmiSyntax::Type type = syntax -> getType();
  if (type != SmiSyntax::ST_NotBaseType) return 0;

  //
  // Check to see if the referring module itself defines the type
  //
  const char *typeref = syntax -> getTyperef();
  const SmiTypedef *typedefinition = referringModule -> getTypedef(typeref);
  if (typedefinition) {
    *definingModule = referringModule;
    return typedefinition;
    }

  //
  // The definition is not local, so we have to find an import for
  // it
  //
  const char *definingModuleName = referringModule -> getModuleImport(typeref);
  if (definingModuleName) {
    *definingModule = moduleList -> getModule(definingModuleName);
    if (*definingModule) {
      return (*definingModule) -> getTypedef(typeref);
      }
    }

  //
  // Nothing found, so it is either an undefined type or a base type
  //
  return 0;
  }

const SmiTypedef *Smi::findHintedTypedef(const SmiModule *referringModule,
                                         const SmiSyntax *syntax) const {

  const SmiModule *definingModule;
  const SmiTypedef *typedefinition = findTypedef(referringModule,
                                                 syntax,
                                                 &definingModule);
  if (typedefinition) {
    referringModule = definingModule;
    while (typedefinition) {
      if (typedefinition -> getDisplayHint()) return typedefinition;

      //
      // Iterate using the type reference of the current syntax
      //
      typedefinition = findTypedef(referringModule,
                                   typedefinition -> getSyntax(),
                                   &definingModule);
      referringModule = definingModule;
      }
    }

  return 0;
  }

const SmiSyntax *Smi::getBaseSyntax(const SmiModule *referringModule,
                                    const SmiSyntax *syntax) const {

  const SmiModule *definingModule;
  const SmiTypedef *typedefinition = findTypedef(referringModule,
                                                 syntax,
                                                 &definingModule);

  while (typedefinition) {
    syntax = typedefinition -> getSyntax();

    typedefinition = findTypedef(referringModule,syntax,&definingModule);
    referringModule = definingModule;
    }

  return syntax;
  }

void Smi::getTypeInfo(const SmiModule& referringModule,const SmiSyntax& syntax,
                      const SmiTypedef **baseTypedef,
                      const SmiSyntax **baseSyntax,const char **hint,
                      const SmiEnumerationList **enumerations) const {
  //
  // The type information is scattered up and down the type chain, so
  // this method scans the chain and remembers the interesting stuff
  //

  //
  // First place to look for enumerations is on the top-most syntax
  //
  if (! *enumerations) {
    *enumerations = syntax.getEnumerationList();
    }

  //
  // Now we see if the syntax has a typeref that refers to a typedef
  // (as opposed to referring to a base type). If it does, then we
  // look at the typedef and its syntax. We then see if the typedef's
  // syntax has a typeref that refers to a typedef, and so on
  //
  const SmiModule *definingModule;
  const SmiSyntax *thisSyntax = &syntax;
  const SmiModule *thisModule = &referringModule;
  const SmiTypedef *typedefinition = findTypedef(thisModule,thisSyntax,
                                                 &definingModule);

  while (typedefinition) {
    //
    // Each typedef could provide a display hint, but we only use the
    // highest one
    //
    if (! *hint) {
      *hint = typedefinition -> getDisplayHint();
      }

    //
    // Fetch the syntax of this type definition
    //
    thisSyntax = typedefinition -> getSyntax();

    //
    // Each typedef could provide an enumeration list, but we only use the
    // highest one
    //
    if (! *enumerations) {
      *enumerations = thisSyntax -> getEnumerationList();
      }

    //
    // Remember this type definition in case it is the deepest, and
    // then fetch the next for the next iteration
    //
    *baseTypedef = typedefinition;
    typedefinition = findTypedef(thisModule,thisSyntax,&definingModule);
    thisModule = definingModule;
    }

  *baseSyntax = thisSyntax;
  }

String *Smi::unsignedIntegerString(const SmiModule *referringModule,
                                   const SmiSyntax *syntax,
                                   unsigned long value) const {
  if (syntax) {
    const SmiTypedef *typedefinition = findHintedTypedef(referringModule,
                                                         syntax);
    if (typedefinition) {
      return SmiUtility::hintedUnsignedInteger(
                              typedefinition -> getDisplayHint(),value);
      }
    }

  return SmiUtility::hintedUnsignedInteger("d",value);
  }

String *Smi::signedIntegerString(const SmiModule *referringModule,
                                 const SmiSyntax *syntax,
                                 const long value) const {
  if (syntax) {
    const SmiTypedef *typedefinition = findHintedTypedef(referringModule,
                                                         syntax);
    if (typedefinition) {
      return SmiUtility::hintedSignedInteger(
                                typedefinition -> getDisplayHint(),value);
      }

    syntax = getBaseSyntax(referringModule,syntax);

    const char *enumerationName = syntax -> findEnumeration(value);
    if (enumerationName) return new String(enumerationName);
    }

  char scratch[64];
  SmiUtility::writeSignedInteger(value,0,10,scratch,sizeof(scratch));
  return new String(scratch);
  }

String *Smi::timeticksString(const SmiModule *referringModule,
                             const SmiSyntax *syntax,
                             unsigned long value) const {
  if (syntax) {
    const SmiTypedef *typedefinition = findHintedTypedef(referringModule,
                                                         syntax);
    if (typedefinition) {
      return SmiUtility::hintedUnsignedInteger(
                                   typedefinition -> getDisplayHint(),value);
      }
    }

  //
  // Ticks are in centiseconds, and represent a difference in time
  // between two epochs
  //
  // Maximum time unit reported here is a day (results in reasonably
  // small numbers)
  //
  StringBuffer buffer;
  char scratch[64];
  std::sprintf(scratch,"(%lu) ",value);
  buffer.append(scratch);

  int days = value / (100 * 60 * 60 * 24);
  value = value - days * (100 * 60 * 60 * 24);
  int hours = value / (100 * 60 * 60);
  value = value - hours * (100 * 60 * 60);
  int minutes = value / (100 * 60);
  value = value - minutes * (100 * 60);
  int seconds = value / (100);
  value = value - seconds * (100);

  if (days) {
    std::sprintf(scratch,"%d days, ",days);
    buffer.append(scratch);
    }
  std::sprintf(scratch,"%02d:%02d:%02d.%02d",hours,minutes,seconds,(int)value);
  buffer.append(scratch);

  return new String(buffer);
  }

static String *hexString(const unsigned char *buffer,int length,int options) {
  int bytes = length;
  if ((options & Smi::ShortValues) && (length > 16)) bytes = 16;

  StringBuffer output;
  output.append("0x");

  char hex[4];
  for (int i = 0 ; i < bytes ; i++ ) {
    std::sprintf(hex,"%02x ",buffer[i]);
    output.append(hex);
    }
  if (bytes != length) output.append("..");

  return new String(output);
  }

static String *hexdumpString(const unsigned char *buffer,int length) {
  StringBuffer output;
  char map[HEXDUMP_OCTETS_PER_LINE + 1];
  char hex[4];

  if (length > HEXDUMP_OCTETS_PER_LINE) output.append("\\\n");
  int mapPos = 0;
  for (int i = 0 ; i < length ; i++ ) {
    std::sprintf(hex,"%02x ",buffer[i]);
    output.append(hex);

//
//  gcc 2.95.2 for ARM doesn't seem to like namespace qualification
//  of the ctype functions
//
//    if (std::isprint(buffer[i])) map[mapPos] = buffer[i];
//
    if (isgraph(buffer[i]) || (buffer[i] == ' ')) map[mapPos] = buffer[i];
    else map[mapPos] = '.';

    if (++mapPos == HEXDUMP_OCTETS_PER_LINE) {
      output.append(" | ");
      output.append(map,mapPos);
      output.append(" |");

      if (i != length - 1) output.append("\n");
      mapPos = 0;
      }
    }

  if (mapPos != 0) {
    if (length > HEXDUMP_OCTETS_PER_LINE) {
      int spaces = HEXDUMP_OCTETS_PER_LINE - (length % HEXDUMP_OCTETS_PER_LINE);
      for (int i = 0 ; i < spaces ; i++ ) output.append("   ");
      }
    output.append(" | ");
    output.append(map,mapPos);
    for (int i = 0 ; i < (HEXDUMP_OCTETS_PER_LINE - mapPos) ; i++ ) {
      output.append(" ");
      }
    output.append(" |");
    }

  return new String(output);
  }

//
// Read a numeric string in network byte order. Note that this function
// is only good for up to 63-bit unsigned integers, or a 64 bit signed
// integer. The RFC doesn't tell us how
// to interpret the octets so I'm just guessing.
//
static long long readNumber(const unsigned char *buffer,int length) {
  long long value = 0;
  for (int i = 0 ; i < length ; i++ ) {
    value = (value << 8) + *buffer++ ;
    }

  return value;
  }

static String *hintedString(const char *hint,const unsigned char *buffer,
                            int length) {
  StringBuffer result;

  //
  // Establish the default format specification (which is supposed
  // to get over-ridden)
  //
  int repeatCount = 1;
  int count = 0;
  int format = 'x';
  int separator = 0;
  int repeatTerminator = 0;
  int repeatIndicator = 0;

  //
  // Apply formatting until there are no more octets to format
  //
  while (length) {
    //
    // If there is another format specification in the hint, pick it up
    // (otherwise, we just reapply the previous format specification). Note
    // that we're careful not to clobber the previous format specification
    // if there's nothing more in the hint string.
    //
    if (*hint) {
      //
      // The first character of a format specification is the (optional)
      // repeat indicator, which says that the repeat count for this
      // specification is given by the next octet of the input string.
      // We have to remember that we saw a repeat indicator because
      // we later need to distinguish between a repeat count that defaults
      // to 1 and a repeat count that is indicated as 1
      //
      repeatCount = 1;
      repeatIndicator = 0;
      if (*hint == '*') {
        hint++ ;
        repeatCount = *buffer++ ; length-- ;
        repeatIndicator = 1;
        }
      }

    if (*hint) {
      //
      // The next part of the format specification is the count of the
      // number of input string octets to be consumed each time we
      // iterated using this format spec
      //
      count = 0;
      while (*hint && isdigit(*hint)) {
        count *= 10;
        count += (*hint - '0');
        hint++ ;
        }
      }

    if (*hint) {
      //
      // The next part of the format specification is the format indicator
      //
      format = *hint++ ;
      }

    if (*hint) {
      //
      // The format indicator is optionally followed by a separator (to
      // be displayed after (not between!) each iteration of this format
      // spec), and if there is a separator, then there may also be an
      // optional terminator to be displayed after this format spec has
      // been repeated the *indicated* number of times (it never follows
      // a defaulted number of repeats)
      // 
      separator = 0;
      repeatTerminator = 0;
      if (*hint && (*hint != '*') && (! isdigit(*hint))) {
        separator = *hint++ ;

        if (repeatIndicator && *hint && (*hint != '*') && (! isdigit(*hint))) {
          repeatTerminator = *hint++ ;
          }
        }
      }

    //
    // Now apply the current format specification
    //
    int iterationsLeft = repeatCount;
    while (iterationsLeft-- && length) {
      //
      // The count is never allowed to exceed the number of octets
      // remaining in the input string, no matter what the format spec
      // says
      //
      if (count > length) count = length;

      //
      // The count can be specified as 0 (which means that no input
      // octets are consumed by this format spec, and the output
      // will consist of separators).
      //
      if (count) {
        switch (format) {
          case 'a':
          case 't': { // UTF-8 is only supported here as a US ASCII superset
            result.append((char *)buffer,count);
            buffer += count ; length -= count;
            break;
            }
          case 'd': {
            long long value = readNumber(buffer,count);
            buffer += count ; length -= count;

            char scratch[64];
            SmiUtility::writeNumber(value,0,10,scratch,sizeof(scratch));
            result.append(scratch);
            break;
            }
          case 'x': {
            long long value = readNumber(buffer,count);
            buffer += count ; length -= count;

            char scratch[64];
            SmiUtility::writeNumber(value,0,16,scratch,sizeof(scratch));
            result.append(scratch);
            break;
            }
          case 'o': {
            long long value = readNumber(buffer,count);
            buffer += count ; length -= count;

            char scratch[64];
            SmiUtility::writeNumber(value,0,8,scratch,sizeof(scratch));
            result.append(scratch);
            break;
            }
          default: {
            length -= count;
            break;
            }
          }
        }

      //
      // Display the separator if there is one, unless we're about to
      // display the repeat terminator (which will be the case if
      // there are no iterations left and the repeat terminator is
      // defined) or there are no more input string octets
      //
      if (separator && length &&
          (! ((! iterationsLeft) && repeatIndicator && repeatTerminator))) {
        result.append(separator);
        }
      }

    //
    // Display the repeat terminator if there is one, and it would
    // not be followed by anything
    //
    if (repeatIndicator && repeatTerminator && length) {
      result.append(repeatTerminator);
      }

    //
    // If the format spec we just used has a count of 0 (i.e. it does
    // not consume any of the input), we have to be careful about an
    // infinite loop. So, we escape if there is no next format spec
    // in the hint string (eventually, a format spec will consume
    // input string octets, or we will run out of new format specs
    // in the hint string).
    //
    if ((count == 0) && (*hint == '\0')) {
      break;
      }
    }

  return new String(result);
  }

String *Smi::octetString(const SmiModule *referringModule,
                         const SmiSyntax *syntax,
                         const unsigned char *buffer,int length,
                         int options) const {

  if (syntax) {
    //
    // If the syntax is the BITS pseudo-type, then use the BITS
    // formatter
    //
    if (std::strcmp(syntax -> getTyperef(),"BITS") == 0) {
      return bitString(referringModule,syntax,buffer,length);
      }

    const SmiTypedef *typedefinition = findHintedTypedef(referringModule,
                                                         syntax);
    if (typedefinition) {
      return hintedString(typedefinition -> getDisplayHint(),buffer,length);
      }

    //
    // If the base syntax is BITS, use the BITS formatter
    //
    syntax = getBaseSyntax(referringModule,syntax);
    if (std::strcmp(syntax -> getTyperef(),"BITS") == 0) {
      return bitString(referringModule,syntax,buffer,length);
      }
    }

  //
  // No hints, so we go by context. If any character is unprintable,
  // we make a hexdump
  //
  for (int i = 0 ; i < length ; i++ ) {
//    if (! std::isprint(buffer[i])) {
    if ((! isgraph(buffer[i])) && (buffer[i] != ' ')) {
      if (options & Smi::ShortValues) {
        return hexString(buffer,length,options);
        }
      else return hexdumpString(buffer,length);
      }
    }

  //
  // All characters are printable, so we can make a printable string
  // directly
  //
  return new String((const char *)buffer,length);
  }

String *Smi::bitString(const SmiModule *module,const SmiSyntax *syntax,
                       const unsigned char *buffer,
                       unsigned long length) const {
  //
  // BITS are encoded left to right in a variable-length octet string.
  // Bit number 0 is the left-most (most significant) bit of the
  // left-most (first) octet.
  //
  // The result format is: "[name[", "...]"
  // Example: "up, running"
  //
  StringBuffer output;
  char scratch[128];
  if (syntax) {
    //
    // Find the base syntax (diving into typerefs until we get a
    // base type).
    //
    syntax = getBaseSyntax(module,syntax);
    }

  int bitNumber = 0;
  int first = 1;
  for (unsigned i = 0 ; i < length ; i++ ) {
    for (unsigned bit = 0 ; bit < 8 ; bit++ , bitNumber++ ) {
      if (buffer[i] & (1 << (7 - bit))) {
        if (! first) {
          output.append(", ");
          }
        first = 0;

        if (syntax) {
          const char *enumerationName = syntax -> findEnumeration(bitNumber);
          if (enumerationName) {
            output.append(enumerationName);
            continue;
            }
          }

        std::sprintf(scratch,"%u",bitNumber);
        output.append(scratch);
        }
      }
    }

  return new String(output);
  }
