/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstring>
#include <unistd.h>
#include <malloc.h>
#include <cerrno>
#include <sys/types.h>
#include <dirent.h>

#include <classlib/DbFlatDir.hpp>
#include <classlib/String.hpp>
#include <classlib/List.hpp>
#include <classlib/ObjectHashMap.hpp>

#include "SmiDb.hpp"
#include "SmiModuleDb.hpp"
#include "SmiDbException.hpp"

class SmiDbPrivate {
  public:
    SmiDbPrivate(int maxOpenDbsArg)
            : dbIndex(0), dbMap(0), dbList(0), maxOpenDbs(maxOpenDbsArg) {

      dbMap = new ObjectHashMap();
      dbList = new List();
      }

    ~SmiDbPrivate() {
      delete dbIndex;
      delete dbMap;
      delete dbList;
      }

    DbFlatDir *dbIndex;
    ObjectHashMap *dbMap;
    List *dbList;
    int maxOpenDbs;
  };

//
// Wrapper class for a Db, so that we can treat it as an SmiObject
// (and put it in an ObjectHashMap). I'm avoiding multiple inheritance,
// so SmiSubDb is not both an SmiObject and a Db
//
class SmiSubDb : public Object {
  public:
    SmiSubDb(Db *dbArg) : db(dbArg) {
      }

    ~SmiSubDb() {
      delete db;
      }

    Db *db;
  };

//
// An SMI database is named by a path to a directory. The directory
// contains a Db database per module plus files for global
// indexes. This makes it easier to manage module addition and
// deletion than putting everything in one database (file).
//
SmiDb::SmiDb(const char *dbPathArg,Mode dbModeArg)
          : Db(dbPathArg,dbModeArg), data(0) {

  if (dbMode == Delete) {
    return;
    }

  //
  // We limit the number of databases that can be open at the same
  // time. The MIB manager only needs one database open at a time,
  // so writers get 1
  //
  data = new SmiDbPrivate(dbMode == Db::Read ? 4 : 1);
  data -> dbIndex = new DbFlatDir(dbPath,dbMode);
  }

SmiDb::~SmiDb() {
  delete data;
  }

static void buildKeyInfo(const char *key,char *moduleName,char *recordName) {

  if (*key == '/') key++ ;

  const char *delimPtr = std::strchr(key,'/');
  if (delimPtr) {
    //
    // The key names a specific module, and a record in the module
    // database
    //
    key = delimPtr + 1;
    delimPtr = std::strchr(key,'/');
    if (delimPtr) {
      int length = delimPtr - key;
      std::memcpy(moduleName,key,length);
      moduleName[length] = '\0';

      std::strcpy(recordName,delimPtr + 1);
      }
    else {
      //
      // There is no key after the module name, so we synthesize a
      // key
      //
      std::strcpy(moduleName,key);
      std::strcpy(recordName,"ModuleInfo");
      }
    }
  else {
    //
    // The key refers to an index record
    //
    *moduleName = '\0';
    std::strcpy(recordName,key);
    }
  }

Db *SmiDb::getDb(const char *moduleName,Mode mode) const {

  //
  // See if we have a database open for this module
  //
  Db *moduleDb = 0;
  ObjectMapping mapping = data -> dbMap -> find(moduleName);
  if (mapping.value) {
    moduleDb = ((SmiSubDb *)mapping.value) -> db;
    }
  else {
    //
    // If we already have the maximum number of databases open,
    // close the oldest
    //
#ifdef REALLY_SMALL_MEMORY
    if (data -> dbList -> count() > data -> maxOpenDbs) {
      ListIterator it = data -> dbList -> iterator();
      String *oldModuleName = (String *)it.next();
      data -> dbMap -> remove(oldModuleName -> getChars());
      it.remove();
      }
#endif

    //
    // No database open for this module, so open it
    //
    char moduleDbPath[256];
    std::sprintf(moduleDbPath,"%s/%s.db",dbPath,moduleName);
    moduleDb = new SmiModuleDb(moduleDbPath,mode);
    if (moduleDb -> exception) {
      THROW(new SmiDbException(moduleDb -> exception -> getMessage(),
                               moduleDb -> exception -> getDiagnostic()),0);
      }

    //
    // Add the database to our module database map
    //
    String *savedModuleName = new String(moduleName);
    data -> dbMap -> add(savedModuleName,new SmiSubDb(moduleDb));
    data -> dbList -> add(savedModuleName);
    }

  return moduleDb;
  }

void *SmiDb::fetch(const char *key,int *length) const {
  void *record = 0;

  char moduleName[256];
  char recordName[256];
  buildKeyInfo(key,moduleName,recordName);

  if (moduleName[0]) {
    Db *moduleDb = getDb(moduleName,dbMode);
    if (exception) {
      return 0;
      }

    record = moduleDb -> fetch(recordName,length);
    if (moduleDb -> exception) {
      THROW(new SmiDbException(moduleDb -> exception -> getMessage(),
                               moduleDb -> exception -> getDiagnostic()),0);
      }
    }
  else {
    //
    // The key is the name of a record in the index database
    //
    record = data -> dbIndex -> fetch(recordName,length);
    if (data -> dbIndex -> exception) {
      THROW(new SmiDbException(data -> dbIndex -> exception -> getMessage(),
                               data -> dbIndex -> exception -> getDiagnostic()),
                               0);
      }
    }

  return (unsigned char *)record;
  }

void SmiDb::store(const char *key,const void *record,const int length) {

  char moduleName[256];
  char recordName[256];
  buildKeyInfo(key,moduleName,recordName);

  if (moduleName[0]) {
    Db *moduleDb = getDb(moduleName,dbMode);
    if (exception) {
      return;
      }

    moduleDb -> store(recordName,record,length);
    if (moduleDb -> exception) {
      THROW(new SmiDbException(moduleDb -> exception -> getMessage(),
                               moduleDb -> exception -> getDiagnostic()),);
      }
    }
  else {
    //
    // The key is the name of a record in the index database
    //
    data -> dbIndex -> store(recordName,(char *)record,length);
    if (data -> dbIndex -> exception) {
      THROW(new SmiDbException(data -> dbIndex -> exception -> getMessage(),
                         data -> dbIndex -> exception -> getDiagnostic()),);
      }
    }
  }

void SmiDb::deleteDatabase() {

  if (dbMode == Db::Read) {
    THROW(new SmiDbException("Can't delete if database is opened read-only!"),);
    }

  DIR *directory = ::opendir(dbPath);
  if (! directory) {
    THROW(new SmiDbException("Failed to delete database"),);
    }

  //
  // After this, this instance can only be destroyed
  //
  delete data;
  data = 0;

  //
  // Delete all database files in the directory (this assumes none
  // of the module or index databases use their own subdirectory).
  //
  struct dirent *dirEntry;
  while ((dirEntry = ::readdir(directory)) != 0) {
    if (dirEntry -> d_name[0] != '.') {
      char path[256];
      std::sprintf(path,"%s/%s",dbPath,dirEntry -> d_name);

      ::unlink(path); // Hmm, wonder if this works in the middle of this loop?
                      // or has some nasty page size dependency
      }
    }

  ::closedir(directory);

  //
  // Delete the database directory
  //
  ::unlink(dbPath);
  }

void SmiDb::deleteRecord(const char *moduleName) {
  if (dbMode != Db::Write) {
    THROW(new SmiDbException("Can't delete if database is opened read-only!"),);
    }

  //
  // Remove the database file for this module
  //
  Db *moduleDb = getDb(moduleName,Delete);
  if (exception) {
    return;
    }
  moduleDb -> deleteDatabase();

  //
  // Delete the module database from the module map
  //
  data -> dbMap -> remove(moduleName);
  }
