/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>

#include "SmiNotFoundException.hpp"

#include "SmiModule.hpp"
#include "SmiModuleDriver.hpp"

#include "SmiAssignmentDriver.hpp"
#include "SmiAssignmentInfoList.hpp"
#include "SmiAssignmentInfoListDriver.hpp"

#include "SmiTypedefDriver.hpp"
#include "SmiTypedefInfoList.hpp"
#include "SmiTypedefInfoListDriver.hpp"

#include "SmiImportsDriver.hpp"
#include "SmiRecordDriver.hpp"

SmiModule::SmiModule(const char *name)
               : smiDb(0), assignmentInfoList(0), assignmentCache(0),
                 typedefInfoList(0), typedefCache(0), importsMap(0) {

  this -> name = new String(name);
  }

SmiModule::SmiModule(const SmiDb *smiDb,const char *name)
               : smiDb(smiDb), assignmentInfoList(0), assignmentCache(0),
                 typedefInfoList(0), typedefCache(0), importsMap(0) {

  this -> name = new String(name);

  SmiModuleDriver driver(this,smiDb,name);
  driver.load();
  if (driver.exception) {
    return;
    }
  }

SmiModule::~SmiModule() {
  delete assignmentInfoList;
  delete assignmentCache;
  delete typedefInfoList;
  delete typedefCache;
  delete importsMap;
  delete name;
  }

const char *SmiModule::getName() const {
  return name -> getChars();
  }

void SmiModule::setLanguage(const char *languageString) {
  }

SmiTypedefInfoList *SmiModule::getTypedefInfoList() const {
  if (typedefInfoList != 0) return typedefInfoList;

  ((SmiModule *)this) -> typedefInfoList = new SmiTypedefInfoList();
  SmiTypedefInfoListDriver driver(typedefInfoList,this,smiDb);
  driver.load();
  if (driver.exception) {
    delete typedefInfoList;
    ((SmiModule *)this) -> typedefInfoList = 0;
    return 0;
    }

  return typedefInfoList;
  }

static void dumpTypedefs(ObjectHashMap *typedefCache) {
  if (! typedefCache) std::printf("Typedef cache is empty\n");
  else {
    ObjectHashMapIterator it = typedefCache -> iterator();
    std::printf("Typedef cache contains:\n");
    while (it.hasNext()) {
      ObjectMapping mapping = it.next();
      std::printf("  %s\n",((SmiTypedef *)mapping.value) -> getName());
      }
    }
  }

void SmiModule::addTypedef(SmiTypedef *typedefinition) {
  if (! typedefCache) typedefCache = new ObjectHashMap();
  typedefCache -> add(new String(typedefinition -> getName()),typedefinition);
  }

const SmiTypedef *SmiModule::getTypedef(const char *name) const {

  //
  // See if the requested typedef is in the cache
  //
  if (typedefCache) {
    ObjectMapping mapping = typedefCache -> find(name);
    if (mapping.value != 0) {
      return (SmiTypedef *)mapping.value;
      }
    }

  //
  // Requested typedef is not in the cache, so load it from the
  // database
  //
  SmiTypedefDriver driver(smiDb,this,name);
  driver.load();
  if (driver.exception) {
    return 0;
    }

  SmiTypedef *typedefinition = driver.takeTypedef();

  //
  // Add this typedef to the cache
  //
  ((SmiModule *)this) -> addTypedef(typedefinition);
  return typedefinition;
  }

SmiAssignmentInfoList *SmiModule::getAssignmentInfoList() const {
  if (assignmentInfoList != 0) return assignmentInfoList;

  ((SmiModule *)this) -> assignmentInfoList = new SmiAssignmentInfoList();
  SmiAssignmentInfoListDriver driver(assignmentInfoList,this,smiDb);
  driver.load();
  if (driver.exception) {
    delete assignmentInfoList;
    ((SmiModule *)this) -> assignmentInfoList = 0;
    return 0;
    }

  return assignmentInfoList;
  }

const SmiAssignment *SmiModule::getAssignment(const char *name) const {
  //
  // See if the requested assignment is in the cache
  //
  if (assignmentCache) {
    ObjectMapping mapping = assignmentCache -> find(name);
    if (mapping.value != 0) {
      return (SmiAssignment *)mapping.value;
      }
    }

  //
  // Requested assignment is not in the cache, so load it from the
  // database
  //
  SmiAssignmentDriver driver(smiDb,this,name);
  driver.load();
  if (driver.exception) {
    return 0;
    }
  SmiAssignment *assignment = driver.takeAssignment();

  //
  // Add this assignment to the cache
  //
  if (! assignmentCache) assignmentCache = new ObjectHashMap();
  assignmentCache -> add(new String(name),assignment);

  return assignment;
  }

const char *SmiModule::getModuleImport(const char *symbolName) const {
  //
  // If we have no imports map yet, read it from the database
  //
  if (! importsMap) {
    SmiImportsDriver driver(smiDb,this);
    driver.load();
    if (driver.exception) {
      return 0;
      }
    importsMap = driver.takeImportsMap();
    }

  //
  // Find the entry
  //
  return importsMap -> getModule(symbolName);
  }

String *SmiModule::toString() const {
  return new String(*name);
  }

String *SmiModule::loadField(const char *key) const {
  char fullKey[1024];
  std::sprintf(fullKey,"/MODULE/%s/%s",name -> getChars(),key);
  SmiRecordDriver driver(smiDb,fullKey);
  driver.load();
  if (driver.exception) {
    return 0;
    }

  return driver.takeValue();
  }
