/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <malloc.h>
#include <cstring>
#include <errno.h>

#include <classlib/String.hpp>
#include <classlib/ObjectHashMap.hpp>
#include <classlib/DbFlatDir.hpp>
#include <classlib/Xml.hpp>

#include "SmiException.hpp"
#include "SmiModuleDb.hpp"

//
// Structure describing the position and size of a record in the
// database. These structures are serialised to the index file
//
class RecordData : public Object {
  public:
    RecordData(long recordOffsetArg,int recordLengthArg)
         : recordOffset(recordOffsetArg), recordLength(recordLengthArg) {
      }

    ~RecordData() {
      }

    long recordOffset;
    int recordLength;
  };

class SmiModuleDbPrivate {
  public:
    SmiModuleDbPrivate() : fd(-1), recordMap(0) {
      recordMap = new ObjectHashMap();
      }

    ~SmiModuleDbPrivate() {
      delete recordMap;
      }

    int fd;
    ObjectHashMap *recordMap;
  };

SmiModuleDb::SmiModuleDb(const char *pathArg,Mode modeArg)
                : Db(pathArg,modeArg), data(0) {

  if (dbMode == Delete) return;

  data = new SmiModuleDbPrivate();

  if (dbMode == Write) {
    data -> fd = ::open(dbPath,O_RDWR | O_CREAT,0644);
    }
  else {
    data -> fd = ::open(dbPath,O_RDONLY);
    }
  if (data -> fd == -1) {
    THROW(new SmiException("Failed to open database"),);
    }

  //
  // Read any existing index
  //
  char dbDirectory[256];
  char *delimPtr = std::strrchr(dbPath,'/');
  int length = delimPtr - dbPath;
  std::memcpy(dbDirectory,dbPath,length);
  dbDirectory[length] = '\0';
  DbFlatDir indexDb(dbDirectory,dbMode);
  if (indexDb.exception) {
    RETHROW(indexDb.exception,);
    }

  char dbName[256];
  std::strcpy(dbName,delimPtr + 1);
  std::strcat(dbName,".idx");
  void *record = indexDb.fetch(dbName,&length);
  if (indexDb.exception) {
    RETHROW(indexDb.exception,);
    }

  if (record) {
    Xml indexXml;
    indexXml.loadXmlFromBuffer((char *)record,length);
    Db::free(record);
    if (indexXml.exception) {
      RETHROW(indexXml.exception,);
      }
    XmlElement root = indexXml.getRootElement();
    int childCount = indexXml.getElementChildCount(root);
    for (int i = 0 ; i < childCount ; i++ ) {
      XmlElement indexElement = indexXml.getElementChild(root,i);
      RecordData *recordData =
        new RecordData(atol(indexXml.getElementAttributeValue(indexElement,1)),
                       atoi(indexXml.getElementAttributeValue(indexElement,2)));
      data -> recordMap ->
          add(new String(indexXml.getElementAttributeValue(indexElement,0)),
              recordData);
      }
    }
  }

SmiModuleDb::~SmiModuleDb() {
  if (data && (data -> fd != -1)) {
    //
    // If this is a writable database, we need to save the index
    //
    if (dbMode == Write) {
      char dbDirectory[256];
      char *delimPtr = std::strrchr(dbPath,'/');
      int length = delimPtr - dbPath;
      std::memcpy(dbDirectory,dbPath,length);
      dbDirectory[length] = '\0';
      DbFlatDir indexDb(dbDirectory,dbMode);
      if (indexDb.exception) {
        RETHROW(indexDb.exception,);
        }

      char dbName[256];
      std::strcpy(dbName,delimPtr + 1);

      Xml indexXml;
      XmlElement root = indexXml.setRootElement("index");
      char scratch[32];
      ObjectHashMapIterator it = data -> recordMap -> iterator();
      while (it.hasNext()) {
        ObjectMapping mapping = it.next();
        RecordData *recordData = (RecordData *)mapping.value;
        XmlElement indexElement = indexXml.addElement(root,"i");
        indexXml.addElementAttribute(indexElement,"n",
                                     ((String *)mapping.name) -> getChars());
        std::sprintf(scratch,"%lu",recordData -> recordOffset);
        indexXml.addElementAttribute(indexElement,"o",scratch);
        std::sprintf(scratch,"%d",recordData -> recordLength);
        indexXml.addElementAttribute(indexElement,"l",scratch);
        }

      char *xmlData = indexXml.xmlString();
      std::strcat(dbName,".idx");
      indexDb.store(dbName,xmlData,std::strlen(xmlData));
      Xml::free(xmlData);
      if (indexDb.exception) {
        ::close(data -> fd);
        RETHROW(indexDb.exception,);
        }
      }

    if (::close(data -> fd) == -1) {
      THROW(new SmiException("Failed to finalise database"),);
      }
    }

  delete data;
  }

void *SmiModuleDb::fetch(const char *recordName,int *length) const {
  //
  // Find the record in the index
  //
  ObjectMapping mapping = data -> recordMap -> find(recordName);
  if (! mapping.value) return 0;
  RecordData *recordData = (RecordData *)mapping.value;

  //
  // Set the file position to the correct offset for the record
  //
  if (::lseek(data -> fd,recordData -> recordOffset,SEEK_SET) == -1) {
    THROW(new SmiException("Failed to locate record in the database"),0);
    }

  int left = recordData -> recordLength;
  char *record = (char *)malloc(left);
  char *recordPtr = record;
  ssize_t consumed;
  while (left) {
    consumed = ::read(data -> fd,recordPtr,left);
    if (consumed == -1) {
      THROW(new SmiException("Failed to read a module record",
                            strerror(errno)),0);
      }

    if (consumed == 0) {
      THROW(new SmiException("Failed to read record (end of file)"),0);
      }

    left -= consumed;
    recordPtr += consumed;
    }

  *length = recordData -> recordLength;
  return record;
  }

void SmiModuleDb::store(const char *recordName,const void *record,int length) {
  if (dbMode == Read) {
    THROW(new SmiException("Can't write to read-only database"),);
    }

  off_t recordOffset = ::lseek(data -> fd,0,SEEK_END);
  int bytesToWrite = length;
  while (bytesToWrite) {
    ssize_t wrote = ::write(data -> fd,record,bytesToWrite);
    if (wrote == -1) {
      THROW(new SmiException("Failed to write record",strerror(errno)),);
      }

    (char *)record += wrote;
    bytesToWrite -= wrote;
    }

  RecordData *recordData = new RecordData(recordOffset,length);
  data -> recordMap -> add(new String(recordName),recordData);
  }

void SmiModuleDb::deleteRecord(const char *recordName) {
  THROW(new SmiException(
                "This database type does not support record deletion"),);
  }

void SmiModuleDb::deleteDatabase() {
  if (dbMode == Read) {
    THROW(new SmiException("Can't delete read-only database"),);
    }

  if (data) ::close(data -> fd);

  char indexPath[256];
  std::sprintf(indexPath,"%s.idx",dbPath);
  if (::unlink(dbPath) == -1) {
    THROW(new SmiException("Failed to delete record file",strerror(errno)),);
    }
  if (::unlink(indexPath) == -1) {
    THROW(new SmiException("Failed to delete record index",strerror(errno)),);
    }

  data = 0;
  }
