/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstring>
#include <cstdlib>

#include <classlib/String.hpp>
#include <classlib/ObjectList.hpp>
#include <classlib/ObjectHashMap.hpp>

#include "SmiDb.hpp"
#include "SmiOid.hpp"
#include "SmiOidTree.hpp"
#include "SmiOidTreeDriver.hpp"

SmiOidTree::SmiOidTree(SmiDb *smiDb) : smiDb(smiDb) {
  root = new SmiOidTreeNode();
  oidToNameMap = new ObjectHashMap();
  nameToOidMap = new ObjectHashMap();

  SmiOidTreeDriver driver(this,smiDb);
  driver.load();
  }

SmiOidTree::~SmiOidTree() {
  //
  // Deleting the root of the OID results in deletion of the entire tree
  //
  delete root;

  //
  // Need to prune the hashmaps to account for the nodes we deleted. We
  // do this by deleting the names from one hashmap, and then invalidate
  // both hashmaps (invalidate() empties a map without deleting the
  // mappings)
  //
  ObjectHashMapIterator it = oidToNameMap -> iterator();
  while (it.hasNext()) {
    ObjectMapping mapping = it.next();
    it.detach();
    delete mapping.value;
    }
  oidToNameMap -> invalidate();
  nameToOidMap -> invalidate();

  //
  // Delete the (empty) hashmaps.
  //
  delete oidToNameMap;
  delete nameToOidMap;
  }

const SmiOidTreeNode *SmiOidTree::add(const char *name,const char *number) {
  SmiOidTreeNode *currentNode = root;
  while (number != 0) {
    int sid = std::atoi(number);

    currentNode = ((SmiOidTreeNode *)currentNode) -> addChild(sid);

    number = strchr(number,'.');
    if (number != 0) number++ ;
    }

  String *nameString = new String(name);
  oidToNameMap -> add(currentNode,nameString);
  nameToOidMap -> add(nameString,currentNode);

  return currentNode;
  }

const SmiOidTreeNode *SmiOidTree::getRootNode() const {
  return root;
  }

const SmiOidTreeNode *SmiOidTree::findDeepest(const SmiOid *oid) const {
  const SmiOidTreeNode *currentNode = root;
  unsigned char length = oid -> getLength();
  const SmiTypes::SubIdentifier *sid = oid -> getSids();

  for (int i = 0 ; i < length ; i++ ) {
    int currentSid = sid[i];
    ObjectList *children = currentNode -> getChildren();

    //
    // No children means we found a leaf node, so it's the deepest
    // node found for the given OID (for example the given OID is
    // an instance of the leaf)
    //
    if (children == 0) return currentNode;

    //
    // Search the children for the current sub-identifier of the OID.
    //
    ListIterator it = children -> iterator();
    int found = false;
    while (it.hasNext()) {
      SmiOidTreeNode *child = (SmiOidTreeNode *)it.next();
      int childSid = child -> getSid();

      //
      // Found the sub-identifier, so we can iterate on this child
      //
      if (childSid == currentSid) {
        currentNode = child;
        found = true;
        break;
        }

      //
      // The children are ordered by sub-identifier, so a larger
      // one means the current sub-identifier cannot be in the
      // list, and so we've found the deepest node for this OID
      //
      if (childSid > currentSid) return currentNode;
      }

    //
    // If we went off the end of the list of children, we've found
    // the deepest node for this OID
    //
    if (! found) return currentNode;
    }

  //
  // Exhausted the OID's sub-identifiers, so we've found the
  // deepest node
  //
  return currentNode;
  }

const SmiOidTreeNode *
         SmiOidTree::findDeepestNamed(const SmiOidTreeNode *node) const {

  while (node) {
    ObjectMapping mapping = oidToNameMap -> get(node);
    if (mapping.name) return node;

    node = node -> getParent();
    }

  //
  // This never happens provided the children of the root node
  // are all named (and the root node should always have a
  // child called 'xxx.iso', and probably another one called
  // xxx.zeroDotZero (0.0)
  //
  return 0;
  }

const SmiOidTreeNode *SmiOidTree::findDeepestNamed(const SmiOid *oid) const {
  const SmiOidTreeNode *node = findDeepest(oid);
  if (node == 0) return 0;
  return findDeepestNamed(node);
  }

const char *SmiOidTree::getName(const SmiOidTreeNode *node) const {
  ObjectMapping mapping = oidToNameMap -> get(node);
  if (mapping.value) {
    return ((String *)mapping.value) -> getChars();
    }

  return 0;
  }

const SmiOidTreeNode *SmiOidTree::getNode(const char *name) const {
  const char *dotPtr = std::strchr(name,'.');
  if (dotPtr) {
    //
    // The name is fully qualified so we can look it up directly
    //
    ObjectMapping mapping = nameToOidMap -> find(name);
    return (SmiOidTreeNode *)mapping.value;
    }

  //
  // For the rare case where an unqualifed name is given (lazy humans)
  // we can do a search of the map. If this turns out to be a problem
  // later, we can re-organise the map into a map of maps where the
  // first one is a module map (this uses less memory than
  // putting both qualifeed and unqualified entries in the same map,
  // or creating both a qualified and unqualified map)
  //
  ObjectHashMapIterator it = nameToOidMap -> iterator();
  while (it.hasNext()) {
    ObjectMapping mapping = it.next();
    const char *mappingName = ((String *)mapping.name) -> getChars();
    dotPtr = std::strchr(mappingName,'.');
    if (std::strcmp(dotPtr + 1,name) == 0) {
      return (SmiOidTreeNode *)mapping.value;
      }
    }

  return 0;
  }
