/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <cstdio>
#include <cstring>
#include <cstdlib>

#include "SmiSyntaxLoader.hpp"

SmiSyntaxLoader::SmiSyntaxLoader(const Xml& xml) : XmlLoader(xml), syntax(0) {
  }

SmiSyntaxLoader::~SmiSyntaxLoader() {
  delete syntax;
  }

static void loadEnumerations(SmiSyntax& syntax,const Xml& xml,
                             XmlElement enumsElement) {
  SmiEnumerationList *enums = new SmiEnumerationList();
  int childCount = xml.getElementChildCount(enumsElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement enumElement = xml.getElementChild(enumsElement,i);
    const char *name = xml.getElementName(enumElement);

    if (strcmp(name,"enumeration") == 0) {
      SmiEnumeration *enumeration =
            new SmiEnumeration(xml.getElementAttributeValue(enumElement,0),
                               xml.getElementAttributeValue(enumElement,1));
      enums -> add(enumeration);
      }
    }

  syntax.setEnumerationList(enums);
  }

static void loadTag(SmiSyntax& syntax,const Xml& xml,XmlElement tagElement) {
  const char *className = 0;
  const char *value = "<bad value>";
  unsigned char implicit = 0;

  int attributeCount = xml.getElementAttributeCount(tagElement);
  for (int i = 0 ; i < attributeCount ; i++ ) {
    const char *name = xml.getElementAttributeName(tagElement,i);
    if (std::strcmp(name,"class") == 0) {
      className = xml.getElementAttributeValue(tagElement,i);
      }
    else if (std::strcmp(name,"value") == 0) {
      value = xml.getElementAttributeValue(tagElement,i);
      }
    else if (std::strcmp(name,"implicit") == 0) {
      if (std::strcmp(xml.getElementAttributeValue(tagElement,i),"true") == 0) {
        implicit = 1;
        }
      }
    }

  SmiAsn1Tag *tag = new SmiAsn1Tag(value);
  tag -> setClass(className);
  tag -> setImplicit(implicit);
  syntax.setTag(tag);
  }

static void loadSizes(SmiSyntax& syntax,const Xml& xml,
                      XmlElement sizesElement) {
  SmiSizeList *sizes = new SmiSizeList();
  int childCount = xml.getElementChildCount(sizesElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement sizeElement = xml.getElementChild(sizesElement,i);
    const char *name = xml.getElementName(sizeElement);

    if (strcmp(name,"size") == 0) {
      SmiSize *size =
         new SmiSize(std::strtoll(xml.getElementAttributeValue(sizeElement,0),
                                  0,10),
                     std::strtoll(xml.getElementAttributeValue(sizeElement,1),
                                  0,10));
      sizes -> add(size);
      }
    }

  syntax.setSizeList(sizes);
  }

static void loadComponents(SmiSyntax& syntax,const Xml& xml,
                           XmlElement componentsElement) {
  SmiComponentList *components = new SmiComponentList();
  SmiSyntaxLoader loader(xml);
  int childCount = xml.getElementChildCount(componentsElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement syntaxElement = xml.getElementChild(componentsElement,i);
    const char *name = xml.getElementName(syntaxElement);

    if (std::strcmp(name,"syntax") == 0) {
      loader.load(syntaxElement);
      components -> add(loader.takeSyntax());
      }
    }

  syntax.setComponentList(components);
  }

static void loadRanges(SmiSyntax& syntax,const Xml& xml,
                       XmlElement rangesElement) {
  SmiRangeList *ranges = new SmiRangeList();
  int childCount = xml.getElementChildCount(rangesElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement rangeElement = xml.getElementChild(rangesElement,i);
    const char *name = xml.getElementName(rangeElement);

    if (strcmp(name,"range") == 0) {
      SmiRange *range =
         new SmiRange(std::strtoll(xml.getElementAttributeValue(rangeElement,0),
                                   0,10),
                      std::strtoll(xml.getElementAttributeValue(rangeElement,1),
                                   0,10));
      ranges -> add(range);
      }
    }

  syntax.setRangeList(ranges);
  }

void SmiSyntaxLoader::load(XmlElement syntaxElement) {
  syntax = new SmiSyntax();

  //
  // First pick up the attributes of the syntax
  //
  int attributeCount = xml.getElementAttributeCount(syntaxElement);
  for (int i = 0 ; i < attributeCount ; i++ ) {
    const char *name = xml.getElementAttributeName(syntaxElement,i);
    const char *value = xml.getElementAttributeValue(syntaxElement,i);

    if (std::strcmp(name,"name") == 0) {
      syntax -> setName(value);
      }
    else if (std::strcmp(name,"module") == 0) {
      syntax -> setModule(value);
      }
    else if (std::strcmp(name,"typeref") == 0) {
      syntax -> setTyperef(value);
      }
    else if (std::strcmp(name,"subtype") == 0) {
      syntax -> setSubtype(value);
      }
    }

  //
  // Now process each child element in turn
  //
  int childCount = xml.getElementChildCount(syntaxElement);
  for (int i = 0 ; i < childCount ; i++ ) {
    XmlElement child = xml.getElementChild(syntaxElement,i);
    const char *name = xml.getElementName(child);

    if (std::strcmp(name,"enumerations") == 0) {
      loadEnumerations(*syntax,xml,child);
      }
    else if (std::strcmp(name,"sizes") == 0) {
      loadSizes(*syntax,xml,child);
      }
    else if (std::strcmp(name,"ranges") == 0) {
      loadRanges(*syntax,xml,child);
      }
    else if (std::strcmp(name,"components") == 0) {
      loadComponents(*syntax,xml,child);
      }
    else if (std::strcmp(name,"tag") == 0) {
      loadTag(*syntax,xml,child);
      }
    }
  }

//
// Return the loaded syntax object, transferring ownership (responsibility
// to delete) to the caller
//
SmiSyntax *SmiSyntaxLoader::takeSyntax() {
  SmiSyntax *returnedSyntax = syntax;
  syntax = 0;
  return returnedSyntax;
  }

String *SmiSyntax::toString() const {
  return 0;
  }
