/*
 *  Copyright (C) 2003  Clive Butler
 *
 *  Author's email address: cebdev@users.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <cstring>

#include <snmp++/smi.h>

#include <classlib/String.hpp>

#include <smidb/SmiObjectType.hpp>

#include "SmiSnmpPPAdapter.hpp"


String *SmiSnmpPPAdapter::snmpVbToString(const Smi &smi,
                                         const SmiModule *module,
                                         const SmiSyntax *syntax,
                                         const Vb &vb,
                                         int options) {

  SmiUINT32 snmpSyntax = vb.get_syntax();
  switch (snmpSyntax) {
    case sNMP_SYNTAX_INT32: { // == sNMP_SYNTAX_INT:
      long value;
      vb.get_value(value);

      return smi.signedIntegerString(module,syntax,value);
      }

    case sNMP_SYNTAX_UINT32: // == sNMP_SYNTAX_GAUGE32:
    case sNMP_SYNTAX_CNTR32: {
      unsigned long value;
      vb.get_value(value);

      //
      // According to RFC 2579, Counter32 can't be hinted but
      // we send it the same route as Unsigned32 anyway..
      //
      return smi.unsignedIntegerString(module,syntax,value);
      }

    case sNMP_SYNTAX_TIMETICKS: {
      unsigned long value;
      vb.get_value(value);
      return smi.timeticksString(module,syntax,value);
      }

    case sNMP_SYNTAX_CNTR64: {
      return new String(vb.get_printable_value());
      }

    case sNMP_SYNTAX_OCTETS: {
      unsigned long length;
      unsigned char buffer[MAX_SNMP_PACKET]; // Better be enough!
      vb.get_value(buffer,length);
      return smi.octetString(module,syntax,buffer,length,options);
      }

    case sNMP_SYNTAX_BITS: {
      //
      // Note that this is the ASN.1 BITS type, not the SNMP SMI BITS
      // pseudo-type!
      //
      unsigned long length;
      unsigned char buffer[MAX_SNMP_PACKET]; // Better be enough!
      vb.get_value(buffer,length);
      return smi.bitString(module,syntax,buffer,length);
      }

    case sNMP_SYNTAX_OPAQUE: {
      unsigned long length;
      unsigned char buffer[MAX_SNMP_PACKET]; // Better be enough!
      vb.get_value(buffer,length);
      return smi.octetString(module,syntax,buffer,length,options);
      }

    case sNMP_SYNTAX_IPADDR: {
      // SmiOCTETS string

      // TBD - check for name
      return new String(vb.get_printable_value());
      }

    case SMI_NSAP: {
      // SmiOCTETS string

      return new String(vb.get_printable_value());
      }

    case sNMP_SYNTAX_OID: {
      Oid value;
      vb.get_value(value);
      SmiLPOID snmpOid = value.oidval();
      SmiOid oid(snmpOid -> ptr,snmpOid -> len);

      return smi.stringValue(oid,options);
      }

    case sNMP_SYNTAX_NULL: {
      return new String("");
      }

    case sNMP_SYNTAX_NOSUCHOBJECT:
      return new String("[Exception: NoSuchObject]");

    case sNMP_SYNTAX_NOSUCHINSTANCE:
      return new String("[Exception: NoSuchInstance]");

    case sNMP_SYNTAX_ENDOFMIBVIEW:
      return new String("[Exception: EndOfMibView]");

    default: return new String("[Dunno]");
    }
  }

String *SmiSnmpPPAdapter::snmpOidToString(const Smi& smi,const Oid &oid,
                                          int options) {
  SmiLPOID oidValue = const_cast<Oid&>(oid).oidval();
  SmiOid smiOid(oidValue -> ptr,oidValue -> len);
  return smi.stringValue(smiOid,options);
  }

SmiOid *SmiSnmpPPAdapter::snmpOid(const Oid& oid) {
  SmiLPOID oidValue = const_cast<Oid&>(oid).oidval();
  return new SmiOid(oidValue -> ptr,oidValue -> len);
  }

String *SmiSnmpPPAdapter::snmpVbToString(const Smi& smi,const Vb& vb,
                                         int options) {

  SmiOid *smiOid = SmiSnmpPPAdapter::snmpOid(vb.get_oid());
  const SmiOidTreeNode *node = smi.getOidTree() -> findDeepest(smiOid);
  if (! node) {
    return new String(vb.get_printable_value());
    }

  const char *nodeName = smi.getOidTree() -> getName(node);
  if (! nodeName) {
    return new String(vb.get_printable_value());
    }

  char *delimPtr = strchr(nodeName,'.');
  int length = delimPtr - nodeName;
  char moduleName[128];
  std::strncpy(moduleName,nodeName,length);
  moduleName[length] = '\0';
  char *assignmentName = delimPtr + 1;

  const SmiModule *module = smi.getModuleList() -> getModule(moduleName);
  if (! module) {
    return new String(vb.get_printable_value());
    }

  const SmiSyntax *syntax = 0;
  const SmiAssignment *assignment = module -> getAssignment(assignmentName);
  if (assignment) {
    if (assignment -> getType() == SmiAssignmentInfo::OBJECT_TYPE) {
      const SmiObjectType *objectType = (SmiObjectType *)assignment;
      syntax = objectType -> getSyntax();
      }
    }

  return SmiSnmpPPAdapter::snmpVbToString(smi,module,syntax,vb,options);
  }

